###############################################################################
# IMPORTS

from classlib.base import *

###############################################################################
class Transport(Task):
    """

    Task that controls the transport of a carrier.
    The task is done when the carrier arrived at its destination
    and when the item is unloaded (if item is not 'empty')
    """
    
    def init_task(self,args):
        """
        Inserts item id into the item data base.
        Then the task waits until the carrier
        Arrived at its destination and unloaded the item,
        even if it is the 'empty' item.
        """

        # store item and destination
        self.item_id=args['item_id']
        self.dest_id=args['dest_id']
        #print self.obj.id,'with status',self.obj.status,\
        #      'init transport task',self.id
        #print '  of',self.item_id,'to',self.dest_id,args['search_item']
        self.mode=args.get('mode','loaded')
        
        # insert task into car_tasks table with carrier id as key
        # IMPORTANT: as a general rule there should be only one task
        # per item at a time.

        # backlink item with task with item id as key
        self.obj.item_tasks[self.item_id]=self

        self.set_status('processing')

        # verify whether there is already an item on board.
        # should this be the case, and this task is a search_item task
        # then the carrier must not be started, but the task must
        # wait to be killed...and the carrier must wait to be started
        # by the task for the already loaded item.
        # Here it is assumed that alway the correct item is loaded.

        #print '\ninit_task ',self.id,'in car',self.obj.id,self.dest_id,self.mode,self.item_id
        if (self.mode=='search_item') &\
           (self.obj.status!='loading')&\
           (len(self.obj.items)==0):
            self.start_car(mode='search_item')
            
        elif (self.mode!='search_item') &\
             (self.obj.status!='loading')&\
             (self.obj.items.keys().count(self.item_id)>0):
            # If item is already on board, then give the start signal
            # Otherwise the start signal is given by the carrier
            # as soon as items are on board, see enter method of carrier.
            # Here one may also check whether accompaning passangers
            # are on board as well.
            self.start_car()
            
        elif (self.mode=='empty'):
            # empty mission
            
            self.start_car(self.mode)
            
        else:
            #print '  wait to be killed or untill item entered carrier'
            pass
            

    def start_car(self,mode='loaded'):
        """
        This function is called when the carrier is loaded and redy to start.
        This function actually starts the carrier and
        initializes some statistical data for this task.
        """
        #print 'task',self.id,\
        #      ':start_car:  start carrier',self.obj.id,'in mode:',mode
        self.info['trippath_start']=self.obj.get_path()
        self.info['triptime_start']=self.obj.sim.time
        
        if self.obj.start(mode):
            print 'WARNING:in task',self.id,': could not start carrier',\
                  self.obj.id,' here.'

            print '  task from client',self.client_id,\
                  'with order_id:',self.order_id 
            #self.obj.sim.stop()
            print '  Task wait to be killed...'
            
    def unloaded(self):
        """
        This function is called by the carrier when the task's item
        has been unloaded.

        This method also completes the statistical data, regarding this task.
        
        Note that this method is also called with the empty item
        when carrier arrived regularly at its destination.
        """
        #print self.obj.id,': Task',self.id,'unloaded and done'

        # pass current location as return info to carman
        self.info['sec_id']=self.obj.sec.id
        self.info['elem_id']=self.obj.sec.elem.id

        # trip statistics
        if self.info.has_key('trippath_start'):
            self.info['trippath']=self.obj.get_path()-\
                                   self.info['trippath_start']
            self.info['triptime']=self.obj.sim.time-self.info['triptime_start']

        # this will terminade task and return all info
        self.set_status('done')

    def found_item(self):
        """
        This function is called by the carrier when the task with
        the 'empty' item found an item.

        The emty item task is therefore done.
        """
        #print self.obj.id,': Task',self.id,'unloaded and done'
        self.mode='done'
        # pass current location as return info to carman
        self.info['sec_id']=self.obj.sec.id
        self.info['elem_id']=self.obj.sec.elem.id
        self.info['found_item']=1
        self.set_status('done')


    def cleanup(self):
        """
        This function cleans up all memory traces of this task in
        the respective object.
        """
        #print self.obj.id,':cleanup task',self.id
        if self.obj.item_tasks.has_key(self.item_id):
            if self.obj.item_tasks[self.item_id]==self:
                del self.obj.item_tasks[self.item_id]

        
            
class Mod(Module):
    """
    Class, containing all data that are specific to a particular carrier
    module (Data and methods that apply to all carrier modules of
    the same type are implemented in the carrier class, below).
    """
    def __init__(self,obj,
                 pos=[0,0,0],
                 status='edit'):
        
        # back-link with simulation carrier object
        #self.obj=obj

        # parameters
        self.length=3.5
        self.width=1.75
        self.accel_comfort=0.2*9.98
        self.accel_emergency=0.5*9.98
        self.accel_failure=1.0*9.98
        self.time_brake=0.1
        
          
        self.modind=-1 # index in carrier array struct
        self.id='No id' # determined by add_mod of the simulation object

        self.sec=None
        self.status=status

        self.pos=pos
        self.ind=-1  # index in Netsim's array structure
        #  initialization, common to all modules
        self.init_common(obj)
        
        # add this module to car simulation object
        self.obj.add_mod(self)

        
        
        # current section of track elementthat is  controlling this carrier
        self.sec=None 
        
        self.car_infront=None
        self.ghost_infront=None
        #self.mark_at=None

        # table with 'physical' items and item id as key
        # can be user, freight, container, cabins...
        self.items={}

        # table mapping item_id to the corresponding task
        self.item_tasks={}

        # link to task classes
        self.task_classes['transport']=Transport
        
        # initial destination id of track element,
        # determined by carrier management
        #self.dest_id=''


        # initialize default costs
        #self.init_costs()

        # initialize all variables that need to be set after each opening
        #self.set_state(obj)
    
        # initialize statistics
        #self.reset_statistics()

        #print 'init car',self.id
        
    def get_state_mod(self,state_image):
        '''
        This function removes all module-specific pointers from the
        state-image, see get_state function.
        To be overwritten by module-specific function.
        '''

        # remove tosec_active as it leads to an endless recursion
        # in the pickle algorithm. List will be rebuild in set_state_mod

        #replace pointer by id
        if self.sec:
            state_image['sec']=self.sec.id

        for itemid in self.items.keys():
            state_image['items'][itemid]=self.items[itemid].id

        if self.car_infront:
            state_image['car_infront']=self.car_infront.id

        if self.ghost_infront:
            state_image['ghost_infront']=self.ghost_infront.id
            
        #del state_image['sec']
        return state_image

    def rebuild_pointers(self):
        """
        Reconstruct pointers that have been taken away in get_state_mod
        """
        if self.obj.netsim:
            #replace id by pointer
            if type(self.sec)==types.IntType:
                self.sec=self.obj.netsim.rebuild_secpointer(self.sec)

        for itemid in self.items.keys():
            self.items[itemid]=self.sim.rebuild_pointer(self.items[itemid])

        if self.car_infront:
            self.car_infront=self.sim.rebuild_pointer(self.car_infront)

        if self.ghost_infront:
            self.ghost_infront=self.sim.rebuild_pointer(self.ghost_infront)
            
    def set_state_mod(self):
        '''
        
        This function is called once during first initialization and
        during the opening process of a simulation file. Here goes all the
        stuff that gets lost during the save process (essentially pointers
        to functions).  


        '''
        #replace id by pointer, if there are ids
        #self.rebuild_pointers()

        if string.atof(self.get_version())<0.3:
           for task in self.tasks.values():
                if not hasattr(task,'mode'):
                    if hasattr(task,'search_item'):
                        if task.search_item==1:
                            task.mode='search_item'
                        else:
                            task.mode='loaded'
    
        # link to task classes
        self.task_classes['transport']=Transport

        #if string.atof(self.get_version())<0.2:
        #print 'old version'
        if  not hasattr(self,'time_brake'):
            self.time_brake=0.0
            # also in netsim the default is zero
            # can only be modifeid in set_params
        
    #def run(self,message):
    #    '''
    #    Message driven operator that makes the carrier running
    #    towards a new destination
    #    '''
    #    self.start()
    #    

    def init_costs(self):
        """
        Define default costs here.

        To be overwritten.
        """
        self.set_costs({'estim_init':{'name':'Estimated initial costs',
                                      'table':{1:     80.0*10**3,
                                               10:    40.0*10**3,
                                               100:   20.0*10**3,
                                               1000:  15.0*10**3,
                                               10000: 10.0*10**3,
                                               100000: 5.0*10**3,},
                                      'mode':'write'},
                        'estim_pa':{'name':'Estimated costs per year',
                                    'table':{1:      1000.0,
                                             10:     900.0,
                                             100:    800.0,
                                             1000:   700.0,
                                             10000:  500.0,
                                             100000: 100.0,},
                                    'mode':'write'},
                        },
                       lifetime_estim=10)
        
     
        
    def start(self,mode='no_task'):
        """
        Give the start signal to the track that the carrier is ready to start.
        There are different start modes:
        'search_item' carrier has a task, but is searching for an item.
        'loaded'      carrier is loaded with an item.
        'no_task'     carrier has currently no task.
        'direct'      start carrier directly with line speed without
                      further cunsultation of the track.
        """
        #print 'start',self.id,mode
        # make sure to have active status
        if not self.set_status('active'):

            # give signal to the section...will be passed to track element
            self.sec.start(self,mode)
        
            self.plot()
            return 0
        else:
            return 1
            
        

    def change_sec(self):
        '''
        Called by Netsim if this carrier changes sections.  First, the
        present section is called and returns the nexts section.  Then, the
        carrier must announce itself to the next section, such that the
        next section does know what to do with this carrier, i.g.  to
        branch, to stop, etc.

        This function must return (the pointer to the) next section.
        
        '''

        # ask section which is the next section
        nextsec, error=self.sec.leave(self)

        # compute mergedist for this section ..even if it is not a merge sec
        #self.obj.dist_merge[self.no]=nextsec.length-nextsec.clearingzone\
        #                              -3*nextsec.velocity_line**2*\
        #                              0.5/self.accel_emergency
        
        self.obj.dist_merge[self.no]=nextsec.length-nextsec.clearingzone\
                                      -(nextsec.velocity_line*self.time_brake+
                                        nextsec.velocity_line**2*\
                                        (0.5/self.accel_emergency-
                                         0.5/self.accel_failure))
        #if self.no==6:
        #    print 'change_sec',self.no,self.obj.dist_merge[self.no]
        #    print '  modind',self.modind
        #    print '  dist_merge',self.obj.dist_merge.arrays
        
        #if error=='unknown_dest': # desination not found in routing table
        #        if self.dest_id !='':
        #            # Warn the carrier management
        #                    
        #        self.send_mes({'to':self.man_id,
        #                       'op':'unknown_dest',
        #                       'dest_id':self.dest_id
        #                       })

        self.sec=nextsec
        self.sec.enter(self)
        return self.sec
  
    #def halt(self):
    #    '''Halt carrier and initiate unloading'''
    #    # Try to unload...carrier will go into load or halt status
    #    self.unload() 

    def unload(self):
        """
        Try to unload items which are arrived.
        Return 1 if something has been unloaded.
        """
        #print 'Unload',self.id,self.status,'at',self.sec.elem.id

        # pure info stuff
        #for item in self.items.values():
        #    if self.item_tasks.has_key(item.id):
        #        print '  Found item',item.id,'with task',\
        #              self.item_tasks[item.id].id
        #    else:
        #        print '  WARNING, Found item without task',item.id

        # make sure to be in loading status before unloading
        self.set_status('loading')
        #self.update_man()

        # current track element
        elem=self.sec.elem  

        unloaded=0
        if len(self.item_tasks.values())>0:
            # check all tasks if the associated item has arrived
            for task in self.item_tasks.values():
                # is it the right destination or no-destination ?
                # onload 'empty' item if search_item flag is set or
                # get also rid of everything if carrier is to be deleted.
                if (elem.id==task.dest_id) | (task.dest_id=='') | \
                   (task.mode=='search_item') | self.del_flag: 
                    if task.item_id!='empty': # is it not the empty item ?
                    
                        # is item physically existing in carrier ?
                        if self.items.has_key(task.item_id): 
                            item=self.items[task.item_id] # get item object

                            # transfer item to elem
                            #print '  start unload',item.id,' to',elem.id
                            #print self.id,'.unload to',elem.id
                            if item.start_unloading(elem):
                            #if  item.transfer(elem,self.get_transform()):
                                print 'WARNING from',self.id,'.unload:'
                                print '  unloading  of',\
                                      item.id,'unsuccessful'
                                
                                self.set_status('halt') # halt if problems

                            else:
                                unloaded=1
                                #print '  unload successfully',\
                                #      ': item',task.item_id,\
                                #      'of task',task.id
                     
                        else:
                            print 'ERROR in unload',self.id,':'
                            print '  item',task.item_id,\
                                  'of task',task.id,'not existent in carrier.'
                            self.set_status('halt') # halt if problems
                            # kill this task
                            task.kill()
                            
                    else:
                        #print '  found carrier is empty'
                        unloaded=1
                        self.load() # load immediately items
                        break

                else: # there is an item but it is not the right destination
                    #print '   restart carrier'
                    self.set_status('halt')
                    self.start('loaded')
        else:
            #print '   no task for this carrier'
            self.load() # load immediately items
        return unloaded

    #def unload_terminated(self,item):
    #    """
    #    this is called by the item when unload completed.
    #    Now it expects the transfer call which in turn will finally
    #    remove the item from the carrier.
    #    """
    #    item.transfer(elem,self.get_transform()):
        
    def exit(self,item):
        '''Item exits from carrier'''
        #print self.id,'.exit',item.id,'at',self.sec.elem.id,self.status
        del self.items[item.id]

        
                                    
        # terminate task, associated with this item
        self.item_tasks[item.id].unloaded()
        
        # change back to normal color if carrier is empty
        if self.obj.netmon:
            if len(self.items)==0:
                self.obj.netmon.itemconfigure(self.id,\
                                              fill=self.color_netmon)

        
        # no more items on carrier? 
        if len(self.items)==0:
            # load after last item unloaded'
            #self.load() # try if current track element has something to load

            # do load factor statistics here
            self.path_loaded+=self.get_path()-self.path_lastload
            self.path_lastload=self.get_path()
            self.time_loaded+=self.obj.sim.time-self.time_lastload
            self.time_lastload=self.obj.sim.time

        # try to get loaded on current
        # current track element
        self.sec.elem.terminate_unloading(self)
        
        
    def load(self,make_idle=True):
        '''Try to load an item from current track element'''

        #print 'car_load',self.id,'at',self.sec.elem.id
        # before loading new stuff, check whether carrier wants to be removed
        if self.del_flag:
            self.set_status('halt')
            return self.remove() # remove carrier

        else:
            # load new items
            #print self.id,'car.load',self.status,'at',self.sec.elem.id

                        
            # is there still a task with 'empty item' 
            for task in self.item_tasks.values():
                #print '   terminate task',task.id,task.order_id,task.item_id
                if task.item_id=='empty':
                    # terminate this task by indicating that an item
                    # has been found
                    task.found_item()
                    
            if len(self.items)>0:
                print 'WARNING: in',self.id,'.load:'
                print '  there is item already loaded!!!',self.items
                self.sec.elem.show_info()
                # stop simulation
                self.obj.sim.time_comp=0

            #print '  try to load from',self.sec.elem.id,self.sec.id
            item=self.sec.elem.load_item(self)
            if item:
                #print '  loaded',item.id
                #print '  car.load:',self.id,' loaded',item.id,\
                # 'wait for calling enter'
                self.set_status('loading')

                
                # inform carrier management  about the loaded item
                self.send_mes({'to':self.man_id,
                               'op':'load_car',
                               'item_id':item.id,
                               })
                # This will cause the carrier management to schedule this
                # carrier, to inform the item's management and
                # finally to order the involved track managements to
                # guide the carrier to the item's destination.
                return 0
            elif make_idle:
                #print '  nothing to load => go idle'
                return self.idle()
            
    def idle(self):
        """
        This method if called when the carrier is at a berth and
        was unable to load an item.
        In this case the carrier is calling directly the idle car function
        of its management in order to generate a task for this carrier.
        """
        
        if self.sec.type=='berth':
            #print self.id,':nothing to load,  tell man that carrier id idle'
            self.set_status('halt')    
            #print '....go idle on',self.sec.type
            self.send_mes({'to':self.man_id,
                           'op':'idle_car',
                           'status':self.status,
                           'elem_id':self.sec.elem.id,
                           'sec_id':self.sec.id,
                           })
                
            
            return 0
        else:
            self.set_status('halt')
            return 1
    
    def enter(self,item):
        '''
        New item has entered the carrier.
        '''
        #print self.id,'.enter item',item.id,'at',self.sec.elem.id,self.status
        #print '  with dest',item.dest_id,self.sim.time

        # has carrier been completely empty before? 
        if len(self.items)>0:
            print 'WARNING: in',self.id,'.enter:'
            print '  there is item already loaded!!!',self.items
            self.sec.elem.show_info()
            # stop simulation!!!
            self.obj.sim.time_comp=0

        # put items into current item list
        self.items[item.id]=item

        
        if len(self.items)==1:
            # do load factor statistics here
            self.time_lastload=self.obj.sim.time
            self.path_lastload=self.get_path()

        # here we stop loading after one item has been loaded
        # this must be subject to changes when several items are loaded
        self.set_status('halt')
        
        #print 'item_tasks while enter',self.item_tasks.keys()
        if self.item_tasks.has_key(item.id):
            if self.item_tasks[item.id].status=='processing':
                # this call will finally start the carrier
                # via the task of the corresponding item 
                self.item_tasks[item.id].start_car()
            else:
                print 'WARNING: task',self.item_tasks[item.id].id,\
                      'in',self.id
                print '   exist but is not processing' 
        else:
            print 'WARNING: from',self.id
            print '  no task for this item, waiting...'
            
        # change color of carrier
        if self.obj.netmon:
            self.obj.netmon.itemconfigure(self.id,\
                                          fill=self.color_loaded_netmon)
        
        
        
        
    def add_man(self):
        '''
        Function to add this carrier to the carrier management database.
        This Function is called already during initialization or changes of
        the carrier management.
        '''
        self.send_mes({'to':self.man_id,
                       'op':'register',
                       'status':self.status})

    def del_man(self):
        '''
        Function to remove this carrier from the carrier management database.
        This Function is called when carrier is going to be deleted
        from the network or changes of the carrier management.
        '''
        self.send_mes({'to':self.man_id,
                       'op':'unregister'})
        
    def update_man(self):
        '''
        This function is called during activation or after changes
        of destinations.
        '''

        # composing the carrier data that needs to be send to the car man
        # For safety reasons we do not send objects (or pointers in general),
        # just strings, numbers or ids (also strings).

        # ids of current section and track element
        sec_id=''
        elem_id=''
        if self.sec: # is current section already available ?
            sec_id=self.sec.id
            if self.sec.elem: # is current track element  already available ?
                elem_id=self.sec.elem.id

        #item_ids=[]
        #for item in self.items: item_ids.append(item.id)
        
        # send status massage to carrier management
        self.send_mes({'to':self.man_id,
                       'op':'update_car',
                       'status':self.status,
                       'elem_id':elem_id,
                       'sec_id':sec_id,
                       #'dest_id':self.dest_id,
                       #'item_ids':item_ids,
                       })
        

    def make_panel(self,panel,**kw):
        """
        Make widgets.
        May be overwritten by specific class.
        """
        # initialize widgets with Non objects such that they are not updated
        # during the make phase.
        
        panel.make_standardpanel()

        # add a load page
        panel.make_page('Load',self.make_loadpage,self.update_loadpage)

        # add a parameters page
        panel.make_page('Parameters',self.make_parampage,self.update_parampage)
        
        # add carrier specific panel infos
        self.make_panel_specific(panel)

    def make_panel_specific(self,panel):
        """
        Add track element specific panel infos.
        To be overwritten.
        """
        pass
    
    def make_loadpage(self,parent,widgets):
        """
        Create panel page with info about loaded items on carrier.
        """

        sel_command=lambda sel_ind=-1,self=self,widgets=widgets:\
                     self._on_panel_load(sel_ind,widgets)
            
        widgets['loadtable']=gui.PanedScrolledTable(parent,
                                                    selectcommand=
                                                                  sel_command,
                                                    #highlightcommand=
                                                    #              sel_command,
                                                    #hullwidth=450,
                                                    #hullheight=150
                                                    )

        
        # fill tables etc. with data
        self.update_loadpage(parent,widgets)

        # pack everything in the rihgt order
        widgets['loadtable'].pack(expand=1,fill='both')
        
   
    def _on_panel_load(self,sel_ind,widgets):
        """
        Open new or update control with selected load.
        """
        item_id=widgets['loadtable'].get_selection(0)
        if item_id!='EMPTY':
            # get module pointer to loaded item
            item=self.obj.sim.split_id(item_id)['mod']
            widgets['controlpanel'].open_controlpanel(item)
            
    def update_loadpage(self,parent,widgets):
        """
        Updates all widgets on load page.
        """
        # self.print_state() 
        header=[{'key':'Item ID','type':'text','width':5},
                {'key':'Dest Name','type':'text','width':5},
                {'key':'Task ID','type':'text','width':2},
                {'key':'Mode','type':'text','width':2},]

        table=[]
        for task in self.tasks.values():
            row={'Item ID':task.item_id}
            row['Task ID']=task.id
            row['Dest Name']=\
                      self.obj.sim.split_id(task.dest_id)['name']
            row['Mode']=task.mode
            table.append(row)
            
            
        widgets['loadtable'].make_table(header,table)

    def make_parampage(self,parent,widgets):
        descriptors,dataset=self.format_params()
        widgets['param_data']=gui.DatasetEditor(descriptors,dataset,
                                               parent,
                                               applyentry_command=\
                                                   self.apply_param_gui,
                                               applybutton_command=\
                                                   self.apply_param_gui)
        widgets['param_data'].pack(expand=1,fill='both')
        
    def update_parampage(self,panel,widgets):
        """
        Update of widgets on costs page. This method is called
        when cost-page needs to be updated.

        """
        descriptor,dataset=self.format_params()
        widgets['param_data'].update_contents(dataset)

    
    def apply_param_gui(self,paramdatawidget,param_data):
        self.set_params(param_data)
            
        descriptor,dataset=self.format_params()
        paramdatawidget.update_contents(dataset)

    def set_params(self,params={}):
        """
        Set parameter data of module.
        """
        #print self.id,'set_params'

        # basic parameters
        self.length=params.get('length',self.length)
        self.width=params.get('width',self.width)

        self.time_brake=params.get('time_brake',self.time_brake)

        self.accel_comfort=params.get('accel_comfort',self.accel_comfort)
        self.accel_emergency=params.get('accel_emergency',self.accel_emergency)
        # set accel_failure to 100000.0 means brick-wall criteria
        self.accel_failure=params.get('accel_failure',self.accel_failure)

        if self.ind>=0:
            #print '  set essential parameters for global network simulator'
            self.obj.netsim.set_car_params(self)

        
        # appearance on the Netmon 2D visulalizer
        self.length_netmon=0.8*self.length
        self.width_netmon=2.0
        self.color_netmon='#ffd700' #gold1
        self.color_loaded_netmon='#ff7f00' #DarkOrange1 #'#ffa500' # orange1
        
        
        
    def format_params(self):
        """
        Returns describtor and dataset of parameters.
        """
        descriptors=[
            {'key':'length',
             'name':'Length',
             'type':'real',
             'printformat':'%4.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
             'units':{'m':1}
            },
            {'key':'accel_comfort',
             'name':'Max. abs. comfort acceleration',
             'type':'real',
             'printformat':'%4.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
             'units':{'m/s^2':1}
            },
            {'key':'accel_emergency',
             'name':'Max. abs. emergency acceleration',
             'type':'real',
             'printformat':'%4.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
             'units':{'m/s^2':1}
            },
            {'key':'accel_failure',
             'name':'Max. abs. failure acceleration',
             'type':'real',
             'printformat':'%4.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
             'units':{'m/s^2':1}
            },
            {'key':'time_brake',
             'name':'Max. brake actuation time',
             'type':'real',
             'printformat':'%10.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
             'units':{'s':1}
            },
            {'key':'velocity_opt',
             'name':'Optimal velocity',
             'type':'real',
             'printformat':'%.2f',
             'valid':{'min':0,'max':10**9},
             'mode':'read',
             'width':0.4,
             'units':{'m/s':1},
            },
            {'key':'headway_opt',
             'name':'Optimal headway',
             'type':'real',
             'printformat':'%.2f',
             'valid':{'min':0,'max':10**9},
             'mode':'read',
             'width':0.4,
             'units':{'m':1},
            },
            {'key':'timeheadway_opt',
             'name':'Optimal time-headway',
             'type':'real',
             'printformat':'%.2f',
             'valid':{'min':0,'max':10**9},
             'mode':'read',
             'width':0.4,
             'units':{'s':1},
            },
            {'key':'capacity_max',
             'name':'Max. line capacity',
             'type':'real',
             'printformat':'%4.2f',
             'valid':{'min':0,'max':10**9},
             'mode':'read',
             'width':0.4,
             'units':{'veh/h':1},
            },
            
            ]

        dataset={'length':self.length,
                 'accel_comfort':self.accel_comfort,
                 'accel_emergency':self.accel_emergency,
                 'accel_failure':self.accel_failure,
                 'time_brake':self.time_brake,
                 'velocity_opt':self.get_velocity_opt(),
                 }
        
        dataset['capacity_max']=self.get_capacity(dataset['velocity_opt'])*3600
        dataset['headway_opt']=self.get_headway(dataset['velocity_opt'])
        dataset['timeheadway_opt']=self.get_timeheadway(\
            dataset['velocity_opt'])

        return descriptors,dataset

    def get_velocity_opt(self):
        """
        Returns the velocity which results in a maximum capacity.
        if emergency accle equals failure accel then the optimum velocity
        is infinite. In this case the function returns the speed of light.
        """
        if self.accel_emergency<self.accel_failure:
            return sqrt(2*self.length/(1/self.accel_emergency\
                                       -1/self.accel_failure))
        else:
            return 299792458.0 # speed of light in m/s

    def get_capacity(self,v):
        """
        Returns line capacity of this carrier, dependent on its velocity.
        """
        if v>0:
            return 1/(self.length/v+self.time_brake+ \
                      v/2*(1/self.accel_emergency-1/self.accel_failure) )
        else:
            #if self.time_brake>0:
            #    return 1/(self.time_brake+ \
            #              v/2*(1/self.accel_emergency-1/self.accel_failure) )
            #else:
            return 0.0

    def get_headway(self,v):
        """
        Returns headway of this carrier, dependent on its velocity.
        """
        return self.length+self.time_brake*v+ \
                   v**2/2*(1/self.accel_emergency-1/self.accel_failure) 
    
    def get_timeheadway(self,v):
        """
        Returns timeheadway of this carrier, dependent on its velocity.
        """
        c=self.get_capacity(v)
        if c==0.0:
            return  self.length/299792458.0
        else:
            return 1/self.get_capacity(v)
            
    #def make_capacitywidgets(self,parent=None,widgets={}):
    #    """
    #    Adding widgets to display and recalc the capacity, headway of
    #    the carrier.
    #    """
    #    widgets['']=Tkinter.Label(parent,
    #                                   text='Add multiple carriers to'+\
    #                                   ' the availlable berths')
    #    
    #    Tkinter.Button(parent,
    #                   text=button_data['text'],
    #                   command=botton_command,
    #                   takefocus=1,
    #                   )
    
    def reset_statistics_mod(self):
        """
        Reset or initialize statistical variables of a general carrier module.
        """
        if self.ind>=0:
            self.path_init=self.get_path()
        else:
            self.path_init=0.0

        # current simulation time is time when statistic starts
        self.time_init=self.obj.sim.time
       
        
        self.time_inmotion=0.0
        self.time_laststart=0.0

        self.path_loaded=0.0
        self.path_lastload=self.get_path()
        
        self.time_loaded=0.0
        self.time_lastload=0.0
        
        #self.reset_statistics_mod_specific()

   

    def format_statistics(self):
        """
        Return descriptor and dataset contents for statistical data
        about the target object.
        """
        #print 'format_statistics for',self.id,self.ind
        #print '  get_path=',self.get_path()
            
        descriptor=[{'key':'Status',
                     'type':'text',
                     'width':1,
                     'mode':'read'
                     },
                    {'key':'Total path',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'km':1},
                     'valid':{'max':10**8},
                     'printformat':'%.1f',
                     },
                    {'key':'Path with load',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'km':1},
                     'valid':{'max':10**8},
                     'printformat':'%.1f',
                     },
                    {'key':'Average velocity',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'m/s':1},
                     'valid':{'max':10**4},
                     'printformat':'%.1f',
                     },
                    {'key':'Average velocity with load',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'m/s':1},
                     'valid':{'max':10**4},
                     'printformat':'%.1f',
                     },
                    {'key':'Path load factor',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'%':1},
                     'valid':{'max':10**2},
                     'printformat':'%.1f',
                     },
                    {'key':'Time load factor',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'%':1},
                     'valid':{'max':10**2},
                     'printformat':'%.1f',
                     },
                    {'key':'elem_id',
                     'name':'Currently on track element',
                     'type':'text',
                     'width':1,
                     'mode':'read',
                     },
                    {'key':'velocity_current',
                     'name':'Current velocity',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'m/s':1},
                     'valid':{'max':10**4},
                     'printformat':'%.2f',
                     },
                    {'key':'timeheadway_min',
                     'name':'Currnet min. time-headway',
                     'type':'real',
                     'printformat':'%4.2f',
                     'valid':{'min':0,'max':10**9},
                     'mode':'read',
                     'width':0.4,
                     'units':{'s':1},
                      },
                    {'key':'capacity_current',
                     'name':'Current line capacity',
                     'type':'real',
                     'printformat':'%4.2f',
                     'valid':{'min':0,'max':10**9},
                     'mode':'read',
                     'width':0.4,
                     'units':{'veh/h':1},
                     },
                    {'key':'id_carinfront',
                     'name':'Current carrier in front',
                     'type':'text',
                     'width':1,
                     'mode':'read',
                     },
                    {'key':'dist_current',
                     'name':'Distance to carrier in front',
                     'type':'real',
                     'width':1,
                     'mode':'read',
                     'units':{'m':1},
                     'valid':{'max':10**8},
                     'printformat':'%.2f',
                     },
                    
                    #{'key':'sec_id',
                    # 'name':'Section ID',
                    # 'type':'text',
                    # 'width':1,
                    # 'mode':'read',
                    # },
                    ]

        # compute important quantities
        if (self.obj.sim.time-self.time_init)>0:
            velocity_average=(self.get_path()-self.path_init)/  \
                              (self.obj.sim.time-self.time_init)
        else:
            velocity_average=0.0
            
        if self.time_inmotion>0.0:
            velocity_average_inmotion=(self.get_path()-self.path_init)/  \
                                       (self.time_inmotion)
        else:
            velocity_average_inmotion=0.0

        #print '  sim.time',self.obj.sim.time
        #print '  time_init',self.time_init
        #print '  time_loaded',self.time_loaded
        
        if self.time_loaded>0.0:
            velocity_average_loaded=self.path_loaded/self.time_loaded
        else:
            velocity_average_loaded=0.0

        if (self.get_path()-self.path_init)>0.0:
            load_factor_path=self.path_loaded/(self.get_path()-self.path_init)
        else:
            load_factor_path=0.0
                    
        if (self.obj.sim.time-self.time_init)>0.0:
            load_factor_temp=(self.time_loaded)/  \
                              (self.obj.sim.time-self.time_init)
                              
        else:
            load_factor_temp=0.0

       
            
        dataset={'Status':self.status,
                 'Total path':
                     (self.get_path()-self.path_init)/1000,
                 'Path with load':self.path_loaded/1000,
                 'Average velocity':velocity_average,
                 'Average velocity with load':
                    velocity_average_loaded,
                 'Path load factor':load_factor_path*100,
                 'Time load factor':load_factor_temp*100,
                 
                  }
        
        if self.sec:
            #dataset['sec_id']=self.sec.id
            dataset['elem_id']=self.sec.elem.id
        else:
            #dataset['sec_id']='None'
            dataset['elem_id']='None'

        if self.ind>=0:
            dataset['velocity_current']=self.get_velocity()
            dataset['timeheadway_min']=self.get_timeheadway(\
                dataset['velocity_current'])
            dataset['capacity_current']=self.get_capacity(\
                dataset['velocity_current'])*3600
        else:
            dataset['velocity_current']=0.0
            dataset['timeheadway_min']=0.0
            dataset['capacity_current']=0.0

        if self.car_infront:
            dataset['id_carinfront']=self.car_infront.id
            dataset['dist_current']=self.obj.dist[self.no][0]
        else:
            dataset['id_carinfront']='None'
            dataset['dist_current']=0.0
        return descriptor,dataset
        
    def control(self,controldata={}):
        '''
        Only function to control behavior of this carrier. This funtion is
        called at discrete events i.e. while entering/leaving a section,
        when leader/ghost changes, or when the array index of leader/ghost
        needs to be updated (call with no argument)
        '''
        #if self.id=='test.car.experim/81':
        #    print '**control:',self.id#,\
        #          #', in sec.id:',self.sec.id,\
        #          #'at',self.get_path_sec()
        #    for key in controldata.keys():
        #        print '  ',key,':',controldata[key]

        if controldata.has_key('decouple'):
            if controldata['decouple']==self.car_infront:
                self.obj.control_mod(self,{'car_infront':None})
                self.car_infront=None
            elif controldata['decouple']==self.ghost_infront:
                self.obj.control_mod(self,{'ghost_infront':None})
                self.ghost_infront=None
            self.obj.control_mod(self,controldata)
            return
        
        

        elif controldata.has_key('start'):
            #print 'control',self.id
            self.set_status('active')

            # do some  statistics
            self.time_laststart=self.obj.sim.time
                
        elif controldata.has_key('stop'):
            self.set_status('halt')
            
            # do some  statistics
            self.time_inmotion+=self.obj.sim.time-self.time_laststart
            self.time_laststart=self.obj.sim.time
            
        # do not change velocity if carrier is not active
        if (controldata.has_key('velocity_line'))&(self.status!='active'):
            del controldata['velocity_line']

        #if self.status!='active':
        #        self.control({'stop':'now'})
        if 1:        
            if controldata.has_key('car_infront'):
                self.car_infront=controldata['car_infront']

            if controldata.has_key('ghost_infront'):
                self.ghost_infront=controldata['ghost_infront']

            # do the settings of the carrier arrays
            # in the carrier simulation object.
            # Empty controldata means index refresh
            self.obj.control_mod(self,controldata)

        else:
            print 'WARNING in',self.id,'.control:'
            print '  could not apply controldata to carrier in status',\
                  self.status
            for key in controldata.keys(): print '   controldata key:',key
            # return 1 
        #self.print_state(5)

    def get_transform(self,trans_rel=identity(4)):
        """
        Get current transformation matrix of carrier.

        With the transformation trans_rel, one can give a exact position
        on the carrier, relative to the middle of the back end of the carrier.
        
        """
        if self.status!='edit':
            
            carseg_ind=self.obj.netsim.segs_ind.arrays[self.ind]+          \
                        int(self.obj.netsim.path_sec[self.ind]/          \
                         self.obj.netsim.length.arrays[self.ind] *           \
                         self.obj.netsim.segs_total.arrays[self.ind])

            cartrans=self.obj.netsim.segs.take([carseg_ind])[0]
            path_seg=self.obj.netsim.path_sec[self.ind] % \
                      self.obj.netsim.seg_length.arrays[self.ind]
            #print 'cartrans=',cartrans
            #print 'path_seg=',path_seg
            return matrixmultiply(matrixmultiply(\
                cartrans, \
                array([[1,0,0,path_seg-self.length],
                       [0,1,0,0],
                       [0,0,1,0],
                       [0,0,0,1]])), trans_rel) 
            
           
        else:
            return matrixmultiply(\
                array([[1,0,0,self.pos[0]-self.length],
                       [0,1,0,self.pos[1]],
                       [0,0,1,self.pos[2]],
                       [0,0,0,1]]),trans_rel) 
        
    def get_position(self,pos_rel=[0,0]):
        """
        Return list with 2D position of the carrier

        With the vector pos_rel=[dx,dy]  one can give the exact
        position on the carrier, relative to the middle of the back end of
        the carrier.
        """
        trans=self.get_transform(array([[1,0,0,pos_rel[0]],
                                        [0,1,0,pos_rel[1]],
                                        [0,0,1,0],
                                        [0,0,0,1]]))
                                 
        return [trans[0,3],trans[1,3]]

    def get_path(self):
        """
        Return total path that this carrier made
        since its initialization.
        """
        if self.ind>=0:
            return self.obj.netsim.path.arrays[self.ind]
        else:
            return 0.0

    def get_path_sec(self):
        """
        Return length of the path that this carrier made on the
        current section.
        """
        if self.ind>-1: # is car in netsim
            return self.obj.netsim.path_sec[self.ind]
        else:
            # return negatuve, impossible value if car is not in netsim
            return -1.0

    def get_velocity(self):
        """
        Return velocity of this carrier.
        """
        return self.obj.netsim.velocity.arrays[self.ind]
    
    def get_velocity_line(self):
        """
        Return desired line velocity of this carrier.
        """
        return self.obj.velocity_line[self.no]
    
    def get_brakedist_comfort(self):
        """
        Return braking distance with comfort decelleration of this carrier.
        """
        return self.obj.netsim.brakedists[self.ind,0]

    def get_brakedist_emergency(self):
        """
        Return braking distance with emergency decelleration of this carrier.
        """
        return self.obj.netsim.brakedists[self.ind,1]

    def get_brakedist_failure(self):
        """
        Return braking distance with failure decelleration of this carrier.
        """
        return self.obj.netsim.brakedists[self.ind,2]

    def get_dist_crit(self,leader):
        """
        Get critical distance to a lead carrier, in emergency case.
        """
        #print 'check for braking distance'
        # theory 22.03.03
        #v_leader=leader.get_velocity()
        #if v_leader!=0.0:
        #    d_crit=(abs(self.get_brakedist_comfort()\
        #                -leader.get_brakedist_comfort())\
        #            -2.0*leader.get_brakedist_comfort()/v_leader\
        #            *abs(self.get_velocity()-v_leader)\
        #            +leader.get_brakedist_emergency()\
        #            -leader.get_brakedist_failure())
        #else:
        #    d_crit=(abs(self.get_brakedist_comfort()\
        #                -leader.get_brakedist_comfort())\
        #            +leader.get_brakedist_emergency()\
        #            -leader.get_brakedist_failure())
             
        #return d_crit
        return self.time_brake*self.get_velocity()+\
               self.get_brakedist_emergency()-leader.get_brakedist_failure()

    def get_dist_min(self,leader):
        """
        Get minimum distance to a lead carrier, the one that must always
        be kept also in normal operation.
        """

        return self.time_brake*self.get_velocity()+\
               max(self.get_brakedist_emergency()\
                   -leader.get_brakedist_failure(),\
                   self.get_brakedist_comfort()-leader.get_brakedist_comfort())
    
               
    

    def print_state(self,debuglevel=0):
        self.show_info(debuglevel)
        
    def show_info(self,debuglevel=0):
        if debuglevel<=self.debug:
            print '---------------------------------------------------------'
            print 'State of',self.id,' status:',self.status,'\n'

            #print 'self.mes_ops:'
            #for key in self.mes_ops.keys(): print '  ',key
            
            if len(self.item_tasks.values())>0:
                print 'Load:'
                for task in self.item_tasks.values():
                    if task.item_id=='empty':
                        print '  Empty'
                    else:
                        if self.items.has_key(task.item_id):
                            item=self.items[task.item_id]
                            print ' ',item.id
                        else:
                            print '  loading',task.item_id
                            
                    print '    Task',task.id,\
                          'from',task.client_id,'order',task.order_id
                    if task.mode=='search_item':
                        print '    Destination: Search item to',task.dest_id
                    else:
                        print '    Destination:',task.dest_id


            
            if self.status=='halt':
                print 'On track elem:',self.sec.elem.id
                print '  Sec:',self.sec.id,self.sec.label,'self.segs_ind:',\
                      self.obj.netsim.segs_ind.arrays[self.ind]
                print 'path_sec=',self.get_path_sec()
                print 'velocity=',self.obj.state.arrays[self.modind]
                print 'velocity_line=',\
                      self.obj.velocity_line.arrays[self.modind]
                if self.obj.point_marked.db.has_key(self.no):
                    print 'marked at',self.obj.point_marked[self.no]
                if self.obj.point_stop.db.has_key(self.no):
                    print 'stop at',self.obj.point_stop[self.no]
                    
                #print 'ind:',self.ind
                #print 'Total path= ',self.obj.netsim.path.arrays[self.ind]
                #print 'path in segment=',self.obj.netsim.path_seg[self.ind]
                #print 'offset=',self.obj.netsim.offset[self.ind]
                #print 'mult=',self.obj.netsim.mult[self.ind]
            if self.status=='active':
                print 'On track elem:',self.sec.elem.id
                print '  Sec:',self.sec.id,self.sec.label,'self.segs_ind:',\
                      self.obj.netsim.segs_ind.arrays[self.ind],\
                      'path_sec=',self.get_path_sec()
                print 'modind=',self.modind
                print 'velocity=',self.obj.state.arrays[self.modind]
                print 'velocity_line=',\
                      self.obj.velocity_line.arrays[self.modind]
                
                print 'brakefor_car_infront=',\
                      self.obj.brakefor_car_infront.arrays[self.modind]
                print 'brakefor_ghost_infront=',\
                      self.obj.brakefor_ghost_infront.arrays[self.modind]
                
                print 'brakedist_comfort=',self.get_brakedist_comfort()
                print 'brakedist_emergency=',self.get_brakedist_emergency()
                print 'brakedist_failure=',self.get_brakedist_failure()

                if self.obj.point_marked.db.has_key(self.no):
                    print 'marked at',self.obj.point_marked[self.no]
                if self.obj.point_stop.db.has_key(self.no):
                    print 'stop at',self.obj.point_stop[self.no]
                    
            if self.car_infront:
                print 'car_infront:',self.car_infront.id,\
                      'modind:',self.car_infront.modind,\
                      'ind:',self.car_infront.ind
                print '  dist=', self.obj.dist[self.no]
                dist_crit=self.get_dist_crit(self.car_infront)
                
                print '  dist_crit=',dist_crit,dist_crit>self.obj.dist[self.no]

                dist_approach=self.get_brakedist_emergency()-\
                      self.car_infront.get_brakedist_failure()+\
                      2.0*0.5/2.0*(self.get_velocity()-\
                                   self.car_infront.get_velocity())**2
                print '  dist_approach=',dist_approach,dist_approach>\
                      self.obj.dist[self.no]
                print '  brakefor_car_infront=',\
                      self.obj.brakefor_car_infront[self.no]
                
                
            if self.ghost_infront:
                print 'ghost_infront:',self.ghost_infront.id,\
                      'modind:',self.ghost_infront.modind,\
                      'ind:',self.ghost_infront.ind
                print '  ghostdist=', self.obj.ghostdist[self.no]
                dist_crit=self.get_dist_crit(self.ghost_infront)
                print '  dist_crit=',dist_crit,\
                      dist_crit>self.obj.ghostdist[self.no]
                print '  brakefor_ghost_infront=',\
                      self.obj.brakefor_ghost_infront[self.no]
                print '  dist_merge',self.obj.dist_merge[self.no]
                
      

    
    def set_status(self,newstatus):
        """
        Central function to change status of carrier$
        """


        #print 'set_status:',self.status,'of carrier',\
        #      self.id,'to',newstatus

        
        
        if (self.status=='edit') & (newstatus=='halt'):
            # register in carrier objects array struct
            # self.obj.add_mod_arrays(self,{'velocity_line':0.0})
            self.obj.add_mod_arrays(self)
            self.status='halt'
            self.add_man() # add carrier to carrier management
            return 0
        
        elif (self.status=='halt') &\
             ((newstatus=='active')|(newstatus=='loading')):
            self.status=newstatus
            self.update_man() # update carrier management
            return 0

        elif (self.status=='loading') & (newstatus=='halt'):
            self.status='halt'
            self.update_man() # update carrier management
            return 0
        
        elif (self.status=='active') &\
             ((newstatus=='halt')|\
              (newstatus=='loading')|\
              (newstatus=='edit')):
            self.status=newstatus
            #if newstatus=='halt':
            #    self.dest_id='' # in halt status, there is no destination
            self.update_man() # update carrier management
            return 0
        
        elif (self.status=='halt') & (newstatus=='edit'):
            # unregister from carrier object array struct
            self.obj.del_mod_arrays(self)
            self.status='edit'
            self.del_man() # delete carrier from carrier management
            return 0

        elif self.status==newstatus: # same status as before, do nothing
            return 0

        else:
            print 'WARNING in',self.id,'.set_status:'
            print '  could not set',self.status,'to',newstatus
            return 1  # status could not be changed
        
    def place(self,elem):
        '''
        Place carrier on track element. The carrier is then registered into
        Netsims arrays by the respective track element and also
        into the carrier sim objects arrays.
        This is the only function that can change the carrier status from
        edit to halt.
        '''

        # can carrier be placed on this element ?
        if( hasattr(elem,'add_car'))&(self.status=='edit'): 
            # no errors with placing ?
            # Here carrier registers also into Netsim's arrays
            ans=elem.add_car(self)
            if ans=='done':
                # change to halt status
                # Here, carrier registeres into carrier sim obj arrays
                self.set_status('halt') 
                self.start('no_task')
                return 0
            
            elif ans=='offtrack': # carrier has been placed beside the track
                # do nothing and wait till it is moved on the track
                return 0
            else:
                return 1
        else:
            return 1
        
    def delete(self):
        """
        Initiate delete operation.
        It sets a delete flag to indicate that this carrier should be removed
        after it is halted and empty.
        """
        self.del_flag=1
        
        # remove immediately if already in edit mode
        #if self.status=='edit':
        return self.remove()
        
    def remove(self):
        '''
        Removal from track element and Netsim and Carrier sim objects arrays.
        It is the only function that can change the carrier status from halt
        to edit.
        Remove is the opposite action of place.
        '''
        #print 'remove',self.id

        #if self.status=='active':

        # try to put carrier into halt mode. only if this is possible
        # the carrie can be deleted
        # do this...
        #self.set_status('halt')
                
        #if self.status=='halt':

        if hasattr(self.sec.elem,'del_car'):
            result=self.sec.elem.del_car(self)
            #print '  ',result
            if result=='deleted':
                self.del_flag=0
                self.set_status('edit')
                # if there is the 2D monitor , delete it from there as well 
                if self.obj.netmon: self.del_plot()
                #else: print 'WARNING car.remove: no netmon'
                
                self.obj.del_mod(self) # delete also form sim obj database
                return 0
            
            elif result=='posponed':
                if self.debug>=3:
                    print self.id,'will be deleted as soon as possible!'
                return 0
            else:
                self.del_flag=0
                if self.debug>=3:
                    print 'ERROR in',self.id,':'
                    print '  ',result
                return 1
        else: 
            if self.debug>=3:
                print self.id,'cannot be deleted here but as soon as possible!'
            return 0
        
        #print self.id,'in status',self.status
        #print 'segs_ind=',self.obj.netsim.segs_ind.arrays
        #print 'path_sec=',self.obj.netsim.path_sec
        #print 'length=',self.obj.netsim.length.arrays
        #print 'segs_total=',self.obj.netsim.segs_total.arrays
        #print 'length.db=',self.obj.netsim.length.db
        #print 'segs_ind.db=',self.obj.netsim.segs_ind.db
        #print 'segs_total.db=',self.obj.netsim.segs_total.db
        #if self.status=='edit':
            
        

    def plot(self,canvas=None):
        #print 'plot',self.id
        if not canvas:
            canvas=self.obj.netmon # =None if netmon not invoked
            add_cartag_flag=1
        else:
            add_cartag_flag=0

        if canvas:
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                canvas.delete(self.id)
            
        

            x_back,y_back=self.get_position([0,0])
            x_nose,y_nose=self.get_position([self.length_netmon,0])

            #print '  plot coords',x_back,y_back,x_nose,y_nose,\
            #      self.length_netmon
            if len(self.items)>0:
                color=self.color_loaded_netmon
            else:
                color=self.color_netmon

            # set section width
            width=self.width_netmon
        
            if (canvas==self.obj.netmon):
                width_scaled=width*self.obj.netmon.canvas_scale
            else:
                width_scaled=width


            canvas.create_line([x_back,y_back,x_nose,y_nose],
                               #capstyle='round',
                               fill=color,
                               width=width_scaled,
                               tag=self.id)

            if canvas==self.obj.netmon:
                #'add section to canvas objects with width=2 group'
                # this means that width will be adjusted dynamically
                # when rescaling, using a fast tk aglorithm
                self.obj.netmon.addtag_withtag('width=%.1f'%(width),
                                               self.id)
        
            self.plot_extra(canvas)

        
            if add_cartag_flag&(self.status!='edit'):
                # all carriers must get the tag car in netmon
                self.obj.netmon.addtag_withtag('car',self.id)
            

    def plot_extra(self,canvas):
        """Extra plots, Carrier specific"""
        # overwrite, in specialized carrier class
        pass
    
    def del_plot(self,canvas=None):
        if not(canvas): canvas=self.obj.netmon# =None if netmon not invoked
        if canvas:
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                canvas.delete(self.id)
        

class Car(Obj):

    def init_obj(self):
        # minimum approach distance between 2 carriers
        self.dist_min=0.3

        # needs to be defined here because needed in set_state
        self.netsim=None

        self.netmon=None
        
        # default management simulation object
        self.man_id='test.carman.experim'
        
        # Message opeartors for modules
        # This is necessary to initialize it here because pointers to
        # functions cannot be pickeled...and modules are safed as a whole.
        # self.modmes_ops={'run':self.run}

                
        # arrays
        #self.modind=Array_db() # index in carrier array struct; with id as key
        self.state=Array_db() # here just velocity
        self.velocity_line=Array_db()

        #self.velocity_desired=Array_db()
        self.brakefor_car_infront=Array_db()
        self.brakefor_ghost_infront=Array_db()
        self.brakefor_stop=Array_db()
        self.velocity_inc=Array_db()
        self.velocity_dec=Array_db()

        

        self.car_infront_ind=Array_db()
        self.modinds_hascar_infront=Array_db()
        self.inds_hascar_infront=Array_db()
        self.dist=Array_db()

        # the distance of the car's current section in which the alignment
        # must be completed during merge. Usually section length-conflict zone
        self.dist_merge=Array_db() 
        
        #self.velocity_line_hascar=Array_db()

        self.ghost_infront_ind=Array_db()
        self.modinds_hasghost_infront=Array_db()
        self.inds_hasghost_infront=Array_db()
        self.ghostdist=Array_db()
        #self.velocity_line_hasghost=Array_db()

        self.modinds_hasstop=Array_db()
        self.inds_hasstop=Array_db()
        self.point_stop=Array_db()
        
        self.inds_marked=Array_db()
        self.modnos_marked=Array_db()

        self.point_marked=Array_db()
        self.pos=array([])
        
        # type of carrier module (Mod is the standard one)
        self.Mod=Mod

        # initialize if simulation in gui mode
        if self.sim.mode=='gui':
            self.multiadd_panel=None
        
        # define initialy a None-module (here single carrier) to be edited 
        self.editmod=None
        self.init_type()

    def set_state(self,state_image):
        self.mods=state_image['mods']
        self.mod_no_current=state_image['mod_no_current']
        self.state=state_image['state']

        self.velocity_inc=state_image['velocity_inc']
        self.velocity_dec=state_image['velocity_dec']

        self.brakefor_car_infront=state_image['brakefor_car_infront']
        self.brakefor_ghost_infront=state_image['brakefor_ghost_infront']
        self.brakefor_stop=state_image.get('brakefor_stop')
        #if len(self.brakefor_car_infront.arrays)>\
        #   len(self.brakefor_stop.arrays):
        #    for mod_no in self.state.db.keys():
        #        self.brakefor_stop[mod_no]=0
        
        self.velocity_line=state_image['velocity_line']

        self.car_infront_ind=state_image['car_infront_ind']
        self.modinds_hascar_infront=state_image['modinds_hascar_infront']
        self.inds_hascar_infront=state_image['inds_hascar_infront']
        self.dist=state_image['dist']
        self.dist_merge=state_image['dist_merge']
        self.ghost_infront_ind=state_image['ghost_infront_ind']
        self.modinds_hasghost_infront=state_image['modinds_hasghost_infront']
        self.inds_hasghost_infront=state_image['inds_hasghost_infront']
        self.ghostdist=state_image['ghostdist']

        #
        self.modinds_hasstop=state_image.get('modinds_hasstop')
        self.inds_hasstop=state_image.get('inds_hasstop')
        self.point_stop=state_image.get('point_stop')
        #print 'set_state',self.type
        #print '  self.ghost_infront_ind',self.ghost_infront_ind.arrays

        #print '  self.inds_hasghost_infront',self.inds_hasghost_infront.arrays
        #print '  self.brakefor_car_infront',self.brakefor_car_infront.arrays
        #print '  self.brakefor_stop.arrays',self.brakefor_stop.arrays
        #print '  self.modinds_hasstop',self.modinds_hasstop
        #print '  self.inds_hasstop',self.inds_hasstop
        #print '  self.point_stop',self.point_stop
        self.inds_marked=state_image['inds_marked']
        self.modnos_marked=state_image['modnos_marked']
        self.point_marked=state_image['point_marked']

        for mod in self.mods.values():
            mod.set_state(self)

            
    def link(self):
        self.netsim=self.try_link_with('base.net.netsim')
        self.netmon=self.try_link_with('base.net.netmon')
        self.netedit_gui=self.try_link_with('base.net.netedit_gui')

        if self.sim.init_flag:
            
            if len(self.state.db)>0:
                # Append simulation object to maincycle
                self.sim.append_obj_iterate(self)
                self.sim.append_obj_output(self)

        if self.sim.init_flag:
            for car in self.mods.values():
                if self.netmon:
                    car.plot()
                 
    def rebuild_pointers(self):
        """
        Rebuild pointers that have been removed before saving.
        This function is called AFTER all objects have been
        initialized and cross-linked.
        """
        if self.sim.init_flag:
            for car in self.mods.values():
                car.rebuild_pointers()
                #print 'link: rebuild_pointers',car.id

                if (car.status!='active')&(car.status!='edit'):
                    #print '  stop now?'
                    #car.show_info()
                    car.control({'stop':'now'})
                
            
    def init_type(self):
        """Type specific initialization. To be overwritten"""
        pass
    
    
    def get_state(self):
        #print 'get_state',self.type
        if self.netmon:
            self.netedit_gui.default()
            
            
        state={'version':self.__version__,
                'mods':self.mods,
                'mod_no_current':self.mod_no_current,
                'state':self.state,
                'brakefor_car_infront':self.brakefor_car_infront,
                'brakefor_ghost_infront':self.brakefor_ghost_infront,
                'brakefor_stop':self.brakefor_stop,
                'velocity_inc':self.velocity_inc,
                'velocity_dec':self.velocity_dec,
                'state':self.state,
                'velocity_line':self.velocity_line,
                'car_infront_ind':self.car_infront_ind,
                'modinds_hascar_infront':self.modinds_hascar_infront,
                'inds_hascar_infront':self.inds_hascar_infront,
                'dist':self.dist,
                'dist_merge':self.dist_merge,
                #'brakedist':self.brakedist,
                'ghost_infront_ind':self.ghost_infront_ind,
                'modinds_hasghost_infront':self.modinds_hasghost_infront,
                'inds_hasghost_infront':self.inds_hasghost_infront,
                'ghostdist':self.ghostdist,
                'inds_marked':self.inds_marked,
                'modnos_marked':self.modnos_marked,
                'point_marked':self.point_marked,
                'modinds_hasstop':self.modinds_hasstop,
                'inds_hasstop':self.inds_hasstop,
                'point_stop':self.point_stop,
                }
        #print state
        return state
    

    
    def control_mod(self,mod,controldata):

        #print 'control_mod:',mod.id#,controldata
        #for key in controldata.keys():
        #    print '  ',key,controldata[key]

        if controldata.has_key('mark_at'):
            if controldata['mark_at']:
                # print 'set marker for',mod.id
                self.inds_marked[mod.no]=mod.ind
                self.modnos_marked[mod.no]=mod.no
                self.point_marked[mod.no]=controldata['mark_at']
                
            else: # remove halt point
                del self.inds_marked[mod.no]
                del self.modnos_marked[mod.no]
                del self.point_marked[mod.no]

        
            
            
        #if (controldata.has_key('velocity_line'))&(mod.status=='active'):
        if controldata.has_key('velocity_line'):
            velocity_line=controldata['velocity_line']
            self.velocity_line[mod.no]=array([velocity_line])
            self.velocity_inc[mod.no]=mod.accel_comfort*self.sim.time_sample
            self.velocity_dec[mod.no]=-mod.accel_comfort*self.sim.time_sample

        if controldata.has_key('stop'):
            # stop carrier instantly !
            # this should only be used if carrier is already at a
            # very low speed !
            self.velocity_line[mod.no]=0.0
            self.velocity_inc[mod.no]=0.0
            self.velocity_dec[mod.no]=0.0
            self.state[mod.no]=0.0

        
        
        
        if controldata.has_key('car_infront'):
            if controldata['car_infront']: # assign new values
                dist_infront=controldata['dist_infront']-self.dist_min
                self.modinds_hascar_infront[mod.no]=mod.modind
                self.inds_hascar_infront[mod.no]=mod.ind
                self.car_infront_ind[mod.no]=controldata['car_infront'].ind
                self.dist[mod.no]=dist_infront
        
            else:  # delete carrier in front from index array
                if self.modinds_hascar_infront.db.has_key(mod.no):
                    del self.modinds_hascar_infront[mod.no]
                    del self.inds_hascar_infront[mod.no]
                    del self.car_infront_ind[mod.no]
                    del self.dist[mod.no]
                    self.brakefor_car_infront[mod.no]=0
                    
        if controldata.has_key('ghost_infront'):
            if controldata['ghost_infront']: # assign new values
                #print 'control_mod: assign ghost'\
                #      ,controldata['ghost_infront'].id,'to',mod.id
                ghostdist_infront=controldata['ghostdist_infront']\
                                   -self.dist_min
                #if self.modinds_hasghost_infront.db.has_key(mod.no):
                self.modinds_hasghost_infront[mod.no]=mod.modind
                self.inds_hasghost_infront[mod.no]=mod.ind
                self.ghost_infront_ind[mod.no]=\
                                           controldata['ghost_infront'].ind
                self.ghostdist[mod.no]=ghostdist_infront
                                                                            
                            
            else:  # delete ghost in front from index array
                if self.modinds_hasghost_infront.db.has_key(mod.no):
                    #print 'control_mod: remove ghost from',mod.id
                    del self.modinds_hasghost_infront[mod.no]
                    del self.inds_hasghost_infront[mod.no]
                    del self.ghost_infront_ind[mod.no]
                    del self.ghostdist[mod.no]
                    self.brakefor_ghost_infront[mod.no]=0

        if controldata.has_key('stop_at'):
            if controldata['stop_at']: # assign new values
                #print 'control_mod: assign stop_at'\
                #      ,controldata['stop_at'],'to',mod.id
                self.modinds_hasstop[mod.no]=mod.modind
                self.inds_hasstop[mod.no]=mod.ind
                self.point_stop[mod.no]=controldata['stop_at']
                            
            else:  # delete stop from index array
                if self.modinds_hasstop.db.has_key(mod.no):
                    #print 'control_mod: remove stop from',mod.id
                    self.brakefor_stop[mod.no]=0
                    del self.modinds_hasstop[mod.no]
                    del self.inds_hasstop[mod.no]
                    del self.point_stop[mod.no]
                
        if not controldata:
            #print '   refresh indexes'
            if self.modinds_hascar_infront.db.has_key(mod.no):
                self.modinds_hascar_infront[mod.no]=mod.modind
                self.inds_hascar_infront[mod.no]=mod.ind
                self.car_infront_ind[mod.no]=mod.car_infront.ind

            if self.modinds_hasghost_infront.db.has_key(mod.no):
                self.modinds_hasghost_infront[mod.no]=mod.modind
                self.inds_hasghost_infront[mod.no]=mod.ind
                self.ghost_infront_ind[mod.no]=mod.ghost_infront.ind

        #print '  brakefor_car_infront',len(self.brakefor_car_infront.arrays)
        #print '  brakefor_stop',len(self.brakefor_stop.arrays)
        #print '  ',self.inds_hasstop.arrays,self.inds_hasstop.arrays
        #print 'self.inds_marked.db=',self.inds_marked.db
        #print 'self.inds_marked.arrays=',self.inds_marked.arrays
        #print 'self.modnos_marked.arrays=',self.modnos_marked.arrays
        #print 'self.point_marked.arrays=',self.point_marked.arrays

    def iterate(self):
        """
        Update of internal states.
        """
        #print 'iterate',self.type
        
                
        # do distance to carriers in front
        if len(self.car_infront_ind.arrays)>0:
            self.v_hascar=take(self.state.arrays,
                                 self.modinds_hascar_infront.arrays)
            self.v_car=take(self.netsim.velocity.arrays,
                              self.car_infront_ind.arrays)
            add(self.dist.arrays,
                (self.v_car-self.v_hascar)*self.sim.time_sample,
                self.dist.arrays)

            self.d_min=take(self.netsim.brakedists,
                                 self.inds_hascar_infront.arrays)[:,1]-\
                                 take(self.netsim.brakedists,
                                      self.car_infront_ind.arrays)[:,2]
        
            put(self.brakefor_car_infront.arrays,
                self.modinds_hascar_infront.arrays,
                greater(self.d_min,self.dist.arrays)\
                |greater(self.v_hascar,self.v_car)\
                &greater(self.d_min+\
                         2.0*take(self.netsim.brake_factors.arrays,
                                  self.inds_hascar_infront.arrays)[:,0]\
                         *(self.v_hascar-self.v_car)**2,
                         self.dist.arrays)
                )
            

        # do distance to carriers with ghosts
        if len(self.ghost_infront_ind.arrays)>0:
            self.v_hasghost=take(self.state.arrays,
                                 self.modinds_hasghost_infront.arrays)
            self.v_ghost=take(self.netsim.velocity.arrays,
                              self.ghost_infront_ind.arrays)
            

            
            add(self.ghostdist.arrays,
                self.sim.time_sample*(self.v_ghost-self.v_hasghost),
                self.ghostdist.arrays)
            #d_approach=take(self.netsim.brakedists,
            #                self.inds_hasghost_infront.arrays)[:,1]\
            #                -take(self.netsim.brakedists,
            #                      self.ghost_infront_ind.arrays)[:,2]

            self.d_min=take(self.netsim.brakedists,
                            self.inds_hasghost_infront.arrays)[:,1]\
                            -take(self.netsim.brakedists,
                                  self.ghost_infront_ind.arrays)[:,2]
                
            self.d_min_crit=greater(self.d_min,
                                    self.ghostdist.arrays)
            
            self.e_min=self.v_hasghost*self.v_hasghost\
                        -(take(self.dist_merge.arrays,
                               self.modinds_hasghost_infront.arrays)\
                          -take(self.netsim.path_sec,
                                self.inds_hasghost_infront.arrays)\
                          )/take(self.netsim.brake_factors.arrays,
                                 self.inds_hasghost_infront.arrays)[:,0]
                              
            
  
            put(self.brakefor_ghost_infront.arrays,
                self.modinds_hasghost_infront.arrays,
                (less(self.ghostdist.arrays,0.0)\
                 &(greater(self.v_hasghost,self.v_ghost)\
                   |greater_equal(self.e_min,0.0)\
                   &(self.d_min_crit\
                     |greater(self.d_min+
                              2.0*take(self.netsim.brake_factors.arrays,
                                       self.inds_hasghost_infront.arrays)[:,0]\
                              *(self.v_hasghost-self.v_ghost)\
                              *(self.v_hasghost-sqrt(abs(self.e_min))),
                              self.ghostdist.arrays)\
                     )\
                   |less(self.e_min,0.0)\
                   &(greater(take(self.netsim.brakedists,
                                  self.inds_hasghost_infront.arrays)[:,1]\
                             -take(self.netsim.brakedists,
                                   self.ghost_infront_ind.arrays)[:,2]\
                             ,self.ghostdist.arrays
                             ))\
                   ))\
                |(greater_equal(self.ghostdist.arrays,0.0)\
                  &(self.d_min_crit\
                    |greater(self.v_hasghost,self.v_ghost)\
                    &greater(self.d_min\
                             +2.0*take(self.netsim.brake_factors.arrays,
                                       self.inds_hasghost_infront.arrays)[:,0]\
                             *(self.v_hasghost-self.v_ghost)**2,
                             self.ghostdist.arrays
                             ))
                  )
                )
            
        # do stops 
        put(self.brakefor_stop.arrays,
            self.modinds_hasstop.arrays,
            greater(take(self.netsim.brakedists,
                         self.inds_hasstop.arrays)[:,0]+
                    take(self.netsim.brakedists,
                         self.inds_hasstop.arrays)[:,1],
                    self.point_stop.arrays-take(self.netsim.path_sec,
                                                self.inds_hasstop.arrays)
                    )
            )
            

        # update state for all
        self.target=choose(self.brakefor_car_infront.arrays |
                           self.brakefor_ghost_infront.arrays|
                           self.brakefor_stop.arrays,
                           (self.velocity_line.arrays,0.0)
                           )
        
        
        add(self.state.arrays,
            choose(greater(self.state.arrays,
                           self.target
                           ),
                   (self.velocity_inc.arrays, self.velocity_dec.arrays)),
            self.state.arrays)

        put(self.state.arrays,
            nonzero(less_equal(self.state.arrays,self.velocity_inc.arrays)&\
                    equal(self.target,0.0)),
            0.0)

        # check which carriers crossed a mark
        self.inds_tomark=nonzero(
            greater(take(self.netsim.path_sec,self.inds_marked.arrays),
                    self.point_marked.arrays)
            )
        
        # treat carriers which crossed a mark
        if len(self.inds_tomark)!=0:
            cars_atmark=[]

            # store carrier which crossed mark to preserve array index
            for ind in self.inds_tomark:
                #car=self.mods[self.modnos_marked.arrays[ind]]
                #print 'crossed mark:',car.id,'at',car.sec.elem.id,car.sec.id
                cars_atmark.append(self.mods[self.modnos_marked.arrays[ind]])

            for car in cars_atmark:
                # section knows what to do with a carrier when it crossed
                # the mark
                car.control({'mark_at':None})
                car.sec.atmark(car)
           
           
        
    def output(self):
        # pass velocity to netsim
        self.netsim.velocity[self.type]=self.state.arrays


            
    def multiadd_gui(self,mod):
        """
        Opens a panel to ask how many and how to add automatically
        multiple modules.
        """
        #print 'multiadd',self.id

        # this is an template module, to get default parameters from 
        self.editmod=mod

        if not self.multiadd_panel:
            self.multiadd_panel=gui.Panel(
                None,
                title=self.name+' - add multiple',
                make_cmd=self.make_multiadd_panel,
                refresh_cmd=self.refresh_multiadd_panel,
                apply_cmd=self.apply_multiadd_panel,
                close_cmd=self.close_multiadd_panel)
        else:     
            self.multiadd_panel.refresh()
        return 1
    
    def make_multiadd_panel(self,parent,widgets):
        """
        Create widgets for panel.
        """
        #print 'make_multiadd_panel'
        widgets['label']=Tkinter.Label(parent,
                                       text='Add multiple carriers to'+\
                                       ' the availlable berths')
        
        descriptors=[
            {'key':'number',
             'name':'How many?',
             'printformat':'%d',
             'type':'integer',
             'width':1,
             'value':{'min':0,max:10**7},
             'mode':'write',
             },
            {'key':'zone',
             'name':'Add to a stop in zone',
             'type':'text',
             'width':1,
             'mode':'write',
             },
            ]

        self.multiadd_data={'number':100,
                            'zone':'everywhere'}

        # get defaults from prototype
        descriptors_params,data_params=self.editmod.format_params()


        descriptors+=descriptors_params
        self.multiadd_data.update(data_params)

        
        
        #print 'update widget with all input fields'
        widgets['multiadd_data']=gui.DatasetEditor(descriptors,
                                                  self.multiadd_data,parent,
                                                   applyentry_command=\
                                                   self.update_multiadd_panel)

        widgets['label'].pack(expand=1,fill='both')
        widgets['multiadd_data'].pack(expand=1,fill='both')
        
    
    def refresh_multiadd_panel(self,parent,widgets):
        """
        Refresh widgets for panel with simulation parameters.
        """
        #print 'refresh_multiadd_panel',widgets.keys()
        widgets['multiadd_data'].update_contents(self.multiadd_data)
        widgets['multiadd_data'].set_size(1.57*350,350)

    def update_multiadd_panel(self,datawidget,dataset):
        """
        Update widgets for panel with new function values.
        This funtion is called when a parameter entry has changed.
        """
        #print 'refresh_multiadd_panel',dataset.keys()
        self.multiadd_data.update(dataset)

        #print '  first update the example carrier'
        self.editmod.set_params(self.multiadd_data)

        #print '  get the new parameter list, with updated functions'
        descriptors_params,data_params=self.editmod.format_params()

        #print '  finally update the  multiadd_data'
        self.multiadd_data.update(data_params)

        #print '  and write new values back into panel'
        datawidget.update_contents(self.multiadd_data)

    def apply_multiadd_panel(self,parent,widgets):
        """
        Apply widgets values for panel with simulation parameters.
        """
        #print 'apply_multiadd_panel',widgets['multiadd_data'].get_contents()
        self.editmod=None
        self.multiadd_data=widgets['multiadd_data'].get_contents()
        self.multi_add()
        return 1

    def close_multiadd_panel(self,parent,widgets):
        """
        Do something before simpanel is closed.
        Here the pointer to the panel is set to None.
        """
        self.editmod=None
        self.multiadd_panel=None

        

    def multi_add(self,**multiadd_data):
        """
        Add automatically multiple modules using self.multiadd_data.
        """
        if multiadd_data:
            self.multiadd_data=multiadd_data
            
        stops=[]
        for id in self.sim.objs.keys():
            id_data=self.sim.split_id(id)
            if (id_data['category']=='track'):
                if string.find(id_data['name'],'stop')!= -1:
                    
                    stops_check=id_data['obj'].mods.values()
                    for stop in stops_check:
                        # decide here whether stoip is suitable for carriers
                        if (stop.status=='active')&\
                           ((stop.zone==self.multiadd_data['zone'])|\
                            (self.multiadd_data['zone']=='everywhere')):
                            #print '>>>>>>>>>> found', stop.id
                            stops.append(stop)

        # compute number of carrier per stop
        if stops:
            cars_per_stop=self.multiadd_data['number']/int(len(stops))
            for stop in stops:
                for n in range(cars_per_stop):
                    car=self.Mod(self)
                    car.set_params(self.multiadd_data)
                    car.place(stop)
                    #car.plot()
                    #print 'multi_add:placed:',car.id,stop.id 

                    
    def add_gui(self):
        self.netmon.set_action('add car',
                               callback_click = self.add_click,
                               callback_motion = self.add_click,
                               callback_rightclick=self.add_end,
                               callback_termination = self.add_termination)
        
        
    def add_click(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:

            # unmagnify coords
            x,y=self.netmon.get_unmagnified_canvas_coords(event)
            self.editmod=self.Mod(self,pos=[x,y,0])
            self.editmod.plot()
            if self.debug>3:
                print 'Initialized module ' + self.editmod.id
            
            
            self.netmon.set_action('add car',
                                   callback_click = self.add_place,
                                   callback_motion = self.add_motion,
                                   callback_pressmotion = self.add_motion,
                                   callback_rightclick=self.add_end,
                                   callback_termination = self.add_termination)

            # this is a bit dirty at the moment...improve also at
            # add_motion
            event.x=canvas.canvasx(event.x)
            event.y=canvas.canvasy(event.y)
            self.event_past = event
            
    def add_motion(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:
            if self.rescaled_flag:
                # reset mod to current pointer position
                x,y=self.netmon.get_unmagnified_canvas_coords(event)
                event.x,event.y=self.netmon.magnify_xy(x,y)
                self.event_past = event
                self.editmod.pos=[x,y,0]
                self.editmod.plot()
                self.rescaled_flag=0
            else:
                x=canvas.canvasx(event.x)
                y=canvas.canvasy(event.y)
                self.netmon.move([(self.editmod.id,
                                   self.event_past.x, x,
                                   self.event_past.y, y)],noscale=1)
                event.x=x
                event.y=y
                self.event_past=event
    
    def add_place(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:
            #x,y=self.netmon.get_unmagnified_canvas_coords(event)
            x=canvas.canvasx(event.x)
            y=canvas.canvasy(event.y)
            error=self.place_gui(x,y)
            if not error:
                self.add_click(event)
            else:
                x,y=self.netmon.get_unmagnified_canvas_coords(event)
                #self.editmod.set_position([x,y])
                event.x,event.y=self.netmon.magnify_xy(x,y)
                self.event_past = event
                self.editmod.plot()


        
    def add_end(self,event):
        self.add_termination(event)
        self.netedit_gui.default()
        
    def add_termination(self,event):
        """
        Terminate add operation.
        """

        # Delete the module in edit mode, if existend.
        #self.del_editmod()
        if self.editmod:
            if self.debug>=5:
                print 'Terminate Add module '+self.editmod.id
            self.del_mod(self.editmod)
            
            # delete from monitor
            self.editmod.del_plot()
            self.editmod=None
       
            
    def place_gui(self,x,y,z=0):
        """
        Try to place carrier on a stop near x,y.
        x,y are the magnified canvas coordinates. 
        """
        self.editmod.del_plot()

        # get track element module on which carrier has been placed
        obj,track_elem_tag=self.netmon.find_obj_at_coords(x,y)
        #canvas_obj_id=self.netmon.canvas.find_closest(x,y)
        #track_elem_tags=self.netmon.canvas.gettags(canvas_obj_id)
        if track_elem_tag:
            track_elem=self.sim.split_id(track_elem_tag)['mod']

            self.editmod.pos=[x,y,0]
            self.editmod.plot()

            # try to place carrier on track_elem
            return self.editmod.place(track_elem)

        else: # no track elements found
            self.editmod.plot()
            return 1
        
            
    def del_gui(self,event,tag):
        # get module with carrier out of tag string
        self.editmod=self.sim.split_id(tag)['mod']

        # try to delete module from data structures
        # if module is occupied, this method sets only a delete flag
        # it will be deleted and disappear as soon as it is empty
        # self.del_mod(self.editmod)
        self.editmod.delete()
        # delete from monitor
        # self.editmod.del_plot()
        self.editmod=None

        # do next delete action
        self.netedit_gui.del_mod()

        
    def show_info(self,event,tag):
        # get module with track element out of tag string
        self.index_check()
        car=self.sim.split_id(tags)['mod']
        car.print_state()
        self.netedit_gui.show_info()
        
    def add_mod(self,mod):
        """
        Insert new module in the module database mods of the simulation object
        Initial module values are already initialized by the  module mod.
        """
        self.mod_no_current=self.mod_no_current+1
        mod.no=self.mod_no_current
        mod.id=self.get_mod_id(mod)
            
        # add module to modules database
        self.mods[self.mod_no_current]=mod
        return 0
    
    def add_mod_arrays(self,mod,controldata=None):
        ''' add carrier to array structure'''
        #print 'add_mod_arrays for',mod.id,mod.no
        self.state[mod.no]=0.0
        self.velocity_line[mod.no]=0.0
        self.brakefor_car_infront[mod.no]=0
        self.brakefor_ghost_infront[mod.no]=0
        self.brakefor_stop[mod.no]=0
        self.velocity_inc[mod.no]=mod.accel_comfort*self.sim.time_sample
        self.velocity_dec[mod.no]=-mod.accel_comfort*self.sim.time_sample
        self.dist_merge[mod.no]=0.0
        self.update_arrays()

        #print 'add_arrays',self.type
        #print '  brakefor_car_infront',len(self.brakefor_car_infront.arrays)
        #print '  brakefor_stop',len(self.brakefor_stop.arrays)
        if controldata:
            mod.control(controldata)
        
        # make sure that this sim object is registered in maincycle now
        self.sim.append_obj_iterate(self)
        self.sim.append_obj_output(self)

    def del_mod_arrays(self,mod,controldata=None):
        ''' remove carrier from array structure'''

        #print 'del_mod_arrays for',mod.id,mod.no
        mod.control({'car_infront':None,'ghost_infront':None,'mark_at':None})

        del self.state[mod.no]
        del self.velocity_line[mod.no]
        del self.velocity_inc[mod.no]
        del self.velocity_dec[mod.no]
        del self.brakefor_car_infront[mod.no]
        del self.brakefor_ghost_infront[mod.no]
        del self.brakefor_stop[mod.no]
        del self.dist_merge[mod.no]
        self.update_arrays()
        mod.control()

                        
    def update_arrays(self):
        ''' update index of arrays'''

        # index list with indexes of all carriers in this simulation object
        # in order to read vectors from  netsim's array structures

        if self.netsim.cars.db.has_key(self.type):
            ind_from,ind_to=self.netsim.cars.db[self.type]
            self.inds=arange(ind_from,ind_to+1)
          
            for no in self.state.db.keys():
                self.mods[no].modind=self.state.db[no][0]

            # marked cars
            for no in self.inds_marked.db.keys():
                self.inds_marked[no]=self.mods[no].modind


            # car in front
            for no in self.modinds_hascar_infront.db.keys():
                self.modinds_hascar_infront[no]=self.mods[no].modind

            for no in self.inds_hascar_infront.db.keys():
                self.inds_hascar_infront[no]=self.mods[no].ind

            for no in self.car_infront_ind.db.keys():
                self.car_infront_ind[no]=self.mods[no].car_infront.ind
                
            # cars with ghosts
            for no in self.inds_hasghost_infront.db.keys():
                self.inds_hasghost_infront[no]=self.mods[no].ind
        
         
            for no in self.modinds_hasghost_infront.db.keys():
                self.modinds_hasghost_infront[no]=self.mods[no].modind

            for no in self.ghost_infront_ind.db.keys():
                self.ghost_infront_ind[no]=self.mods[no].ghost_infront.ind

            # cars with stop
            for no in self.inds_hasstop.db.keys():
                self.inds_hasstop[no]=self.mods[no].ind
        
            for no in self.modinds_hasstop.db.keys():
                self.modinds_hasstop[no]=self.mods[no].modind

            
            #print 'update_arrays: ',
        
    def index_check(self):
        '''Index checker (Debugging function)'''
        # print '!!!!!!!!!!!!! INDEX CHECK !!!!!!!!!!!!!!!!!!!!'
        for mod in self.mods.values():
            if mod.status=='active':
                if mod.car_infront:
                    if mod.car_infront.ind!=\
                       self.car_infront_ind[mod.no][0]:
                        print 'Index check: bad car infront index:',mod.id
                        #mod.print_state()
                        self.sim.sim_gui_stop()
                        return 0
                   
    def del_mod(self,mod):
        """
        Delete module
        """
        #print 'del_mod',mod.id,mod.no,mod.status

        # now delete carrier instance when edit status and in mods db
        if (mod.status=='edit')&(self.mods.has_key(mod.no)):
            # remove module from sim object array structures
            self.del_mod_arrays(mod)

            # delete from carrier obj's database
            del self.mods[mod.no] 
                
            # if there are no more cars to simulate,
            # remove carrier sim object from maincycle
            if len(self.mods)==0:
                # remove completely simulation object from current simulation
                
                #print 'remove Car object from maincycle:',self.type
                #self.sim.remove_obj_iterate(self)
                #self.sim.remove_obj_output(self)
                self.sim.remove_obj(self)
                
            return 0

        else: return 1 

            
        
            
