from classlib.user import *
import tkFileDialog,tkMessageBox,os
from tkColorChooser import askcolor
#from RandomArray import *

class Genericmod(Mod):

    def init_module_specific(self):
        """
        module specific init.
        """
        # appearance on the Netmon 2D visulalizer
        self.diameter_netmon=1.5  # diameter in points of the circle
        self.color_netmon_hair='#8b4513'#chocolate4 color of the circle outline
        self.color_netmon='#cd661d' # chocolate3 color of the circle inner

    def set_state_mod_specific(self):
        '''
        This function is called during the opening process of a simulation
        file. Here goes all the stuff that gets lost during the save
        process (essentially pointers to functions).  It is also called
        during the module initialization. 

        '''
        if not hasattr(self,'diameter_netmon'):
            # appearance on the Netmon 2D visulalizer
            self.diameter_netmon=1.5  # diameter in points of the circle
            self.color_netmon_hair='#8b4513' # color of the circle inner
            self.color_netmon='#cd661d' # color of the circle outline
            
    def init_params(self,param={}):
        """
        Initialize parameters. Is called only once during init_common.
        Here all constant attributes should be defined.
        To oververwrite.
        """
        
        
        # set default parameters
        self.velocity_walk=3.0/3.6 # default walking speed in m/s
        self.time_init=0
        self.time_boarding=10.0
        self.time_unboarding=10.0
        
        
        # desired origin and destination of  trip
        self.orig_id=''
        self.dest_id=''

   
            
    def set_params(self,param={}):
        """
        Set arbitrary parameters and make them effective, if possible.
        """
        #print 'set_params for',self.id
        if param.has_key('time_boarding'):
            # has explicit boarding time
            self.time_boarding=param['time_boarding']
            
        elif param.has_key('time_boarding_min'):
            # has random defined boarding time
            self.time_boarding=uniform(param.get('time_boarding_min',10),
                                       param.get('time_boarding_max',20))
        


        if param.has_key('time_unboarding'):
            # has explicit boarding time
            self.time_unboarding=param['time_unboarding']
            
        elif param.has_key('time_unboarding_min'):
            # has random defined boarding time
            self.time_unboarding=uniform(param.get('time_unboarding_min',10),
                                         param.get('time_unboarding_max',20))

        if param.has_key('orig_id'):
            # desired origin and destination of  trip
            self.orig_id=param['orig_id']
            #print '  orig_id=',self.orig_id

        if param.has_key('dest_id'):
            self.dest_id=param['dest_id']
            #print '  dest_id=',self.dest_id
        
        
        #  time to initialize trip
        if param.has_key('time_init'):
            # has explicit time to initialize trip
            self.time_init=param['time_init']
            self.obj.time_init[self.no]=self.time_init
            #print 'user',self.id,param
            #print 'self.time_init:',self.obj.time_init.arrays
            
        elif param.has_key('time_init_min'):
            # has random defined when to initialize trip
            self.time_init=uniform(param.get('time_init_min',10),
                                       param.get('time_init_max',20))
            self.obj.time_init[self.no]=self.time_init

        if param.has_key('color_netmon'):
            #print 'user',self.id,param['color_netmon']
            self.color_netmon=param['color_netmon']
            #self.color_netmon_hair='grey'

        if param.has_key('diameter_netmon'):
            self.diameter_netmon=param['diameter_netmon']
            
            
        
    def format_params(self):
        """
        Return formatted costdata.
        """
        descriptors=[
            {'key':'time_init',
             'name':'Start time',
             'type':'real',
             'value':{'min':0.0,'max':24.0*3600.0},
             'units':{'s':1},
             'printformat':'%8.2f',
             'width':1,
             'mode':'write',
             },
            {'key':'time_boarding',
             'name':'Boarding time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'units':{'s':1},
             'printformat':'%6.2f',
             'width':1,
             'mode':'write',
             },
            {'key':'time_unboarding',
             'name':'Exit time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'units':{'s':1},
             'printformat':'%6.2f',
             'width':1,
             'mode':'write',
             },
            {'key':'velocity_walk',
             'name':'Walking speed',
             'type':'real',
             'value':{'min':0,'max':900},
             'printformat':'%3.2f',
             'units':{'m/s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'color_netmon',
             'name':'Color',
             'type':'text',
             'width':1,
             'mode':'write',
             'button':{'text':'Pick...',
                       'command':self.obj.get_color_gui}
             },
            {'key':'diameter_netmon',
             'name':'Aura diameter',
             'type':'real',
             'value':{'min':0.0,'max':10.0},
             'printformat':'%.2f',
             'units':{'m':1},
             'width':1,
             'mode':'write',
             },
            ]
            
 
        dataset={'time_init':self.time_init,
                 'time_boarding':self.time_boarding,
                 'time_unboarding':self.time_unboarding,
                 'velocity_walk':self.velocity_walk,
                 'color_netmon':self.color_netmon,
                 'diameter_netmon':self.diameter_netmon,}

        return descriptors,dataset
        
    def transfer(self,location,trans=identity(4)):
        '''
        General method to enter 
        User changes to a new location.

        
        '''

        # so let's first update x/y position from transformation
        self.set_position([trans[0,3],trans[1,3]])

        #print self.id,'transfers to',location.id,location.status
        #print 'transfers:self.position=',self.get_position()
        #if self.location:
        #    print 'from',self.location.id

        # OK, what kind of location is it ?
        category=self.obj.sim.split_id(location.id)['category']
        #print self.id,'transfers to',location.id,',category:',category

        
        # can user enter this location ?
        # if user is still in edit mode it can only enter a track element
        if hasattr(location,'enter')&\
               (location.status not in ('passive','edit'))&\
               (not ( (self.status=='edit') & (category!='track') )):

            # make sure that user is active
            self.set_status('active')

            #print self.id,'transfers to',location.id,location.status
            # exit old location, if existent
            if self.location:
                #print '  self.location:',self.location.id
                if self.debug>=5:
                    print '***Transfer',self.id,\
                          '\n   from',self.location.id,'to',location.id
                    
                self.location.exit(self)
                # this is to trap a bug...
                #try:
                #    self.location.exit(self)
                #except:
                #    print 'Error transferring',self.id,\
                #          'from',self.location.id,'to',location.id
                #    return 1
            #else:
            #    print '  self.location:',self.location
                
            # enter new location.
            # This is just to signal the new location that there is a
            # user entering
            location.enter(self)


            # Now, dependent on the new and old location, the userer
            # will trigger an action chain
            
            #print 'transferred',self.id,self.status,'to',location.id,category
            if  category=='track': # is it a track element i.e. stop

                if self.location_category=='car':
                    # last location has been in a carrier
                    # this means user just arrived from a trip
                    
                    # statistics:
                    # compute traveling time
                    self.triptime_total+=self.obj.sim.time-\
                                         self.time_start_traveling
                    # increase number of trips by one
                    self.trips_total+=1

                    
                    # walk to the exit of the stop...and do nothing anymore
                    # print self.id,'transfer to',location.id
                    self.walkto(location,'exit')
                    # store new location as current location
                    self.location=location
                    self.location_category=category
                    return 0

                else:
                    # we are at a stop, so let's
                    # make a trip to a new destination,

                    # Are we already inscribed in a user management ?
                    # ... in case it is out first trip
                    if self.man_id=='':
                        # let's inscribe in default  user management.
                        #print  self.id,'transfer:inscribe at ',self.obj.man_id
                        self.man_id=self.obj.man_id
                        self.send_mes({'to':self.man_id,
                                       'op':'register'
                                       })

                    # walk to a random place near the platform
                    # print self.id,'transfer to',location.id
                    #self.walkto(location,'platform')

                    # store new location as current location
                    self.location=location
                    self.location_category=category
                    
                    # check if there is already a destination
                    if self.dest_id=='':
                        # get a random destination, using meta information
                        stops=self.sim.get_stops()
                        self.dest_id=stops.values()[randint(len(stops))].id
                        
                       
                    # immediately do a trip
                    self.start_trip()
                    return 0
                
            elif category=='car':
                # now user is about to enter a car
                self.car=None
                #print '<<<  transfer',self.id,'at',self.obj.sim.time,self.time_start_waiting
                
                # initialize traveling time
                self.time_start_traveling=self.obj.sim.time

                
                if self.obj.netmon: self.del_plot()

                # store new location as current location
                self.location=location
                self.location_category=category
                return 0
                
            else:
                return 1
            
            
        
        else:
            #print self.id,'transfer to',location.id,' failed for',self.id
            return 1

    def exit_at(self,elem):
        """
        A track element tells user that he arrived at the exit.
        He may do  something, just as with the waiting_at call...
        """
        #print 'exit_at',self.id,elem.id
        self.del_plot()
    
    def plot(self,canvas=None):
        '''General plot function'''
        
        if not canvas: canvas=self.obj.netmon # =None if netmon not invoked

        if canvas:
            # first remove from canvas if existent
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                canvas.delete(self.id)

            x,y=self.get_position()
            d=self.diameter_netmon
            canvas.create_oval(x-d/2,y-d/2,x+d/2,y+d/2,
                               outline=self.color_netmon_hair,
                               fill=self.color_netmon,
                               tag=self.id)
            
    def show_info_specific(self):
        """
        Show mod specific info about driver.
        """
        print 'Start time=',self.time_init

class Generic(User):
    def init_type(self):
        self.name='Generic driver'
        self.info="""Makes a single trip.
        If used with select multiple:
        - Origin/destination can be specified by
          a file containing the origin-to-detination demand matrix.
        - Time when trip starts can be chosen randomly within an interval.
        - Boarding time and exit time  can be chosen randomly
          within an interval.
        - walking speed can be specified.
        """
        self.data_sep='='
        self.time_init=Array_db()
        
        self.Mod= Genericmod

    
    def iterate_specific(self):
        """
        Additional Object specific iterations.
        """
        #print 'iterate_specific'
        #if (len(self.time_init.db)>0):
        #if 1:

        # get array inds of users who must initialize a trip
        self.inds_init=nonzero(self.sim.time>self.time_init.arrays)
        #print 'self.time_init:',self.time_init.arrays
        #print 'self.inds_init=',self.inds_init
        if len(self.inds_init)>0:
            
            for ind in self.inds_init:
                # get user module
                user=self.mods[self.time_init.get_key_with_index(ind)]
                
                # and transfer it to the desired station of origin
                stop=self.sim.get_mod(user.orig_id)

                    
                # compute initial coordinates somewhere
                # near entrance of stop
                x,y=stop.get_pos_entrance(uniform(0,1),uniform(0,-1))

                user.transfer(stop,array([[1,0,0,x],
                                          [0,1,0,y],
                                          [0,0,1,0],
                                          [0,0,0,1]]))

                #print 'user',user.id,'transferred to',stop.id
                # set a high time_init
                # to prevent initializing another time
                self.time_init[user.no]=9**8

    def format_multiadd(self):
        """
        Create widgets for panel.
        """
        descriptors=[
            {'key':'odfile_name',
             'name':'File with O/D matrix',
             'type':'text',
             'width':1,
             'mode':'write',
             'button':{'text':'Browse...',
                       'command':self.get_odfile_gui}
             },
            {'key':'time_init_min',
             'name':'Start trip earliest',
             'type':'real',
             'value':{'min':0,'max':24*3600},
             'printformat':'%.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'time_init_max',
             'name':'Start trip latest',
             'type':'real',
             'value':{'min':0,'max':24*3600},
             'printformat':'%.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'time_boarding_min',
             'name':'Minimum boarding time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'time_boarding_max',
             'name':'Maximum boarding time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'time_unboarding_min',
             'name':'Minimum exit time',
             'type':'real',
             'value':{'min':0,max:3600},
             'printformat':'%.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'time_unboarding_max',
             'name':'Maximum exit time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'velocity_walk',
             'name':'Walking speed',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%.2f',
             'units':{'m/s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'color_netmon',
             'name':'Color',
             'type':'text',
             'width':1,
             'mode':'write',
             'button':{'text':'Pick...',
                       'command':self.get_color_gui}
             },
            {'key':'diameter_netmon',
             'name':'Aura diameter',
             'type':'real',
             'value':{'min':0.0,'max':10.0},
             'printformat':'%.2f',
             'units':{'m':1},
             'width':1,
             'mode':'write',
             },
            {'key':'profile_name',
             'name':'Load user-profile',
             'type':'text',
             'width':1,
             'mode':'read',
             'button':{'text':'Load...',
                       'command':self.get_profile_gui}
             },
            
            {'key':'saved_profile',
             'name':'Save user-profile',
             'type':'text',
             'width':1,
             'mode':'read',
             'button':{'text':'Save...',
                       'command':self.save_profile_gui}
             },
            ]
        dataset={'odfile_name':'',
                 'time_init_min':5*60,
                 'time_init_max':3600+5*60,
                 'time_boarding_min':5.0,
                 'time_boarding_max':10.0,
                 'time_unboarding_min':5.0,
                 'time_unboarding_max':10.0,
                 'velocity_walk':3.0/3.6,
                 'color_netmon':'#8b4513',
                 'diameter_netmon':1.5,
                 'profile_name':'unknown',
                 'saved_profile':'',}

        return descriptors,dataset
    
    def make_multiadd_panel(self,parent,widgets):
        """
        Create widgets for panel.
        """
        #descriptors,data=self.format_multiadd()
        widgets['label']=Tkinter.Label(parent,
                                       text='Add multiple users to'+\
                                       ' the availlable stops')
        
        descriptors,self.multiadd_data=self.format_multiadd()

        self.editmod=None
        
        widgets['multiadd_data']=gui.DatasetEditor(descriptors,
                                                   self.multiadd_data,parent)

        widgets['label'].pack(expand=1,fill='both')
        widgets['multiadd_data'].pack(expand=1,fill='both')


    def get_color_gui(self,datawidget,descriptor,value):
        """
        Calls color picker.
        """
        (triple,hexstr)=askcolor()
        
        if hexstr:
            # update widget
            dataset=datawidget.get_contents()
            dataset['color_netmon']=hexstr
            datawidget.update_contents(dataset)
        
    def get_odfile_gui(self,datawidget,descriptor,value):
        """
        Calls browser and inserts OD file into the multiadd data.
        """
        odfile_name =tkFileDialog.askopenfilename(
                initialdir=self.sim.simdir,
                filetypes=[('Text files','*.txt'),('All files','*')])

        # update widget
        dataset=datawidget.get_contents()
        dataset['odfile_name']=odfile_name
        datawidget.update_contents(dataset)

    def get_profile_gui(self,datawidget,descriptor,value):
        """
        Reads data from user-profile.
        """
        profile_name =tkFileDialog.askopenfilename(
                initialdir=self.sim.simdir,
                filetypes=[('User profiles','*user*.txt'),\
                           ('Text files','*.txt'),('All files','*')])


        # update widget
        dataset=datawidget.get_contents()

        dataset_func=gui.DataSet()
        descriptors,d=self.format_multiadd()
        dataset=dataset_func.get_dataset_from_file(profile_name,descriptors,
                                                   dataset,
                                                   self.data_sep)
        
        
        dataset['profile_name']=os.path.basename(profile_name)
        dataset['saved_profile']=''
        
        datawidget.update_contents(dataset)

    def save_profile_gui(self,datawidget,descriptor,value):
        """
        Calls browser to save user-profile.
        """
        profile_name =tkFileDialog.SaveAs(
                initialdir=self.sim.simdir,
                filetypes=[('User profiles','*user*.txt'),\
                           ('Text files','*.txt'),('All files','*')]).show()

        try:
            profile=open(profile_name,'w')
            dataset=datawidget.get_contents()
            for key in dataset.keys():
                if key not in ('','saved_profile','profile_name'):
                    profile.write(key+self.data_sep+str(dataset[key])+'\n')
            
            dataset['profile_name']=os.path.basename(profile_name)
            dataset['saved_profile']=''
            profile.close()
            datawidget.update_contents(dataset)

        except:
            print 'WARNING in save: could not open',profile_name
            return 1
        
    
    
    def multi_add(self,**multiadd_data):
        """
        Add automatically multiple modules using self.multiadd_data.
        """
        if multiadd_data:
            self.multiadd_data=multiadd_data
            if not self.multiadd_data.has_key('odfile_name'):
                if self.multiadd_data.has_key('profile_name'):
                    profile_name=self.multiadd_data['profile_name']
                    
                    dataset_func=gui.DataSet()
                    descriptors,d=self.format_multiadd()
                    #print 'multi_add',profile_name,descriptors
                    
                    data=dataset_func.get_dataset_from_file(\
                    profile_name,descriptors,self.multiadd_data,self.data_sep)
                    
                    #self.multiadd_data=dataset_func.get_dataset_from_file(\
                    #profile_name,self.multiadd_data,self.data_sep)
                
                
            
        # create a database with all currently availlable stops
        # with stop number in string-format as key
        data_sep=','   # column separator od od file
            
        if (self.multiadd_data['odfile_name']):
            
            #print 'multi_add with ODM',self.multiadd_data['odfile_name']
            stops=self.sim.get_stops()
            try:
                odfile=open(self.multiadd_data['odfile_name'],'r')
                od_text=odfile.readlines()
                odfile.close()
            except:
                print 'WARNING: file not found:',\
                      self.multiadd_data['odfile_name']
                return


            #print '  len(od_text)=', len(od_text)   
            # is there enough data in the file
            if len(od_text)>1:
            
                # create od database:
                # key is stop number,
                # value is a list with number of users, where the
                # entry corresponds to one of the destinations in dests
                dests_no=string.split(od_text[0],data_sep)
                
                # convert all stop identifiers to stop numbers
                # print 'dests_orig=',dests_no
                # print 'stops.keys()=',stops.keys()
                for n in range(len(dests_no)):
                    dests_no[n]=string.strip(dests_no[n])
                    #=self.get_stop_no(dests_no[n],stops)
                #print 'dests_no=',dests_no
                
                for line in od_text[1:]:
                    orig_data=string.split(line,data_sep)
                    #print n,'---orig_data[n]=',orig_data
                    orig_no=string.strip(orig_data[0])
                    #=self.get_stop_no(orig_data[0],stops)

                    # make sure that this sim object is registered
                    # in maincycle now
                    self.sim.append_obj_iterate(self)
                    
                    for n in range(len(orig_data))[1:]:
                        if orig_data[n]: # avoid empty entries
                            #print 'place',string.atoi(orig_data[n]),\
                            #      'at',orig_no,\
                            #      'with dest',dests_no[n]
                            
                            for m in range(string.atoi(orig_data[n])):
                                
                                user=self.Mod(self)
                                #print '  add',self.mod_no_current,\
                                #      user.id,user.no
                                if stops.has_key(orig_no)&\
                                   stops.has_key(dests_no[n]):
                                    # complete multiadd_data with
                                    #origin and dest
                                    self.multiadd_data['orig_id']=\
                                                       stops[orig_no].id
                                    self.multiadd_data['dest_id']=\
                                                       stops[dests_no[n]].id
                                    # test:
                                    #self.multiadd_data['time_init']=20
                                
                                    # and configure user
                                    user.set_params(self.multiadd_data)
                                else:
                                    print 'WARNING:',orig_no,'or',dests_no[n],\
                                          'inexistant stop.'
                            
                
            else:
                if self.sim.mode=='gui':
                    tkMessageBox.showerror('Insufficient information in:\n'+\
                                           self.multiadd_data['odfile_name'])
                else:
                    print 'Insufficient information in:\n'+\
                                           self.multiadd_data['odfile_name']
        else:
            if self.sim.mode=='gui':
                tkMessageBox.showerror('No such O/D patter file found:\n'+\
                                       self.multiadd_data['odfile_name'])
            else:
                print 'WARNING: no file',self.multiadd_data['odfile_name']

    def get_stop_no(self,stop_ident,stops):
        """
        returns number of stop in string format
        obtained from the stop identifier string, which can be either the
        stop-number itself or the stop name.
        """
        
        if stop_ident:
            # remove end of line and other
            stop_ident=string.split(stop_ident,'\n')[0]
            stop_ident=string.split(stop_ident,' ')[0]

            # first check if it is a stop number:
            if stops.has_key(stop_ident):
                return stop_ident
            else:
                # now check if names exist
                for stop in stops.values():
                    if stop.name==stop_ident:
                        return str(stop.no)
                return ''
        else:
           return '' 


    
    def get_state_specific(self,state_image):
        """
        For specific object.
        This function is called with state_image as argument
        which will be completed with specific data.
        """
        state_image['time_init']=self.time_init

        
        
    def set_state_specific(self,state_image):
        """
        For specific object.
        """
        self.time_init=state_image['time_init']

        #print 'set_state_specific',self.id,len(self.mods)
        if self.mods:
            users=self.mods.values()
            
            # recover initial trip time from older version
            if not hasattr(users[0],'time_init'):
                for user in users:
                    # recover initial trip time from matrix
                    #print self.obj.time_init.db,self.obj.time_init.arrays
                    if self.time_init.db.has_key(user.no):
                        user.time_init=self.time_init[user.no][0]
                    else:
                        user.time_init=0.0
        
    def del_mod_specific(self,mod):
        """
        Object specific Delete module. 
        """
        if self.time_init.db.has_key(mod.no):
            del self.time_init[mod.no]
