from classlib.track import *



class Stop4berth_short_right(Stop):

    def make_track(self):
        """
        This function sets essential parameters for the specific track
        element. It needs to be called once during initialization.
        """

        # ATTENTION!!
        # length of acceleration and brake section must be adapted to speeds
        
        # main line velocity
        velocity_main=30.0/3.6 # m/s
        
        # velocity in berths section
        velocity_berth=5.0/3.6 # m/s

        # number of berths in berths section
        n_berth=4

        # length of one berth (must be greater than lenght of car)
        length_berth=3.8
        
        # total lengtth of berthsec
        length_berthsection=n_berth*length_berth+1.0
        
        
        # berth that must be kept clear in case there are no users waiting
        berths_keep_clear=1


        #######################################################################
        # topology
        addedlength=length_berthsection-15.0
        self.element=[\
            # s0 entry
            {'label':'entry',
             'shape': ( 'flexline', {'offset':(0.0,0.0,0.0),
                                     'endcoords':(5.0,0.0,0.0),
                                     'alpha0':0.0,
                                     'alpha1':0.0,
                                     }),
             'topo' : (1,2),
             'velocity_line':0.75*velocity_main,
             },
            # s1 main at entry
            {'label':'main_entry',
             'shape': ('line', {'offset':(5.0,0.0,0.0),
                                'length':20.0}),
             'clearingzone':18.0,
             'topo' : (2,3),
             'velocity_line':velocity_main,
             },
            # s4 main at exit
            {'label':'main_exit',
             'shape': ('line', {'offset':(25.0,0.0,0.0),
                                'length':35.0+addedlength}),
             'topo' : (3,4),
             'velocity_line':velocity_main,
             },
            # s4 exit
            {'label':'exit',
             'shape': ('flexline', {'offset':(60.0+addedlength,0.0,0.0),
                                     'endcoords':(65.0+addedlength,0.0,0.0),
                                     'alpha0':0.0,
                                     'alpha1':0.0,
                                     }),
             'topo' : (4,5),
             'velocity_line':velocity_main,
             },
            # s5 brake-line
            {'label':'brake',
             'velocity_line': velocity_berth,
             'shape': ( 'flexline', {'offset':(5.0,0.0,0.0),
                                     'endcoords':(30.0,2.0,0.0),
                                     'alpha0':0.0,
                                     'alpha1':15.0,
                                     }),
             'clearingzone':18.0,
             'topo' : (2,6)
             },
            # buffer-line
            {'label':'buffer',
             'velocity_line': velocity_berth,
             'shape': (  'flexline',{'offset':(30.0,2.0,0.0),
                                     'endcoords':(35.0,4.0,0.0),
                                     'alpha0':15.0,
                                     'alpha1':0.0,
                                    }),
             'topo' : (6,8)
             },
            # s6 berths
            {'label':'berths',
             'type':'berth',
             'side_plat':'right',
             'berth_length':length_berth,
             'berth_no':    n_berth,
             'berths_keep_clear': berths_keep_clear,
             'shape':         ('line', {'offset':(35.0,4.0,0),
                                        'length':15.0+addedlength}),
             'velocity_line': velocity_berth,
             'topo' :         (8,7),
             },
            #s7 acceleration-line
            {'label':'accel',
             'shape': ('scurve', {'offset':(50.0+addedlength,0.0,0.0),
                                  'length':10.0,
                                  'displace':-4.0,
                                  'turn'  :'left'
                                  }),
             'topo' : (7,4),
             'velocity_line':velocity_main,
             },
            ]

        self.nodes_connect={1:'in', 5:'out'}

        #self.info='Off-line stop with 3 berth and 60m of total length.\n'+\ 
        #'Main line velocity is 8.3m/s'
        self.info= 'Stop with 4 off-line berths'+\
                   'on the right side.'+\
                   'Use in low traffic density areas \n'+\
                   'because optimized for low space usage \n'+\
                   'The stop is designed for a main line velocity of '+\
                   '8.3 m/s.'+\
                   'DO NOT EXCEED THIS VELOCITY !'+\
                   'Both extremities can be extended by dragging the nodes.'
        # define class for this track element module
        self.Mod=ShortStopmod




                    
                    