#from classlib.track import *
from classlib.base import *
from classlib.man import *
import copy
from RandomArray import *

class Transport(Task):
    """ 
    Task to transport a carrier to its  destination:
    1. insert destination of carrier into management's cars table
    2. inform destination element of the carrier arrival
    3. wait until destination element reported that carrier arrived or passed
    """
    
    def init_task(self,car):
        """
        Read and store information about the carrier to be transported
        and schedule track elements that are on the carrier's route.
        """

        # store carrier data from order
        self.car_id=car['car_id']
        self.dest_id=car['dest_id']
        self.mode=car['mode']
        
        #print 'Trackman,Transport.init_task ',self.id,'for',car['car_id']
        #print '  from carman order',self.order_id
        
        if self.dest_id=='': # is destination defined ?
            # currently: choose random dn=len(self.obj.dests)
            # number of of availlable destinations
            n=len(self.obj.dests)
            if n>0:
                # get new random destination
                self.dest_id=self.obj.dests.keys()[randint(n)]

        # schedule track elements that are on the route to dest_id,
        # starting from sec_id, the section on which the carrier is
        # currently located.
        #print '--- Sched',self.car_id,'from',car['sec_id'],\
        #      'to',self.dest_id
        self.schedule(car['sec_id'])
        
        # back-link to task in carrier db
        # This is the db, where diverge points can look up
        # what to do with a carrier (see man.get_nextsec)
        # This  means next section determined dependent
        # on carrier destination.
        # Later this method should be replaced by a
        # better scheduling scheme.
        self.obj.car_tasks[self.car_id]=self
        
        # Now wait untill the done method is called
        # by the management
        

    def schedule(self,sec_id):
        """
        (re)schedule all track elements that are on the route
        of the carrier to its destination. The route start with
        sec_id, the current section of the carrier and ends with
        on the dest_id track element of this task.

        All diverge points and the destination track element
        will be scheduled.
        """
        #print 'ssssssssssssssssssssssssssssss by Task',self.id
        #print 'Schedule track for',self.car_id
        #print '   from sec',sec_id,'to',self.dest_id

        # prevent task from being killed during this algorithm
        self.operating=1

        #for proc in self.procs.values():
        #    mod=self.obj.sim.split_id(proc.res_id)['mod']
        #    for task in mod.tasks.values():
        #        print '---------------------------------------------'
        #        print mod.id,'task',task.id,\
        #              'from',task.client_id,'/',task.order_id,\
        #              'status=',task.status
        #        print 'in tasks_by_client ?',task.obj.tasks_by_client.has_key(
        #            task.client_id+task.obj.order_sep+task.order_id)
            
        # first, kill all pending scheduling processes
        for proc in self.procs.values():
            #print self.obj.id,'.scheduling: task',self.id,\
            #      'kills',proc.id,'with order',proc.order_id
            proc.kill()

        # check track elements on the route
        terminated=0
        self.set_status('processing')

        # save id of initial element. This will be required to prevent
        # the task removal if this element is a later part of the path
        # and the carrier leaves this element.
        initelem_id=self.obj.secs[sec_id]['elem_id']
        
        nextelem_id=-1 # current next track element, -1 means not defined
        nextsec_id=-1 # current nextsec, -1 means not defined

        
        # path and time estimations
        trippath_est=0.0
        triptime_est=0.0
        
        # list with tuples with (sec_id,nextsec_id), defined at diverge points
        nextsec_ids={} # initialization of routing table for diverge elements
        target_flag=0 # set to 1 when search arrives at destination berth
        while not terminated:
            if self.obj.secs.has_key(sec_id): # valid section of this cluster ?
                sec=self.obj.secs[sec_id]
                elem_id=sec['elem_id']


                #print '  check:',elem_id,sec_id,'branches:',\
                #      len(sec['tosec_ids'])
                if len(sec['tosec_ids'])>1: # diverge point ?
                    #print '    found diverge point:',elem_id,sec_id
                    # OK, but is there already a routing table ?
                    
                    if sec.has_key('routing_table'):
                        routing_table=sec['routing_table']
                        #print '    has already a routing table'#,routing_table

                        # does the routing table contain the desired dest_id ?
                        if routing_table.has_key(self.dest_id):
                            nextsec_id=routing_table[self.dest_id]
                            #print '  found nextsec in routing tab:',nextsec_id
                            # store nextsec_id in nextsec_ids table
                            nextsec_ids[sec_id]=nextsec_id

                    if nextsec_id==-1: # no next section has been found ?
                        #print '   no next section:take first one, if existent'
                        if  len(sec['tosec_ids'])==1: 
                            nextsec_id=sec['tosec_ids'][0]
                            
                        
                # is section a berth-section at the destination ?
                elif (sec['type']=='berth') & (elem_id==self.dest_id) &\
                     (nextelem_id!=-1):
                    #print '    Terminate for',self.car_id,\
                    #      sec['type'],elem_id,self.dest_id
                    terminated=1 # search terminated
                    target_flag=1
                    
                # is section any section with  one nextsection (no diverge) ?
                elif len(sec['tosec_ids'])==1:
                    nextsec_id=sec['tosec_ids'][0]
                    #print 'just one nextsec:',nextsec_id


                # print 'But does nextsec exist in cluster ?',nextsec_id
                if self.obj.secs.has_key(nextsec_id):
                    nextelem_id=self.obj.secs[nextsec_id]['elem_id']

                elif not target_flag: 
                    #if self.debug>=3:
                    print 'WARNING: interrupted loop, no connections found'
                    self.set_status('waiting') # wait for rerouting
                    terminated=1

                
                # Schedule track element if we terminated or
                # if we are about to leave a diverge element
                if terminated |\
                       ((nextelem_id!=elem_id) & (len(nextsec_ids)>0)):
                    #if (nextelem_id!=elem_id):
                    #    print ' =>schedule',elem_id,'change elem'
                    #else:
                    #    print ' =>schedule',elem_id
                    #print '      nextsec_ids=',nextsec_ids
                    #print '      terminated=',terminated,\
                    #      'target_flag=',target_flag
                    self.add_proc(Sched_elem,
                                  elem_id,
                                  nextsec_ids,
                                  target_flag)
                    nextsec_ids={}


                # path and time estimations, don't count last section
                if not terminated:
                    trippath_est+=sec['length']
                    triptime_est+=sec['length']/sec['velocity_line']
                   
                # prepare next step
                sec_id=nextsec_id
                nextsec_id=-1
            
            else:
                # sec_id is not a valid section of this cluster or is
                # currently inactivated...just wait for rerouting
                self.set_status('waiting')
                terminated=1

        # path and time estimations
        if (trippath_est>0)&(self.status=='processing'):
            self.info['trippath_est']=trippath_est
            self.info['triptime_est']=triptime_est

        self.info['initelem_id']=initelem_id
        
        self.return_info()
        self.operating=0
            

    def update(self,proc):
        """
        Check whether carrier arrived,i.e. a schedule process from
        a berth element is done.

        Do also statistics according to the current status.
        """
        # print 'task.update',proc.id,'proc.res_id=',proc.res_id
        if proc.res_id!='':
            elem=self.obj.elems[proc.res_id]
        else:
            elem=None
        # if target_flag is set means that task is terminated
        # as soon as process proc  is terminated
        #print self.obj.id,':update proc',self.id,'.',proc.id,proc.status 
        if proc.status=='done':
            # check whether task is also terminated
            if proc.target_flag |\
                   ((self.mode=='search_item') &\
                    (proc.info['action']=='halt')):
               #print '  update:',proc.order_id,'done for',self.car_id

               # do statistics
               elem['cars_switched']+=1
               elem['cars_arrived']+=1

               for proc_to_kill in self.procs.values():
                   if proc_to_kill!=proc:
                       #print '  kill proc',proc_to_kill.id
                       proc_to_kill.kill()
               
            else:
                # do statistics
                elem['cars_switched']+=1

        elif proc.status=='error':
            # check if carrier has been deviated
            if proc.info['action']=='deviate':
                # do statistics
                elem['cars_switched']+=1
                if self.mode!='search_item':
                    elem['cars_deviated']+=1
                
                #print '!!!!!!! Deviation of',self.car_id,\
                #      'to',proc.info['nextsec_id']
                # reschedule carrier, starting from nextsec_id
                #if not self.search_item:
                self.schedule(proc.info['nextsec_id'])
                
    
    def cleanup(self):
        """
        This function cleans up all memory traces of this task in
        the respective managing object.
        It is called after the task has been done, failed or killed.
        """
        #print self.obj.id,'cleanup task',self.id,'order',self.order_id
        #print '   for',self.car_id
        if self.obj.car_tasks.has_key(self.car_id):
            if self.obj.car_tasks[self.car_id]==self:
                del self.obj.car_tasks[self.car_id]
            
    def kill(self):
        """
        Start to kill all runnung processes. Task will remain untill all
        processes have terminated successfully.
        May be overwritten by specific management.
        """
        #print 'Transport Task.kill',self.id,'carman order_id:',self.order_id
        # avoid double killing
        #if (self.status!='kill')&(self.status!='killed'):
        if self.procs:
            self.set_status('kill')
            for proc in self.procs.values():
                # avoid double killing
                #if (proc.status!='kill')&(proc.status!='killed'):
                #    print '  kills Proc',proc.id
                proc.kill()
        else:
            self.set_status('killed')
        

            
class Sched_elem(Proc):
    """
    Schedules track element.
    Current implementation:
    Process that orders a track destination to report when the 
    carrier of the respective task has arrived. 
    """

    def init_proc(self,elem_id,nextsec_ids={},target_flag=0):
        """
        Make a transport order to track element elem_id.

        Optionally, the ids of the section and next section for the
        a diverge element's routing table can be provided for diverges.

        The target_flag indicates that this track element is the destination.

        The initelem_flag indicates whether the carrier trip starts on
        this element.
        """
        #print 'Proc',self.task.obj.id,'Task',self.task.id,'.',self.id,':'
        #print '   Schedule',elem_id,'for',self.task.car_id
              
                    
        # if target_flag is set means that task is terminated
        # as soon as this process is terminated
        self.target_flag=target_flag

        self.set_status('processing')
        self.make_order(elem_id,'transport',
                        {'car_id':self.task.car_id,
                         'nextsec_ids':nextsec_ids,
                         'target_flag':target_flag,
                         'mode':self.task.mode,
                         })
        
        # stored for statistic purposes
        #self.nextsec_ids=nextsec_ids

        self.res_id=elem_id
        #print 'Proc',self.id,'has res_id:',elem_id
        
    def kill(self):
        """
        Terminate this process.
        overwritten by specific process.
        """
        #print 'kill proc',self.task.obj.id,self.task.id,self.id
        if self.res_id:
            # avoid double killing
            if (self.status!='kill')&(self.status!='killed'):
                # send task of resource the kill command
                self.task.obj.send_mes({'to':self.res_id,
                                        'op':'kill_task',
                                        'order_id':self.order_id,
                                        })
                
                self.set_status('kill')
            else:
                print 'WARNING: proc.kill of Sched_elem in'\
                      ,self.task.obj.id,self.task.id,self.id
                print '  Still resource',self.res_id,\
                      'but already in status',\
                      self.status
        else:
            #print '  set_status killed'
            self.set_status('killed')
        
class Get_destinations(Task):
    """
    Task to return all destinations of this cluster.

    This could be done more simple and more direct,
    but it is used to test the task->process mechanism.
    """
    def init_task(self,args=None):
        """
        Call retuzrn_info function to send back immediately
        all current destinations.
        """
        print 'Get_destinations from',self.obj.id
        self.obj.userman_task_getdests=self
        
        self.set_status('processing') # this will also return info
        self.return_info()
        
    def return_info(self):
        """
        Central function to send  information back to client.
        
        Returns the current available destinations in this cluster.
        Task will not be terminated since the put_info function will
        be called at the client side.

        This function will be called by the management each time the
        destination change.
        """
        
        

        # do not return to auto-generated tasks
        if self.client_id!=self.obj.id:
            category=self.obj.sim.split_id(self.client_id)['category']
            #print 'return_info of',self.obj.id,'to',category
            self.info['dest_ids']=[]
            if category=='userman':
                for id_dest in self.obj.dests.keys():
                    if self.obj.elems[id_dest]['zone']!='maintenance':
                        self.info['dest_ids'].append(id_dest)
                
                
            elif category=='carman':
                #print 'init_task,status=',self.status
                self.info['dest_ids']=self.obj.dests.keys()
        
            # make sure that info contains status
            self.info['status']=self.status
            #print self.info['dest_ids']
            self.obj.send_mes({'to':self.client_id,
                               'op':'put_info',
                               'order_id':self.order_id,
                               'info':self.info
                               })
            self.info={}
            
class Experim(Man):
    '''
    Track management simulation object
    (Expetimental version)
    '''
    
    def init_obj_specific(self):
        #print 'init_obj_specific of',self.id
        self.__version__=self.sim.__version__
        self.output_interval=2.0 # second update

        # make at least one reroute when simulation starts running
        self.reroute=1
        
        self.elems=self.resources
        
        self.secs={}
        self.dests={}
        self.car_tasks={} # maps carrier ids to their task

        # task with userman polling for current destinations
        # return_info methof must be called each time destination changes
        self.userman_task_getdests=None 

        #self.routes={}
        
        # add operations that can be executed by an incoming message
        self.mes_ops['add_elem']=self.add_elem
        self.mes_ops['del_elem']=self.del_elem
        self.mes_ops['update_elem']=self.update_elem
        self.mes_ops['get_dest']=self.get_dest
        self.mes_ops['reschedule_cars']=self.reschedule_cars

        
        # The table task_classes is mapping task names to task classes.

        # task to return all available destinations
        self.task_classes['get_destinations']=Get_destinations

        # generic task to coordinate the transport of a carrier
        self.task_classes['transport']=Transport

        self.costmatrix=array([])
        #print '  task_classes=\n',self.task_classes

    def init_costs_specific(self):
        """
        Management specific cost initialization.
        To be overwritten.
        """
        pass

       
    
    def get_state(self):
        '''Prepare state image for save operation'''
        #print self.id,'get_state:'
        state_image=self.get_state_man({})
        state_image['elems']=self.elems
        state_image['secs']=self.secs
        #for sec_id in state_image['secs'].keys(): print '  sec_id=',sec_id
        state_image['dests']=self.dests
        self.car_tasks
        #state_image['']=self.
        state_image['car_tasks']=self.car_tasks
        state_image['userman_task_getdests']=self.userman_task_getdests
        state_image['reroute']=self.reroute
        state_image['costmatrix']=self.costmatrix 
        #state_image['']=self. 

        #for key in state_image.keys():
        #    print '  state:',key
        #    if key=='costmatrix': print len(state_image['costmatrix']),\
        #       len(self.costmatrix)
        
        return state_image
        
    def set_state(self,state_image):
        '''Set state while opening simulation file'''
        #print self.id,'set_state:'
        #for state_name in state_image.keys(): print '  state:',state_name
        self.set_state_man(state_image)
        self.elems=state_image['elems']
        self.secs=state_image['secs']
        #for sec_id in self.secs.keys(): print '  sec_id=',sec_id
        #self.nextsec_ids=state_image['nextsec_ids']
        #if state_image.has_key('stops'):
        #    self.dests=state_image['stops']
        #else:
        self.dests=state_image['dests']

        if state_image.has_key('reroute'):
            self.reroute=state_image['reroute']

        if state_image.has_key('costmatrix'):
            self.costmatrix=state_image['costmatrix']
            
        #self.=state_image[''] 
        self.car_tasks=state_image['car_tasks']
        self.userman_task_getdests=state_image['userman_task_getdests']

         
        #print '  task_classes=\n',self.task_classes
        if string.atof(self.get_version())<0.3:

            # will be done by update_man of elem
            flag_stop=0
            for dest in self.dests.values():
                for sec in dest['secs']:
                    if sec['type']=='berth':flag_stop=1
                    if not sec.has_key('coords_in'):
                        sec['coords_in']=(0.0,0.0)
                        sec['coords_out']=(0.0,0.0)
                
                    
            for task in self.tasks.values():
                if not hasattr(task,'mode'):
                    if hasattr(task,'search_item'):
                        if task.search_item==1:
                            task.mode='search_item'
                        else:
                            task.mode='loaded'
            # delete previoud update elem messages without class key
            messages_new=[]
            for message in self.messages:
                if message['op']=='update_elem':
                    if message.has_key('class'):
                        messages_new.append(message)
                else:
                    messages_new.append(message)
            self.messages=messages_new
                
    def get_resource_name(self):
        """
        Returns string with name of resource that is managed by this management
        To be overwritten.
        """
        return 'Track-element'

    def reset_statistics_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular type.

        To be overwritten.
        """
        # do through resources here as this function may be called
        # from the general management class
        #print 'reset_statistics_specific for',self.id
        
        for resource in self.resources.values():
            self.reset_statistics_resource(resource)

    def reset_statistics_resource(self,elem):
        """
        Initialize and reset all statistical data for this track element. 
        """
        #print 'reset_statistics_user for ',user['id']
        elem['cars_arrived']=0
        elem['cars_switched']=0
        elem['cars_deviated']=0
        
        
    def format_statistics_specific(self,descriptor,dataset):
        """
        Format management specific contents for statistical data.
        """
        descriptor_specific=[
            {'key':'Total # of switch operations',
             'type':'integer',
             'printformat':'%9d',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
            'width':0.2,
             },
            {'key':'Total # of deviations',
             'type':'integer',
             'printformat':'%9d',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.2,
            },
            {'key':'Deviation rate',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**2},
             'mode':'read',
             'width':0.2,
             'units':{'%':1}
             }
            ]

        descriptor+=descriptor_specific

        n=0
        switched_sum=0
        deviated_sum=0

        # calculating statistics over all elements
        for elem in self.elems.values():
            switched_sum+=elem['cars_switched']
            deviated_sum+=elem['cars_deviated']

        dataset['Total # of switch operations']=switched_sum
        dataset['Total # of deviations']=deviated_sum
        if switched_sum>0:
            dataset['Deviation rate']=100.0*deviated_sum/switched_sum
        else:
            dataset['Deviation rate']=0
        
    def format_resources_specific(self,header,table):
        """
        Updates all management specific widgets on resource page and table.
        """
        header.append({'key':'cars_arrived',
                       'name':'arrivals',
                       'type':'integer',
                       'printformat':'%d',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.3,
                       })
        #header.append({'key':'cars_departed',
        #               'type':'integer',
        #               'printformat':'%d',
        #               'valid':{'min':0,'max':10**8},
        #               'mode':'read',
        #               'width':0.2,
        #               })
        header.append({'key':'cars_switched',
                       'name':'switches',
                       'type':'integer',
                       'printformat':'%d',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.3,
                       })
        header.append({'key':'cars_deviated',
                       'name':'deviated',                       
                       'type':'integer',
                       'printformat':'%d',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.3,
                       })
        header.append({'key':'cars_deviated_rel',
                       'name':'deviated',
                       'type':'real',
                       'printformat':'%.2f',
                       'valid':{'min':0,'max':10**8},
                       'units':{'%':1},
                       'mode':'read',
                       'width':0.4,
                       })
        # do necessary computations for extra columns
        for row in table:
            if row['cars_switched']>0:
                row['cars_deviated_rel']=100.0*row['cars_deviated']/\
                                          row['cars_switched']
            else:
                row['cars_deviated_rel']=0
        

    def get_dest(self,message):
        '''
        Function returns all currently availlable destinations on
        this track cluster, using the put_dest of the calling management.  
        '''
        self.send_mes({'to':message['from'],
                       'op':'put_dest',
                       'dest_ids':self.dests.keys()
                       })
        
        
    def add_elem(self,message):
        elem_id=message['from']
        self.elems[elem_id]={'id':elem_id,
                             'status':message['status'],
                             'sec_ids':[]
                             }
        self.reset_statistics_resource(self.elems[elem_id])
        
        
    def del_elem(self,message):
        elem_id=message['from']
        if self.elems.has_key(elem_id):
            for sec_id in self.elems[elem_id]['sec_ids']:
                #print self.id,': Delete section',sec_id,'from cluster'
                if self.secs.has_key(sec_id):
                    del self.secs[sec_id]
            
                
            if self.dests.has_key(elem_id):
                del self.dests[elem_id]
            
            del self.elems[elem_id]

            # remove itself from simulation objects when there are no more
            # track elements
            # ...currently omitted because of problems with messages
            #if len(self.elems)==0:
            #    self.sim.remove_obj(self)
            
    def update_elem(self,message):
        
        elem_id=message['from']
        #print self.id,'.update_elem ',elem_id
        #if message.has_key('secs'):
        #    print 'secs:',message['secs']
        
        if not self.elems.has_key(elem_id):
            #print '  add element to db'
            self.elems[elem_id]={'id':elem_id,
                                 'status':message['status'],
                                 'sec_ids':[]
                                 }
            self.reset_statistics_resource(self.elems[elem_id])

        
        # reroute if topology has changed:
        if self.elems.has_key(elem_id):
            if self.elems[elem_id]['status']!=message['status']:
                self.reroute=1 # status has changed
        else:
            #print '  reroute1'
            self.reroute=1  # element is added

        # store new status
        #elf.elems[elem_id]['status']=message['status']
 
        # update all element keys except section specific
        for key in message.keys():
            #print '  key',key
            if key not in ('secs','from','op','to','reroute'):
                self.elems[elem_id][key]=message[key]

        # do sections...
        self.elems[elem_id]['sec_ids']=[]
        for sec in message['secs']:
            sec_id=sec['id']
            #print self.id,': section',sec_id,'\n',sec
            self.elems[elem_id]['sec_ids'].append(sec_id)
                
            # reroute if topology has changed:
            if self.secs.has_key(sec_id):
                if ((self.secs[sec_id]['tosec_ids']!=sec['tosec_ids'])| \
                    (self.secs[sec_id]['status']!=sec['status']) )\
                    &(message['reroute']):
                    #print '  reroute2'
                    self.reroute=1
            else:
                #print '  reroute3'
                self.reroute=1
 
            # do berth
            #print 'type',sec['type']
            #print 'status',sec['status']
            #print 'reroute',message['reroute']
            if sec['type']=='berth':
                #print 'add berth new active',elem_id
                if self.dests.has_key(elem_id): # dest already in db ?
                    # has status gone inactive ?
                    if sec['status']!='active':
                        # remove destination
                        del self.dests[elem_id]
                        #print '  reroute4'
                        self.reroute=1
                    else:
                        if sec['status']=='active':
                            # add destination
                            self.dests[elem_id]=1
                            #print '  reroute4'
                            self.reroute=1
                else:
                    if sec['status']=='active':
                        # add destination
                        self.dests[elem_id]=1
                        #print '  reroute6'
                        self.reroute=1

            #print '  Save section',sec_id,' in db'
            if self.secs.has_key(sec_id): # update section data
                self.secs[sec_id].update(sec)
            else: # create sec_id in db
                self.secs[sec_id]=sec

        
    def update_costsmatrix(self):
        """
        Computes costs between all destinations
        """
        if self.debug>=3:
            print 'Trackman: update_costsmatrix...'
            
        ## add row ind of costmatrix to dests dict
        n_dests=len(self.dests.keys())
        self.costmatrix=zeros((n_dests,n_dests))

        # first assign matrix index to each stop
        ind_orig=0
        for id_orig in self.dests.keys():
            self.elems[id_orig]['ind_costmatrix']=ind_orig
            ind_orig+=1

        # now compute the costs for each stop to each other stop
        n=0
        for id_orig in self.dests.keys():
            #print '  do',id_orig,self.elems[id_orig]['ind_costmatrix'],\
            #      'of',n_dests,self.elems[id_orig]['zone']
            print 'update_costsmatrix_stop: %.2f completed'%(100*n/len(self.dests))
            n+=1
            self.update_costsmatrix_stop(id_orig)

        #print 'self.costmatrix=\n',self.costmatrix

    def get_cost(self,id_orig,id_dest):
        """
        Returns costs to get from origin to destination. 
        """
        #print 'get_cost',self.elems[id_orig]['ind_costmatrix'],self.elems[id_dest]['ind_costmatrix'],self.costmatrix[self.elems[id_orig]['ind_costmatrix'],\
        #                      self.elems[id_dest]['ind_costmatrix']]
        if len(self.costmatrix)>0:
            return self.costmatrix[self.elems[id_orig]['ind_costmatrix'],\
                                   self.elems[id_dest]['ind_costmatrix']]
        else:
            #print 'WARNING in',self.id,':costmatrix not assigned'
            #print '  => returning random value'
            return uniform(1,600)
    
    def update_costsmatrix_stop(self,id_orig):
        """
        Computes costs between all destinations
        """
        #print 'update_costsmatrix_stop',id_orig,len(self.dests)
        ## add row ind of costmatrix to dests dict
        ind_orig=self.elems[id_orig]['ind_costmatrix']
        # start section is accel section of stop
        secid_orig=self.elems[id_orig]['sec_ids'][-1]

        n=0
        for id_dest in self.dests.keys():#
            ind_dest=self.elems[id_dest]['ind_costmatrix']
            #print ind_orig,ind_dest

            
            
            if (ind_dest!=ind_orig)&\
                   (self.elems[id_orig]['zone']!=\
                    self.elems[id_dest]['zone']):
                # route and save costs in matrix
                route=self.route(secid_orig,id_dest)
                #print '  secid_orig=',secid_orig,ind_orig,\
                #      'id_dest=',id_dest,ind_dest
                #print '      route=',route
                #self.costmatrix[ind_orig,ind_dest]=self.route(
                #    secid_orig,id_dest)['cumcost'][-1]
                if route:
                    self.costmatrix[ind_orig,ind_dest]=route['cumcost'][-1]
                else:
                    print 'WARNING from update_costsmatrix_stop:'
                    print '  No route between',id_orig,'and',id_dest
                    self.costmatrix[ind_orig,ind_dest]=10.0**8
                #print 'cumcost=',self.route(\
                #    secid_orig,id_dest)['cumcost']
            else:
                self.costmatrix[ind_orig,ind_dest]=1.0 # one second only
        
    
    def update_cluster(self):
        '''Complete routing from all diverges to all current destination'''
        #return
        # make a list with diverge section ids
        divsec_ids=[]
        for sec_id in self.secs.keys():
            if len(self.secs[sec_id]['tosec_ids'])>1:
                # clear routing table
                self.secs[sec_id]['routing_table']={}
                divsec_ids.append(sec_id)
        if self.debug>=3:
            print 'Trackman: Generate routing tables...this may take a while.'

        n=0
        for divsec_id in divsec_ids:
            if self.debug>=3:
                print 'update_cluster: %.2f completed'%(100.0*n/len(divsec_ids))
            n+=1
            for dest_id in self.dests.keys():
                #print '\n\n********************************************'
                #print 'route from',divsec_id,' to ',dest_id,':'
                if not self.secs[divsec_id]['routing_table'].has_key(dest_id):
                    
                    theroute=self.route(divsec_id, dest_id)
                    #if theroute:
                    #    print '\n Theroute = ',theroute['sec_ids']
                    
                    if not theroute:
                        divsec_id_error=divsec_id
                        dest_id_error=dest_id
                        #'!!!!!!!!!! Routing unsuccessful !!!!!!!!!'
                                        
                    else:
                        # check if there are diverge sections on this route
                        # and fill in routing table accordingly
                        for sec_id in theroute['sec_ids']:
                            if divsec_ids.count(sec_id)>0:
                                ind=theroute['sec_ids'].index(sec_id)
                                nextsec_id=theroute['sec_ids'][ind+1]
                                self.secs[sec_id]['routing_table'][dest_id]=\
                                                                   nextsec_id

                else:
                    # print 'Routing already done!!'
                    pass

        # reschedule carrier because net topology may have changed:
        # send a message to all track elements and poll all carrier.
        # the elements will send back a message, calling the
        # reschedule_cars method with a directory that
        # maps carrier ids to section ids.
        # 
        #print 'update_cluster:'        
        for elem_id in self.elems.keys():
            # print '  poll for car',elem_id
            self.send_mes({'to': elem_id,
                           'op':'poll_cars',
                           })

        # inform usermanagements about currently availlable destinations
        # by returning info from the permanent get_destination task
        if self.userman_task_getdests:
            #print '!!!userman_task_getdests.return_info'
            self.userman_task_getdests.return_info()
        else:
            print 'WARNING:  no userman_task_getdests!!'

        # make/update matric with costs between all possible OD pairs
        self.update_costsmatrix()
        
    def route(self,sec_id,dest_id):

        # define initial route at initial section
        routes=[{'sec_ids': [sec_id], # sequence of sections
                 'cumcost': [0.0],    # cummulative costs
                 'length':  0.0,      # total route length
                 'arrived': 0         # =1 if route has arrived at destination
                 }]
        
        terminated=0
        route_arrived=None
        while not terminated:
            newroutes=[]
            for route in routes:
                #print '--------------- Route-section ids:\n',route['sec_ids']

                #print ' get current  section id',route['sec_ids'][-1]
                sec_id=route['sec_ids'][-1]

                # get current costs
                cost=route['cumcost'][-1]
                    
                # check first whether this route deserves continuation...

                cont=1 # be positive !

                # discontinue if the route closes a loop AND
                # if we are not yet arrived 
                if (route['sec_ids'].count(sec_id)==2) &\
                   (not route['arrived']):
                    #print '!!!!!!!!!! Loop'
                    cont=0 # discontinue
                else:
                    # compare route with other routes
                    for comproute in routes:
                        if comproute!=route:
                            # comproute contains the future sections of route? 
                            if comproute['sec_ids'].count(sec_id)>0:
                                ind=comproute['sec_ids'].index(sec_id)
                                # and had lower costs at this section ?
                                if comproute['cumcost'][ind] < cost:
                                    #print '!!!!!!!!!! Part of other route'
                                    cont=0 # discontinue

                            # comproute did arrive and has lower costs &
                            if comproute['arrived'] &\
                               (comproute['cumcost'][-1] <cost):
                                #print '!!!!!!!!!! Other route arrived'
                                cont=0 # discontinue
                                
                        if not cont:
                            break # terminate loop
                        
                #print '\n .......... Cont=',cont,'\n'
                
                if cont:
                    if route['arrived']:
                        #print '  -----------> Arrived'
                        # take over old route, we already arrived
                        newroutes.append(copy.deepcopy(route))
                        
                    else: # make next step
                        #print '  -----------> Iterate......'

                        # get ids of next section(s)
                        if self.secs.has_key(sec_id):
                            nextsec_ids=self.secs[sec_id]['tosec_ids']
                        else:
                            nextsec_ids=[]

                        # generate a new route  for each new branch
                        for nextsec_id in nextsec_ids:
                            #print 'create new route for nextsec_id',nextsec_id                         
                            # make sure that nextsec is in sections table
                            if (self.secs.has_key(nextsec_id)):
                                
                                #print ' first make a copy of old route'
                                newroute=copy.deepcopy(route)

                                # then iterate...
                                newroute['sec_ids'].append(nextsec_id)
                                    
                                # add length
                                newroute['length']=newroute['length']+\
                                               self.secs[nextsec_id]['length']
                            
                                # add costs
                                newroute['cumcost'].append(
                                    cost+self.cost(nextsec_id))

                                # check whether we are arrived
                                if (self.elems[dest_id]['sec_ids'].count(
                                    nextsec_id)==1)&\
                                    (self.secs[nextsec_id]['type']=='berth'):
                                    newroute['arrived']=1
                                    route_arrived=newroute
                                    
                                                                
                                newroutes.append(newroute)
                                #print '= Newroutes=\n',newroutes,nextsec_id,self.secs[nextsec_id]['length'] 
                            else:
                                #print '!!!!!! No continuation for ',nextsec_id
                                pass
                            
            #print '\n=======> newroutes:\n',newroutes
            #print 'routes:',routes              
            

            # check termination criteria...

            # no route can be found
            if len(newroutes)==0: 
                terminated=1
                theroute=None

            # there is only one route left, which has arrived
            #elif (len(newroutes)==1) & newroutes[0]['arrived']:
            elif route_arrived:
                terminated=1
                theroute=route_arrived

            
            else:
                #print '  do another iteration',len(newroutes)
                routes=copy.deepcopy(newroutes)
                    
        return theroute
    
    def reschedule_cars(self,message):
        """
        Call the scheduling procedure of the task of all carriers
        in the cars_sec_ids directory. The cars_sec_ids directory must map the
        carrier id to its current section id.
        """
        sec_ids=message['car_sec_ids']
        for car_id in sec_ids.keys():
            # has carrier a task?
            #print 'reschedule carrier',car_id,'from',sec_ids[car_id]
            if self.car_tasks.has_key(car_id):
                #print '  found task!'
                # tell task to reschedule carrier from currect section
                self.car_tasks[car_id].schedule(sec_ids[car_id])
                
            
    def cost(self,sec_id):
        '''
        Function to compute the additional path-costs of this section.
        This function is called by the routing algorithm.
        '''

        # this is the estimated travel time for this section,
        # assuming the carrier has always line speed.
        return self.secs[sec_id]['length'] / \
               self.secs[sec_id]['velocity_line']
        

    def link(self):
        """
        Here we check simply whether it is necessary to reroute
        """
        #print 'link in',self.id,len(self.costmatrix)
        self.reroute=0
        if len(self.costmatrix)==0:
            self.reroute=1
        #else: 
        #    stops=self.sim.get_stops()
        #    for id_dest in self.dests.keys():
        #        if not stops.has_key(id_dest):
        #            self.reroute=1

                
        
    def output(self):
        #print '--------------------------- UPDATE',self.type,self.reroute
        #print '  dests:',self.dests
        #print '  Messages:\n',self.messages
        #print '  task_classes=\n',self.task_classes
        self.read_mes()
        
        # if there have been changes to the topology
        # or availlability of branches the update the whole cluster
        # and inform all track-elements of these changes
        if self.reroute:
            self.update_cluster()
            self.reroute=0
            #print '*********************************************************'
            #for sec in self.secs.values():
            #    print sec['elem_id'],sec['id'],'------------'
            #    print '  tosec_ids:',sec['tosec_ids']
            #    if sec.has_key('routing_table'):
            #        print '  routing_table',sec['routing_table']
            #        print '\n'
        #print 'Messages after...:',self.messages 
