
"""
This script opens a predefined its simulation file,
it adds a specified number of vehicles
it adds a users according to origin-to-desination matrices (ODMs),
it runs the simulation for a specified amout of time,
it safes simulation and export results, if desired.

The number of vehicles and the names of the ODMs are specified in the
double for-loop declarations. It is assumed that the simulation file
and the file containing the ODM are in the same directory.

The script will produce a simulation file for each set of parameters.

"""
# defines file path relative to the position of this script file
# Note: under windows use double-backslash (\\) to separate directories. 
simdir='projects'

# the simulation file with the (empty) network.
simfile='fiera_track'

# load the simulator package into memory
import its

# import OS independent directory path operations
from os import path

# loop though set of first parameter
for n_cars in [100,200,300]:

    # loop through set of second parameter
    for profile in ('fiera_user_family_asym','fiera_user_commuter_peak'):

        print 'Simulate',simfile,\
              'with',n_cars,'vehicles and user profile',profile
        
        # initialize the simulator with the simulation file
        sim=its.Sim( path.join(simdir,simfile+'.its') )
        
        # make sure the track is active
        sim.activate_all()
        
        # initialize the vehicle simulation object of the desired carrier type
        object_vehicle=sim.append_obj('test.car.experim')
        
        # this instruction adds an equal number of vehicle-modules
        # to each stop of the network, using the given vehicle parameters
        # You can place the carriers to stops that belong to a specified zone
        # by modifying the zone keyword to 'maintenance' for example.
        object_vehicle.multi_add(number=n_cars,
                                 zone='everywhere',
                                 length=3.0,
                                 accel_comfort=2.5,
                                 accel_emergency=5.0,
                                 accel_failure=10.0,
                                 time_brake=0.1)

        # initialize the user simulation object
        object_user=sim.append_obj('base.user.generic')

        # and add users according to the ODM and other parameters.
        object_user.multi_add(profile_name=path.join(simdir,profile+'.txt'))
                          
        # start the simulation with the end-of-simulationtime
        # in seconds as parameter
        sim.start(time_end=600)

        # save simulation with the name extended by the parameter names
        simfile_output=simfile+'_cars'+str(n_cars)+'_'+profile
        print 'Saving',simfile_output+'.its'
        sim.save( path.join(simdir,simfile_output+'.its') )

        # if you also want a report written and saved for this simulation add
        sim.export_results(path.join(simdir,simfile_output+'.txt'))

        # before running the simulation with the next set of parameters,
        # delete the old simulation in order to safe memory
        del sim
