###############################################################################
# IMPORTS
from classlib.base import *
from LinearAlgebra import *
from RandomArray import *
from copy import *
from string import find
import Tkinter,Pmw,gui


###############################################################################
class Transport(Task):
    """

    Task that controls the transport of a carrier over a track element:
    the track management announces a carrier to this track element
    together with trip information.  The task replies when this carrier:
    (i) left this track element or (ii) arrived at a stop, in case this
    track element is the destination.

    """
    
    def init_task(self,args):
        """
        Inserts carrier id into the carrier data base.
        Then the task waits untill the carrier left this track element
        or arrived at its destination.
        """
        self.car_id=args['car_id']
        
        # no more used self.initelem_flag=args['initelem_flag']
        
        
        # insert task into car_tasks table with carrier id as key
        # if there is already a task, just append the task to the list
        if not self.obj.car_tasks.has_key(self.car_id):
            self.obj.car_tasks[self.car_id]=[self]
        else:
            self.obj.car_tasks[self.car_id].append(self)


        self.set_status('processing')

        # now lets store the parameters that tell us what to do with the car
        self.target_flag=args['target_flag']
        self.nextsec_ids=args['nextsec_ids']
        #print 'nextsec_ids=',self.nextsec_ids
        #if args['car_id']=='test.car.experim/230':
        #    print 'Trans task',self.id,
        #    'at',self.obj.id,'ord',self.order_id#,'schedule',self.car_id
        #    print '  nextsec_ids=',self.nextsec_ids
        #
        self.mode=args.get('mode','loaded')

        
    def deviatedto_sec(self,sec,nextsec):
        """
        
        This function is called by the respective section when carrier
        wanted to enter a blocked section, but instead, it has been
        deviated to section nextsec.

        Currently, the task is simply declaired in error, which will
        trigger rescheduling of the carrier.
        """
        #print self.obj.id,'Task',self.id
        #print '  Deviate',self.car_id,'at',self.obj.id,'to',nextsec.id
        self.info['action']='deviate'
        self.info['nextsec_id']=nextsec.id
        self.set_status('error') # task got an error

    def get_nextsec(self,sec_id):
        """
        This function is called by the respective section when the concerned
        carrier left this section.

        Currently, the task is simply declaired as done.
        """
        if self.nextsec_ids.has_key(sec_id): # is section in table ?
                nextsec_id=self.nextsec_ids[sec_id]
                #print 'Task',self.id,\
                #      '.get_nextsec',nextsec_id,'at',self.obj.id
                #print '     for',self.car_id
                
                # take out this nextsec [probably not necessary??,just confuse]
                del self.nextsec_ids[sec_id]

                return nextsec_id
            
        else: # section not found by this task
            return -1
        
    def leave_sec(self,sec):
        """
        This function is called by the respective section when the concerned
        carrier left this section.

        Currently, the task is simply declaired as done.
        """
    
        # did carrier leave also the track element ?
        if sec.tonode.has_key('direction'):
            if (sec.tonode['direction']=='out'):

                # declaire task as done only if...
                if (len(self.nextsec_ids)==0)& (not self.target_flag):
                    self.info['action']='exit'

                    #print self.car_id,'exits at',self.obj.id,\
                    #      'Sec',sec.id,'target_flag',self.target_flag
                
                    # terminate task
                    self.set_status('done')

                

    def haltin_sec(self,sec):
        """
        This function is called by the respective section when the concerned
        carrier halted in section sec.

        Currently, the task is simply declaired as done.
        """
        #print self.car_id,'halted in',self.obj.id,'Sec',sec.id

        if self.target_flag & (sec.type=='berth'):
            self.info['action']='halt'
            self.set_status('done') # task is done
        #else:
            #print '...element internal change of sections'


    def cleanup(self):
        """
        This function cleans up all memory traces of this task in
        the respective management.
        """
        #if self.obj.id=='test.track.diverge_flexible/1':
        #    print 'cleanup',self.id,'at',self.obj.id,'ord',self.order_id,\
        #          self.car_id
        if self.obj.car_tasks.has_key(self.car_id):
            car_tasks=self.obj.car_tasks[self.car_id]
            if car_tasks.count(self)>0:
                #print '  remove task from car_tasks'
                car_tasks.remove(self)
                if len(car_tasks)==0:
                    del self.obj.car_tasks[self.car_id]
      

    #def kill(self):
    #    """
    #    Start to kill all running processes. Task will remain untill all
    #    processes have terminated successfully.
    #    May be overwritten by specific management.
    #    """
    #    
    #    if self.procs:
    #        self.set_status('kill')
    #        for proc in self.procs.values():
    #            proc.kill()
    #    else:
    #        print 'Kill',self.id
    #        self.set_status('killed')

   
class Sec:
    """
    Contains all information of a specific section. The methods are common
    to all section and linked to functions that are implemented in Netsim.
    """
    def __init__(self,elem,defaultlocalsec):

        localsec=deepcopy(defaultlocalsec)
        self.shape, self.param = localsec['shape']

        # has this section a particular label,
        # defined in the make_track function of the track element
        if localsec.has_key('label'):
            self.label=localsec['label']
        else:
            self.label=''

        if localsec.has_key('velocity_line'):
            self.velocity_line=localsec['velocity_line']
        else:
            self.velocity_line=elem.obj.velocity_line

        if localsec.has_key('car_max'):
            self.car_max=localsec['car_max']
        else:
            self.car_max=int(10.0**4) # just a high integer
            
        if localsec.has_key('type'):
            self.type=localsec['type']
        else:
            self.type='io' # normal input/output line (no store or berth)

        # the clearing zone is a zone at the beginning of the section,
        # where there can be only one carrier in diverge elements.
        # This means in this zone there is an overlap with a diverging section.
        if localsec.has_key('clearingzone'):
            self.clearingzone=localsec['clearingzone']
        else:
            self.clearingzone=10.0 # meter

        # minimum length that carriers must be separated
        #if localsec.has_key('length_carsep'):
        #    self.length_carsep=localsec['length_carsep']
        #else:
        #    self.length_carsep=0.2
        #    # should not depend on where the car is!!
        #    #if self.label=='berths':
        #    #    self.length_carsep=0.2
        #    #else:
        #    #    self.length_carsep=1.0
        
        fromnode_id,tonode_id=localsec['topo'] # get local node ids
        self.fromnode=elem.nodes[fromnode_id]
        self.tonode=elem.nodes[tonode_id]

        self.elem=elem         # save track element,
        self.netsim=elem.obj.netsim # netsim
        self.localsec=localsec # and the local section data (still necessary??)

        self.status='passive'
        self.segs=None  # Multidimensional array with geometry info of segments

        # Lists with connected sections
        self.fromsecs=[]    
        self.tosecs=[]      

        # Lists with connected active sections
        self.fromsecs_active=[]    
        self.tosecs_active=[]

      
        # carrier FIFO buffer (also list) with all carriers on this section
        self.cars=[]

        
        # will be determined by Netsim
        self.id=-1      #  unique id within Netsim
        self.seg_ind=0 #  index to first segment in Netsim's segs multi-array

        # this will compute the transformation matrices of each segment
        self.update()

        # initialize statistics
        self.reset_statistics()

    def reset_statistics(self):
        """
        Reset or initialize statistical variables
        """
        # total number of cars entred this section
        self.cars_entered=0

        # average time headway when they cars enter this section
        self.headway_mean=0.0

        # number over how many samples the headway average is computed
        headway_average_number=100
        self.headway_mean_factor=1.0/headway_average_number

        
        
        # sum of number of cars * time interval
        # this serves to compute the average occupation which is
        # self.cars_time/time
        self.cars_time=0.0

        # time instant the last car entered. Here negative to indicate
        # that no car has ever entered.
        self.time_entered=-1.0
        
    def update_statistics(self,*args):
        """
        Update statistical parameters.
        """
        
        self.headway_mean_factor=args.get('headway_mean_factor',
                                          self.headway_mean_factor)
        
        
    def start(self,car,mode):
        """
        Start this carrier.
        """
        controldata={}


        if car.status=='active':
            # set default desired carrier velocity
            if not self.tosecs_active: # no active connected section ?
                controldata['velocity_line']=0.0 # brake carrier

            elif self.tosecs_active[0].type=='berth':# next section is a berth?
                # brake carrier down to speed defined by this berth section
                controldata['velocity_line']=\
                              self.tosecs_active[0].velocity_line
            else:
                controldata['velocity_line']=self.velocity_line
                
            # pass control data through the track element wide
            # control algorithm.
            # !!! for future improvements: it would be more realistic
            # if only car.id and mode were passed to the element-wide control
            self.elem.control_car_on_start(self,car,mode,controldata)
         
            # Low-level control of carrier
            #car.control(controldata)

            
    def enter(self,car):
        '''Carrier enters this section. Returns control data for carrier'''

        
        
        if car.status=='active':
            # print car.id,'entering',self.elem.id,self.id
            controldata={'velocity_line':
                         self.velocity_line}

            # cases when car can get new leader 
            if (len(self.tosecs_active)>1)|(len(self.fromsecs_active)):
                controldata=self.search_car_infront(car,controldata)

            if not self.tosecs_active: # no active connected section ?
                controldata['velocity_line']=0.0 # brake carrier

            elif self.tosecs_active[0].type=='berth':# next section is a berth?
                # brake carrier down to walking speed
                controldata['velocity_line']=\
                              self.tosecs_active[0].velocity_line

            # insert carrier in sections carrie FIFO buffer 
            self.cars.append(car)
        
            # pass control data through the track element wide
            # control algorithm.
            controldata = self.elem.control_car_on_enter(self,car,controldata)

            # Low-level control of carrier
            car.control(controldata)

            # do statistics
            self.cars_entered+=1
            if self.time_entered>0.0:
                time_headway=self.elem.obj.sim.time-self.time_entered
            else:
                # no car has ever entered this section.
                # set some initial estimation
                time_headway=self.length*self.velocity_line
                
            self.headway_mean=(1-self.headway_mean_factor)*self.headway_mean+\
                               self.headway_mean_factor*time_headway

            self.cars_time+=time_headway*len(self.cars)
            #if required, the average car occupation = self.cars_time/time

            self.time_entered=self.elem.obj.sim.time

        else:
            # strange, however, insert carrier in sections carrie FIFO buffer 
            self.cars.append(car)

        # may be worth doing in the future...
        # insert car also in element-wide carrier database
        #if self.tonode.has_key('direction'): # entered element?
        #    if self.tonode['direction']=='in':
        #        self.elem.cars[car.id]={'mod':car,'sec':self}
        #    else: # update sec
        #        self.elem.cars[car.id]['sec']=self
        #else: # update sec
        #    self.elem.cars[car.id]['sec']=self
            
        
    def leave(self,car):
        '''Carrier leaves this section. Return next section'''
        #if self.elem.id=='test.track.stop10berth_rightside/1':
        #    print 'leave:',self.label,car.id
            
        if self.cars:
            self.cars.remove(car)

            # will be removed, because we don't block any more
            
            # decide if this section must reopen input
            # because there is no more overflow:
            #if len(self.cars)==self.car_max-1:
            #    for sec in self.fromsecs:
            #        sec.unblock_tosec(self)

        # call element wide control function for carrier
        controldata = self.elem.control_car_on_leave(self,car,{})
        
        # Low-level control of carrier
        car.control(controldata)

        # remove car from element-wide carrier database
        #if self.tonode.has_key('direction'): # left element?
        #    if self.tonode['direction']=='out':
        #        del self.elem.cars[car.id]
        

        # determine next section...
        if len(self.tosecs_active)==1: # only one active nextbranche ?
            return self.tosecs_active[0], '0' # use the only active section
        
        elif len(self.tosecs_active)>1: # more than one branch ?
            # call element to ask for a decision
            nextsec,error=self.elem.get_nextsec(self,car)
            return nextsec,error
        
        else:
            
            # if there is no active next section move the carrier back
            # to the beginning of this section.
            # This is just a simple solution to avoid an error
            #print 'Dead end',self.id
            return self, 'no_sec'


    def atmark(self,car):
        '''
        Carrier when crossing the mark point of this  section.
        '''
        # inform track element which must control carrier
        # print 'sec.atmark:',self.elem.id,self.id,car.id
        controldata={}
        self.elem.control_car_at_mark(self,car,controldata)

        # Low-level control of carrier
        car.control(controldata)

        
    def is_congested(self,threshold=0.2,startsec=None):
        """
        Tries to detect whether section is congested.
        """
        # here the wy to detect congestion is to compare the car's 
        # actual velocity with the line velocity
        # for speed reasons use just last car of queue as sample
        #print '   -----  is_congested',self.id
        if len(self.cars)>0:
            congested=(self.cars[-1].get_velocity()/self.velocity_line) < threshold
            #print '           ',self.cars[-1].get_velocity(),self.velocity_line,\
            #        (self.cars[-1].get_velocity()/self.velocity_line),congested
        else:
            congested=False
            #print '           No car'
        
        if not congested:
            
            # continue checking downstream
            
            if (len(self.tosecs_active)>1)|(self==startsec):
                # stop checking whether next section is congested
                return False
                
            else:
                # now check downstream
                if startsec==None:
                    startsec=self
                if len(self.tosecs_active)==1:
                    return self.tosecs_active[0].is_congested(threshold,startsec)
                else:
                    # no next active section (dead end)
                    return True
        else:
            return True  
        
        
        
        
    def search_car_infront(self,car,controldata,cumuldist=0,secs_total=0):
        """
        Search for the next carrier in front in the network
        and generate control data for the given carrier.
        """
        #IDEA: use minimum comfort braking dist of car as limit for search
        
        buffer_ind=0
        car_itself_flag=self.cars.count(car)
        if car_itself_flag:
            buffer_ind=self.cars.index(car)
        
        if buffer_ind>0: # this carrier is not the first car in section
            # car in front is on the same section in front
            car_infront=self.cars[buffer_ind-1]
            # compute distance between carriers
            controldata['dist_infront']=\
                               car_infront.get_path_sec()-\
                               car_infront.length-\
                               car.get_path_sec() #-self.length_carsep
            controldata['car_infront']=car_infront
            return controldata

        # carrier in front is in this section and it is not this car itself?
        elif (len(self.cars)>0)&(not car_itself_flag): 
            # compute distance between carriers
            controldata['dist_infront']=cumuldist+\
                                   self.cars[-1].get_path_sec()-\
                                   self.cars[-1].length-\
                                   car.get_path_sec() #-self.length_carsep
            controldata['car_infront']=self.cars[-1]
            return controldata

        else: # search carrier in front in the next active sections 
            controldata['car_infront']=None
            if self.tosecs_active: # is there any connected active section ?
                # look in connected active sections... 
                # if still in the range of braking distance
                cumuldist=cumuldist+self.length

                if secs_total<=4:
                    mindist=10.0**10
                    car_infront_closest=None
                    dist_closest=0.0
                    
                    for sec in self.tosecs_active:
                        data=sec.search_car_infront(car,controldata,
                                                    cumuldist,secs_total+1)
                        
                        if data['car_infront']:
                            
                            car_infront=data['car_infront']
                            dist_crit=data['dist_infront']+\
                                 car_infront.get_brakedist_failure()
                            #print '  crit:',dist_crit,mindist
                            if dist_crit<mindist:
                                car_infront_closest=car_infront
                                dist_closest=data['dist_infront']
                                mindist=dist_crit
                                

                        controldata['car_infront']=car_infront_closest
                        controldata['dist_infront']=dist_closest

            else:
                # dead end in sight -> set target line speed to cawl speed
                controldata['velocity_line']=0.5

        return controldata
                            
    def get_root(self):
        """
        Returns a list [x,y,z] with the root coordinates of this section,
        this means the coordinates of the first segment.
        """
        return ravel(self.segs[0,:3,3]).tolist()

    def polar(self,dx,dy):
        """
        Returns correct distance and angle in rad.
        """
        if dx==0.0:
            angle=sign(dy)*pi/2
        else:
            angle=arctan(dy/dx)
            if dx<0.0:
                angle+=pi
            
        return sqrt(dx**2+dy**2),angle
    
    def transform_polyline(self,poly):
        """
        Transforms a polyline, given by the vector
        poly=[x1,y1, x2,y2, ... ] 
        into the coordinate system of this section, which means
        the coordinate system of the first segment of this section.
        """
        p=array(poly)
        x=p[0:len(p):2]
        y=p[1:len(p):2]

        segs=self.transform_segs(self.segs[0], self.xyz_to_segs(x,y))
        #segs= transpose(matrixmultiply(
        #    self.segs[0],self.xyz_to_segs(x,y)),(1,0,2))

        return ravel(segs[:,:2,3]).tolist()

    def start_stretch_gui(self,event,node_id):
        """
        Initializes callbacks to stretch nodes.
        
        """
        #print 'start_stretch_node_gui at node',node_id
        self.editnode_id=node_id
        self.elem.obj.netmon.set_action('stretch',
                               callback_click = self.stretch_place,
                               callback_motion = self.stretch_motion,
                               callback_pressmotion = self.stretch_motion,
                               callback_rightclick=self.elem.obj.move_end
                               )
        
    def stretch_motion(self,event):
        canvas=event.widget
        if canvas==self.elem.obj.netmon.canvas:

            # Coordinates x,y are the coordinates at 100% magnification.
            # The magnified coordinates are event.x, event.y
            x,y=self.elem.obj.netmon.get_unmagnified_canvas_coords(event)


            # compute coordinates in the coordinate system of the
            # track element

            phi=0.0*pi/180
            editnode_trans=array([[cos(phi), -sin(phi),    0, x],
                                  [sin(phi),  cos(phi),    0, y],
                                  [0,                0,    1, 0.0],
                                  [0,                0,    0, 1]])

            
            trans0=matrixmultiply(inverse(self.elem.trans),editnode_trans)


            #print '\n\n------------------------------------------------------'
            #print 'editnode_trans=\n',editnode_trans
            #print 'elem.trans=\n',self.elem.trans
            #print 'inverse(self.elem.trans)=\n',inverse(self.elem.trans)
            #print 'trans0=\n',trans0
            #print 'elem.trans*trans0=\n',matrixmultiply(self.elem.trans,
            #                                            trans0)
            
            
            if self.stretch(trans0):
                self.elem.plot()
                
            #x0,y0=matrixmultiply(inverse(self.elem.trans),v)[0:2,0]
            
    def stretch_place(self,event):
        """
        Method, called when the track element is placed after the section
        has been stretched.
        """
        canvas=event.widget
        #if canvas==self.elem.obj.netmon.canvas:
        x=canvas.canvasx(event.x)
        y=canvas.canvasy(event.y)

        #print 'stretch_node_place to:',x,y
        #print '  trans editnode=\n',self.elem.nodes[self.editnode_id]['trans']
        self.elem.obj.netsim.update_elem(self.elem)
        # get the global open node that is closest to the global node,
        # corresponding to the edit node of track element
        node_closest=self.elem.obj.netsim.get_closest_node(
            self.elem.nodes_open[self.editnode_id])

        if node_closest:
            #print '  trans closestnode=\n',node_closest.localnode['trans']
            trans0=matrixmultiply(inverse(self.elem.trans),
                                  node_closest.localnode['trans'])
            self.stretch(trans0)

        self.elem.obj.netsim.update_elem(self.elem)
            
        if self.editnode_id==self.elem.rootnode:
            #self.elem.set_root(self.editnode_id,[x,y,0.0],0.0)
            self.elem.set_root(self.editnode_id,trans=
                               self.elem.nodes[self.editnode_id]['trans'],
                               swapnodes=1)
                               

        

        #print 'stretch_node_place trans=',\
        #      self.elem.nodes[self.editnode_id]['trans']

        
        
        #self.elem.set_root(self.editnode_id,trans=identity(4))
        #self.elem.nodes[self.elem.rootnode]['trans'])
        

        # put track element module into passive mode
        self.elem.set_status('passive')
        self.elem.plot()

        
        #print '  rootnode  trans0=\n',\
        #      self.elem.nodes[self.elem.rootnode]['trans0']
        #print '  rootnode  trans=\n',\
        #      self.elem.nodes[self.elem.rootnode]['trans']
            

        self.elem.obj.netedit_gui.move_mod()
        
    def stretch(self,trans):
        """
        Stretch local node with id  self.editnode_id to match
        the transformation trans, which is in the coordinate system
        of the track element.

        Return 0 if no stretching is possible for this target transform,
        otherwise 1.

        To be overwritten by specific line class.
        """
        return 0
            
    def set_velocity_line(self,v):
        """
        Sets velocity on this section.
        """
        self.velocity_line=v
        
    def update(self):
        #print 'sec.update',self.id,self.elem.id

        if self.elem.status=='edit':
            # calculate trasformation matrices of each segment
            #self.elem.obj.shape_functions[self.shape](self)
            self.update_segs()
            
            

    def update_segs(self,param=None):
        """
        Method to call if the geometry of the segments need to be
        transformed or regenerated with different parameters.
        """
        
        if not param:
            param=self.param
            
        # transform segments in final coordinates
        segs0=self.make_segs(param)
        if len(segs0)>0:
            self.segs0=segs0
            self.param.update(param)
            self.segs=self.transform_segs(self.elem.trans, self.segs0)

            # compute geometry of nodes
            self.make_connodes_geom()
        
            # determin section length by adding up the length of each segment
            self.length=sum(
                sqrt(sum((self.segs[1:,:3,3]-self.segs[:-1,:3,3])**2,(1)))
                )
            
            # total number of segments
            self.segs_total=len(self.segs)-1
            
            # length per segment, assuming equally spaced segments
            self.seg_length=self.length/self.segs_total

            return 1
        return  0 # failed to update with given parameters
    
    def transform_segs(self,trans,segs):
        """
        Transforms all segments in the segs array of 4x4 transformation
        matrices with the 4x4 transformation matrix trans.
        """
        self._trans_segs=trans
        return array(map(self._transform_seg,segs))
        
    def _transform_seg(self,seg):
        """
        Transform a single segment, expressed in 4x4 matrix.
        """
        #print 'trans_seg-----------------------'
        #print seg
        return matrixmultiply( self._trans_segs,seg)
        #print seg
        
    def make_segs(self,param):
        """
        Generation of the 3D segment data.
        Use parameters in self.param.
        The coordinate system is the one of the track element.
        The calling fuctions take care of
        the necessary transformations to make the track element fit
        into the network.
        """
        
        x,y,z=param['offset']
        length=param['length']
        
        fromseg=self.points_to_trans( (x-1,y,z), (x,y,z) )
        toseg=self.points_to_trans( (x,y,z),(x+length,y,z) )

        return concatenate((array([fromseg]),array([toseg])))

        

    def xyz_to_segs(self,x,y,z=None,phi=None):
        if not z:
            z=0.0*zeros(x.shape)
        if not phi:
            phi=0.0*zeros(x.shape)

        segs=0.0*zeros((len(x),4,4))
 
        for i in range(len(x)):
            segs[i]=array([[cos(phi[i]), -sin(phi[i]),  0.0, x[i]],
                           [sin(phi[i]),  cos(phi[i]),  0.0, y[i]],
                           [0.0,                  0.0,  1.0, z[i]],     
                           [0.0,                  0.0,  0.0, 1.0   ]])

        return segs
    
    def points_to_trans(self,frompoint,topoint):
        """ computes a transformation matrix out of two point (x,y,z)"""
        x,y,z=topoint
        x0,y0,z0=frompoint
        dx=x-x0
        dy=y-y0
        l=sqrt(dx**2+dy**2)
        return  array([[dx/l, -dy/l,  0,x],
                       [dy/l,  dx/l ,0,y],
                       [0,     0,    1,0],
                       [0,     0,    0,1]])

       
    
        
    def make_connodes_geom(self):
        """
        Checks whether this section has connecting nodes. If so, the
        position and orientation of those are computed from the
        extremities of the segments.
        """
        if self.fromnode.has_key('direction'):
            #if segs0:
            self.fromnode['trans0']=self.segs0[0]

            self.fromnode['trans']=self.segs[0]
            
        if self.tonode.has_key('direction'):
            #if segs0:
            self.tonode['trans0']=self.segs0[-1]
            
            self.tonode['trans']=self.segs[-1]
            

    
    def activate(self):
        status_old=1*self.status
        if self.status=='passive':
            # set active only if both ends are connected somewhere
            if not ((self.fromsecs==[]) | (self.tosecs==[])):
                # try to set status to active
                self.status='active'

                if self.elem.obj.netsim.activate_sec(self):
                    # did not work, set status back to old value
                    self.status=status_old
                    return 1
                else:
                    # did work, tell to neighbours that section is active
                    for sec in self.fromsecs:
                        sec.unblock_tosec(self,permanent=1)
                    for sec in self.tosecs:
                        sec.unblock_fromsec(self,permanent=1)
                    return 0
            else:
                return 1
        
        elif self.status=='active':
            return 0
        else:
            return 1

    def inactivate(self):
        if self.status=='active':
            if self.elem.obj.netsim.inactivate_sec(self):
                return 1
            else:
                self.status='passive'
                # tell to neighbours that section is inactive
                for sec in self.fromsecs:
                    sec.block_tosec(self,permanent=1)
                for sec in self.tosecs:
                    sec.block_fromsec(self,permanent=1)
                return 0
            
        elif self.status=='passive':
            return 0
        else:
            return 1
    
    def unblock_tosec(self,tosec,permanent=0):
        '''
        This function is called by the neighbour sections when
        they get activated.
        '''
        if self.tosecs_active.count(tosec)==0:
            self.tosecs_active.append(tosec)

        # call track element wide control function
        self.elem.unblock_tosec(self,tosec,permanent)
        
    def block_tosec(self,tosec,permanent=0):
        '''
        This function is called by the neighbour sections when
        they get inactivated.
        '''
        if self.tosecs_active.count(tosec)==1:
            self.tosecs_active.remove(tosec)

        # brake down all carriers on this section
        # if there is no more active tosection
        if not self.tosecs_active:
            for car in self.cars:
                car.control({'velocity_line':0.0})

        # call track element wide control function
        self.elem.block_tosec(self,tosec,permanent)
        

    def unblock_fromsec(self,sec,permanent=0):
        '''
        This function is called by the neighbour sections when
        they get activated.
        '''
        if self.fromsecs_active.count(sec)==0:
            self.fromsecs_active.append(sec)

    def block_fromsec(self,sec,permanent=0):
        '''
        This function is called by the neighbour sections when
        they get inactivated.
        '''
        if self.fromsecs_active.count(sec)==1:
            self.fromsecs_active.remove(sec)



class Linesec(Sec):
    def make_segs(self,param):
        """
        Generation of the 3D segment data.
        Use parameters in self.param.
        The coordinate system is the one of the track element.
        The calling fuctions take care of
        the necessary transformations to make the track element fit
        into the network.
        """
        x,y,z=param['offset']
        length=param['length']
        
        fromseg=self.points_to_trans( (x-1,y,z), (x,y,z) )
        toseg=self.points_to_trans( (x,y,z),(x+length,y,z) )

        #print '\n----make line'
        #print 'fromseg',fromseg
        #print 'toseg',toseg
        
        return concatenate((array([fromseg]),array([toseg])))

    def stretch(self,trans0):
        """
        Stretch local node with id  self.editnode_id to match
        the transformation trans0, which is in the coordinate system
        of the track element.
        
        Return 0 if no stretching is possible for this target transform,
        otherwise 1.
        """
        
        param=deepcopy(self.param)
        x0,y0=trans0[0:2,3]
        #print 'stretch_node to',x0,y0
        x_offset,y_offset,z_offset=param['offset']

        if self.tonode['id']==self.editnode_id:
            length=x0-x_offset
            if length>1.1*self.clearingzone: 
                param['length']=length
                if self.update_segs(param):
                    self.elem.obj.sim.messagebar.message(
                        'userevent',
                        'Section length=%02dm'%(length))
                    return 1
                return 0
            return 0
                    
        elif self.fromnode['id']==self.editnode_id:
            length=x_offset-x0+param['length']
            if (length>1.1*self.clearingzone): 
                param['length']=length
                param['offset']=[x0,y_offset,z_offset]
                if self.update_segs(param):
                    self.elem.obj.sim.messagebar.message(
                        'userevent',
                        'Section length=%02dm'%(length))
                    return 1
                else:
                    return 0
            else:
                return 0
        else:
            return 0

class Scurvesec(Sec):
    def make_segs(self,param):
        """
        Generation of the 3D segment data.
        Use parameters in self.param.
        The coordinate system is the one of the track element.
        The calling fuctions take care of
        the necessary transformations to make the track element fit
        into the network.
        """
        
        x,y,z=param['offset']
        length=param['length']
        displace=param['displace']
        if param['turn']=='left':
            turn=-1.0
        elif param['turn']=='right':
            turn=+1.0
        
                    
        phi_max=pi/16
        if param.has_key('angle_max'):
            phi_max=pi*param['angle_max']/180

        number_of_segments=int(length/self.elem.obj.length_seg)
        x_step=length/number_of_segments
        #x_step=self.elem.obj.length_seg
        
        
        xx_multiplier=tan(turn*phi_max)/(turn*displace/pi)
        xx=arange(-length/2,+length/2+x_step,x_step)
        yy=arctan(xx_multiplier*xx)
        yy_offset=-min(yy)
        stretch_factor=turn*displace/max(yy+yy_offset)
        
        
        x_curve=x+length/2+xx
        y_curve=y+stretch_factor*(yy+yy_offset)
        z_curve=z*ones(xx.shape)
        phi_curve=arctan(
            xx_multiplier*stretch_factor/(1.0+xx*xx)
            )
        
        # this is necessary to get extremities exactly straight
        phi_curve[0]=0.0
        phi_curve[-1]=0.0
        
        return self.xyz_to_segs(x_curve,y_curve,z_curve,phi_curve)

    def stretch(self,trans0):
        """
        Stretch local node with id  self.editnode_id to match
        the transformation trans0, which is in the coordinate system
        of the track element.
        
        Return 0 if no stretching is possible for this target transform,
        otherwise 1.
        """
        param=deepcopy(self.param)
        
        x,y,z=param['offset']
        xn,yn=trans0[0:2,3]

        if self.tonode['id']==self.editnode_id:
            
            param['length']=xn-x
            param['displace']=yn-y
          
            if self.update_segs(param):
                self.elem.obj.sim.messagebar.message(
                    'userevent',
                    'Length=%02dm, Displace=%02dm'%(self.param['length'],
                                                    self.param['displace']))
                
                return 1
            else:
                return 0
        
        #elif self.fromnode['id']==self.editnode_id:
       # 
        #    length=x-xn+self.param['length']
        #    displace=(-y+yn+self.param['displace'])
            #if displace>0:
            #    self.param['turn']='left'
            #else:
            #    self.param['turn']='right'
            
        #    if (length>1.1*self.clearingzone):
        #        # (abs(displace)>1): 
        #        self.param['length']=length
        #        self.param['displace']=abs(displace)
        #        self.param['offset']=[xn,yn,z]
        #        self.update_segs()
        #        
        #        print xn,yn,displace,'displace=',self.param['displace'],\
        #              'offset=',self.param['offset'],self.param['turn']
        #        return 1
        #    return 0
        #    x,y,z=self.param['offset']
        #    dx=x-xn
        #    dy=y-yn
        #    self.param['offset']=[x-dx,y-dy,z]
        #    self.param['length']=self.param['length']+dx
        #    self.param['displace']=self.param['displace']+dy
        #  
        #    self.update_segs()
        #    self.elem.obj.sim.messagebar.message(
        #        'userevent',
        #        'Length=%02dm, Displace=%02dm'%(self.param['length'],
        #                                        self.param['displace']))
            
        #    return 1
        
        else:
            return 0
        
        
class Flexlinesec(Sec):
    def make_segs(self,param):
        """
        Generation of the 3D segment data.
        Use parameters in self.param.
        The coordinate system is the one of the track element.
        The calling fuctions take care of
        the necessary transformations to make the track element fit
        into the network.
        """
        x0,y0,z0=param['offset']
        x1,y1,z1=param['endcoords']
        alpha0=1.0*param.get('alpha0',0.0)
        alpha1=1.0*param.get('alpha1',0.0)

        #print '\n\n------------------------------------------------------'
        #print 'offset=',param['offset']
        #print 'endcoords=',param['endcoords']

        
        delta_x=1.0*(x1-x0)
        delta_y=1.0*(y1-y0)
        delta_abs,delta_angle=self.polar(delta_x,delta_y)
        
        #print '\ndelta_x,delta_y,=',delta_x,delta_y
        #print 'delta_abs,delta_angle=',delta_abs,delta_angle*180/pi

        beta0=alpha0*pi/180-delta_angle

        #if (beta0*180/pi<-90)|(beta0*180/pi>+90):
        #    # conditions too hard
        #    return array([])

        tan_beta0=tan(beta0)
        tan_beta1=tan((alpha1*pi/180-delta_angle))

        #print '\nalpha0,alpha1=',alpha0,alpha1
        #print 'beta0,beta1=',(alpha0*pi/180-delta_angle)*180/pi,\
        #      (alpha1*pi/180-delta_angle)*180/pi

        a1=tan_beta0
        a2=-(2*a1+tan_beta1)/delta_abs
        a3=(tan_beta1+a1)/delta_abs**2
        #print '==========='
        #print 'a1,a2,a3=',a1,a2,a3
        
        number_of_segments=int(delta_abs/self.elem.obj.length_seg+0.5)
        ds=delta_abs/(1.0*number_of_segments)
        
        
        
        s=arange(0.0,delta_abs+ds,ds)

        # make sulre that s has the corrent number of segments
        # and not one too much because of rounding errors
        s=s[0:number_of_segments+1]
        
        s0=ones(s.shape)
        ss=s*s
        sss=ss*s

        u_curve=s
        v_curve=a1*s+a2*ss+a3*sss
        z_curve=z0*s0

        beta_curve=arctan(a1+2*a2*s+3*a3*ss)


        segs_pre=self.xyz_to_segs(u_curve,v_curve,z_curve,beta_curve)
        
        phi=delta_angle
        
        trans= array([[cos(phi), -sin(phi),    0,x0],
                      [sin(phi),  cos(phi),    0,y0],
                      [0,                0,    1,z0],
                      [0,                0,    0,1]])

        return self.transform_segs(trans,segs_pre)
        # return transpose(matrixmultiply(trans,segs_pre),(1,0,2))
    
       
    def stretch(self,trans0):
        """
        Stretch local node with id  self.editnode_id to match
        the transformation trans0, which is in the coordinate system
        of the track element.
        
        Return 0 if no stretching is possible for this target transform,
        otherwise 1.
        """
        #print '\n\n---------------------------------------------------------'
        #print 'trans0=\n',trans0
        param=deepcopy(self.param)
        
        xn,yn=trans0[0:2,3]
        #print 'stretch_node to',xn,yn,trans0
        if trans0[0,0]==0.0:
            if sign(trans0[1,0])>0:
                alpha=0.0
            else:
                alpha=180.0
        else:
            alpha=180.0/pi*arctan(trans0[1,0]/trans0[0,0])
        #print 'alpha=',alpha
        
        #x_offset,y_offset,z_offset=self.param['offset']
        alpha0=param.get('alpha0',0)
        alpha1=param.get('alpha1',0.0)

        if self.tonode['id']==self.editnode_id:
            if abs(alpha0-alpha)<=90.0:
                param['endcoords']=[xn,yn,0.0]
                param['alpha1']=alpha
                if self.update_segs(param):
                    x0,y0,z0=self.param['offset']
                    x1,y1,z1=self.param['endcoords']
                    #self.elem.obj.sim.messagebar.message(
                    #'userevent',
                    #'Bounds: x=%02dm, y=%02dm'%(x1-x0,y1-y0))
                    return 1
                else:
                    return 0
            else:
                #print 'stretch: condition violation alpha0,alpha=',\
                #      alpha0,alpha,abs(alpha0-alpha)
                return 0
                    
        elif self.fromnode['id']==self.editnode_id:
            if abs(alpha1-alpha)<=90.0:
                param['offset']=[xn,yn,0.0]
                param['alpha0']=alpha
                if self.update_segs(param):
                    return 1
                else:
                    return 0
            else:
                return 0
        return 0
 
    
class Curvesec(Sec):
    def make_segs(self,param):
        """
        Generation of the 3D segment data.
        Use parameters in self.param.
        The coordinate system is the one of the track element.
        The calling fuctions take care of
        the necessary transformations to make the track element fit
        into the network.
        """

        x,y,z=param['offset']
        r=param['radius']
        if param.has_key('angle'):
            phi=pi*self.param['angle']/180
        else:
            phi=pi/2
            param['angle']=90

        # kuhn:phi_step=phi/(r*phi/self.elem.obj.length_seg)
        number_of_segments=int(r*phi/self.elem.obj.length_seg)

        phi_step=0.5*pi/number_of_segments

        if param['turn']=='right':
            y_center=y+r
            x_center=x
            phi0=-pi/2
            phi_vec=arange(phi0,phi0+phi,+phi_step)
            phi_vec[-1]=phi0+phi
            phi_curve=arange(0,phi,phi_step)
            phi_curve[-1]=phi

        elif param['turn']=='left':
            y_center=y-r
            x_center=x
            phi0=pi/2
            phi_vec=arange(phi0,phi0-phi,-phi_step)
            phi_vec[-1]=phi0-phi
            phi_curve=arange(0,-phi,-phi_step)
            phi_curve[-1]=-phi

        elif param['turn']=='from_left':
            y_center=y-r
            x_center=x
            phi_vec=arange(pi,pi-phi,-phi_step)
            phi_vec[-1]=pi-phi
            phi_curve=arange(phi,0.0,-phi_step)
            phi_curve[-1]=0.0

        elif param['turn']=='from_right':
            y_center=y+r
            x_center=x
            phi_vec=arange(pi,pi+phi,phi_step)
            phi_vec[-1]=pi+phi
            phi_curve=arange(-phi,0.0,phi_step)
            phi_curve[-1]=0.0
        
        x_curve=x_center+r*cos(phi_vec)
        y_curve=y_center+r*sin(phi_vec)
        z_curve=z*ones(phi_vec.shape)

        #print '\n----make curve'
        #print 'x,y,r',x,y,r
        #print 'x_center,y_center',x_center,y_center
        #print 'x_curve',x_curve
        #print 'y_curve',y_curve
        #print 'phi_vec', phi_vec
        #print 'phi_curve', phi_curve
        
        return self.xyz_to_segs(x_curve,y_curve,z_curve,phi_curve)

    def stretch(self,trans0):
        """
        Stretch local node with id  self.editnode_id to match
        the transformation trans0, which is in the coordinate system
        of the track element.
        
        Return 0 if no stretching is possible for this target transform,
        otherwise 1.
        """
        param=deepcopy(self.param)
        
        xn,yn=trans0[0:2,3]
        x,y,z=param['offset']
        r=param['radius']
        
                
        if self.tonode['id']==self.editnode_id:
            dy=y-yn
            if abs(dy)<self.clearingzone:
                if dy==0.0:
                    dy=self.clearingzone
                else:
                    dy=sign(dy)*self.clearingzone

            if dy>0:
                param['turn']='left'
            else:
                param['turn']='right'
            
            param['radius']=abs(dy)
            if self.update_segs(param):
                self.elem.obj.sim.messagebar.message(
                    'userevent',
                    'Radius=%02dm'%(abs(dy)))
            
                return 1
            else:
                return 0
                
        elif self.fromnode['id']==self.editnode_id:
            
            if param['turn'] in ('left','right'):
                if param['turn']=='left':
                    y_to=y-r
                    x_to=x+r
                elif param['turn']=='right':
                    y_to=y+r
                    x_to=x+r
                
                dy=y_to-yn

                if abs(dy)<self.clearingzone:
                    if dy==0.0:
                        dy=-self.clearingzone
                    else:
                        dy=sign(dy)*self.clearingzone
        
                if dy<0:
                    param['turn']='left'
                    param['offset']=[(x_to+dy),(y_to-dy),0]
                else:
                    param['turn']='right'
                    param['offset']=[(x_to-dy),(y_to-dy),0]
            
                param['radius']=abs(dy)


                #print 'x,y,x_to,y_to',x,y,x_to,y_to
                #print 'x,y,xn,yn',x,y,xn,yn
                #print '\n  -> offset',param['offset']
                #print '  -> radius',param['radius'],param['turn'],dy
            
                if self.update_segs(param):
                    self.elem.obj.sim.messagebar.message(
                        'userevent',
                        'Radius=%02dm'%(abs(dy)))
            
                    return 1
                else:
                    return 0
                
            elif  param['turn'] in ('from_right','from_left'):
                dy=-(y-yn)
                if abs(dy)<self.clearingzone:
                    if dy==0.0:
                        dy=self.clearingzone
                    else:
                        dy=sign(dy)*self.clearingzone
                        
                if dy<0:
                    param['turn']='from_left'
                else:
                    param['turn']='from_right'
                
                param['radius']=abs(dy)
                if self.update_segs(param):
                    self.elem.obj.sim.messagebar.message(
                        'userevent',
                        'Radius=%02dm'%(abs(dy)))
            
                    return 1
                else:
                    return 0
            

            else: # no recognized turn parameter
                return 0

        else:
            return 0
        

       

    
class Mod(Module):
    """
    Class, containing all data that are specific to a particular module
    of a Track element  (Data and methods that apply to all Track element
    modules of the same type are implemented in the Track class below).
    
    The position and orientation of the track module is determined by the
    transfromation matrix (transform) and the node (rootnode) of the track
    element.

    """
    def __init__(self,obj,
                 localnode=1,
                 pos=[0,0,0],
                 alpha=0,
                 status='edit'):

        
        # software user specified name
        self.name=''
        
        # back-link with simulation track object
        #self.obj=obj
        
        # administered by track simulation obj...
        self.rootnode=localnode
        
        self.status=status

        # defined by  add_mod function...
        self.no=-1      # module number,
        self.id='No id' # complete module id
        
        # create local nodes and sections
        self.nodes={}
        self.secs=[]

        # initialization, common to all modules
        self.init_common(obj)

        #print '__init__',self.id
        
        # get transform for some default values
        self.make_trans(pos,alpha)
        
        for localsec in obj.element: 
            
            # create nodes data base with local node id  as key
            fromnode,tonode = localsec['topo'] # get local node ids
            if not(self.nodes.has_key(fromnode)):
                node={'id':fromnode}
                if obj.nodes_connect.has_key(fromnode):
                    node['direction']=obj.nodes_connect[fromnode]
                self.nodes[fromnode]=node
                
            if not(self.nodes.has_key(tonode)):
                node={'id':tonode}
                if obj.nodes_connect.has_key(tonode):
                    node['direction']=obj.nodes_connect[tonode]
                self.nodes[tonode]=node

            # create section and append to sections list
            shape, param = localsec['shape']            
            sec=self.obj.sec_classes[shape](self,localsec)

            self.secs.append( sec )
            
        # administered by Netsim...
        self.nodes_open={}  # maps open local nodes (key) to
                            # global open nodes (value)

        
        
        #print '  redo transform for defined position'
        self.make_trans(pos,alpha)

        

        # Tasks stuff
        # create carrier->task table, with a link to the task that is
        # responsible for tha carrier that will pass by this element.
        self.car_tasks={} # controlled by Transport task, key = carrier id

        # message stuff
        # add operations that can be executed by an incoming message
        self.mes_ops['poll_cars']=self.poll_cars
        
        # link to task classes
        self.task_classes['transport']=Transport
        
        obj.add_mod(self) # add module to the respective track simulation obj
        #self.set_root(localnode,pos,alpha)

        # execute also track element specific initializations
        self.init_specific()
        
        # initialize default costs
        #self.init_costs()

        # initialize statistics
        #self.reset_statistics()

        #self.set_state(self.obj)

    def get_state_mod(self,state_image):
        '''
        This function removes all module-specific pointers from the
        state-image, see get_state function.
        To be overwritten by module-specific function.
        '''

        # remove tosec_active as it leads to an endless recursion
        # in the picle algorithm. List will be rebuild in set_state_mod
        self.get_state_mod_specific(state_image)

        if state_image.has_key('cars_noghost'):
            for secid in self.cars_noghost.keys():
                self.cars_noghost[secid]=self.sim.get_ids(\
                    self.cars_noghost[secid])
            state_image['cars_noghost']=self.cars_noghost
        
        for sec in self.secs:
            if sec.id>=0: # is it already in netsims structure
                #print '      sec:remove sec pointers from',sec.id
                sec.tosecs_active=self.get_id_list(sec.tosecs_active)
                sec.fromsecs_active=self.get_id_list(sec.fromsecs_active)
                sec.tosecs=self.get_id_list(sec.tosecs)
                sec.fromsecs=self.get_id_list(sec.fromsecs)
                sec.cars=self.sim.get_ids(sec.cars)
                
            del sec.netsim
            del sec.elem
            #del sec.cars

                

        state_image['secs']=self.secs
        
        return state_image

    def get_state_mod_specific(self,state_image):
        '''
        Same as get_state_mod, but module specific, to be overwritten.
        '''
        pass
    
    def get_id_list(self,pointer_list):
        """
        Return list with id of pointer' target.
        """
        id_list=[]
        for pointer in pointer_list:
            id_list.append(pointer.id)
        return id_list

    def get_secs_from_sec_id_list(self,sec_id_list):
        """
        Get back pointers to sections from a list of section ids.
        This requires the presence of netsim.
        """
        secs=[]
        for item in sec_id_list:
            if type(item)==types.IntType:
                secs.append(self.obj.netsim.rebuild_secpointer(item))
            else:
                secs.append(item)
        return secs
    
    def set_state_mod(self):
        '''

        This function is called during the opening process of a simulation
        file. Here goes all the stuff that gets lost during the save
        process (essentially pointers to functions).  It is also called
        during the module initialization. Hence, it is also possible to
        define atributes that has been added in the new version, but that
        need to be initialized for objects that have been saved under an
        older version.

        '''
        # define cost-functions
        #self.cost_functions={}
        #self.cost_functions['get_costs_estim']=self.get_costs_estim

        
        
        # add operations that can be executed by an incoming message
        self.mes_ops['poll_cars']=self.poll_cars
        
        # link to task classes
        self.task_classes['transport']=Transport

        # only if initialized in simulation env, not in preview
        #if self.obj.netsim: 
        #    self.rebuild_pointers()

        if string.atof(self.get_version())<0.3:
            # return more man update information
            #self.update_man()
            pass
            # introduce the element-wide cars dictionary
            # with the scope to keep track of low-level elemend-wide
            # information of the car.
            
            # The key is the car.id value is a database with varios information
            # here construct dictionary:
            #self.cars={}
            #for sec in self.secs:
            #    for car in sec.cars:
            #        self.cars[car.id]={'mod':car,'sec':sec}
            # currently abandonned as general strucrure, only used in stopmod
            # but would be useful, also for task point and other info
            
            

        # currently no extra separation...make it a car property...
        #for sec in self.secs:
        #    if sec.localsec.has_key('length_carsep'):
        #        sec.length_carsep=sec.localsec['length_carsep']
        #    else:
        #        sec.length_carsep=0.0
        #        sec.localsec['length_carsep']=sec.length_carsep
        #        #if sec.label=='berths':
        #        #    sec.length_carsep=0.2
        #        #else:
        #        #    sec.length_carsep=1.0
        
        self.set_state_mod_specific()

    def set_state_mod_specific(self):
        '''
        Same as set_state_mod, but module specific, to be overwritten.
        '''
        pass

    def rebuild_pointers(self):
        # rebuild sec links lists which is removed during saveing
        if hasattr(self,'cars_noghost'):
            for secid in self.cars_noghost.keys():
                self.cars_noghost[secid]=self.sim.rebuild_pointers(\
                    self.cars_noghost[secid])
            
        for sec in self.secs:
            if sec.id>=0:
                
                sec.tosecs_active=\
                         self.get_secs_from_sec_id_list(sec.tosecs_active)
                sec.fromsecs_active=\
                         self.get_secs_from_sec_id_list(sec.fromsecs_active)
                sec.tosecs=self.get_secs_from_sec_id_list(sec.tosecs)
                sec.fromsecs=self.get_secs_from_sec_id_list(sec.fromsecs)

                sec.cars=self.sim.rebuild_pointers(sec.cars)

                
            sec.netsim=self.obj.try_link_with('base.net.netsim')
            sec.elem=self

        # no longer needed
        # now rebuild the pointers in netsim
        #if self.obj.netsim:
        #    self.obj.netsim.rebuild_secpointers(self.secs)

        self.rebuild_pointers_specific()

    def rebuild_pointers_specific(self):
        """
        Convert ids to object pointers in specific module.
        """
        pass
                
    def init_costs(self):
        """
        Define default costs here.

        To be overwritten.
        """
        self.set_costs({'estim_init_ex':
                        {'name':'Estim. initial extra costs',
                         'table':{1:    40.0*10**3,
                                  10:   20.0*10**3,
                                  100:  10.0*10**3,
                                  1000: 5.0*10**3},
                         'mode':'write'},
                        'estim_init_pm':
                        {'name':'Estim. initial costs per meter',
                         'table':{1:    10000.0,
                                  100:  5000.0,
                                  1000: 2500.0,
                                  10000:1500.0,
                                  100000:500.0},
                         'mode':'write',
                         'currency':self.obj.sim.currency+'/m',},
                        'estim_pa_ex':
                        {'name':'Estim. extra costs per year',
                         'table':{1:   40.0,
                                  10:  20.0,
                                  100: 10.0,
                                  1000:5.0},
                         'mode':'write',
                         'currency':self.obj.sim.currency+'/m',},
                        'estim_pa_pm':
                        {'name':'Estim. costs per meter per year',
                         'table':{1:   80.0,
                                  10:  40.0,
                                  100: 20.0,
                                  1000:10.0},
                         'currency':self.obj.sim.currency+'/(ma)',
                         'mode':'write'},
                        
                        'estim_init':{'name':'Estim. initial costs',
                                      'value':0.0,# set in do_costs_current
                                      'mode':'read'},
                        'estim_pa':{'name':'Estim. costs per year',
                                    'value':0.0,# set in do_costs_current
                                    'currency':self.obj.sim.currency+'/m',
                                    'mode':'read'},
                        },
                        lifetime_estim=10)

        
    def do_costs_current_mod(self):
        """
        Calculates module spacific current costs
        and modifies the costs dictionary.
                
        """
        # quantity would here be the number of these track elements,
        # not the length of the track
        # so add up total length...could be done better in trackman...
        self.length=0.0
        for obj in self.obj.sim.objs.values():
            iddata=self.obj.sim.split_id(obj.id)
            if (iddata['category']=='track')&(iddata['family']=='test'):
                for mod in obj.mods.values():
                    if mod.man_id==self.man_id:
                        self.length+=mod.get_length()

        #print 'get_costs_estim:'
        #print '  quantity',quantity
        #print '  length of elem',self.get_length()
        #print '  total length',length

        quantity=self.get_quantity()
        self.set_cost_value('estim_init',
                            self.get_cost_value('estim_init_ex',quantity)+ \
                            self.get_cost_value('estim_init_pm',self.length)* \
                            self.get_length()
                            )
                   
        self.set_cost_value('estim_pa',
                            self.get_cost_value('estim_pa_ex',quantity)+ \
                            self.get_cost_value('estim_pa_pm',self.length)* \
                            self.get_length()
                            )

    
    def format_costs_mod(self,quantity,descriptors,table):
        """
        Do module specific cost formatting
        """
        # overwrite cost per meter stuff because quantity are the total length
        
        table['estim_init_pm']=self.get_cost_value(\
            'estim_init_pm',self.length)
        
        table['estim_pa_pm']=self.get_cost_value(\
            'estim_pa_pm',self.length)
        descriptors.append({'key':'length',
                            'name':'Length of track element',
                            'type':'real',
                            'printformat':'%.2f',
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'units':{'m':1},
                            'mode':'read',
                            })
        descriptors.append({'key':'length_total',
                            'name':'Total track length',
                            'type':'real',
                            'printformat':'%.2f',
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'units':{'m':1},
                            'mode':'read',
                            })
        
        table['length']=self.get_length()
        table['length_total']=self.length
        
    def add_man(self):
        '''
        Function to add this track element to the track management database.
        This Function is called already during initialization or changes of
        the track management.
        '''
        self.send_mes({'to':self.man_id,
                       'op':'add_elem',
                       'status':self.status})

    def del_man(self):
        '''
        Function to remove this track element fom the track management
        database.
        This Function is called when track element is going to be deleted
        from the network or changes of the track management.
        '''
        self.send_mes({'to':self.man_id,
                       'op':'del_elem'})
        
    def update_man(self,reroute=1):
        '''
        Update track management with connection info of track element.
        This function is called during activation or after topological
        changes.
        '''
        #print 'update_man:',self.id
        # composing the section data that needs to be send to the track man
        # For safety reasons we do not send objects (or pointers in general),
        # just strings, numbers or ids (also strings).
        
        secs_info=[] # a list with directories, each containg info about sec
        for sec in self.secs:
            tosecs_ids=[]
            for tosec in sec.tosecs:
                #print ' ',sec.id,'>>> tosec',tosec.id,tosec.status
                # include only active sec ids, however they may be
                # temporarily blocked (not part of the list tosecs_active)
                if tosec.status=='active': 
                    tosecs_ids.append(tosec.id) 
            #print '  append sec.id',sec.length,sec.transform_polyline((0,0))
            secs_info.append({'id':sec.id,
                              'type':sec.type,
                              'status':sec.status,
                              'elem_id':sec.elem.id,
                              'velocity_line':sec.velocity_line,
                              'length':sec.length,
                              'tosec_ids':tosecs_ids,
                              'coords_in':sec.transform_polyline((0,0)),
                              'coords_out':\
                              sec.transform_polyline((0,sec.length))
                              }
                             )
        #print 'secs_info',secs_info
        #print 'update_man',self.id
        message={'to':self.man_id,
                 'op':'update_elem',
                 'class':'default',# may be overwritten with 'stop'
                 'reroute':reroute,
                 'status':self.status,
                 'secs':secs_info}
        
        self.update_man_specific(message)
        #for key in message: print '  ',key
        self.send_mes(message)

    def update_man_specific(self,message):
        '''
        Update track management with connection info of track element.
        This function is called during activation or after topological
        changes.
        '''
        pass

    def poll_cars(self,message):
        """
        track management is wants to know the section id of all
        carriers on this track element.
        The results must be put in a directory called car_sec_ids, which
        maps the carrier id to its current section id and send back to sender.
        This will trigger a carrier rescheduling at the track management.
        """
        #print 'poll_cars at',self.id
        car_sec_ids={}
        for sec in self.secs:
            for car in sec.cars:
                #print '  found',car.id
                car_sec_ids[car.id]=sec.id
                
        if car_sec_ids:
            #print '  send_mes with carrier',len(car_sec_ids)
            self.send_mes({'to': message['from'],
                           'op':'reschedule_cars',
                           'car_sec_ids':car_sec_ids,
                           })

    def get_length(self):
        """
        Helper function that returns the total length of all sections.
        """
        length=0.0
        for sec in self.secs:
            length+=sec.length
        return length
    
    def make_panel(self,panel,**kw):
        """
        Make widgets.
        May be overwritten by specific class.
        """
        # initialize widgets with Non objects such that they are not updated
        # during the make phase.
        
        panel.make_standardpanel()

        # add a sections page, which is common to all track elements
        panel.make_page('Sections',self.make_secspage,self.update_secspage)

        # add track element specific panel infos
        self.make_panel_specific(panel)

    def make_panel_specific(self,panel):
        """
        Add track element specific panel infos.
        To be overwritten.
        """
        pass

    
    def make_secspage(self,parent,widgets):
        #self.show_info()
        secs_group= Pmw.Group(parent,
                              tag_text='Table of all sections')
        secs_frame=secs_group.interior()
        sec_group= Pmw.Group(parent,
                            tag_text='Info on selected section')
	sec_frame=sec_group.interior()

        # Highligh or selection in this table should show the list
        # of vehicles in the correponding section.
        # Thus, it is necessary to pass the widgetset of the panel
        # to the list command _on_list_cars.
        # first argument, sel_ind, will be replaced by selected line
        # by the PanedScrolledTable instance.
        self.sec_selected=None
        sel_command=lambda sel_ind=1,self=self,widgets=widgets:\
                     self._on_update_sec(sel_ind,widgets)
            
        widgets['secstable']=gui.PanedScrolledTable(secs_frame,
                                                    highlightcommand=
                                                                  sel_command,
                                                    selectcommand=
                                                                  sel_command,
                                                    hullwidth=1.57*150,
                                                    #hullheight=150
                                                    )
        # create section specific data
        headers=[{'key':'label',
                  'name':'Label',
                  'type':'text',
                  'width':1,
                  'mode':'read',
                  },
                 {'key':'velocity_line',
                  'name':'Line speed',
                  'type':'real',
                  'printformat':'%3.3f',
                  'width':1,
                  'mode':'write',
                  'valid':{'min':0,'max':10000},
                  'units':{'m/s':1},
                  },
                 {'key':'length',
                  'name':'Length',
                  'type':'real',
                  'printformat':'%3.3f',
                  'width':1,
                  'mode':'read',
                  'valid':{'min':0,'max':1000000},
                  'units':{'m':1},
                  },
                 #{'key':'car_max',
                 # 'name':'Max # of carrier',
                 # 'type':'integer',
                 # 'printformat':'%3d',
                 # 'width':1,
                 # 'mode':'write',
                 # 'valid':{'min':0,'max':10},
                 # },
                 ]
        
        widgets['secs_param']=gui.DatasetEditor(headers,{},
                                                sec_frame,
                                                applyentry_command=\
                                                   self._on_secs_param_entry,
                                                height=100)
        
        # create lest lf carriers
        panel_car_command=lambda self=self,widgets=widgets:\
                     self._on_panel_car(widgets)
        widgets['carslist']=Pmw.ScrolledListBox(
            sec_frame,
            items=(),
            labelpos='nw',
            label_text='Vehicles',
            #listbox_height = 6,
            selectioncommand=panel_car_command,
            dblclickcommand=panel_car_command,
            hscrollmode='none',
            usehullsize = 1,
            hull_width = 180,
            hull_height = 120,
            )
        
        

        

        # fill tables etc. with data
        self.update_secspage(parent,widgets)
        
        # pack everything in the rihgt order
        widgets['secstable'].pack(expand=1, fill='y',side='left')

        widgets['secs_param'].pack(expand=1, fill='y',side='top',
                                 padx = 3, pady = 3)
        
        widgets['carslist'].pack(expand=1, fill='y',side='bottom',
                                 padx = 3, pady = 3)
        
        
        secs_group.pack(fill = 'both', expand = 1,
                        padx = 3, pady = 3,side='left')

        sec_group.pack(fill = 'both', expand = 1,
                       padx = 3, pady = 3,side='right')
        
    def _on_update_sec(self,sec_id,widgets):
        """
        List cars of section with index sec_id in carslist widget of panel
        """
        self.update_sec_gui(widgets,sec_id)
        

    def update_sec_gui(self,widgets,sec_id=-1):
        """
        Updates all section related widgets.
        """
        # update parameters of selected section
        if sec_id==-1:
            sec_id=widgets['secstable'].get_selection_ind()
        sec=self.secs[sec_id]
        self.sec_selected=sec
        
        widgets['secs_param'].update_contents({'label':sec.label,
                                               'length':sec.length,
                                               'velocity_line':\
                                               sec.velocity_line,
                                               'car_max':sec.car_max,
                                               })


        # update cars of selected section
        if sec_id>=0:
            sec=self.secs[sec_id]
            carslist=[]
            for car in sec.cars:
                carslist.append(car.id)
            widgets['carslist'].setlist(carslist)
        
    def _on_panel_car(self,widgets):
        """
        Open new or update control with selected carrier.
        """
        selections=widgets['carslist'].getcurselection()
        if selections:
            car=self.obj.sim.split_id(selections[0])['mod']
            widgets['controlpanel'].open_controlpanel(car)

    def _on_secs_param_entry(self,secs_param_widget,secs_param):
        if self.sec_selected:
            self.sec_selected.set_velocity_line(secs_param['velocity_line'])
            # apply this velocity to all cars on the sec
            for car in self.sec_selected.cars:
                if self.sec_selected.label not in ('berths',):
                    if car.status=='active':
                        car.control({'velocity_line':\
                                     self.sec_selected.velocity_line})


    
    def update_secspage(self,parent=None,widgets={}):
        """
        Updates all widgets on section page.
        """
        #print 'update_secspage'
        # update secstable
        
        headers=[{'key':'ID','type':'integer','width':0.2},
                {'key':'Label','type':'text','width':1},
                {'key':'Status','type':'text','width':0.5},
                {'key':'Cars','type':'integer','width':0.5},]

        table=[]
        for sec in self.secs:
            table.append({'ID':sec.id,
                          'Label':sec.label,
                          'Status':sec.status,
                          'Cars':len(sec.cars)})
        
        widgets['secstable'].make_table(headers,table)
        

        # update section related widgets only
        self.update_sec_gui(widgets)
        
        
    def reset_statistics_mod(self):
        """
        Reset or initialize statistical variables.
        """
        #print 'Reset or initialize statistical variables.',self.id
        # current simulation time is time when statistic starts
        self.time_init=self.obj.sim.time
        
        for sec in self.secs:
            sec.reset_statistics()
        self.reset_statistics_mod_specific()

    
    
    def reset_statistics_mod_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular track element.

        To be overwritten.
        """
        pass

    
        
    def format_statistics(self):
        """
        Return descriptor and dataset contents for statistical data
        about the target object.
        """
        # compute important quantities
        outputs_total=0
        inputs_total=0
        cars_total=0
        cars_entered_total=0
        cars_left_total=0
        for sec in self.secs:
            #print '  sec',sec.id
            cars_total+=len(sec.cars)
            if sec.fromnode.has_key('direction'):
                cars_entered_total+=sec.cars_entered
                inputs_total+=1

            if sec.tonode.has_key('direction'):
                cars_left_total+=sec.cars_entered
                outputs_total+=1

        if (self.obj.sim.time-self.time_init)>0:
            throughput_in=\
              cars_entered_total/((self.obj.sim.time-self.time_init)/3600.0)
            throughput_out=\
              cars_left_total/((self.obj.sim.time-self.time_init)/3600.0)
        else:
            throughput_in=0.0
            throughput_out=0.0
            
        descriptor=[{'key':'Status',
                     'type':'text',
                     'width':1,
                     'mode':'read'},
                    {'key':'Current number of vehicles',
                     'type':'integer',
                     'width':1,
                     'mode':'read'},
                    {'key':'Entered number of vehicles',
                     'type':'integer',
                     'width':1,
                     'mode':'read'},
                    {'key':'Number of entered vehicles per hour',
                     'type':'real',
                     'printformat':'%5.0f',
                     'width':1,
                     'mode':'read',
                     'units':{'1/h':1}},
                    ]
        dataset={'Status':self.status,
                 'Current number of vehicles':cars_total,
                 'Entered number of vehicles':cars_entered_total,
                 'Number of entered vehicles per hour':throughput_in,
                 }

        if inputs_total>1:
            descriptor.append({'key':'Entered vehicles\n'+\
                                       'per hour per input line',
                               'type':'real',
                               'printformat':'%5.0f',
                               'width':1,
                               'mode':'read', # or 'write' if editable
                               'units':{'1/h':1}
                               })
            dataset['Entered vehicles\n'+\
                    'per hour per input line']=throughput_in/inputs_total

        if outputs_total>1:
            descriptor.append({'key':'Leaving Vehicles\n'+\
                               'per hour per output line',
                               'type':'real',
                               'printformat':'%5.0f',
                               'width':1,
                               'mode':'read', # or 'write' if editable
                               'units':{'1/h':1}
                               })
            
            dataset['Leaving Vehicles\n'+\
                     'per hour per output line']=throughput_in/outputs_total

        self.format_statistics_specific(descriptor,dataset)    
        return descriptor,dataset

    
    
    def show_info(self):
        """
        Shows general information about this track element
        """
        print '===============',self.id,self.no,'=================='

        
        if len(self.tasks)>0:
            print '  Tasks:'
            for task in self.tasks.values():
                print '  Task:',task.id,\
                      'status=',task.status,'order',task.order_id,\
                      'for carrier',task.car_id,'\n    ',\
                      'mode=',task.mode,\
                      'target_flag=',task.target_flag
                if task.nextsec_ids:
                    for fromsec_id in task.nextsec_ids.keys():
                        print '    Route',fromsec_id,'->',\
                              task.nextsec_ids[fromsec_id]

        print '  Sections:'
        for sec in self.secs:
            print '    Sec:',sec.id,sec.label,' Status=',sec.status
            if sec.cars:
                for car in sec.cars:
                    print '      ',car.id,car.get_path_sec()
                    
            if self.debug >= 2:
                for tosec in sec.tosecs:
                    #'Check if pointers in tosecs and fromsecs really exist'
                    elem_data=self.obj.sim.split_id(tosec.elem.id)
                    if elem_data['mod'].secs.count(tosec)>0:
                        found= 'in '+elem_data['name']
                    else:
                        found='Not found'
                    print '      tosec',tosec.id,tosec.status,found
                    
                    
                for tosec in sec.tosecs_active:
                    #'Check if pointers in tosecs and fromsecs really exist'
                    elem_data=self.obj.sim.split_id(tosec.elem.id)
                    print 'tosecs_active',tosec.id#,tosec
                    if elem_data['mod'].secs.count(tosec)>0:
                        found= 'in '+elem_data['name']#+str(tosec.segs[0])
                    else:
                        found='Not found'
                    print '      tosec_active_active',\
                          tosec.id,tosec.status,found
                    is_congested=tosec.is_congested()
                    print '         congested=',is_congested
                    
                        
                for fromsec in sec.fromsecs:
                    print '      fromsec',fromsec.id,fromsec.status
                for fromsec in sec.fromsecs_active:
                    print '      fromsec_active',fromsec.id,fromsec.status
                    
        self.show_info_specific()

        #print '=============== GLOBAL INFO ==============='


        
        #for mod in self.obj.mods.values():
        #    print mod.id,mod.no
            
    def show_info_specific(self):
        """
        Shows track element specific information
        """
        pass
        
    def block_tosec(self,sec,tosec,permanent=0):
        #print 'block',self.id
        # update track man, only temorary, so do not reroute
        self.update_man()
        pass
            
    def unblock_tosec(self,sec,tosec,permanent=0):
        #print 'unblock',self.id

        # update track man, only temorary, so do not reroute
        self.update_man()
        
        # activate control of all carriers on this section
        for car in sec.cars:
            controldata={'velocity_line':sec.velocity_line}
            if sec.cars[0]==car: # first car in the section
                controldata=sec.search_car_infront(car,controldata)
                
                
            car.control(controldata)

    def probe(self):
        return self.id=='test.track.diverge_flexible/3'
    
    def get_nextsec(self,sec,car):
        """
        This function returns the next section object when carrier car
        leaves section sec. This means that sec has at least two
        possible active branches.

        The decition in which section to branch the carrier depends
        on many things like: instructions from task, whether carrier is
        empty and searches for an item to load, if the desired nextsec is
        clear etc.
        
        """
        #if self.probe():
        #    print 'get_nextsec',self.id,'with',car.id
        # check whether carrier is loaded and a task on the track side
        # this is only a check and should be eliminated, when empty carriers
        # get tasks as well.
        if (len(car.items)>0) & (not self.car_tasks.has_key(car.id)):
            print 'WARNING: in',self.id,'.get_nextsec:'
            print '  no task for loaded ',car.id
            # stop simulation
            #self.obj.sim.stop()
        
        # let's see if there is a task for this carrier
        if self.car_tasks.has_key(car.id):

            # go through all available tasks, oldest first
            for task in self.car_tasks[car.id]:
                nextsec_id=task.get_nextsec(sec.id)
                if nextsec_id != -1:
                    break # fond a valid nextsec_id
                
            # print 'Found sec id',nextsec_id,'in table'
            # get object address of next active section from next section id
            found=0
            is_nextsec_congested=False
            are_all_nextsecs_congested=False
            nextsec_uncongested=None
            for tosec in sec.tosecs_active:
                is_congested=tosec.is_congested()
                are_all_nextsecs_congested=are_all_nextsecs_congested|is_congested
                if not is_congested:
                    nextsec_uncongested=tosec
                
                if nextsec_id == tosec.id:
                    nextsec=tosec
                    is_nextsec_congested=is_congested
                    found=1
                    
                   
            
            if found:
                #print 'Nextsec:',car.id,'from',self.id,'->',nextsec.id
                # But is it safe to enter ?
                accept_nextsec = (not is_nextsec_congested)|are_all_nextsecs_congested
                
                if self.enter_sec_crit(nextsec,car) & accept_nextsec:
                    #if self.probe():
                    #    print '  sec',nextsec.id,' safe to enter'
                    return nextsec, '0'
                else:
                    #if self.probe():
                    #    print '  sec',nextsec.id,' not safe to enter or congested!!'
                        
                    #print '!!!--- not safe to enter: deviate carrier',car.id
                    # get any active and clear next section
                    nextsec=self.find_active_nextsec(sec,car)
                    
                    if nextsec.id != tosec.id:
                        # the desired sec is either congested or unsafe to enter
                        
                        # tell element that there has been a deviation
                        # of a carrier...this will trigger a rescheduling
                        self.control_car_on_deviation(sec,nextsec,car,task=task)
                        if self.probe():
                            print '  deviate to',nextsec.id,accept_nextsec
                        return nextsec, 'blocked'
                    
                    else:
                        # may happen that the congested one is the only
                        # safe one
                        if self.probe():
                            print '  OK congested but safe',nextsec.id,accept_nextsec
                        return nextsec, '0'
                    
            else:
                #print '!!! nextsec inactive  for car',car.id
                nextsec=self.find_active_nextsec(sec,car)
                self.control_car_on_deviation(sec,nextsec,car,task=task)
                return nextsec, 'blocked'
                
        else:
            #print '!!! did not find a task for this carrier',car.id
            # use first free active section
            #print 'unknown_dest'
            #print 'Found no task for',car.id
            #if nextsec_uncongested:
            #    if self.enter_sec_crit(nextsec_uncongested,car):
            #        nextsec=nextsec_uncongested
            #    else:
            #        nextsec=self.find_active_nextsec(sec,car)
            #else:
            #    nextsec=self.find_active_nextsec(sec,car)
            
            nextsec=self.find_active_nextsec(sec,car)
            #if self.probe():
            #    print '  random safe sec',nextsec.id
            return nextsec, 'unknown_dest'
    
    def find_active_nextsec(self,sec,car):
        '''
        Returns best active next section to section sec for carrier car.

        The nextsection is chosen that maximizes
        the remaining distance after a brake between
        carrier car and the last carrier in each of the branches. 

        Here it is supposed that there are at least 1 active next section.
        '''

        #if self.probe():
        #    print 'find_active_nextsec:try random branch',car.id
    
        crit=0
        dist_max=-10.0**8
        dist_max_unsafe=-10.0**8
        nextsec=None
        for checksec in sec.tosecs_active:
            save,dist_delta=self.enter_sec_crit_dist(checksec,car)
            if save:
                if dist_delta>dist_max:
                    #if self.probe():
                    #    print '  dist_max_safe=',dist_delta,\
                    #          'nextsec',checksec.id
                    dist_max=dist_delta
                    nextsec=checksec
                    
            elif dist_delta>dist_max_unsafe:
                dist_max_unsafe=dist_delta
                nextsec_unsafe=checksec
                
        # if none satisfied safety crit, take the least critical
        if not nextsec:
            #if self.probe():
            #    print '  no save sec found....take the least critical'
            nextsec=nextsec_unsafe
      
        return nextsec

    
    def enter_sec_crit(self,sec,car):
        '''
        This function verifies whether the carrier car can enter in
        section sec. This depends on the brakeing distance of the carrier
        and the position and speed of the last carrier in this section.

        Carrier can enter if function returns 1
        '''
        crit,dist_delta=self.enter_sec_crit_dist(sec,car)
        return crit

    def enter_sec_crit_dist(self,sec,car):
        '''
        This function verifies whether the carrier car can enter in
        section sec. This depends on the brakeing distance of the carrier
        and the position and speed of the last carrier in this section.

        Carrier can enter if function returns 1
        It returns also the distance that remains after braking.
        '''
    
        crit=1
        dist_delta=10.0**8
        if len(sec.cars)>0: # is section occupied ?
            lastcar=sec.cars[-1]
            dist=lastcar.get_path_sec()-lastcar.length
            
        elif car.get_brakedist_comfort()>sec.length:
            # the comfort braking dist is longer than the length of the
            # section where the car wants to enter.
            # In this case, check beyon 
            cd=sec.search_car_infront(car,{},0,0)
            lastcar=cd['car_infront']
            if lastcar:
                dist=cd['dist_infront']
        else:
            lastcar=None
            #if self.probe():
            #    print 'enter_sec_crit of sec',sec.id,' with',car.id
            #    print '  empty sec'
            

        if lastcar:
            if lastcar.modind!=-1: # last car already in array structure
                # is there sufficient space for braking ?
                #if self.probe():
                #    print 'enter_sec_crit of sec',sec.id,' with',car.id,\
                #          '\n  lastcar=',lastcar.id
                #    print '  d=',dist,\
                #          'd_min=',car.get_dist_min(lastcar)
                #    print '  v=',car.get_velocity(),\
                #          'v_l=',lastcar.get_velocity()
                #print 'enter_sec_crit for',car.id,lastcar.id
                dist_delta=dist-car.get_dist_min(lastcar)
                       
                if lastcar!=car.car_infront:
                    if dist_delta<0:
                        #if self.probe():
                        #    print '  braking distance too long!'
                        crit=0
                              
            else: # is a carrier in this branch which in not yet active
                crit=0
        
                
        return crit,dist_delta

    def control_car_on_start(self,sec,car,mode,controldata):
        '''
        Track element wide carrier control when carrier is ready to start.
        The suggested controldata from the respective section may be
        modified and the carrier must be controlled.
        To overwrite with track element specific control instructions
        '''
        
        car.control(controldata)
    

    def control_car_on_enter(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier enters section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        To overwrite with track element specific control instructions
        '''
        return controldata
    
    
    def control_car_on_leave(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier leaves section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        To overwrite with track element specific control instructions
        '''

        #if controldata.has_key['ghost_infront']:
        #print 'control_car_on_leave at',self.id,sec.id
        #    print '   found',car.id,'with ghost!!!'

        # check if a task has been completed
        if self.car_tasks.has_key(car.id):
            #print '  call task.leave_sec of',self.car_tasks[car.id][0]
            self.car_tasks[car.id][0].leave_sec(sec)
            
        return controldata

    def control_car_at_mark(self,sec,car,controldata={}):
        '''
        Track element wide carrier control when carrier crosses a mark
        on section sec.

        To be overwritten by track elements that use marks.
        '''
        #print 'control_car_at_mark:',self.id,sec.id,car.id
        
        pass

    

    def control_car_on_deviation(self,sec,nextsec,car,
                                 controldata={},task=None):
        '''
        Track element wide carrier control when carrier wanted to enter 
        a blocked section, but instead, it has been deviated to
        section nextsec. This function will fail the present track scheduling
        task for this carrier. This will initiate a complete rescheduling.
        '''

        # check if a task has been completed
        #if self.car_tasks.has_key(car.id):
        #    # deviate task
        #    self.car_tasks[car.id].deviatedto_sec(sec,nextsec)
        if task:
            task.deviatedto_sec(sec,nextsec)

        elif self.car_tasks.has_key(car.id):
            # use oldest task
            self.car_tasks[car.id][0].deviatedto_sec(sec,nextsec)
        
        # controldata currently no functionality
        return controldata
    
    def plot(self,canvas=None):
        if not canvas: canvas=self.obj.netmon# =None if netmon not invoked

        if canvas:
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                canvas.delete(self.id)

            # set section width
            width=self.obj.width_netmon
            if (canvas==self.obj.netmon):
                width_scaled=width*self.obj.netmon.canvas_scale
            else:
                width_scaled=width

            for sec in self.secs:
                # plot segments from section transforms
                if sec.status!='active':
                    canvas.create_line(ravel(sec.segs[:,:2,3]).tolist(),
                                       capstyle='round',
                                       tag=self.id,width=1.5,
                                       fill=self.obj.color_netmon_edit)
                else:
                    
                    id=canvas.create_line(ravel(sec.segs[:,:2,3]).tolist(),
                                          capstyle='round',
                                          tag=self.id,width=width_scaled,
                                          fill=self.obj.color_netmon,smooth=1)
                    if canvas==self.obj.netmon:
                        # add section to canvas objects with width=x group
                        # this means that width will be adjusted dynamically
                        # when rescaling, using a fast tk aglorithm 
                        self.obj.netmon.addtag_withtag('width=%.1f'%(width),id)
            
                # get local node ids
                md=3 # marker diameter
                fromnode=sec.fromnode['id']
                tonode = sec.tonode['id']

                if self.status!='edit':
                    # put markers at open nodes
                    if self.nodes_open.has_key(fromnode):
                        node_name=self.id+self.obj.sim.part_sep+\
                                   'node'+self.obj.sim.part_sep+`fromnode`
                        x,y=self.nodes_open[fromnode].pos[0:2].tolist()
                        canvas.create_oval(x-md,y-md,x+md,y+md,
                                           #fill=self.obj.color_netmon_edit,
                                           tags=(self.id,node_name)
                                           )
                    if self.nodes_open.has_key(tonode):
                        node_name=self.id+self.obj.sim.part_sep+\
                                   'node'+self.obj.sim.part_sep+`tonode`
                        x,y=self.nodes_open[tonode].pos[0:2].tolist()
                
                        canvas.create_oval(x-md,y-md,x+md,y+md,
                                           #fill=self.obj.color_netmon_edit,
                                           tags=(self.id,node_name),
                                           fill='black')
                else:
                    if self.nodes[tonode].has_key('direction'):
                        x,y=ravel(self.nodes[tonode]['trans'][:2,3]).tolist()
                        canvas.create_oval(x-md,y-md,x+md,y+md,
                                           tags=(self.id),
                                           fill='black'
                                           )
                    
                    if self.nodes[fromnode].has_key('direction'):
                        x,y=ravel(self.nodes[fromnode]['trans'][:2,3]).tolist()
                        canvas.create_oval(x-md,y-md,x+md,y+md,
                                           fill=self.obj.color_netmon_edit,
                                           tags=(self.id)
                                           )
            
                
            self.plot_extra(canvas)

            # put tracks in background ...under cars and users
            canvas.lower(self.id)
        
            # not very clean method...
            # this will put an eventually existing background map on netmon
            # in the background.
            if (canvas==self.obj.netmon):
                self.obj.netmon.lower_maps()
                
        
                
        
    def del_plot(self,canvas=None):
        if not(canvas): canvas=self.obj.netmon# =None if netmon not invoked

        if canvas:
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                # print 'del_plot',self.id,canvas
                canvas.delete(self.id)
            
    def plot_extra(self,canvas):
        """Extra plots, apart from sections. Track element specific"""
        # overwrite, in specialized track element class
        pass

    def get_trans_param(self,trans):
        """
        Returns position and angle in xy plane. 
        """
        pos=trans[0:3,3].tolist()
        dx=trans[0,0]
        dy=trans[1,0]
        phi=arctan2(dy,dx)*180/pi
        return pos,phi
    
    def get_pos(self):
        """
        Get current position of root node of track element. 
        """
        return self.trans[0:3,3].tolist()

    def get_angle0(self):
        """
        Get current root angle in degrees of root node of track element. 
        """
        if self.nodes.has_key(self.rootnode):
                if self.nodes[self.rootnode].has_key('trans0'):
                    pos0,alpha0=self.get_trans_param(\
                    self.nodes[self.rootnode]['trans0'])
                    return alpha0
                else:
                    return 0.0
        else:
            return 0.0

        
    def get_angle(self):
        """
        Get current angle in degrees of root node of track element. 
        """
        dx=self.trans[0,0]
        dy=self.trans[1,0]
        phi=arctan2(dy,dx)
        #if dx==0.0:
        #    phi=sign(dy)*pi/2
        #else:
        #    phi=arctan(dy/dx)
        #    if dx<0.0:
        #        phi+=pi
        return phi*180/pi
    
    def set_root(self,
                 localnode=1,pos=[0,0,0],alpha=0,trans=None,swapnodes=1):
        """
        Setting the reference coordinate system for this track element
        by choosing the root node, its position and direction.
        """
        #print 'set_root',localnode,pos,alpha,trans
        #print '  angle before:',self.get_angle()
        if self.status=='edit':
            self.rootnode=localnode
            if trans:
                self.make_trans(trans=trans,swapnodes=swapnodes)
            else:
                
                self.make_trans(pos,alpha,swapnodes=swapnodes)

            if self.obj.netsim:
                self.obj.netsim.update_elem(self)

            #print '  angle after:',self.get_angle()
            if self.nodes.has_key(self.rootnode):
                if self.nodes[self.rootnode].has_key('trans0'):
                    pos0,alpha0=self.get_trans_param(\
                    self.nodes[self.rootnode]['trans0'])
                    #print '  alpha0=',alpha0,alpha0-self.get_angle()
            return 0
        else:
            return 1

    def set_status(self,newstatus):
        '''Central function to change status of track element'''
        # connect track element in network if it goes from an
        # 'edit' to 'passive'  status
        if (self.status=='edit') & (newstatus=='passive'):
            self.status=newstatus
            error=self.obj.netsim.connect_elem(self)
            if not error:
                # add element to track management
                self.add_man()
                return 0
            else:
                self.status='edit'
                return 1
            
        
        elif (self.status=='passive') & (newstatus=='edit'):
            # isolate track element before entering edit mode
            self.status=newstatus
            if self.obj.netsim.isolate_elem(self):
                self.status='passive'
                return 1
            else:
                # delete element from  track management
                self.del_man()
                
                return 0

        elif (self.status=='passive') & (newstatus=='active'):
            # try to activate all sections
            still_passive=1
            for sec in self.secs:
                still_passive=still_passive & sec.activate()

            # update track management
            self.update_man()
            
            # if one section is active then the whole element is active
            if still_passive:
                return 1
            else:
                self.activate_specific()
                self.status=newstatus
                return 0
        
        elif (self.status=='active') & (newstatus=='passive'):
            # try to inactivate all sections
            still_active=1
            for sec in self.secs:
                still_active=still_active & sec.inactivate()

            # update track management
            self.update_man()
            
            # if at least one section remains active then
            # the whole element is still active
            if still_active:
                return 1
            else:
                self.unactivate_specific()
                self.status=newstatus
                return 0
            
        elif self.status==newstatus: # same status as before, do nothing
            return 0
        else:
            return 1  # status could not be changed

    def activate_specific(self):
        """
        Track element specific activation.
        """
        pass

    def unactivate_specific(self):
        """
        Track element specific disactivation.
        """
        pass
    
    def make_trans(self,pos=[0,0,0],alpha=0,trans=None,swapnodes=0):
        #print self.id,'elem.make_trans',pos,alpha,trans
        if trans:
            pass
        else:
            x,y,z=pos
            alpha0=self.get_angle0()
            phi=(alpha+alpha0)*pi/180
            trans=array([[cos(phi), -sin(phi),    0,x],
                         [sin(phi),  cos(phi),    0,y],
                         [0,                0,    1,z],
                         [0,                0,    0,1]])
            
        if self.nodes.has_key(self.rootnode):
            if self.nodes[self.rootnode].has_key('trans0'):
                self.trans=matrixmultiply(
                    trans,inverse(self.nodes[self.rootnode]['trans0']))
            else:
                self.trans=trans
        else:
            self.trans=trans
            
        #print '  trans',self.trans
    
  
class Divmod(Mod):
    '''
    This is the module that must be used for general
    1:N diverge track elements.
    '''
    def control_car_on_enter(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier leaves section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        To overwrite with track element specific control instructions
        '''
        #print 'Divmod.control_car_on_enter',car.id,'at',self.id,sec.id
        #print 'controldata=',controldata
        #if controldata.has_key('ghost_infront'):
        #    print 'before: gost',controldata['ghost_infront'].id,\
        #          '        dist=',controldata['ghostdist_infront']
        # only control if carrier is entering one of the diverging branches
        if sec!=self.secs[0]:
            #print ' go through all parallel sections, but not the entry line'
            for branchsec in self.secs[1:]:
                
                if  branchsec !=  sec : 
                    #print ' detected a parallel branch to sec',branchsec.id
                    controldata_check=self.diverge(branchsec)
                    #if controldata_check.has_key('ghost_infront'):
                    #    print 'after call:gost'
                    #          'dist=',controldata_check['ghostdist_infront']
                    # check whether distance to ghost is shorter than
                    # detected in a previous branch
                    if controldata.has_key('ghost_infront'):
                        if controldata['ghostdist_infront']<\
                           controldata_check['ghostdist_infront']:
                            controldata.update(controldata_check)
                    else:
                            controldata.update(controldata_check)

                    #print 'controldata_check=',controldata_check
        #if controldata.has_key('ghost_infront'):
        #    print 'gost',controldata['ghost_infront'].id,\
        #          'dist=',controldata['ghostdist_infront']
        return controldata
    
    
    def control_car_at_mark(self,sec,car,controldata={}):
        '''
        Track element wide carrier control when carrier crosses a mark
        on section sec.

        '''
        #print 'DIVcontrol_car_at_mark',sec.elem.id,sec.id,car.id
        if sec!=self.secs[0]:
            #print ' remove ghost carrier',car.id
            controldata['ghost_infront']=None
            #car.control(controldata)
    
    def diverge(self,sec):
        """
        Check whether there is a carrier in the clearing zone
        of this section.
        If yes, return the carrier as ghost in the control data and
        compute distance from diverge point.
        """
        controldata={}
        #if controldata.has_key('ghost_infront'):
        #    print 'before search_ghost:gost',controldata['ghost_infront'].id,\
        #          'dist=',controldata['ghostdist_infront']
            
        if len(sec.cars)>0:

            lastcar=sec.cars[-1]
            #print 'search_ghost: found',lastcar.id,'on',sec.id
            dist=sec.netsim.path_sec[lastcar.ind]-lastcar.length
            if dist<sec.clearingzone:
                controldata['ghost_infront']=lastcar
                controldata['ghostdist_infront']=dist
                controldata['mark_at']=sec.clearingzone
        
            
        return controldata

    
class Mergemod(Mod):
    '''This is the module that must be used for merge track elements'''

    def init_specific(self):
        """
        Module specific initialization
        """
        # this call must be made by all othe module types using Mergemod
        # as mixin
        #self.init_specific_mergemod()
        # default implementation where sections 0,1 must be entry secs
        self.define_merge_entrysecs(self.secs[0:2])
        
    #def init_specific_mergemod(self):
    #    """
    #    Merge module specific initialization
    #    """
    #    # very specific implementation where sections 0,1 must be entry secs
    #    self.define_merge_entrysecs(self.secs[0:2])

    def define_merge_entrysecs(self,mergesecs):
        """
        Define which of the sentions of the track element enter the merge zone.
        """
        #'initialize a list with cars that are not a ghost for other cars
        # for each entrysection
        self.cars_noghost={}

        self.mergesecs=mergesecs
        for sec in mergesecs:
            self.cars_noghost[sec.id]=[]

        
    def control_car_on_enter(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier enters section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        Here the control data is modified by the elements merge algorithm.
        '''
        self.control_car_on_enter_mergemod(sec,car,controldata)
        return controldata
    
    def control_car_on_enter_mergemod(self,sec,car,controldata):
        """
        Mergemod specific control_car_on_enter.

        Designed for 2 merge branches only.
        """
        #if (self.id=='test.track.merge_curve/1')&(sec==self.mergesecs[0]):
        #    print '========================================================'
        #    print 'control_car_on_enter_mergemod',sec.id,'@',self.obj.sim.time
        #    car.show_info()
        #    
        #    if car.car_infront:
        #        print '------------------>>'
        #        
        #        car.car_infront.show_info()
        
        
        if sec == self.mergesecs[1]:
            #print '   msec 1 detected'
            if self.mergesecs[1].length>self.mergesecs[0].length:
                #'  do merge decision later under equal distance conditions'
                controldata['mark_at']=self.mergesecs[1].length-\
                                        self.mergesecs[0].length
            else:
                self.append_cars_noghost(car)
                #'   merge this carrier with carriers on section 0'
                self.merge(car,self.mergesecs[0],controldata)
                
        elif sec == self.mergesecs[0]:
            #print '   msec 0 detected'
            if self.mergesecs[0].length>self.mergesecs[1].length:
                #'   do merge decision later under equal distance conditions'
                controldata['mark_at']=self.mergesecs[0].length-\
                                        self.mergesecs[1].length
            else:
                self.append_cars_noghost(car)
                #'    merge this carrier with carriers on section 1'
                self.merge(car,self.mergesecs[1],controldata)

        #else:
        #    # output section detected
        #    self.reassign_car_infront_after_merge()
            
            
        return controldata

    #def reassign_car_infront_after_merge(self):
    #    # lead carrier may have changed for most advanced carriers in
    #    # input sections
    #    #print 'reassign_car_infront_after_merge',self.id
    #    
    #    for msec in self.mergesecs:
    #        if msec.cars:
    #            # make entering carrier car in front
    #            #print '  found car in sec',msec.id
    #            c=msec.search_car_infront(msec.cars[0],{})
    #            if c:
    #                #if c['car_infront']:
    #                #    print '  for car.id',msec.cars[0].id
    #                #    print '  with car_infront.id',c['car_infront'].id
    #                #    print '  dist_infront',c['dist_infront']
    #                msec.cars[0].control(c)
                    
    def control_car_at_mark(self,sec,car,controldata={}):
        '''
        Track element wide carrier control when carrier crosses a mark
        on section sec.

        '''
        self.control_car_at_mark_mergemod(sec,car,controldata)
        return controldata
    
    def control_car_at_mark_mergemod(self,sec,car,controldata={},
                                     no_mergecar_behind=0):
        """
        Mergemod specific control_car_at_mark
        """
        #'control_car_at_mark_mergemod',sec.id,car.id
        self.append_cars_noghost(car)
        # very specific implementation where sections 0,1 must be entry secs
        if (sec == self.mergesecs[1])& (len(self.mergesecs[0].cars)>0):
            # merge this carrier with carriers on section 0
            self.merge(car,self.mergesecs[0],controldata,no_mergecar_behind)
            
        elif (sec == self.mergesecs[0])& (len(self.mergesecs[1].cars)>0):
            # merge this carrier with carriers on section 1
            self.merge(car,self.mergesecs[1],controldata,no_mergecar_behind)
            

    
    def control_car_on_leave(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier leaves section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        Here the ghost carrier will be removed from the control data.
        '''
        self.control_car_on_leave_mergemod(sec,car,controldata)
        return controldata
    
    def control_car_on_leave_mergemod(self,sec,car,controldata):
        """
        Mergemod specific control_car_on_leave
        """
        #print 'control_car_on_leave_mergemod',sec.id,car.id
                
        # remove ghost carrier, if there is any
        controldata['ghost_infront']=None

        self.remove_cars_noghost(car)

    
    def append_cars_noghost(self,car):
        """
        Append car to the cars_noghost list of section sec.
        These are lists with cars that are not a ghost for other cars.
        The dictionary contains such a list for each  entrysection,
        with section id as key
        """
        #print 'append_cars_noghost',car.id,car.sec.id,
        if self.cars_noghost.has_key(car.sec.id):
            self.cars_noghost[car.sec.id].append(car)
            #self.obj.netmon.itemconfigure(car.id,fill='white')
            
    def remove_cars_noghost(self,car):
        """
        Remove cars from the cars_noghost list of section sec. 
        """
        sec_ind=car.sec.id
        if self.cars_noghost.has_key(sec_ind):
            if self.cars_noghost[sec_ind].count(car)>0:
                # remove this car and all previously entered from list
                ind=self.cars_noghost[sec_ind].index(car)
                del self.cars_noghost[sec_ind][0:ind+1]
        #self.obj.netmon.itemconfigure(car.id,fill='orange')
                
    def merge(self,car,mergesec,controldata={},
              no_mergecar_behind=0):
        '''
        This is the central merge algorithm which merges the carrier on
        section car.sec into the chain of vehicles on the target section
        on the same track element.  This algorithm modifies the control
        data of the carrier (i.e. puts a ghost carrier in front) and
        issues also control commands to the carriers on the target
        section, if necessary!
        '''
        car_dist_to_merge=car.sec.length-car.get_path_sec()
        car_dist_to_merge_after_brake=car_dist_to_merge-\
                                       car.get_brakedist_comfort()
        mergesec_cars=self.cars_noghost[mergesec.id][:]
        mergesec_cars.reverse()
        
        mergecar_behind=None
        mergecar_infront=None

        #if self.id=='test.track.merge_flexible/1':
        #    print '>>>>>>>>Merge',car.id,'with sec',mergesec.id
        #    for mcar in mergesec_cars:
        #        print '        ',mcar.id
                
        for mergecar in mergesec_cars: # take last entered first
            mergecar_path_sec=mergecar.get_path_sec()
            mergecar_dist_to_merge_after_brake=mergesec.length-\
                                             mergecar_path_sec-\
                                             mergecar.get_brakedist_comfort()

            if ((car_dist_to_merge_after_brake>\
                 mergecar_dist_to_merge_after_brake)|\
                (mergecar.ghost_infront!=None))\
                &(mergecar.get_velocity_line()>0):
                mergecar_infront=mergecar
                break
            
            elif not no_mergecar_behind: # yes we can merge it behind car
                mergecar_behind=mergecar

        
        
        #self.cars_noghost,car.sec.id
        if mergecar_infront:
            #print '        mergecar_infront',mergecar_infront.id
            # create ghost mergecar for car
            self.remove_cars_noghost(mergecar)
            mergecar_dist_to_merge=mergesec.length-\
                                    mergecar_infront.get_path_sec()
            controldata['ghost_infront']=mergecar_infront
            controldata['velocity_line']=car.sec.velocity_line
            controldata['ghostdist_infront']=car_dist_to_merge-\
                                              mergecar_dist_to_merge-\
                                              mergecar.length
            #mergecar_infront.control({'velocity_line':mergesec.velocity_line})
            
            
        if mergecar_behind:
            #print '        mergecar_behind',mergecar_behind.id
            # create ghost car for mergecar
            self.remove_cars_noghost(car)
            mergecar_dist_to_merge=mergesec.length-\
                                    mergecar_behind.get_path_sec()
            mergecar_behind.control({'ghost_infront':car,
                                     'velocity_line':mergesec.velocity_line,
                                     'ghostdist_infront':
                                     mergecar_dist_to_merge-\
                                     car_dist_to_merge-\
                                     car.length
                                     }
                                    )
            
    def get_controldata_to_make_ghost(self,car,ghostcar):
        """
        Returns a dictionary with control data that is necessary to create
        a ghostcar for the carrier car.
        """
        ghostcar_dist_to_merge=ghostcar.sec.length-ghostcar.get_path_sec()
        car_dist_to_merge=car.sec.length-car.get_path_sec()
        self.remove_cars_noghost(ghostcar)
        return {'ghost_infront':ghostcar,
                'ghostdist_infront':car_dist_to_merge-ghostcar_dist_to_merge-\
                                    ghostcar.length
                }
            
    
    

class Stopmod(Divmod,Mergemod):
    
    def init_specific(self):
        
        #print 'init_specific',self.id
        
        # name of stop and zone in which this stop is located
        self.name=self.obj.sim.get_stop_name() # get unique stop name
        self.zone='residential'
        self.n_users_queuing=0
        
        #print '__dict__=',self.__dict__
        # create a section directory with label as key
        # This will faciliate the acces to certain sections later on...
        self.secs_dir={}
        for sec in self.secs:
            if sec.label:
               self.secs_dir[sec.label]=sec

        # find out whether left or right side berth
        if self.secs_dir['berths'].localsec.has_key('side_plat'):
          s=self.secs_dir['berths'].localsec['side_plat']
          if s=='right':
            self.side=-1
          else:
            self.side=+1

        else:
          # get berth side out of shape info
          shape,param=self.secs_dir['brake'].localsec['shape']
          #print self.secs_dir['brake']
          if param['turn']=='right':
            self.side=-1
          else:
            self.side=+1

        #print 'side',self.side
        #self.placesec=self.secs[6]
        #self.halt_point=self.placesec.localsec['halt_point']
        self.users={}

        # define merge sections of stop
        self.define_merge_entrysecs([self.secs_dir['main_exit'],
                                     self.secs_dir['accel']])
        
        # create a central queue for waiting users.
        # The first user in this queue will proceed to the next 
        # availlable carrier.
        #self.queue=[]

        # introduce the element-wide cars dictionary
        # with the scope to keep track of low-level elemend-wide
        # information of the car.
        self.cars={}
        
        # create data-structure for berth...
        self.berths=[]
        self.berthssec=self.secs_dir['berths']

        
        # extract berth length from section shape info
        line_type,line_param=self.berthssec.localsec['shape']
        berthsec_length=line_param['length']

        # extract also number of berths
        berth_no=self.berthssec.localsec['berth_no']

        berth_length=self.berthssec.localsec.get('berth_length',
                                                 berthsec_length/berth_no)
        offset=0.5*(self.berthssec.length-(berth_no)*berth_length)+berth_length


        for i in range(berth_no):
            halt_point=offset+i*berth_length
            self.berths.append({'ind':i,
                                'queue':[],
                                'users_approaching':0,
                                'status':'liberated',
                                'allocated':0,
                                'car':None,
                                'halt_point':halt_point,
                                'n_loads':0,
                                'n_unloads':0,
                                }
                               )
                               
                
        self.berthssec.berth_no=berth_no
        self.berthssec.berth_length=berth_length

        
        # print 'self.berths',self.berths

        # this is a a list with berth index that demand carriers
        # first berth index in the list will be served first
        self.berths_demand=[]

        if not self.berthssec.localsec.has_key('berths_keep_clear'):
            self.berthssec.localsec['berths_keep_clear']=int(berth_no/2)
            
        self.berths_occupy=berth_no-\
                                self.berthssec.localsec['berths_keep_clear']

        # if this stop is near a bus or rail station or p&r
        # this can be indicated here
        self.mode_exch=''
        
        # number of carriers that are initially outside the berth region
        self.cars_offtrack=[]
       
        # initially switch off sync mode.
        # it will be switched on (within the load_item method)
        self.time_sync=0.0
        # berth allocation index for synchronous mode
        self.ind_berth_alloc=0
        self.ind_berth_firstalloc=int(self.berthssec.berth_no/2)-1

        self.flag_cars_blocked=0
        
        self.is_blocked_unloadzone=True
        self.is_blocked_loadzone=True
            
    def get_state_mod_specific(self,state_image):
        '''
        Same as get_state_mod, but module specific, to be overwritten.
        '''
        # this is done to avoid links to objects while saving
        #print 'get_state_mod_specific',self.id
        for car in self.berthssec.cars:
            # replace  berth objects in cars dict by their berth ind
            if self.cars[car.id]['berth_allocated']!=None:
                self.cars[car.id]['berth_allocated']=\
                       state_image['cars'][car.id]['berth_allocated']['ind']
            else:
                state_image['cars'][car.id]['berth_allocated']=-1
                
            if self.cars[car.id]['berth_occupied']!=None:
                state_image['cars'][car.id]['berth_occupied']=\
                       self.cars[car.id]['berth_occupied']['ind']
            else:
                state_image['cars'][car.id]['berth_occupied']=-1

        # reduce berth structure
        for berth in self.berths:
            ids_queue=[]
            for user in berth['queue']:
                ids_queue.append(user.id)
            state_image['berths'][berth['ind']]['queue']=ids_queue
            if berth['car']:
                state_image['berths'][berth['ind']]['car']=berth['car'].id

        ids_berths_demand=[]
        for berth in self.berths_demand:
            ids_berths_demand.append(berth['ind'])
        state_image['berths_demand']=ids_berths_demand

        # offtrack.
        ids_cars_offtrack=[]
        for car in self.cars_offtrack:
            ids_cars_offtrack.append(car.id)
        state_image['cars_offtrack']=ids_cars_offtrack

        
                
        # delete states that can be recovered from other states
        del state_image['secs_dir']
        del state_image['berthssec']

        return state_image
    
    def rebuild_pointers_specific(self):
        """
        reconstruct pointers to object that have been removed by
        get_state_mod_specific, before saving data.
        """
        # create a section directory with label as key
        # This will faciliate the acces to certain sections later on...
        for berth in self.berths:
            queue=[]
            for user in berth['queue']:
                queue.append(self.sim.rebuild_pointer(user))
            berth['queue']=queue
            if berth['car']:
                berth['car']=self.sim.rebuild_pointer(berth['car'])


        cars_offtrack=[]
        for car in self.cars_offtrack:
            cars_offtrack.append(self.sim.rebuild_pointer(car))
        self.cars_offtrack=cars_offtrack
        
        berths_demand=[]
        for berth in self.berths_demand:
            if type(berth)==types.IntType:
                berths_demand.append(self.berths[berth])
        self.berths_demand=berths_demand
        
        self.secs_dir={}
        for sec in self.secs:
            if sec.label:
               self.secs_dir[sec.label]=sec
        # define merge sections of stop
        #self.define_merge_entrysecs([self.secs_dir['main_exit'],
        #                             self.secs_dir['accel']])
        self.berthssec=self.secs_dir['berths']
        #print 'rebuild_pointers_specific',self.id  
        for car in self.berthssec.cars:
            if type(self.cars[car.id]['berth_allocated'])==types.IntType:
                # replace berth ind in cars dict by their objects
                #print '  car=',car.id,\
                #      'berth_allocated',self.cars[car.id]['berth_allocated'],\
                #      'berth_occupied=',self.cars[car.id]['berth_occupied']
                if self.cars[car.id]['berth_allocated']>=0:
                    self.cars[car.id]['berth_allocated']=\
                         self.berths[self.cars[car.id]['berth_allocated']]
                else:
                    self.cars[car.id]['berth_allocated']=None
                    
                if self.cars[car.id]['berth_occupied']>=0:
                    self.cars[car.id]['berth_occupied']=\
                         self.berths[self.cars[car.id]['berth_occupied']]
                else:
                    self.cars[car.id]['berth_occupied']=None
        
        
            
            
                    
    def set_state_mod_specific(self):
        '''
        Same as set_state_mod, but module specific, to be overwritten.
        '''
        #if self.probe():
        #print 'set_state_mod_specific',self.id

        if string.atof(self.get_version())<0.3:
            for task in self.tasks.values():
                if not hasattr(task,'mode'):
                    if hasattr(task,'search_item'):
                        if task.search_item==1:
                            task.mode='search_item'
                        else:
                            task.mode='loaded'

               
        if string.atof(self.get_version())<0.2:
            #print 'old version'
            # for newly initialized elements this will be done in init_specific
            
            # introduce the element-wide cars dictionary
            # with the scope to keep track of low-level elemend-wide
            # information of the car.
            # The key is the car.id value is a database with varios information
            # here construct dictionary:

            self.cars={}
            for car in self.berthssec.cars:
                self.cars[car.id]={ #'mod':car,
                                   'state':'passthrough',
                                   'berth_allocated':None,
                                   'berth_occupied':None,
                                   'found_in':'',
                                   'loaded':0
                                   }
                status=self.get_status_car_in_berthssec(car)

                self.cars[car.id]['state']=status['state']
                self.cars[car.id]['loaded']=status['loaded']
                if status['state'] in ('waiting','loading'):
                    self.cars[car.id]['berth_occupied']=self.berths[\
                        status['ind_berth']]
                if status['state'] == 'allocated':
                    self.cars[car.id]['berth_allocated']=self.berths[\
                        status['ind_berth']]
                    

                    
            # this is a a list with berth index that demand carriers
            # first beth index in the list will be served first
            self.berths_demand=[]

            # time to control uload/unload phase in synch mode
            self.ind_berth_firstalloc=int(self.berthssec.berth_no/2)-1
            if self.is_start_sync():
                # set berth to synchronous mode with a time proportional
                # to the number of berth
                self.start_sync_mode()
                # in case all berth are liberated ...
                if self.are_all_berth_liberated():
                    self.start_allocation()
            else:
                # switch off sync mode for small stops 
                self.time_sync=0.0
                # berth allocation index for synchronous mode
                self.ind_berth_alloc=self.berthssec.berth_no-1
            
            # new statistics
            self.n_loads=0
            self.n_unloads=0
            self.waitingtime_total=0.0
        
            # for older versions without name
            if  not hasattr(self,'name'):
                self.name=''
            
            #if self.obj.netsim: 
            if self.name=='':
                self.name=self.obj.sim.get_stop_name()

            
            
            if hasattr(self,'berthssec'):
                if not self.berthssec.localsec.has_key('berths_keep_clear'):
                    self.berthssec.localsec['berths_keep_clear']=\
                                                   int(self.berth/2)
            
            # delete no more used variables
            if hasattr(self,'berths_alloc'):
                del self.berths_alloc
                del self.berth_alloc_ind
                del self.berth_firstalloc_ind
                del self.berths_occupied
        #else:
            # rebuild specific pointers ...already done
            #self.rebuild_pointers()
            
        if string.atof(self.get_version())<0.5:
            # reduce velocity in main entry section to prevent entry rejections
            #print '  ',string.atof(self.get_version())
            for sec in self.secs:
                #print '  ',sec.label,sec.velocity_line
                if sec.label=='entry':
                    sec.velocity_line=0.75*sec.velocity_line
                    #print '  check:',sec.label,sec.velocity_line
                    
            #for sec in self.secs:
            #    print '  check again:',sec.label,sec.velocity_line
            
                
            # key mod in users database no longer used
            if hasattr(self,'berths_alloc'):
                for user in self.users.values():
                    if user.has_key('mod'):
                        del user['mod']          
                        
            
            # temporary
            if not hasattr(self,'waitingtime_total'):
                self.waitingtime_total=0.0
            
            if not hasattr(self,'waitingtime2_total'):
                self. waitingtime2_total =0.0
                self. waitingtime_min =10.0**8
                self. waitingtime_max =0.0
                
                for berth in self.berths: 
                    berth['n_loads']=0
                    berth['n_unloads']=0.0
                    
                    
            if hasattr(self,'cars'):
                for carid in self.cars.keys():
                    if not self.cars[carid].has_key('found_in'):
                        #print 'get_state_mod: add found_in to',carid
                        self.cars[carid]['found_in']=''
    
            if not hasattr(self,'n_users_queuing'):           
                self.n_users_queuing=len(self.get_users_queued())
    
            if not hasattr(self,'flag_cars_blocked'):           
                self.flag_cars_blocked=0
    
            if not hasattr(self,'mode_exch'):           
                self.mode_exch=''
                
            if not hasattr(self,'is_blocked_unloadzone'):     
                self.is_blocked_unloadzone=True
                
                
                
            if not hasattr(self,'is_blocked_loadzone'):     
                self.is_blocked_loadzone=True
            
        
    def init_costs(self):
        """
        Define default costs here.

        Stops have higher extra costs for platforms, elevators, etc.
        """
        self.set_costs({'estim_init_ex':
                        {'name':'Estim. initial extra costs',
                         'table':{1:    80.0*10**3,
                                  10:   40.0*10**3,
                                  100:  20.0*10**3,
                                  1000: 10.0*10**3},
                         'mode':'write'},
                        'estim_init_pm':
                        {'name':'Estim. initial costs per meter',
                         'table':{1:    10000.0,
                                  100:  5000.0,
                                  1000: 2500.0,
                                  10000:1500.0,
                                  100000:500.0},
                         'mode':'write'},
                        'estim_pa_ex':
                        {'name':'Estim. extra costs per year',
                         'table':{1:   80.0,
                                  10:  40.0,
                                  100: 20.0,
                                  1000:10.0},
                         'mode':'write'},
                        'estim_pa_pm':
                        {'name':'Estim. costs per meter per year',
                         'table':{1:   80.0,
                                  10:  40.0,
                                  100: 20.0,
                                  1000:10.0},
                         'mode':'write'},
                        
                        'estim_init':{'name':'Estim. initial costs',
                                      'value':0.0,# set in do_costs_current
                                      'mode':'read'},
                        'estim_pa':{'name':'Estim. costs per year',
                                    'value':0.0,# set in do_costs_current
                                    'mode':'read'},
                        },
                        lifetime_estim=10)
    def activate_specific(self):
        """
        Track element specific activation.
        """
        if self.time_sync>0:
            self.start_sync_mode()
        #else:
        #    self.stop_sync_mode()

    def unactivate_specific(self):
        """
        Track element specific activation.
        """
        self.obj.del_timer_arrays(self)
    
    def update_cars(self,car,data={}):
        """
        Update track element wide carrier database with data.
        If there is no data entry for the respective carrier,
        then it will be created.
        """
        if not self.cars.has_key(car.id):
            #if self.probe():
            #    print 'update_cars: create data for',car.id
            self.cars[car.id]={#'mod':car,
                               'state':'passthrough',
                               'berth_allocated':None,
                               'loadberth_allocated':None,
                               'berth_occupied':None,
                               'found_in':'',
                               'loaded':0
                               }
            self.cars[car.id].update(data)

        else:
            #if self.probe():
            #    print  'update_cars: ', car.id
            #    print  '  state:',self.cars[car.id]['state']
            #    print  '  loaded',self.cars[car.id]['loaded']
                #for key in   self.cars[car.id].keys():
                #    print '   ',key,':',self.cars[car.id][key]
            self.cars[car.id].update(data)

    def update_man_specific(self,message):
        '''
        Update track management with connection info of track element.
        This function is called during activation or after topological
        changes.
        '''
        #print 'update_man_specific',self.id
        message['class']='stop'
        message['zone']=self.zone
        message['name']=self.name
        message['n_berth']=len(self.berths)
        
    def delete_from_cars(self,car):
        """
        Delete car from track element wide carrier database 
        """
        del self.cars[car.id]
        
    def start_sync_mode(self,time_sync=5.0):
        """
        start synchronized mode.
        """
        if self.probe():
            print '\n\n------start_sync_mode for',self.id,'at t=',self.sim.time
        
        #if self.time_sync<=0.0:
        #    # compute automatically synctime
        #    # this is a somewhat arbitrary equation
        #    #self.time_sync=240.0+2.0*self.berthssec.berth_no
        #else:
        
        
        
            
        #if 1:
            
        #if not self.is_blocked_unloadzone: 
        # now check if there are searching, unallocated cars
        carlist=copy(self.berthssec.cars)
        #carlist.reverse()
        for car in carlist:
            #if car!=triggercar:
            if (self.cars[car.id]['state']=='allocated'):
                    berth=self.cars[car.id]['berth_allocated']
                    path_sec_min=car.get_path_sec()+car.get_brakedist_comfort()+0.1
                    
                    if path_sec_min < self.berths[0]['halt_point']:
                        self.unalloc_berth(berth,car,{})
                        cd={'velocity_line':
                            self.secs_dir['berths'].velocity_line,
                            'mark_at':self.berths[0]['halt_point']-1.0,
                            'stop_at':self.berths[0]['halt_point']-0.5}
                        self.cars[car.id]['state']='searching'
                        cd.update(self.alloc_berth_with_car_sync(\
                            car,cd))
                        car.control(cd) 
                
                        
                        
                        if self.probe():
                            print '  unallocated car',car.id,\
                                'from berth',berth['ind'],path_sec_min<berth['halt_point']
                        
        
        #insert synch time in array structure
        # in the new scheme after time_sync the waiting cars
        self.time_sync=time_sync 
        # in unload zone will be asked to move to load zone 
        self.obj.add_timer_arrays(self,self.time_sync)
        
        
        
        # initialize synch mode: block everything, then
        # call functions that verify and set the rest..
        self.is_blocked_unloadzone=True
        
        self.is_blocked_loadzone=True
        self.ind_berth_alloc=self.berthssec.berth_no-1 
        #self.ind_berth_alloc=self.ind_berth_firstalloc
        
        self.is_unloadzone_free()
        self.is_loadzone_free()
        
    
        #self.unblock_loadzone()
        
        #
        #self.ind_loadberth_alloc=0
        #self.liberate_all=1
                                 
        # start with the liberation phase
        #self.start_liberation()

    def stop_sync_mode(self,triggercar=None):
        """
        start synchronized mode.
        Only used from parameter change
        """
        #print 'stop_sync_mode for',self.id,'at t=',self.sim.time
        if self.probe():
            print '\n\n------stop_sync_mode for',self.id,'at t=',self.sim.time
        # remove sync specific values
        self.time_sync=0.0
        #self.obj.del_timer_arrays(self)
        
        # put stop in normal allocation phase
        #self.start_allocation()
        # old:self.ind_berth_alloc=self.ind_berth_firstalloc
        #self.ind_berth_alloc=self.berthssec.berth_no-1 
        
        
        # now check if there are searching, unallocated cars
        triggercd={}
        for car in self.berthssec.cars:
            if self.cars[car.id]['state']=='searching':
                # use here non-synchronous allocation!
                cd=self.alloc_berth_with_car(car,{})
                # there may be still some cars at zero speed...
                cd.update({'velocity_line':\
                           self.secs_dir['berths'].velocity_line})
                if car!=triggercar:
                    car.control(cd)
                else:
                    triggercd=cd.copy()
    
        # there may be a previous berth demand list which is no more valid
        self.berths_demand=[]
        # get berths with berth that have a "long" queue
        berths=self.get_berths_with_queue(length_queue=1)
        ## reverse list such that remotest berth gets first priority 
        berths.reverse()
        # create berths_demand with berth that are not allocated already
        for berth in berths:
            if not berth['allocated']:
                self.berths_demand.append(berth)

        # clear all berth from waiting carriers
        # self.clear_berths(0)

    def is_start_sync(self):
        """
        Verify if it is time to switch to sync mode.
        """
        n_berths_with_queue = len(self.get_berths_with_queue(length_queue=1))
        n_berths_keep_clear = self.berthssec.localsec['berths_keep_clear']
        crit= n_berths_with_queue >= self.berthssec.berth_no - n_berths_keep_clear
        if self.probe():
            print '************* is_start_sync', self.id
            print '  n_berths_with_queue,n_berths_keep_clear,crit',\
                    n_berths_with_queue,n_berths_keep_clear,crit
            
        
        if self.time_sync==0.0: 
            return crit
                    
        else:
            return False
        
        # old:
        #((len(self.berthssec.cars)\
        #  +len(self.secs_dir['brake'].cars) ) \
        # > 1*self.berthssec.berth_no)

    def is_stop_sync(self):
        """
        Verify if it is time to switch off sync mode.
        """
        n_berths_with_queue = len(self.get_berths_with_queue(length_queue=1))
        crit= n_berths_with_queue < self.berthssec.berth_no - self.ind_berth_firstalloc-1
        
        if self.probe():
            print '************* is_stop_sync',self.id
            print '  n_berths_with_queue,ind_berth_firstalloc,crit',\
                    n_berths_with_queue,self.ind_berth_firstalloc,crit
            
        return crit
        
        #old
        #        &\
        # ((len(self.berthssec.cars)+len(self.secs_dir['brake'].cars) ) \
        #  < 0.5*self.berthssec.berth_no)
    
        
    def start_liberation_old(self):
        """
        start berth liberation phase of sync mode.
        from now on no new allocation will be accepted unless
        all berths are liberated.
        """
        
        do_liberate=0
        if self.liberate_all:
            do_liberate=not self.are_all_berth_liberated()
        else:
            do_liberate=not self.are_all_unloadberth_unallocated()
        
        
        
        
        if self.probe():
            print  'start_liberation: ind_berth_alloc=',\
                  self.ind_berth_alloc,'at time',self.sim.time
            print '  liberate_all,do_liberate=',self.liberate_all,do_liberate
            print '  are_all_berth_liberated=',self.are_all_berth_liberated()
            print '  are_all_unloadberth_unallocated=',self.are_all_unloadberth_unallocated()
        
        
        
        # check if all berth are liberated?
        if do_liberate: 
            # by setting berth allocation counter to -1
            #self.ind_berth_alloc=-1

            # set timer on a very high value, so that it will not get triggered
            self.obj.add_timer_arrays(self,10.0**9)
            
            # and liberating waiting carriers
            if self.liberate_all:
                # all berth
                if self.probe():
                        print ' start_liberation_old2'
                self.clear_berths(0)
                self.liberate_all=0
            else:
                # only load berth
                if self.probe():
                        print '  start_liberation_old'
                self.clear_berths(self.ind_berth_firstalloc+1)

            # reduce line velocity at mein exit section
            # in order to favour
            #self.secs_dir['main_exit'].set_velocity_line(
            #    0.5*self.secs_dir['main_exit'].velocity_line)
        else:
            #print 'WARNING in start_liberation of',self.id
            #print '  stop are already in liberation phase...start_allocation'
            #self.start_allocation()
            pass
        
    def start_allocation_old(self,triggercar=None):
        """
        start berth allocation phase of sync mode.
        from now on allocations are accepted until:
        - all berths are allocated or
        - sync timer runs out. 
        """
        if self.probe():
                print 'start_allocation: ind_berth_alloc=',\
                      self.ind_berth_alloc,'at time',self.sim.time

        # after the first allocation phase, only load berth must be liberated 
        #self.liberate_all=0

        triggercd={}

        
        #if len(self.get_berths_with_queue(length_queue=1))>=\
        #   self.ind_berth_firstalloc-1:
        if 1:    
            # set berth allocation counter 
            
            #self.ind_berth_alloc=self.berthssec.berth_no-1 
            
            # old          
            #self.ind_berth_alloc=self.ind_berth_firstalloc
            
            # start timer
            #if self.probe():
            #    print '  start timer with',self.time_sync
            #self.obj.add_timer_arrays(self,self.time_sync)
            
            # restore line velocity at main exit section
            # in order to favour
            #self.secs_dir['main_exit'].set_velocity_line(
            #    2.0*self.secs_dir['main_exit'].velocity_line)

            
            # now check if there are searching, unallocated cars
            for car in self.berthssec.cars:
                if self.cars[car.id]['state']=='searching':
                    cd=self.alloc_berth_with_car_sync(car,{})
                    cd.update({'velocity_line':\
                              self.secs_dir['berths'].velocity_line})
                    if car!=triggercar:
                        car.control(cd)
                    else:
                        triggercd=cd.copy()
            
            
        #else:
        #    # too low demand at berth...
        #    triggercd=self.stop_sync_mode(triggercar)

        return triggercd
        
        
    def handle_timer(self):
        """
        Called on timeout of timer during iterate.
        """
        
        if self.time_sync>0.0:
            # only in synch mode
            if self.probe():
                print '\n*** handle_timer',self.id
            #self.unblock_loadzone()
            self.is_loadzone_free()
            self.is_unloadzone_free()
            self.obj.add_timer_arrays(self,self.time_sync)
            
        # is stop in allocation phase?
        
        #if self.ind_berth_alloc>=0:
        #    self.start_liberation()
        #else:
        #    # stop is in liberation phase
        #    # set timer on a very high value, so that it will not get triggered
        #    self.obj.add_timer_arrays(self,10.0**9)
            
  
    def alloc_car_for_berth(self,berth,only_loaded=0):
        """
        Returns the next upstream searching carrier for a given berth.
        """
        #if self.probe():
        #        print 'alloc_car_for_berth:'
        #        print '  Try to allocate berth',berth['ind'],\
        #              'at',berth['halt_point'],not berth['allocated']
                      
        car_found=None
        if not berth['allocated']:
            

            # get next upstream carrier with respect to this berth
            ind_nextcar=self.get_ind_nextcar_upstream(berth)
            
            # go through all upstream carriers with valid index
            # ind_nextcar=-1 means that no upstream carrier has been found
            while (not car_found) &\
                  (ind_nextcar < len(self.berthssec.cars))&\
                  (ind_nextcar != -1): 
                nextcar=self.berthssec.cars[ind_nextcar]
                state_nextcar=self.cars[nextcar.id]['state']
                #status_nextcar=self.get_status_car_in_berthssec(nextcar)
                #print '      nextcar.id=',nextcar.id
                #print '          status=',status_nextcar['state']
                       
                # is this the car still present at berth
                if nextcar==berth['car']:
                    # try next car upstream
                    ind_nextcar+=1
                           
                elif (state_nextcar=='searching')&\
                         (not
                          ((self.cars[nextcar.id]['loaded']==0)&\
                           only_loaded)):
                    car_found=nextcar
                    #if 1:#self.probe():
                    #    print '    Allocate berth',berth['ind'],\
                    #          'with searching', nextcar.id,\
                    #          'at',nextcar.get_path_sec()

                    # ..and make sure the carrier is not stopped   
                    cd=self.alloc_berth(berth,nextcar,
                                        {'velocity_line':\
                                         self.berthssec.velocity_line},
                                        'berths_demand')
                    nextcar.control(cd)
                    
                    

                elif (state_nextcar=='waiting')&(not only_loaded):
                    car_found=nextcar
                    berth_found=self.cars[nextcar.id]['berth_occupied']
                    #if 1:#self.probe():
                    #    print '    Liberate berth',berth_found['ind']
                    #    print '    Allocate berth',berth['ind'],\
                    #          'with waiting', nextcar.id,\
                    #          'at',nextcar.get_path_sec()
                    self.unoccupy_berth(berth_found)
                    self.alloc_berth(berth,car_found,None,'berths_demand')
                else:
                    #if self.probe():
                    #    print '    carrier unsuitable for allocation',\
                    #          nextcar.id
                    #    print '      with path_sec',nextcar.get_path_sec()
                    # try next car upstream
                    ind_nextcar+=1

            if (not car_found)&(not only_loaded ):
                # no car has been allocated for this berth
                # make sure that berth is in berths_demand list
                if self.berths_demand.count(berth)==0:
                    #if 1:#self.probe():
                    #    print '   put berth',berth['ind'],berth['allocated'],\
                    #          'in berths_demand list'
                    self.berths_demand.append(berth)
                #else:
                #    print 'WARNING: in alloc_car_for_berth', self.id
                #    print '  berth',berth['ind'],'already in berths_demand'

            else: # a car has been allocated for this berth
                # make sure this stop is no more in berths_demand
                if self.berths_demand.count(berth)>0:
                    self.berths_demand.remove(berth)
        else:
            # OK, beth just tried, without checking whether
            # it is already allocated
            #print 'WARNING in alloc_car_for_berth',self.id
            #print '  berth',berth['ind'],'already allocated.'
            pass
        
        #if self.probe():
        #    print '  berths_demand=',self.berths_demand,
        #    if car_found: print car_found.id
        #    else: print 'no car found'
        
        return car_found


    
    def probe(self):
        """
        Returns 1 if we want output.
        """
        return False
        
        # modena 1
        #return  (self.id=='test.track.stop5berth_leftside/38')
        #return  (self.id=='test.track.stop10berth_rightside/10')
        #return  (self.id=='test.track.stop5berth_leftside/5')
        #return  (self.id=='test.track.stop5berth_rightside/1')
        #return 1
        
        #return  (self.id=='test.track.stop10berth_short_left/1')
        
        # 2 stop  test
        #return  (self.id=='test.track.stop10berth_short_left/4')
        #return  (self.id=='test.track.stop10berth_short_left/5')
    
    
        #return  (self.id=='test.track.stop20_right_v16/3')
        #return  (self.id=='test.track.stop20_right_v16/6')
        #eturn  (self.id=='test.track.stop3berth_leftside/1')
        #return (self.id=='test.track.stop5berth_leftside/1')
        #if find(self.id,'10berth'):
        #    return 1
        #else:
        #   return 0 
        #return  (self.id=='test.track.stop10berth_rightside/3')

    def move_waiting_cars(self,ind_berth):
        """
        Check if waiting cars will possibly get a new allocation,
        from the given berth downstream.
        Send also away carriers if there are too many.
        """
        # in this case check whether there are waiting carriers
        # which could or should move on...
        #state=self.cars[car.id]['state']

        
        # self.berthssec.berth_no
        #if self.probe():
        #    print 'move_waiting_cars:',self.id,ind_berth,n_sendaway
            
        #if state in ('waiting','loading'):
        #    berth=self.cars[car.id]['berth_occupied']
        #else:
        #    berth=self.get_nextberth_downstream(car)
            
        if self.time_sync<=0.0: # only in async mode, otherwise timeout is used
            # make a rough calc in how many waiting
            # carriers should be send away
            n_sendaway=len(self.berthssec.cars)- self.berths_occupy-1
            #if self.probe():
            #    print '  Try to make waiting cars moving',berth['ind']
            # go through downstream berth and check for waiting car
            # start with last berth
            
            
            #for ind in range(self.berthssec.berth_no-1,0,-1):
            for ind in range(self.berthssec.berth_no-1,ind_berth-1,-1):
                checkberth=self.berths[ind]
                checkcar=checkberth['car']
                # go ahead if carrier is waiting 
                if (checkberth['status']=='waiting'):
                    #if self.probe():
                    #  print '  check berth',checkberth['ind'],n_sendaway
                    if n_sendaway>0:
                        self.unoccupy_berth(checkberth)
                        # make sure car will leave berth
                        #print 'move_waiting_cars: move',checkcar.id
                        self.update_cars(checkcar,{'state':'passthrough'})
                        n_sendaway-=1
                    else:
                        # there is a waiting car, try to find a new berth
                        #berth_found,found_in=self.find_berth_downstream(\
                        #    checkberth)
                        berth_found,found_in=\
                             self.find_accessible_berth_downstream(checkberth)
                                
                        #if found_in in ('berths_demand','liberated'):
                        #    # great, found a berth ahead
                        #    # but allocate only if freely accessible
                        #    
                        #    if self.has_free_berthaccess(checkcar,
                        #                                     berth_found):
                        #        if self.probe():
                        #            print '  Liberate berth',checkberth['ind']
                        #            print '    allocate berth',\
                        #                 berth_found['ind'],'with',checkcar.id
                        if berth_found:
                            self.unoccupy_berth(checkberth)
                            self.alloc_berth(\
                                          berth_found,checkcar,None,found_in)
                                
        
    def has_free_berthaccess(self,car,berth):
        """
        Returns 1 if there is no other carrier between car and berth.
        """

        
        n=self.berthssec.cars.index(car)
        free=1
        if n>0:
            status_nextcar=self.cars[self.berthssec.cars[n-1].id]
            if status_nextcar['berth_occupied']:
                if status_nextcar['berth_occupied']['ind']<berth['ind']:
                    free=0
            elif status_nextcar['berth_allocated']:
                if status_nextcar['berth_allocated']['ind']<berth['ind']:
                    free=0
            
        return free

        #nextcar=self.berthssec.cars[n-1]
        #return (nextcar.get_path_sec()-nextcar.length)>berth['halt_point']

            
    def alloc_berth_with_car(self,car,controldata=None,mode=''):
        """
        Searches a suitable berth ahead of car, allocates and returns the
        aprobriate control data.
        """
        state=self.cars[car.id]['state']
        
        #if self.probe():
        #print self.id,'alloc_berth_with_car:',car.id,state,'at',int(self.sim.time),'s'
        #print '  simtime:',self.sim.time
        
        # check if car will possibly get a new allocation:
        #if 1: #state in ('searching','waiting'):
        cd={}   
                            
        # is carrier loaded at a berth and wants to go ahead?
        # or has a mission as an empty vehicle
        if state in ('loading','empty'):
            berth=self.cars[car.id]['berth_occupied']

            
            
            if state=='loading':
                # because loading is aparently terminated
                self.update_cars(car,{'loaded':1})
            else:
                #print 'alloc_berth_with_car: empty mission at',self.name,car.id
                # carrier got a mission empty-> treat now as passthrough
                self.update_cars(car,{'state':'passthrough'})

            # put berth explicitely in waiting mode, otherwise it worn't be
            # liberated
            berth['status']='waiting'
            
            #if self.probe():
            #print '  unoccupy_berth',berth['ind']
            #    print '  start car'
                
            cd=self.unoccupy_berth(berth,{})

            # is this not the last berth
            if berth['ind']!=self.berthssec.berth_no-1:
                if self.probe():
                    print '  5clear berth from ind',berth['ind']+1
                self.clear_berths(berth['ind']+1)

                # NO!! can't just unallocate allocated cars
                # but this is not enough, we also need to unallocate
                # already allocated empty berth, otherwise an allocated carrier
                # in front will stop at an empty berth and wait....!!!
                #self.unallocate_empty_berths(car)
        
        
        
        elif state=='passthrough':
            #if self.probe():
            #    print '  accelerate car to linespeeed'

            cd=self.berthssec.search_car_infront(
                car,{'velocity_line':self.berthssec.velocity_line})

            #cd={'velocity_line':
            #    self.secs_dir['berths'].velocity_line}

            # is there a berth with possibly waiting cars ahead
            nextberth=self.get_nextberth_downstream(car)
            if nextberth:
                if self.probe():
                    print '  4:clear berth from ind',nextberth['ind']+1
                self.clear_berths(nextberth['ind'])
                
        elif state=='waiting':
            # a waiting carrier must always try to allocate a berth
            # further ahead.
            berth=self.cars[car.id]['berth_occupied']
            

            # are there berths looking for a carrier?
            berth_found,found_in=self.find_accessible_berth_downstream(berth)
            
            # first check whether there is still demand for cars,
            # with a criteria that is slightly more generous than
            # the one in get_nextsec()
            passthrough=0

            #if (len(self.berthssec.cars) < self.berths_occupy+1)|\
            #   (len(self.berths_demand)>len(self.berthssec.cars)-1):
            #    # there is demand
            #    passthrough=0
            #else:
            #    print '  passthrough because no demand'
            #    passthrough=1

            # now check if there is a pass-through carrier,
            # or a searching car with load behind.
            # in this case the carrier must be passthrough or it must
            # find a demand berth

            ind=self.berthssec.cars.index(car)
            #cars=self.berthssec.cars
            while (ind<len(self.berthssec.cars))&(not passthrough):
                car_status=self.cars[self.berthssec.cars[ind].id]
                passthrough=(car_status['state']=='passthrough')|\
                             ((car_status['state']=='searching')&\
                              car_status['loaded'])
                
               
                ind+=1
                # set also passthrough if a vehicle allocated a berth
                # in front of this berth..should have been avoided by
                # move_waiting_cars call of allocating car
                #if (not passthrough)&(car_status['state']=='allocated'):
                #     passthrough=car_status['berth_allocated']['ind']>=\
                #                  berth['ind']
                                 
                    
                
            
            if (berth_found==berth):
                # found same berth...uninteresting
                found_in=''
                berth_found=None
                            
            if found_in=='berths_demand':
                # great, found a berth in demand list
                #self.berths_demand.remove(berth_found)
                #if self.probe():
                #    print '  Liberate berth',berth['ind']
                #    print '  allocate demand berth',berth_found['ind']
                cd=self.unoccupy_berth(berth,{})
                cd=self.alloc_berth(berth_found,car,cd,found_in)
                #self.berths_demand.remove(berth_found)
                
            elif  (found_in=='liberated')&(not passthrough):
                # found a free berth ahead,
                # but allocate only if there are no passthrough cars behind
                #if self.has_free_berthaccess(car,berth_found):
                #if self.probe():
                #    print '  Liberate berth',berth['ind']
                #    print '  allocate liberated berth',berth_found['ind']
                cd=self.unoccupy_berth(berth,{})
                cd=self.alloc_berth(berth_found,car,cd)
                #else:
                #    found_in=''
                #    berth_found=None
            
            elif  (found_in=='occupied')&(not passthrough):
                # found an occupied berth ahead...
                # better remain 
                found_in=''
                berth_found=None
            

                    
            elif passthrough:
                # there are passthrough cars behind, but
                # no demand berth ahead could be found
                #if self.probe():
                #    print '  Liberate berth',berth['ind']
                #    print '  passthrough carrier and clear',car.id

                cd=self.unoccupy_berth(berth,{})
                self.update_cars(car,{'state':'passthrough'})
                found_in=''
                berth_found=None

                # in front of a passthrough car cannont be a waiting car
                if berth['ind']!=self.berthssec.berth_no-1:
                    if self.probe():
                        print '  alloc_berth_with_car:finally clear berth3',berth['ind']
                    self.clear_berths(berth['ind']+1)
                
                
            else:
                #if self.probe():
                #    print 'remain waiting at berth???',berth['ind']
                # there is still a chance to swap allocations
                # SEE BELOW..
                
                found_in=''
                berth_found=None



            # now to prevent that a carrier behind the carrier, has allocated
            # a berth in front of the carrier
            if berth_found:
                # try to find car to swap allocation with berth_found
                swapcar,swapberth=self.find_car_to_swap(car,berth_found)
                if swapcar:
                    #if self.probe():
                    #    print '  swap allocations:'
                    #    print '    unalloc found berth',berth_found['ind'],\
                    #          'from',car.id
                    #    print '    unalloc swapberth',swapberth['ind'],\
                    #          'from',swapcar.id
                    #    print '    alloc berth',swapberth['ind'],'with',car.id
                    #    print '    allocate found berth',berth_found['ind'],\
                    #          'with',swapcar.id
                    
                    cd=self.unalloc_berth(berth_found,car,cd)
                    swapcd=self.unalloc_berth(swapberth,swapcar,{})
                    cd=self.alloc_berth(swapberth,car,cd)
                    swapcd=self.alloc_berth(berth_found,swapcar,swapcd)
                    # apply control data for swapcar here
                    swapcar.control(swapcd)

                    # swap also found_in info:
                    found_in=self.cars[car.id]['found_in']
                    self.cars[car.id]['found_in']=\
                               self.cars[swapcar.id]['found_in']
                    self.cars[swapcar.id]['found_in']=found_in

            # car is still in berth means there is no passthrough
            elif  self.cars[car.id]['berth_occupied']:
                # now try to find a car behind that allocates a berth in front
                # if this is the case let the found car allocate this berth
                # and the car of this berth will allocate tha one of the
                # found car
                swapcar,swapberth=self.find_car_to_swap(car,berth)
                if swapcar:
                    # allocate car to the allocated berth of swapcar
                    # and allocate swapcar to the liberated berth of car
                    #if self.probe():
                    #    print '  swap allocations:'
                    #    print '    unoccupy present berth',berth['ind'],\
                    #          'from',car.id
                    #    print '    unalloc swapberth',swapberth['ind'],\
                    #          'from',swapcar.id
                    #    print '    alloc berth',berth['ind'],'with',swapcar.id
                    #    print '    alloc swapberth',berth['ind'],\
                    #          'with',car.id
                    
                    cd=self.unoccupy_berth(berth,cd)
                    swapcd=self.unalloc_berth(swapberth,swapcar,{})
                    swapcd=self.alloc_berth(berth,swapcar,swapcd)
                    cd=self.alloc_berth(swapberth,car,cd)
                    # apply control data for swapcar here
                    swapcar.control(swapcd)

                    # swap also found_in info:
                    found_in=self.cars[car.id]['found_in']
                    self.cars[car.id]['found_in']=\
                               self.cars[swapcar.id]['found_in']
                    self.cars[swapcar.id]['found_in']=found_in
                    
                else:
                    pass
                    #if passthrough:
                    #    # no stop found, but cannot remain....
                    #    cd=self.unoccupy_berth(berth,cd)
                    #else:
                    #    # waiting carrier found no berth and cannot swap
                    #    #...remain waiting
                    #    pass
            
            if self.cars[car.id]['state']!='passthrough':
                # now, if car has found a motivated allocation,
                # it must make sure that there are no waiting cars in the way
                if self.cars[car.id]['state']=='allocated':
                    if not self.has_free_berthaccess(car,\
                                         self.cars[car.id]['berth_allocated']):
                        
                        if berth['ind']!=self.berthssec.berth_no-1:
                            if self.probe():
                                print '  alloc_berth_with_car2:finally clear berth',berth['ind']
                            self.clear_berths(berth['ind']+1)
                        
                # in case carrier remained in berth, verify if there is
                # still a car behind with an allocation in front.
                if self.cars[car.id]['state']=='waiting':
                    passthrough=0
                    ind=self.berthssec.cars.index(car)
                    while (ind>=0)&(not passthrough):
                        status_checkcar=self.cars[self.berthssec.cars[ind].id]
                        if status_checkcar=='allocated':
                            if (status_checkcar['berth_allocated']['ind']>\
                                berth['ind']):
                                 passthrough=1
                        #elif status_checkcar['state']=='passthrough':
                        #    passthrough=1
                        ind-=1  
                                
                    if passthrough:
                        #if self.probe():
                        #    print '  finally Liber. berth',berth['ind']
                        #    print '  passthrough carrier and clear',\
                        #          car.id
                        cd=self.unoccupy_berth(berth,{})
                        self.update_cars(car,{'state':'passthrough'})
                        if berth['ind']!=self.berthssec.berth_no-1:
                            if self.probe():
                                print 'alloc_berth_with_car:clear_berths'
                            self.clear_berths(berth['ind']+1)

            if (berth['status']=='liberated')&(not berth['allocated']):
                # berth has been liberated from a waiting carrier
                # means that there is no demand here
                # however, there may be carriers without allocation waiting

                # so try to allocate berth with a searching car
                self.alloc_car_for_berth(berth,only_loaded=1)
                #pass
                
                
                        
                    
        elif state=='searching':
            # get first berth where car can possibly halt
            berth=self.get_nextberth_downstream(car)


            
            if berth:
                # try to move on waiting carriers
                #self.move_waiting_cars(berth)
                cd=self.try_alloc_berth_with_searching_car(car,berth)
                if not cd:
                    print 'WARNING in alloc_berth_with_car',self.id
                    print '  searching car',car.id,\
                          '  could not be allocated at time',\
                          int(self.sim.time),'s'
                    print '  ...blocked  until allocated later'
                    #self.sim.stop()
                    #print '  halt vehicle...and wait' 
                    self.update_cars(car,{'state':'blocked'})
                    self.flag_cars_blocked=1
                    cd={'velocity_line':0}
            else:
                print 'WARNING in alloc_berth_with_car',self.id
                print '  searching car',car.id,\
                      '  could not find a berth at time',\
                      int(self.sim.time),'s'
                print '  ...blocked  until allocated later'
                #self.sim.stop()
                #print '  halt vehicle...and wait' 
                self.update_cars(car,{'state':'blocked'})
                self.flag_cars_blocked=1
                cd={'velocity_line':0}
          
        
        else: # strange state???
            # passthrough: do nothing
            cd={}

         
                    
        # immediately control carrier, if no controldata provided
        if cd:
            if controldata==None:
                controldata={}
                controldata.update(cd)
                car.control(controldata)
            else:
                # just update controldata from input
                controldata.update(cd)

            #if self.probe():
            #    for key in controldata.keys():
            #        print '      ',key,'=',controldata[key]

        return controldata

    def try_alloc_berth_with_searching_car(self,car,berth):
        """
        Used by  alloc_berth_with_car in case of a searching car.
        Returns control data if a berth could be allocated,
        otherwise empty.
        """
        #if self.probe():
        #    print 'try_alloc_berth_with_searching_car',car.id,berth['ind']
        cd={}
        
        if berth:
            # test:
            #inds=[]
            #for b in self.berths_demand:
            #    inds.append(b['ind'])
            #print '  berths_demand indexes:',inds

            # are there berths looking for a carrier?
            
            berth_found=None
            #if self.cars[car.id]['loaded']:
            #    # try to get a freely accessible berth
            #    berth_found,found_in=self.find_accessible_berth_downstream(\
            #        berth,included=1)
                
            if not berth_found:
               
                    
                berth_found,found_in=self.find_berth_downstream(\
                    berth,included=1)
                
                #if berth_found:
                #    if self.probe():
                #        print 'try_alloc_berth_with_searching_car',self.id
                #        print '  find_berth_downstream',berth_found['ind'],\
                #              ' from berth',berth['ind'],found_in
                #else:
                #print '  find_berth_downstream -',\
                #      ' from berth',berth['ind']
        else:
            #print '  no next accessible berth detected'
            berth_found=None
            found_in=''

        if found_in=='berths_demand':
            #self.berths_demand.remove(berth_found)
            #if 1:#self.probe():
            #    print '  allocate demand berth',berth_found['ind']
            cd=self.alloc_berth(berth_found,car,\
                                {'velocity_line':
                                 self.berthssec.velocity_line},found_in)
            #self.berths_demand.remove(berth_found)
                
        elif  found_in in ('liberated','occupied'):
            #if 1:#self.probe():
            #    print '  allocate liberated berth',berth_found['ind']
            cd=self.alloc_berth(berth_found,car,\
                                {'velocity_line':\
                                 self.berthssec.velocity_line})

        # clear berth if found berth is not freely accessible or if
        # no berth could be found
        if berth_found:
            #if self.probe():
            #    print '  check free access from',berth['ind'],\
            #        ' to berth',berth_found['ind'],\
            #        self.has_free_berthaccess(car,berth_found)
            if not self.has_free_berthaccess(car,berth_found):
                if self.probe():
                    print 'try_alloc_berth_with_searching_car:clear_berths'
                self.clear_berths(berth['ind'])
        elif berth:
            if berth['ind']!=self.berthssec.berth_no-1:
                if self.probe():
                    print 'try_alloc_berth_with_searching_car:clear_berths'
                self.clear_berths(berth['ind']+1)
        return cd
                
    
    def alloc_berth_with_car_sync(self,car,controldata=None,mode=''):
        """
        Searches a suitable berth ahead of car in synchronous mode,
        allocates and returns the aprobriate control data.
        """
        state=self.cars[car.id]['state']
        if self.probe():
            print 'alloc_berth_with_car_sync:',car.id,state,self.is_blocked_unloadzone,self.is_blocked_loadzone
        #    # are we in the liberation phase?
        #    if self.ind_berth_alloc==-1:
        #        print '  liberation phase'
        #    else:
        #        print '  allocation phase'

           
                            
        # is carrier loaded at a berth and wants to go ahead?
        if state in ('loading','empty'):
            berth=self.cars[car.id]['berth_occupied']

            if state=='loading':
                # because loading is aparently terminated
                self.update_cars(car,{'loaded':1})
            else:
                # carrier got a mission empty-> treat now as passthrough
                self.update_cars(car,{'state':'passthrough'})
                #cd=self.berthssec.search_car_infront(
                #    car,{'velocity_line':self.berthssec.velocity_line})

            
            # are we in the liberation phase?
            #if self.ind_berth_alloc==-1:
            # because loading is aparently terminated
            # put berth explicitely in waiting mode, otherwise it worn't be
            # liberated
            berth['status']='waiting'
            
            #if self.probe():
            #print '  unoccupy_berth',berth['ind']
            #    print '  start car'
            
            cd=self.unoccupy_berth(berth,{})

            # waiting cars in front of a leaving loaded car are not tolerated 
            # in load area if this not the last berth
            if  berth['ind']>self.ind_berth_firstalloc:
                if berth['ind']!=self.berthssec.berth_no-1:
                    if self.probe():
                        print '  clear berth from ind',berth['ind']+1
                    self.clear_berths(berth['ind']+1)

        
        
        elif state=='passthrough':
            #if self.probe():
            #    print '  accelerate car to linespeeed'

            cd=self.berthssec.search_car_infront(
                car,{'velocity_line':self.berthssec.velocity_line})

            #cd={'velocity_line':
            #    self.secs_dir['berths'].velocity_line}

            # is there a berth with possibly waiting cars ahead
            #nextberth=self.get_nextberth_downstream(car)
            #if nextberth:
            #    #if self.probe():
            #    #    print '  clear berth from ind',nextberth['ind']+1
            #    self.clear_berths(nextberth['ind']+1)
                
        elif state =='waiting':
            
            berth=self.cars[car.id]['berth_occupied']
            
            # when berth is in unload region
            if  berth['ind']<=self.ind_berth_firstalloc:
                
                if self.cars[car.id].has_key('loadberth'):
                    allocberth =self.cars[car.id]['loadberth']
                    if allocberth['allocated']:
                        # the desired berth in loadzone is already allocated
                        # now try to move car ahead within UNloadzone:
                        allocberth=self.get_nextberth_unloadzone_synch(berth=berth) 
                else:
                    # get next possible nonallocated berth in loadzone to allocate
                    allocberth=self.get_nextberth_loadzone_synch() 
                    
                   
                #else:
                #    if allocberth['allocated']: 
                #        # get next possible nonallocated berth in loadzone to allocate
                #        #??? just leave it and wait till required beth can be allocated
                #        allocberth=None  
                
                if allocberth:
                    # found alloc berth means not blocked
                    if self.probe():
                        print '  ind_berth_alloc=',self.ind_berth_alloc
                        print '  current berth',berth['ind'],' is in unload area'
                        print '  try to allocate berth',allocberth['ind']
                              
                              
                    if not allocberth['allocated']:
                        
                        
                        
                        cd=self.unoccupy_berth(berth,{})
                        
                        cd=self.alloc_berth(allocberth,car,cd)
                        if self.probe():
                            print '  unoccupy_berth',berth['ind']
                            print '  allocate berth',allocberth['ind']
                            print '  result for car:',car.id,self.cars[car.id]['state'],cd
                       
                    else:
                        cd={'velocity_line':0.0,}
                        
                        if self.probe():
                            print '  berth',allocberth['ind'],'cannnot be allocated'
                            print '  ...stop',car.id
                            print '  result for car:',car.id,self.cars[car.id]['state'],cd
                        #if self.probe():
                        #print '  berth',allocberth['ind'],'in load area still allocated...'
                        #print '  wait until it gets unoccupied'
                    
                        
                
                else:
                    # loadzone blocked do nothing 
                    
                        
                    cd={'velocity_line':0.0,}  
                    if self.probe():
                        print '  no allocation possible...remains waiting',car.id,cd
                    
                        
                    
            else: # car is waiting in loading area
                # is stop in allocation phase?
                #if self.ind_berth_alloc>=0:
                #if self.probe():
                #print '  berth',berth['ind'],' is in load area wait to be loaded or sent off'
                cd={}

                    
        elif state=='searching':
            
            # first check if there are other searching cars in front
            ind_car=self.berthssec.cars.index(car)
            blocked=False
            if  car != self.berthssec.cars[0]:
                if self.probe():
                    print '  check if searching cars in front',ind_car,range(ind_car-1,-1,-1)
                
                for ind in range(ind_car-1,-1,-1):
                    car_check=self.berthssec.cars[ind]
                    # is car allocated in loadzone and in front of berth
                    if self.cars[car_check.id]['state']=='searching':
                        blocked=True
                        # block unload zone here otherwise ity won't get unblocked
                        self.is_blocked_unloadzone=True
                        break
                    if self.probe():
                        print '  check car',car_check.id,self.cars[car_check.id]['state']
                
                if self.probe():
                    print '  -> no searching cars in front.'
                
            if not blocked:
                # no searching car in front    
                berth=self.get_nextberth_unloadzone_synch()
                
            else:
                berth=None
                
            # try to unblock waiting cars in unloadzone
            # this may happen if they became waiting while loadzone was blocked
            #self.unblock_loadzone(car)
                
            
            
            # are we in the allocation phase?
            if berth!=None:
                # can carrier stop for the next to be allocated berth
                
            
                path_sec_min=car.get_path_sec()+car.get_brakedist_comfort()+0.1
                if self.probe():
                    print '  try to allocate berth',berth['ind'],path_sec_min<berth['halt_point']
                #    print '  halt_point=',berth['halt_point'],\
                #          'path_sec_min=',path_sec_min

                if path_sec_min<berth['halt_point']:
                    #if self.probe():
                    #print '  allocate berth',berth['ind']

                    cd=self.alloc_berth(berth,car,{})
                    
                    # define here loadberth
                    self.update_cars(car,{'loadberth':self.berths[self.ind_berth_alloc]})
                
                    # decrease alloc counter
                    # counter now used to allocate berth in load zone
                    self.ind_berth_alloc-=1
                    if self.ind_berth_alloc==self.ind_berth_firstalloc:
                        # attention this is not a real allocation
                        # and hence no blocking of alloc here
                        # allocation is performed when car is in waiting mode
                        self.ind_berth_alloc=self.berthssec.berth_no-1
                        
                else:
                    #if self.probe():
                    print 'WARNING in ',self.id,\
                          '.alloc_berth_with_car_sync: cannot stop',car.id,\
                          ' for berth',berth['ind'],\
                          'path_sec_min=',path_sec_min,\
                          'path_sec=',car.get_path_sec(),\
                          'brakedist_comfort=',car.get_brakedist_comfort(),\
                          'v=',car.get_velocity()
                    
                    print '   set passthrough state and deviate'
                    cd={'velocity_line':
                        self.secs_dir['berths'].velocity_line}
                    # set passthrough state
                    self.update_cars(car,{'state':'passthrough'})
                    if self.car_tasks.has_key(car.id):
                        # deviate car. the passed secs have no influence
                        self.control_car_on_deviation(self.secs_dir['berths'],
                                                      self.secs_dir['accel'],
                                                      car,
                                                      self.car_tasks[car.id][0])
                    self.show_info()
                    # stop simulation
                    self.obj.sim.time_comp=0
                
            else:
                if self.probe():
                    print '  unable to allocate '
                    print '  accelerate carrier',car.id,' but make it stop in front of first berth'

                
                cd={'velocity_line':
                    self.secs_dir['berths'].velocity_line,
                    'mark_at':self.berths[0]['halt_point']\
                    -1.0,
                    'stop_at':self.berths[0]['halt_point']\
                    -0.5}
                
            
             
        # immediately control carrier, if no controldata provided
        if cd:
            if controldata==None:
                controldata={}
                controldata.update(cd)
                car.control(controldata)
            else:
                # just update controldata from input
                controldata.update(cd)

            #if self.probe():
            #    for key in controldata.keys():
            #        print '      ',key,'=',controldata[key]

        return controldata


    
                
    
    def get_ind_nextcar_upstream(self,berth):
        """
        Get the index of next upstream carrier from haltpoint of berth.
        The carrier can located with self.berthssec.cars[index]
        If there is no upstream carrier then the returned index=-1
        """
        # start searching backwards with last carrier (first entered)
        # of berthssec and check which carrier is first able to stop
        # in front of haltpoint of the berth?
        ind=0
        ind_found=-1
        while (ind_found==-1)&(ind<len(self.berthssec.cars)):
            # can this car stop at the haltpoint of the berth?
            if (self.berthssec.cars[ind].get_path_sec()\
               +self.berthssec.cars[ind].get_brakedist_comfort()\
               <=berth['halt_point']):
                # get index of this carrier
                ind_found=ind
            else:
                ind+=1

            #print 'get_ind_nextcar_upstream:',\
            #      ind,self.berthssec.cars[ind_found].id,\
            #      self.berthssec.cars[ind_found].get_path_sec(),\
            #      berth['halt_point']

        return ind_found

    def get_nextberth_unloadzone_synch(self,triggercar=None,berth=None,respect_waiting=False):
        """
        Returns an  accessible berth in unload zone in synchronous mode 
        for a carrier  which has just entered the berthsec.
    
         ???this would brake allocation for loading zone
        A carrier entered berth section in allocation mode.
        
        Now check out the farest berth without obstacles.
        If farest berth without obstacles is behind the next regular beth
        then return the regular berth.
    
       
        """
        if self.probe():
            print 'get_nextberth_unloadzone_synch:'#,berth['ind']
        #self.berths[self.ind_berth_alloc]
        
        # first check if unloadzone is blocked
        
        #if self.is_blocked_unloadzone:
        #    self.is_unloadzone_free(triggercar)
        
        blocked=False
        if berth==None:
            n=-1
        else:
            n=berth['ind']
            
            
        while (not blocked)&(n<=self.ind_berth_firstalloc):
            n+=1
            berth_check=self.berths[n]
            if respect_waiting:
                blocked= berth_check['allocated'] & (berth_check['status']!='liberated')
            else:
                blocked= berth_check['allocated']
            if self.probe():
                print '  berth',n,'allocated=', berth_check['allocated'],\
                        'status=', berth_check['status'],'blocked=',  blocked
        
        if ((n-1)>=0)&(self.berths[n-1]!=berth):
            berth_found=self.berths[n-1]
            if self.probe():
                print '  >> Alloc berth',berth_found['ind']
            
            #if 
            #self.is_blocked_unloadzone=False
            #self.unblock_unloadzone(triggercar)
        else:
            berth_found=None
            self.is_blocked_unloadzone=True
            if self.probe():
                print '  >> block unloadzone, is_blocked_unloadzone',self.is_blocked_unloadzone
            # first beth has been allocated
            # switch on liberatin fase
            #self.start_liberation()
            
        return berth_found
    
    def get_nextberth_loadzone_synch(self):
        """
        Returns an  accessible berth in unload zone in synchronous mode 
        for a carrier  which is waiting in unloadzone.
        """
        if self.probe():
            print '\n\nget_nextberth_loadzone_synch, self.ind_berth_alloc=', self.ind_berth_alloc
            for ind in range(self.ind_berth_firstalloc+1,len(self.berths)):
                if self.berths[ind]['status'] in ('loading','waiting'):
                    print '  berth status', ind, self.berths[ind]['status'],self.berths[ind]['car'].id
                else:
                    print '  berth status', ind, self.berths[ind]['status']
        
        if not self.is_blocked_loadzone:
            ind_alloc=self.ind_berth_alloc
            is_firstloop=True
            while   self.berths[ind_alloc]['allocated'] \
                    & is_firstloop:
                        ind_alloc-=1 
                        # skip load sequence in this case
                        if ind_alloc == self.ind_berth_firstalloc:   
                            ind_alloc=self.berthssec.berth_no-1
                        if ind_alloc ==self.ind_berth_alloc:
                            is_firstloop=False
            if is_firstloop:
                # found a fee berth
                allocberth=self.berths[ind_alloc]
                if self.probe():
                    print '  >> allocate berth',allocberth['ind']
                return allocberth
            
            else: 
                # last allocable berth in loadzone is allocated
                # block and wait till loadzone is free
                self.is_blocked_loadzone=True
                # moved to unalloc()
            
                if self.probe():
                    print '  Allocation refused all allocated'
                    #print '  Block loadzone! is_blocked_loadzone',self.is_blocked_loadzone
                
                return None
        
        else:
            if self.probe():
                print '  loadzone blocked!'
            return None
        
                        
    def is_unloadzone_free(self,triggercar=None):
        """
        Function that retuns 1 if all unloadbeths have been liberated.
        """
        if self.probe():
            print 'is_unloadzone_free?'
        # check if all berths are liberated so that
        ind=0
        free=True
        while free & (ind<=self.ind_berth_firstalloc):
            #free= (not self.berths[ind]['allocated'])
            free=self.berths[ind]['status']=='liberated'
            if self.probe():
                print '  check berth=',ind,self.berths[ind]['allocated'],self.berths[ind]['status'],
                if self.berths[ind]['status']=='occupied':
                    print 'with',self.berths[ind]['car'].id
                else:
                    print '.'
            ind+=1
        
        if self.probe():
            print '  Result=',free
            print '     Unblock now?',free & self.is_blocked_unloadzone  
            
        if free:
            if self.is_blocked_unloadzone:
                if self.probe():
                    print '  set  is_blocked_unloadzone = False '
                self.is_blocked_unloadzone = False 
                self.unblock_unloadzone(triggercar)
        
              
        return free
    
    def unblock_unloadzone(self,triggercar=None):
        """
        Unblock searching, unallocated cars who wants
        to allocate in unloadzone
        """
        
        if self.probe():
            print 'unblock_unloadzone!!'
            print '     is_blocked_unloadzone=',self.is_blocked_unloadzone
        
        # unblock only if is_blocked_unloadzone is set to False 
        # by is_unloadzone_free 
        if not self.is_blocked_unloadzone: 
            # now check if there are searching, unallocated cars
            carlist=copy(self.berthssec.cars)
            #carlist.reverse()

            for car in carlist:
                if car!=triggercar:
                    if self.cars[car.id]['state']=='searching':
                        cd=self.alloc_berth_with_car_sync(car,{})
                        #cd.update({'velocity_line':\
                        #          self.secs_dir['berths'].velocity_line})
                        
                        car.control(cd)
                        if self.probe():
                            print '         unblocked:',car.id,self.cars[car.id]['state']
                        #else:
                        #    triggercd=cd.copy()
                    
                    if self.cars[car.id]['state']=='passthrough':
                        #print 'unblock_unloadzone: move blocked passthrough',car.id
                        car.control({   'stop_at':None,
                                        'mark_at':None,
                                        'velocity_line':
                                        self.secs_dir['berths'].velocity_line})
        
    def is_loadzone_free(self,triggercar=None):
        """
        Function that retuns 1 if all loadbeths have been liberated.
        """
        
        if self.probe():
            print '\nis_loadzone_free? '        
        # here we simply prevent a situation where all berths in the 
        # load area have been liberated, but the first cer in the 
        # unload area cannot allocate in the load area.
        ind=self.ind_berth_firstalloc+1
        free=1
        while free & (ind<self.berthssec.berth_no):
            #free= (not self.berths[ind]['allocated'])
            if self.probe():
                print '   berth',ind,self.berths[ind]['allocated'],self.berths[ind]['status'],
                if self.berths[ind]['status']=='occupied':
                    print 'with',self.berths[ind]['car'].id
                else:
                    print '.'
            #free= not self.berths[ind]['allocated']
            free= self.berths[ind]['status']=='liberated'
            ind+=1
        if self.probe():
            print '\n   Result=',free
        if free:
            # There are no allocations in loadzone ! Here alloc counter is reset
            # try to allocate all unallocated free cars in unload zone
            if 1: #self.is_blocked_loadzone:
                self.is_blocked_loadzone=False
                
                # in this case it ia necessary to reset allocation index!!
                
                # DONT'Touch allocation index!! done only in unalloc
                # in case allocation index is at first birth
                # put index on remotes berth
                #self.ind_berth_alloc=self.berthssec.berth_no-1
                
                if self.probe():
                    print '  unblock_loadzone!!'
                    print '     is_blocked_loadzone=',self.is_blocked_loadzone
                    print '     ind_berth_alloc=',self.ind_berth_alloc
                # let's unoccupy all cars in unload zone
                self.unblock_loadzone(triggercar) 
        
        
                                  
        return free
    
    def unblock_loadzone(self,triggercar=None):
        """
        Unblock waiting cars in unloadzone who want
        to allocate in loadzone
        """
        
        if self.probe():
            print 'unblock_loadzone!!'
            print '     is_blocked_loadzone=',self.is_blocked_loadzone
        
        # unblock only if is_blocked_unloadzone is set to False 
        # by is_unloadzone_free 
        if not self.is_blocked_loadzone: 
            
            # try to dend awai waiting cars
            ind=self.ind_berth_firstalloc
            while (ind>=0):
                if self.berths[ind]['status']=='waiting':
                    if self.berths[ind]['car']!=triggercar:
                        berth=self.berths[ind]
                        car=berth['car']
                        if self.probe():
                            print '  try to unblock', car.id
                        
                        if self.cars[car.id].has_key('loadberth'):
                            loadberth =self.cars[car.id]['loadberth']
                            if loadberth['allocated']:
                                # the desired berth in loadzone is already allocated
                                # now try to move car ahead within UNloadzone:
                                loadberth=self.get_nextberth_unloadzone_synch(berth=berth) 
                        else:
                            # get next possible nonallocated berth in loadzone to allocate
                            loadberth=self.get_nextberth_loadzone_synch() 
                            
                        if loadberth:
                            if not loadberth['allocated']:
                                cd=self.unoccupy_berth(berth,{})
                                cd=self.alloc_berth(loadberth,car,cd)
                                # apply control data for swapcar here
                                car.control(cd)
                                if self.probe():
                                    print '   move',car.id,car.id,self.cars[car.id]['state']
                                    print '      from berth', berth['ind'],'to',loadberth['ind']
                            else:
                                if self.probe():
                                    print '   loadberth',loadberth['ind'],'found is allocated'
                        
                        
                        else:
                            # this is strange, should be free,
                            # are there more waiting cars in unloadzone than
                            # availlable berth in loadzone?
                            if self.probe():
                                    print '   currently no loadber availlable'
                # get previous berth in unload zone            
                ind-=1
                        
                        
    def are_all_berth_liberated(self):
        """
        Function that retuns 1 if all beths are liberated.
        """
        
        # check if all berths are liberated so that
        ind=0
        liberated=1
        while liberated & (ind<self.berthssec.berth_no):
            liberated=(self.berths[ind]['status']=='liberated')&\
                       (not self.berths[ind]['allocated'])
            ind+=1
        return liberated

    def are_all_unloadberth_unallocated(self):
        """
        Function that retuns 1 if all loadbeths are not allocated.
        """
        # check if all berths are liberated so that
        ind=0
        unalloc=1
        while unalloc & (ind<=self.ind_berth_firstalloc):
            unalloc=(not self.berths[ind]['allocated'])
            ind+=1
        return unalloc
                
    def find_accessible_berth_downstream(self,berth,included=0):
        """
        Returns an liberated accessible berth which is looking for a carrier.
        The found berth will be downstream with respect to the given berth.

        """
        #if self.probe():
        #    print 'find_accessible_berth_downstream:',berth['ind']
        berth_found=None
        found_in=''
        n=berth['ind']
        #
        #if included|((not included)&(n<self.berthssec.berth_no))
        if not berth_found:
            if not included:
                n+=1
            #if self.probe():
            #    print '  try to find a a freely accessible liberated berth'
                
            blocked=0
            while (not blocked)&(n<self.berthssec.berth_no)&\
                  (found_in!='berths_demand'):
                berth_check=self.berths[n]
                #if self.probe():
                #    print '  check berth',berth_check['ind'],\
                #          berth_check['status']
                # check if there are no allocations loading or waiting
                blocked=berth_check['allocated'] |\
                         (berth_check['status']!='liberated')
                if (not blocked)&(not berth_check['allocated']):
                    berth_found=berth_check
                    if self.berths_demand.count(berth_found)>0:
                        found_in='berths_demand'
                    else:
                        found_in='liberated'
                        
                # get next berth and check whether it is unblocked too
                n+=1
                
        return berth_found,found_in


                    
    def find_berth_downstream(self,berth,included=0):
        """
        Returns a berth which is looking for a carrier.
        The found berth will be downstream with respect to the given berth.

        The function will first try to find a suitable berth in
        the berths demand list. If there is no demand, it will return the
        remotest liberated beth of the stop.
        If there is no suitable berth, None will be returned.
        """
        
        # take first reachable in the berth demand list
        found_in=''
        berth_found=None
        n=berth['ind']
       
        notincluded=not included
        #if self.probe():
        #    print 'find_berth_downstream:',berth['ind']+notincluded
        n=0
        #inds_remove=[]
        while (not berth_found)&(n<len(self.berths_demand)):
            if self.berths_demand[n]['ind']>=berth['ind']+notincluded:
                if not self.berths_demand[n]['allocated']:
                    berth_found=self.berths_demand[n]
                    found_in='berths_demand'
                else:
                    print 'WARNING in find_berth_downstream',self.id,':\n',\
                          '  berth',self.berths_demand[n]['ind'],\
                          'in berths_demand list already allocated'
                    #self.sim.stop()
                    #print '->remove from list'
                    #inds_remove.append(n)
                    n+=1
            else:
                n+=1

        # if no berth found, then try to find a a freely accessible 
        # liberated berth
        if not berth_found:
            #print '  try to find a a freely accessible liberated berth'
            n=berth['ind']+notincluded
            blocked=0
            while (not blocked)&(n<self.berthssec.berth_no):
                berth_check=self.berths[n]
                # check if there are no allocations loading or waiting
                blocked=berth_check['allocated'] |\
                         (berth_check['status']!='liberated')
                if (not blocked)&(not berth_check['allocated']):
                    berth_found=berth_check
                    found_in='liberated'
                # get next berth and check whether it is unblocked too
                n+=1
                
        # search for next liberated berth?
        #if not berth_found:
        #    # search for next free berth that is closer than the found one
        #    n=berth['ind']
        #    while (not nearerberth_found)&(n<=berth_found['ind']):
        #        berth_check=self.berths[n]
        #        # free and can it be allocated
        #        if (not berth_check['allocated'])&\
        #               (berth_check['status']=='liberated'):
        #            nearerberth_found=berth_check
        #        else:
        #            # get next berth 
        #            n+=1
                    
        # search for remotest liberated berth
        if not berth_found:
            #print ' takethe remotest berth which is liberated and unallocated'
            n=self.berthssec.berth_no-1
            while (not berth_found) & (n>=0):
                berth_check=self.berths[n]
                if (not berth_check['allocated'])&\
                   (berth_check['status']=='liberated')&\
                   (berth_check['ind']>berth['ind']+notincluded):
                    berth_found=berth_check
                    found_in='liberated'
                elif berth_check['ind']==berth['ind']:
                    # berth to start with reached-> terminate
                    n=-1
                else:
                    # go back one berth
                    n-=1

        

        
        if not berth_found: # no unallocated liberated berth has been found,
            #print '  so try to find an unallocated occupied berth'
            ind=berth['ind']+notincluded
            while (berth_found==None)&\
                      (ind<self.berthssec.berth_no):
                if not self.berths[ind]['allocated']:
                    # berth is in waiting or occupied status
                    berth_found=self.berths[ind]
                    found_in='occupied'
                else:
                    ind+=1
            
        #if berth_found:
        #    if berth_found['allocated']:
        #        print 'WARNING in find_berth_downstream',self.id,':\n',\
        #              '  berth',berth_found['ind'],\
        #              'already allocated'        
            
        return berth_found,found_in

    #def update_berths_demand(self):
    #    """
    #    Updates the berths_demand list.
    #    All berth with an unserved  queue are put in this list.
    #    """
    #    self.berths_demand=[]
    #    for berth in self.berths:
    #        if len(berth['queue'])>0:
    #            if (not berth[]
    
    
    
    def get_nextberth_downstream(self,car):
        """
        Get the next downstream berth for which the carrier is able to stop.
        """
        
        # print 'get_nextberth_downstream:'

        nextberth=None
        ind=0
        # point in sec where carrier would stop when braking now
        path_sec_min=car.get_path_sec()+car.get_brakedist_comfort()

        while (not nextberth)&(ind<self.berthssec.berth_no):
            #print '  check berth',ind,\
            #      'with halt_point',self.berths[ind]['halt_point'],\
            #      'path_sec_min=',path_sec_min
            # can carrier stop before the halt_point of this stop?
            if self.berths[ind]['halt_point']>path_sec_min:
                # get this berth
                nextberth=self.berths[ind]
            else:
                # check next stop
                ind+=1
                
        return nextberth
    
    def alloc_berth(self,berth,car,controldata=None,found_in=''):
        """
        Allocate berth with carrier and generate carrier control data.
        Control carrier if no controldata is provided.
        """
        # allocated berth must be removed from the berths_demand list
        #if self.berths_demand.count(berth)>0:
        #    self.berths_demand.remove(berth)
        if self.probe():
            print 'alloc_berth:',self.id,'berth',berth['ind'],\
                  '<',car.id,found_in

        if not berth['allocated']:
            # allocate it 
            berth['allocated']=1
            
        else:
            print 'WARING in alloc_berth',self.id,':\n',\
                  '  car',car.id,'tries to allocate the allocated berth',\
                  berth['ind']
            
        # make default state change
        if found_in=='berths_demand':
            #print 'alloc_berth at',self.id,'with',car.id
            if self.berths_demand.count(berth)>0:
                #print '  remove berth',berth['ind'],'from demand list'
                self.berths_demand.remove(berth)
            self.update_cars(car,{'state':'allocated',
                                  'berth_allocated':berth,
                                  'found_in':found_in})
        else:
            self.update_cars(car,{'state':'allocated',
                                  'berth_allocated':berth})
        
        
        # provide carrier control data:
        # tell carrier to stop at the corresponding point
        # target stop is a little bit behind the mark just to be sure that
        # carrier rolls over mark
        cd={'mark_at':berth['halt_point']-0.1,
            'stop_at':berth['halt_point']+0.1 ,
            'velocity_line':
                self.secs_dir['berths'].velocity_line}
        
        # control carrier if no controldata must be retuned
        if controldata==None:
            controldata={}
            controldata.update(cd)
            car.control(controldata)
        else:
            # just update controldata from input
            controldata.update(cd)
    
        return controldata

    def unalloc_berth(self,berth,car,controldata=None):
        """
        Remove allocation mark from berth and generate carrier control data.
        Control carrier if no controldata is provided.
        """
        if self.probe():
            print '\n unalloc_berth:',berth['ind'],'from',car.id
            
        if berth['allocated']:
            # unallocate it 
            berth['allocated']=0
            
        else:
            print 'WARING in unalloc_berth',self.id,':\n',\
                  '  car',car.id,'tries to unallocate the unallocated berth',\
                  berth['ind']

        self.update_cars(car,{'berth_allocated':None})
        
        # provide carrier control data:
        cd={'stop_at':None,'mark_at':None} # remove marks

        # is stop in synchronous mode?
        if self.time_sync>0.0:
            
            if  berth['ind']<=self.ind_berth_firstalloc:
                # berth is in unload zone
                #self.is_unloadzone_free()
                pass
            else:
                # berth is in loadzone
                
                # this will cause the allocation in loadzone to allocate
                # only to the right of this berth
                #if (berth['ind']-1)<self.ind_berth_alloc:
                    #self.ind_berth_alloc=berth['ind']-1
                
                    #if self.ind_berth_alloc <= self.ind_berth_firstalloc:
                    #    # last beth in load zone has been unallocated
                    #    # bloch allocation here
                    #    self.is_blocked_loadzone=True
                        
                # this will check if there are no allocations in load zone
                # if there are non it will also move waiting cars in 
                # unloadzone   
                pass
                #self.is_loadzone_free(car)
                    
                
                # check also unloadzone, just in case
                #self.is_unloadzone_free(car)
                # no too time consuming wait till timer is doing this job
                
        # control carrier if no controldata must be retuned
        if controldata==None:
            controldata={}
            controldata.update(cd)
            car.control(controldata)
        else:
            # just update controldata from input
            controldata.update(cd)
        
        return controldata

    def occupy_berth_with_car(self,car,controldata=None):
        """
        Occupy previously allocated berth with car.
        """
        if not car:
                print 'WARNING: in',self.id,'.occupy_berth_with_car:'
                print '  No car given for',berth['ind']
                self.show_info()
                # stop simulation!!!
                self.obj.sim.time_comp=0
                controldata={'stop_at':None,'stop':'now'}
                return controldata
        
        berth=self.cars[car.id]['berth_allocated']
        if self.probe():
            print 'occupy_berth_with_car:',berth['ind'],'with',car.id,self.cars[car.id]['state']
        if controldata:
            self.unalloc_berth(berth,car,controldata)
        else:
            self.unalloc_berth(berth,car)
        berth['car']=car

        # default berth status, which may be changed to 'occupied'
        # if the load_item function is called
        if self.cars[car.id]['loaded']:
            # loaded car occupied berths
            # in this case unloading takes place
            berth['status']='occupied'
            self.update_cars(car,{'state':'loading','berth_occupied':berth})
        else:
            berth['status']='waiting'
            self.update_cars(car,{'state':'waiting','berth_occupied':berth})
        
        # default carrier control data
        cd={'stop_at':None,'stop':'now'} # stop carrier

        # control carrier if no controldata must be retuned
        if controldata==None:
            controldata={}
            controldata.update(cd)
            car.control(controldata)
        else:
            # just update controldata from input
            controldata.update(cd)
        
        return controldata
    
    def unoccupy_berth(self,berth,controldata=None):
        """
        Liberate previously occuied berth with car.
        """
        
        if berth['status'] == 'waiting':
            car=berth['car']

            if not car:
                print 'WARNING: in',self.id,'.unoccupy_berth:'
                print '  No car in waiting berth',berth['ind']
                self.show_info()
                # stop simulation!!!
                self.obj.sim.time_comp=0
                controldata={}
                return controldata
                

            if self.probe():
                print 'unoccupy_berth',self.id,berth['ind'],'from',car.id,self.cars[car.id]['state']
            
            # liberate berth...
            berth['status']='liberated'
            berth['car']=None
            # change status of carrier in trackelement
            if (self.cars[car.id]['loaded'])|\
                   (self.cars[car.id]['state']=='passthrough'):
                self.update_cars(car,{'state':'passthrough',
                                      'berth_occupied':None,
                                      'found_in':''})
                # in non-synchronous mode, an empty waiting carrier
                # starts searching for a new berth
                #elif self.time_sync<=0.0:
            else:
                self.update_cars(car,{'state':'searching',
                                      'berth_occupied':None,
                                      'found_in':''})
            # in synchronous mode, an empty waiting carrier
            # gets kicked out.
            #elif self.time_sync>0.0:
            #    self.update_cars(car,{'state':'passthrough',
            #                          'berth_occupied':None})

            # ... by accelerating carrier
            cd={'start':1,
                'stop_at':None,
                'mark_at':None,
                'velocity_line':
                self.secs_dir['berths'].velocity_line}
        

            # is stop in synchronous mode?
            if self.time_sync>0.0:
                
                
                
                #is berth in the loading area?
                if berth['ind']>self.ind_berth_firstalloc:
                    # if there are non it will also move waiting cars in 
                    # unloadzone   
                    self.is_loadzone_free(car)    
                    # check also unloadzone, just in case
                    #self.is_unloadzone_free(car)
            else:
                
                # async mode
                pass
            
            # check out whether it is possible to allocate blocked carriers
            if self.flag_cars_blocked:
                still_cars_blocked=0   
                for blockedcar in self.berthssec.cars:
                    if self.cars[blockedcar.id]['state']=='blocked':
                        #print '  try to allocate blocked car ',blockedcar.id
                        self.update_cars(blockedcar,{'state':'searching'})
                        still_cars_blocked=1
                        # is stop in synchronous mode?
                        if self.time_sync>0.0:
                            blockedcar.control(\
                                self.alloc_berth_with_car_sync(\
                                blockedcar,{'velocity_line':
                                     self.secs_dir['berths'].velocity_line}))
                        else:
                            # allocate a berth for carrier, if possible
                            blockedcar.control(self.alloc_berth_with_car(\
                                blockedcar,{'velocity_line':
                                     self.secs_dir['berths'].velocity_line}))
            
                self.flag_cars_blocked=still_cars_blocked

            
            # control carrier if no controldata must be retuned
            if controldata==None:
                controldata={}
                controldata.update(cd)
                car.control(controldata)
            else:
                # just update controldata from input
                controldata.update(cd)
            
        else:
            print 'WARNING: unoccupy_berth',berth['ind'],' in',self.id,\
                  'failed.'
            print '         status of berth="',berth['status'],\
                  '" instad of "waiting"'
            controldata={}  
        
        return controldata
    
    def get_status_car_in_berthssec(self,car):
        """
        Attention: this function is only used to recover the car status
        from earlier versions. Used in init_specific
        
        Determins the status of the car within the berthsection.
        Returns a dictionary with the following possible key, value pairs:
        1. path_sec:x, state:'searching', loaded:0, ind_berth: ind
        2. path_sec:x, state:'searching', loaded:1, ind_berth: ind
        3. path_sec:x, state:'passthrough', ind_berth: ind
        4. path_sec:x, state:'allocated', ind_berth: ind
        5. path_sec:x, state:'waiting', ind_berth: ind
        6. path_sec:x, state:'loading', ind_berth: ind
        7. path_sec:x, state:'offtrack', ind_berth: -1

        Where path_sec is the position within the berth section and
        n is the berth index. n=0 is index of first berth in stop from
        carrier point of view.

        if state is searching or passthrough, ind_berth is the index of the
        next downstream reachable berth.
        It returns -1 if there is no more a berth.

        Contents of the info key is a string with arbitrary value. 
        """
        # initialize return variable
        status={'info':'',
                'ind_berth':-1,
                'loaded':0}

        # get position
        #if hasattr(car,'obj'):
        #    status['path_sec']=car.get_path_sec()
        #else:
        #    status['path_sec']=0

        
        

        # is carrier already allocated?
        if self.berths_alloc.has_key(car.id):
            status['state']='allocated'
            status['ind_berth']=self.berths_alloc[car.id]['ind']

        # is carrier at a berth?
        elif self.berths_occupied.has_key(car.id):
            berth=self.berths_occupied[car.id]
            status['ind_berth']=berth['ind']
            # reinsure that car is at berth:
            if berth['car']==car:
                if berth['status']=='waiting':
                    status['state']='waiting'
                elif berth['status']=='occupied':
                    status['state']='loading'
            else:
                print 'WARNING: inconsistant berths_occupied at',self.id,\
                      'for',car.id
                # give it the passthrough state
                status['state']='passthrough'
          
        # if not allocated or at a berth,
        # so car is running. Is there a task for this carrier?
        elif self.car_tasks.has_key(car.id):
            # use always oldest  task
            task=self.car_tasks[car.id][0]
            if task.mode=='search_item':
                # carrrier is empty and searching for load
                status['state']='searching'
            elif task.target_flag:# car has a someone to unload at this stop
                status['state']='searching'
            else: # car has a someone, but not for this stop
                # give it the passthrough state
                status['state']='passthrough'

        
        # check if car is still off the track?
        elif self.cars_offtrack.count(car)==1:
            status['state']='offtrack'
            
        # car is not allocated at no berth and has no task!!        
        else: 
            # give it the passthrough state
            status['state']='passthrough'
            status['info']='no_task'

        # set loaded flag
        if self.car_tasks.has_key(car.id):
            # use always oldest  task
            task=self.car_tasks[car.id][0]
            if not task.mode=='search_item':
                status['loaded']=1
            
        return status


    def find_car_to_swap(self,car,berth):
        """
        Find an upstream car to swap allocation with a car at a berth.
        The  upstream car must be allocated to a berth behind the
        respective berth.
        """
        
        n=self.berthssec.cars.index(car)+1
        n_max=len(self.berthssec.cars)
        foundcar=None
        foundberth=None
        while (not foundcar)&(n<n_max):
            checkcar=self.berthssec.cars[n]
            checkberth=self.cars[checkcar.id]['berth_allocated']
            if checkberth:
                # is allocated berth behind current berth
                if checkberth['ind']>berth['ind']:
                    # get it
                    foundcar=checkcar
                    foundberth=checkberth
                else:
                    n+=1
            else:
                n+=1 

        return  foundcar,foundberth 
        
    def check_user(self,car,message='',signal_false_enter=False):
        """
        Check the programmed destination of user
        """
        items=car.items.values()
        if items:
            user=items[0]
            if (signal_false_enter&(user.dest_id!=self.id))| \
                (not signal_false_enter)&(user.dest_id==self.id):
                print 'WARNING from check_user:',user.id,'in',car.id,'with'
                print '  ',user.dest_id, 'at \n  ',self.id,message,
                if signal_false_enter:
                    print 'FALSE ENTER'
                else:
                    print 'FALSE PASS'
                
    def control_car_on_start(self,sec,car,mode,controldata):
        """
        Track element wide carrier control when carrier is ready to start.

        """
        if self.probe():
            print self.id,'control_car_on_start:',car.id,sec.label
                
        if sec.label=='berths':
            #self.check_user(car,message='at start berth')
            #if self.probe():
            #print 'control_car_on_start:',self.id,car.id,mode
            if mode=='empty':
                self.update_cars(car,{'state':mode})
                
            # are we in synchronous mode?
            if self.time_sync>0.0:
                if self.cars[car.id]['berth_occupied']:
                    self.alloc_berth_with_car_sync(car)
                elif self.cars[car.id]['berth_allocated']:
                    #if self.probe():
                    #    print '  carrier has already been allocated'
                    pass
                
                else:
                    #if self.probe():
                    #print '  car has been started while not\n',\
                    #          ' sitting on a berth and without allocation->go'
                    self.update_cars(car,{'state':'passthrough'})
                    car.control(sec.search_car_infront(car,controldata))
              
            else: # non- synchronous mode
                if self.cars[car.id]['berth_occupied']:
                    self.alloc_berth_with_car(car)
                
                elif self.cars[car.id]['berth_allocated']:
                    # carrier has already been allocated
                    pass
                
                else:
                    #print 'WARNING: control_car_on_start in',self.id
                    #print '  car',car.id,\
                    #      ' has been started while not sitting on a berth'
                    #print '  simtime:',self.sim.time
                    #print '    state:',self.cars[car.id]['state']
                    #print '    loaded:',self.cars[car.id]['loaded']
                    
                    # and without allocation
                    self.update_cars(car,{'state':'passthrough'})
                    car.control(sec.search_car_infront(car,controldata))

            
      
                
            
    def control_car_on_enter(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier enters section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        Here, command carrier to halt at halt_point.
        '''
        
        # is car entering the berth section ?
        #if sec.label=='entry':
        #    pass

        if sec.label=='berths':
            #if self.probe():
                #print 'control_car_on_enter',car.id,sec.label
                #print '   at',self.id,sec.id,sec.label
            #if self.probe():
            #    print 'control_car_on_enter',car.id,sec.label
            
            # insert car into track elemet wide cars dict
            # currently only used for berths section
            self.update_cars(car)
            
            # determine current state of car
            if self.car_tasks.has_key(car.id):
                # use always oldest  task
                task=self.car_tasks[car.id][0]
                #print 'control_car_on_enter:'
                # carrier has a load for this stop or is empty 
                if ((task.mode=='search_item') | task.target_flag):
                    #print '  make car searching for berth'
                    self.update_cars(car,{'state':'searching',
                                          'loaded': task.mode=='loaded'})
                    
                else:
                    #print '  carrier does not want to stop at a berth'
                    self.update_cars(car,{'state':'passthrough',
                                          'loaded':task.mode=='loaded' })

                
                
            else: # carrier has no task
                # make car searching
                self.update_cars(car,{'state':'searching'})


            #print 'enter',self.id,'car',car.id,\
            #      'state:',self.cars[car.id]['state'],\
            #      'loaded:',self.cars[car.id]['loaded']
                
            # is stop in synchronous mode?
            if self.time_sync>0.0:
                #print 'synchronous'
                controldata.update(self.alloc_berth_with_car_sync(\
                    car,controldata))
            else:
                #print 'aasynchronous'
                # allocate a berth for carrier, if possible
                controldata.update(self.alloc_berth_with_car(car,controldata))
            
        
        # check whether to diverge at carrier entrance at brake line
        # with a carrier on main_entry line
        elif (sec.label=='brake'):
            if (len(self.secs_dir['main_entry'].cars)>0):
                controldata.update(self.diverge(self.secs_dir['main_entry']))

            # hold on to main line speed untill out of clearing zone
            controldata['velocity_line']=\
                                  self.secs_dir['main_entry'].velocity_line
            controldata['mark_at']=sec.clearingzone

            # check out here already whether there is an empty carrier
            # at an empty berth, blocking the other berth.

            self.move_waiting_cars(0)

            #if self.queuing_user()==0:
            #    #if len(self.get_users_queued())==0:
            #    self.clear_berths(0)

                        
        # check whether to diverge at carrier entrance at the main_entry line
        # with a carrier on brake line
        elif (sec.label=='main_entry')&(len(self.secs_dir['brake'].cars)>0):
            controldata.update(self.diverge(self.secs_dir['brake']))

        elif  sec.label=='accel':
            # do here special merge for accelerating carrier
            self.append_cars_noghost(car)
            mergesec=self.secs_dir['main_exit']
            mergesec_cars_num=len(mergesec.cars)
            #mergesec_cars_noghost=self.cars_noghost[mergesec.id][:]
             
            if mergesec_cars_num==0:
                # go ahead and accelerate to line speed
                pass
            else:
                self.control_car_on_enter_mergemod(sec,car,controldata)

            #elif mergesec_cars_num==len(mergesec_cars_noghost):
            #    # mergecar is last entered in merge zone
            #    controldata.update(self.get_controldata_to_make_ghost(\
            #         car,mergesec_cars_noghost[-1]))
                
            #elif (self.time_sync>0):#&(self.ind_berth_alloc<0):
            #    #self.append_cars_noghost(car)
            #    #self.merge(car,mergesec,controldata,no_mergecar_behind=1)
            #    self.control_car_on_enter_mergemod(sec,car,controldata)
            #elif (self.time_sync>0)&\
            #     (((mergesec.length-mergesec.cars[0].get_path_sec()\
            #        -mergesec.cars[0].get_brakedist_comfort())-\
            #       (c)
            #      )
            
                
            #else:
            #    # stop car and wait untill it gets accelerated by a car on
            #    # the main line
            #    controldata['velocity_line']=0.0


                

        elif sec.label=='main_exit':
            d_mark=sec.length-self.secs_dir['accel'].length\
                    -car.time_brake*(sec.velocity_line-\
                                     self.secs_dir['berths'].velocity_line)\
                    -0.5/car.accel_comfort*(sec.velocity_line**2-\
                                     self.secs_dir['berths'].velocity_line**2)\
                    -sec.velocity_line**2*\
                    (0.5/car.accel_emergency-0.5/car.accel_failure)
            
                    
                                     
            # without TB
            #d_mark=sec.length-sec.velocity_line*(\
            #    (sec.velocity_line-self.secs_dir['berths'].velocity_line)
            #    /car.accel_comfort\
            #    +self.secs_dir['accel'].length/sec.velocity_line\
            #   -(sec.velocity_line-self.secs_dir['berths'].velocity_line)**2\
            #    /(2*car.accel_comfort*sec.velocity_line))\
            #    -0.5*sec.velocity_line**2/car.accel_emergency\
            #    +0.5*sec.velocity_line**2/car.accel_failure

            
            #print 'control_car_on_enter at',self.id,d_mark
            if d_mark>0:
                controldata['mark_at']=d_mark
            else:
                # check for mergecontrol instructions
                self.control_car_on_enter_mergemod(sec,car,controldata)

        return controldata


    def control_car_on_leave(self,sec,car,controldata):
        '''
        Track element wide carrier control when carrier leaves section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        Here the ghost carrier will be removed from the control data.
        '''
        #if self.probe():
        #    print 'control_car_on_leave: car',car.id,sec.label
            
        if sec.label=='exit':
            # check if a task has been completed
            if self.car_tasks.has_key(car.id):
                self.car_tasks[car.id][0].leave_sec(sec)

        elif sec.label=='berths':
            
            # delete car from element wide cars dict 
            # currently only used for berths section
            self.delete_from_cars(car)

            #self.check_user(car,message='leave')
                
            # make sure that stop in syncmode gets switched to allocation phase
            # if there are no more cars on the berths section

            # is stop in synchronous mode?
            #if self.time_sync>0.0:
            #    # is stop in liberation phase?
            #    #if self.ind_berth_alloc<0:
            #    if len(sec.cars)==0:
            #        #if self.probe():
            #        #    print 'control_car_on_leave: car',car.id
            #        #    print '  stop in synchronous mode.'
            #        #    print '  len(sec.cars)=',len(sec.cars),\
            #        #          'ind_berth_alloc=',self.ind_berth_alloc
            #        self.start_allocation()#

            #        # if synch mode has just been started, try this
            #        if self.liberate_all:
            #           if self.are_all_berth_liberated():
            #              self.start_allocation()
                        

        elif sec.label=='accel':
            if self.cars_offtrack:
                self.obj.ready_to_add_offtrack_cars(self)

        elif sec.label=='main_exit':
            # check if we can accelerate carriers waiting on the
            # acceleration section
            if len(sec.cars)==0:
                accelsec=self.secs_dir['accel']
                for car in accelsec.cars:
                    car.control({'velocity_line':accelsec.velocity_line})
            
        # check for merge control instructions 
        self.control_car_on_leave_mergemod(sec,car,controldata)
            
        return controldata
    
    def control_car_at_mark(self,sec,car,controldata={}):
        '''
        Track element wide carrier control when carrier comes to a
        halt on section sec.
        '''
        #if self.probe():
        #    print 'control_car_at_mark ',self.id,car.id,sec.label
                
        if sec.label=='berths':
            
            
            
            if self.probe():
                print 'control_car_at_mark',car.id,self.cars[car.id]['state']


            if self.cars[car.id]['berth_allocated']:
                berth=self.cars[car.id]['berth_allocated']
                #if self.probe():
                
                self.occupy_berth_with_car(car,controldata)
                #print '  occupied berth',berth['ind'],berth['status'],\
                #      'with',car.id,self.cars[car.id]['state'],\
                #      'berth_occupied:',\
                #      self.cars[car.id]['berth_occupied']['ind']
                
                # berth is now in waiting mode

                # check if a task has been completed
                if self.car_tasks.has_key(car.id):
                    task=self.car_tasks[car.id][0]
                    if task.target_flag:
                        task.haltin_sec(self.berthssec)
                #else:
                #    print 'WARNING: in control_car_at_mark',self.id
                #    print '  car',car.id,'has stopped here without task'


                # is stop in sync mode?
                if self.time_sync>0.0:
                    # when berth is in unload area
                    if berth['ind']<=self.ind_berth_firstalloc:
                        if  self.cars[car.id]['loaded']:
                            #get unloaded
                            car.unload()
                        # car is unloaded in unload area, but there
                        # is a queue at this berth. This can happen
                        # during the transition to synch mode
                        elif (len(berth['queue'])>0)&\
                                 (not self.cars[car.id]['loaded']):
                            car.load()

                        # car is unloaded in unload area and queue is empty..
                        else: 
                            #if self.probe():
                            #    print 'control_car_at_mark:'
                            #    print '  empty car',car.id,\
                            #          'remains waiting in unload area'

                            # try to send car to a berth in load area...
                            # prevent carrier to get in halt mode
                            #del controldata['stop_at']
                            #del controldata['stop']
                            self.alloc_berth_with_car_sync(car,controldata)
                            # if there has not been an allocation, make it stop
                            #if self.cars[car.id]['state']!='allocated':
                            #    if self.probe():
                            #        print '  car could not be reallocated'
                            #        print '  halt it in waiting mode',\
                            #              controldata
                            #    #car.control(controldata)
                            #else:
                            if self.cars[car.id]['state']=='allocated':
                                # remove start and stop flags from
                                # control data from previous assignments
                                # in order guarantee a smooth continuation
                                del controldata['start']
                                del controldata['stop']
                                
                                #if self.probe():
                                #    print ' send car along without putting it'
                                #    print '  in halt mode before',controldata
                                #car.control(cd)
                            #else:
                            #    # this means allocation has not been successful
                            #    if self.probe():
                            #       print '  car',car.id,'remains in unload zone'
                            #        print '  controldata', controldata
                            #        #print ' ***send car idle'
                            #    #car.idle()
                            
                    else: # car is waiting in loading area
                        if  not self.cars[car.id]['loaded']:
                            # try to load carrier here 
                            car.load()
                        else:
                            # this can happen while switching beween modes
                            #print 'WARNING control_car_at_mark of',self.id
                            #print '  unloaded car',car.id,\
                            #      'in loading area at berth',berth['ind']
                            car.unload()
                    
                    
                    
                else: # non sync mode
                    # is carrier loaded and wants to get unloaded
                    #print 'control_car_at_mark of',
                    #self.id,car.id,self.cars[car.id]['loaded']
                    #print '  cardata',self.cars[car.id] 
                    if  self.cars[car.id]['loaded']:
                        #print '  call unload of',car.id
                        car.unload()
                    elif (len(berth['queue'])>0):
                        # try to load carrier here
                        #print '   try to load carrier',car.id,\
                        car.load()
                    else:
                        #print '  try to get a new allocation downstream'
                        # otherwise stay here and wait
                        #print '   alloc_berth_with_car for',car.id
                        self.alloc_berth_with_car(car,controldata)
                        if self.cars[car.id]['state']=='allocated':
                            #print '  OK,remove start and stop flags from'
                            # control data from previous assignments
                            # in order guarantee a smooth continuation
                            if controldata.has_key('start'):
                                del controldata['start']
                            if controldata.has_key('stop'):
                                del controldata['stop']
                                
                        elif self.cars[car.id]['state']=='passthrough':
                            # remove stop flags from
                            # control data from previous assignments
                            # in order guarantee a smooth continuation
                            #if controldata.has_key('start'):
                            #    del controldata['start']
                            #if controldata.has_key('stop'):
                            del controldata['stop']
                        else:
                            #print '  nothing to load, no berth downstream
                            #>make it idle'
                            car.idle()
                
            else:
                if self.probe():
                    print 'WARNING in control_car_at_mark',self.id 
                    print '  no berth has been allocated for',car.id,self.cars[car.id]['state'],\
                              ' => wait alloc phase'
                #car.control({'velocity_line':sec.velocity_line})
                #car.control({'velocity_line':0.0})
                controldata={'velocity_line':0.0}
                
        elif sec.label=='main_entry':
            #print 'remove ghost'
            controldata['ghost_infront']=None
            controldata['velocity_line']=sec.velocity_line
            #car.control(controldata)    

        elif sec.label=='brake':
            #print 'remove ghost'
            controldata['ghost_infront']=None

            # now start braking down to linespeed of berth
            controldata['velocity_line']=sec.velocity_line
            #car.control(controldata)
            
        elif sec.label=='main_exit':
            # special stop merge instructions
            ind=sec.cars.index(car)
            #print 'control_car_at_mark',ind,sec.cars
            if ind==len(sec.cars)-1:
                # car is last carrier entered on main_exit...loads of space
                self.control_car_at_mark_mergemod(sec,car,controldata)

            #elif (car.get_path_sec()-sec.cars[ind+1].get_path_sec())>\
            #     (0.5*sec.velocity_line**2/car.accel_comfort):
            elif (car.get_path_sec()-sec.cars[ind+1].get_path_sec())>\
                 (2.0*(sec.velocity_line*car.time_brake+\
                       0.5*sec.velocity_line**2/car.accel_emergency-\
                       0.5*sec.velocity_line**2/car.accel_failure)):
                # enough space behind entering car and following car
                self.control_car_at_mark_mergemod(sec,car,controldata)
            else:
                #self.control_car_at_mark_mergemod(sec,car,controldata)
                self.control_car_at_mark_mergemod(sec,car,controldata,
                                                  no_mergecar_behind=1)
        elif sec.label=='accel':
                self.control_car_at_mark_mergemod(sec,car,controldata)

        #if self.probe():
        #    print '  leave control_car_at_mark',car.id
            
    def softclear_berths(self):
        """
        Make all waiting carriers moving that are empty and do not pass by
        berth with waiting users.

        """
        #return
        #if self.probe():
        #    print 'softclear_berths: from ',berth_ind
                    
        # start liberating carriers with laste berth
        n=self.berthssec.berth_no-1
        aboard=False
        while (n>=0)&(not aboard):
            berth_check=self.berths[n]

            berth_check=self.berths[n]
            # check if there are no allocations loading or waiting
            if berth_check['status']=='waiting':
                car=berth_check['car']
                self.unoccupy_berth(berth_check)
                # make sure that it does not get reallocated out
                self.update_cars(car,{'state':'passthrough'})

                if self.probe():
                    print 'softclear_berths: cleared',berth_check['ind'],\
                          'from car',car.id
                          
                # test: stop kicking out cars
                aboard = True    
            
            # aboard if there are waiting users at a berth
            aboard = len(berth_check['queue'])>0
                   
            # switch to next berth
            n-=1
            
    def clear_berths(self,berth_ind=0):
        """
        Make all waiting carriers moving that are on and in front of berth
        with berth_ind.

        """
        #if self.probe():
        #            print 'clear_berths: from ',berth_ind
        while berth_ind<self.berthssec.berth_no:
            berth_check=self.berths[berth_ind]
            
            if berth_check['status']=='waiting':
                car=berth_check['car']
                self.unoccupy_berth(berth_check)
                # make sure that it does not get reallocated out
                self.update_cars(car,{'state':'passthrough'})

                if self.probe():
                    print 'clear_berths: clear',berth_check['ind'],\
                          'from car',car.id
            # switch to next berth
            berth_ind+=1

    def unallocate_empty_berths(self,car):
        """
        unallocate already allocated empty berth for all cars in fron of car.
        """
        #print 'unallocate_empty_berths',car.id
        ind=self.berthssec.cars.index(car)+1
        while (ind<len(self.berthssec.cars)):
            car_check=self.berthssec.cars[ind]
            if self.cars[car_check.id]['berth_allocated']:
                berth_alloc=self.cars[car_check.id]['berth_allocated']
                if len(berth_alloc['queue'])==0:
                    self.unalloc_berth(berth_alloc,car_check)
                    self.update_cars(car_check,{'state':'passthrough'})
            ind+=1


    def find_berth(self):
        """
        Find a liberated and unallocated berth at this stop.
        This is used by the add carrier function.
        """
        #print 'find_berth'
        # check for an unallocated berth, start with first berth
        i=0
        terminated=0
        berth=None

        berthsec=self.berthssec

        # The position of the last carrier on the berths section is required
        # to check if tghis last carrier is entirely
        # beyon the halt point of the found berth
        pos_lastcar=berthsec.length # default in case there is no other carrier
        if len(berthsec.cars)>0:
            lastcar=berthsec.cars[-1] # there is one
            # last car in array structure?
            if (lastcar.status=='active')|(lastcar.status=='halt'): 
                pos_lastcar=berthsec.netsim.path_sec[lastcar.ind]-\
                             lastcar.length
                #print 'find_berth: berthsec.netsim.path_sec[lastcar.ind]',\
                #      berthsec.netsim.path_sec[lastcar.ind]
            
        
        #print '  check berth at',self.berths[0]['halt_point'],\
        #      'pos_lastcar=',pos_lastcar,self.berths[0]['allocated'],\
        #      self.berths[0]['status']
        
        # check first berth is free
        if (not self.berths[0]['allocated'])&\
           (self.berths[0]['status']=='liberated')&\
           (self.berths[0]['halt_point']<pos_lastcar):

            # check if there are more berth free and return the
            # last free berth
            
            while not terminated:
                i=i+1
                if i<self.berthssec.berth_no:
                    #print '   check berth at',\
                    #      self.berths[i]['halt_point'],\
                    #      self.berths[i]['allocated'],\
                    #      self.berths[i]['status']
                    
                    # if berth i is already allocated or occupied
                    if (self.berths[i]['allocated'])|\
                       (pos_lastcar < self.berths[i]['halt_point'])|\
                       (self.berths[i]['status']!='liberated'):
                        #print '     return free berth i-1'
                        berth=self.berths[i-1]
                        terminated=1
                        
                    elif i==(self.berthssec.berth_no-1): # last berth free
                        #print '     return last free berth'
                        berth=self.berths[i]
                        terminated=1

                    
                        
                else: # no more berth ?
                    #print  '  first one was only free berth'
                    berth=self.berths[0] 
                    
                    terminated=1
            
        return berth



    def get_nextsec(self,sec,car):
        """
        This function returns the next section object when carrier car
        leaves section sec. This means that sec has at least two
        possible active branches.

        Here the standard algorithm is adapted to the branch requirements
        at the diverge section at the entrance of the stop.
        
        """
        if not self.cars_offtrack:
            berthssec=self.secs_dir['berths']
            if len(berthssec.cars)>0:
                is_congested=(berthssec.cars[-1].get_velocity()  \
                                / berthssec.velocity_line) < 0.2
            else:
                is_congested=False
                
            #determine whether berth are (will be) occupied
            are_all_berth_accupied=(len(self.berthssec.cars)\
                                    +len(self.secs_dir['brake'].cars)-1)\
                                    >= self.berthssec.berth_no
                                
            if self.time_sync>0:
                # demand in synchronous mode
                #if not self.is_blocked_unloadzone:
                demand= (not are_all_berth_accupied) & (not is_congested)
                #else:
                #    # no demand during liberation phase
                #    demand=0
            else:
                # demand for non-synchronous mode
                demand=((len(self.berthssec.cars)+\
                         len(self.secs_dir['brake'].cars))<self.berths_occupy-1)\
                         | (len(self.berths_demand)>len(self.berthssec.cars))&(not is_congested)

        else:
            # absolutely no demand if there are still carriers off-track
            demand=-1
            are_all_berth_accupied=True
            
        #if True:#self.probe():
        #    print 'get_nextsec',self.id,' for',car.id,' demand=',demand
        #    print '  cars in berth',len(self.berthssec.cars),\
        #          'cars in brake',len(self.secs_dir['brake'].cars),\
        #          'berths_occupy',self.berths_occupy

        
        
        # let's see if there is a task for this carrier
        
        #if self.car_tasks.has_key(car.id):
        #if 1: 
        if self.car_tasks.has_key(car.id):
            # use always oldest  task
            task=self.car_tasks[car.id][0]
            search_item=(task.mode == 'search_item')
            target_flag=task.target_flag
            nextsec_id_task=task.get_nextsec(sec.id)

            #print self.id,'.get_nextsec for',car.id,'task',task.id
            #print '  mode=',task.mode,'target_flag',task.target_flag,self.sim.time
        else:
            #print  '  there is no task??'
            # define the following variables in a way that
            # the car is entering if the stop needs empty cars
            # this means it will behave as a car with a search item task
            task=None
            search_item=1
            target_flag=0
            
        #print self.id,'.get_nextsec: ',car.id,\
        #print '  target_flag',target_flag,'search_item=',search_item,\
        #       'demand=',demand

        #if task:
        #    print '  Task',task.id,'mode=',task.mode,\
        #          'nextsec_id of task',nextsec_id_task
        
        # recomment to enter the brake line if carrier is
        # searching for an item and if there are items in the queue
        if  (target_flag & (not search_item)&(demand!=-1))\
               | (target_flag & search_item & (demand==1) )\
               | (search_item & (demand==1) &\
                  (len(self.secs_dir['brake'].cars)\
                   <self.secs_dir['brake'].car_max) ):
            
            nextsec_id=self.secs_dir['brake'].id
            #print '  try to go in',nextsec_id
            
        elif (task!=None)&(not search_item):
            # car has a task and a load, but this stop is not a target.
            # so it will probably go straight...
            
            # look up in the nextsection
            # table whether there is a valid entry.

            
            if nextsec_id_task==-1:
                nextsec_id=self.secs_dir['main_entry'].id
                #print '  try to go streight',nextsec_id
            else:
                nextsec_id=nextsec_id_task
                #print '  try to go (streight)', nextsec_id_task
                
            #if self.probe():
            #print '  not search_item, nextsec_id',nextsec_id
            
        else: # default means go straight
            nextsec_id=self.secs_dir['main_entry'].id
            #print '  default means go straight nextsec_id',nextsec_id

        
        #if self.probe():
        #        print '  preselection nextsec_id',nextsec_id,\
        #              'target_flag=',target_flag

        # if there was no demand but the carrier had to get in
        # anyway because the stop is a target, then start
        # already here clearing the berths
        if (demand<=0)&(nextsec_id==self.secs_dir['brake'].id)\
                &(self.time_sync<=0.0):
            #if self.probe():
            #print '  get_nextsec: clear_berths,self.time_sync',self.time_sync
            self.clear_berths()
        
        #print """  Now check if desired nextsec is free, safe or 
        #           if a branching would disturb the vehicle flow---."""

        
        # is desired nextsec the brake section ?
        error='0'
        if nextsec_id==self.secs_dir['brake'].id:
            nextsec=self.secs_dir['brake']
            #print '  Verify BRAKE', self.secs_dir['brake'].id
            #print ' crit,len(nextsec.cars)',\
            #      self.enter_sec_crit(nextsec,car),\
            #      len(nextsec.cars),nextsec.car_max

            # it is required that no carriers are on the brake section.
            if len(nextsec.cars)==0:
                #print ' OK,  no carriers are on the brake section'
                #return nextsec, '0'
                pass
            
            # check if it is safe to enter and if last car is still moving
            # above line speed
            elif self.enter_sec_crit(nextsec,car)&\
                     (len(nextsec.cars)<nextsec.car_max):
                #print ' OK,   safe to enter and not overcrowded'
                #return nextsec, '0'
                pass
            
            # allow also to enter brake section if there is only the
            # current carrier in front of this carrier
            #elif len(nextsec.cars)>0:
            #    return nextsec, '0'

            else: # now try main-line
                #if self.probe():
                #if self.enter_sec_crit(nextsec,car):
                #    print '  FAILED enter brake, safety critical'
                #else:
                #    print '  FAILED enter brake, overcrowded',len(nextsec.cars),nextsec.car_max
                #print '  now try main-line'
                # OK to enter if last car of nextsec
                # is current car in front
                nextsec=self.secs_dir['main_entry']
                if len(nextsec.cars)>0:
                    if nextsec.cars[-1]!=car.car_infront:
                        nextsec=self.safest_nextsec([\
                            self.secs_dir['main_entry'],
                            self.secs_dir['brake']])
                        #if self.probe():
                        #print ' send car on safest nextsec',nextsec.id
   
        # is desired nextsec_id the main line ?
        elif nextsec_id==self.secs_dir['main_entry'].id:
            #print '  Verify MAIN', self.secs_dir['main_entry'].id
            nextsec=self.secs_dir['main_entry']

            # OK to enter if nextsec is free
            if len(nextsec.cars)==0:
                #return nextsec, '0'
                #print ' OK,   free'
                pass
            
            # OK to enter if last car of nextsec is current car in front
            elif nextsec.cars[-1]==car.car_infront:
                #return nextsec, '0'
                #print '  OK, car in front is safe' 
                pass
            
            # OK if safety criteria is satisfied
            elif self.enter_sec_crit(nextsec,car):
                #print '+++ safe to enter'
                #return nextsec, '0'
                #print '  OK, safe' 
                pass
            else:
                nextsec=self.safest_nextsec([self.secs_dir['main_entry'],
                                             self.secs_dir['brake']])
                #if self.probe():
                #print 'get_nextsec at ',self.id,sec.id,'for',car.id
                #print '  FAILED, try to send carrier on safest nextsec',\
                #          nextsec.id
                
                
        else: # task did not provide a valid routing information
            #print '  Verify ... no task found' 
            #if self.probe():
            #        print '  task did not provide a valid routing'
            # try to branch it into berth
            nextsec=self.secs_dir['brake']
            # it is required that no carriers are on the brake section.
            if len(nextsec.cars)==0:
                #return nextsec, '0'
                pass
            else:
                # try to send carrier on safest nextsec
                nextsec=self.safest_nextsec([self.secs_dir['main_entry'],
                                             self.secs_dir['brake']])
                #nextsec=self.secs_dir['main_entry']
                #return nextsec, '0'

        # now deviate task if desired routing could not be satisfied
        # but only in case the carrier is not searching an item...
        # in this case it wants to get deviated :)
        if (task!=None)&(not search_item):
            if nextsec.id!=nextsec_id_task:
                #print 'get_nextsec at ',self.id,sec.id,'for',car.id
                #print '  demand=',demand,\
                #      'target_flag=',target_flag,'search_item=',search_item,\
                #      'order_id',task.order_id,\
                #      'had',len(self.car_tasks[car.id]),'tasks'
                #print '  deviated from',nextsec_id_task,\
                #      'to nextsec',nextsec.id
                error='blocked'
                self.control_car_on_deviation(sec,nextsec,car,task)
                
        #if self.probe():
        #print '  =>nextsec=',nextsec.id,error
        
        if nextsec==self.secs_dir['brake']:
            self.check_user(car,message='get_nextsec',signal_false_enter=True)
        else:
            self.check_user(car,message='get_nextsec')    
        return   nextsec, error
    
    def safest_nextsec(self,secs):
        """
        This function returns the safest section to enter
        at a diverge point.
        """
        
        dist_max=-10.0**10
        nextsec_best=secs[0]
        for nextsec in secs:
            # if sec.enter_sec_crit(car,nextsec): 
            if len(nextsec.cars)>0: # is there a car on next section  ?
                lastcar=nextsec.cars[-1]
                if lastcar.modind!=-1: # last car in array structure?
                
                    dist_lastcar=lastcar.get_path_sec()
                    dist= dist_lastcar+lastcar.get_brakedist_failure()
                    
                    if dist>dist_max:
                        dist_max=dist
                        nextsec_best=nextsec
                        

            else: # best is, if nextsec is clear
                dist_max=+10.0**10
                nextsec_best=nextsec

        return nextsec_best
        
    def add_car(self,car):
        ''' Add a carrier to this Track element'''
        #print self.id,'.add_car'
        # check whether there is space for one more carrier on the
        # berth section of this stop
        berth=self.find_berth()
        if berth:
            berthsec=self.berthssec
            
            # for safety reasons, check if
            # there is no carrier arriving on the brake line.
            if len(self.secs_dir['brake'].cars)==0:
                # tell netsim to add array structures for this carrier
                # and put it at the beginning of the berth section
                if  self.obj.netsim.add_car(car,berthsec,berth['halt_point']):
                    if self.debug>=3:
                        print 'WARNING: something went wrong with netsim'
                    return 'error'
                else:
                    # add carrier to berth section's carrier buffer
                    berthsec.cars.append(car)

                    # add carrier to track element wide carrier dict
                    self.update_cars(car,{'state':'searching'})

                    if self.debug>=5:
                        print 'Carrier added on berth ',self.id
                    return 'done'
                        
        #print '  no free berth found'

        # place car off-track if not already there
        if self.cars_offtrack.count(car)==0:
            self.cars_offtrack.append(car)
            x,y=self.berthssec.transform_polyline(
                [-1-1.5*car.length*len(self.cars_offtrack),-5*self.side])
            car.pos=[x,y,0]
            car.sec=self.secs_dir['brake']
            car.plot()
        return 'offtrack'

    def add_cars_offtrack(self):
        """
        Take off-track carriers and try to add them to the stop
        """
        # try to add car until it fails i.e. berth are full or not safe
        added='done'
        while (len(self.cars_offtrack)>0)&(added=='done'):
            car_off=self.cars_offtrack[-1]
            added=self.add_car(car_off)
            # print 'add_cars_offtrack:',car_off.id,added
            if added=='done':
                car_off.set_status('halt')
                car_off.start('no_task')
                del self.cars_offtrack[-1]

        # are there still cars offtrack
        if len(self.cars_offtrack)>0:
            # make sure that there is no waiting car blocking the insertion
            self.clear_berths(0)
            

        
    def del_car(self,car):
        '''
        Remove  carrier from  this track element, if possible
        The calling module must assure that carrier is empty.
        '''

        if self.cars_offtrack.count(car)==1:
            del self.cars_offtrack[self.cars_offtrack.index(car)]
            return 'deleted'
        
        # is carrier on berth section? and is it in halt mode?
        elif (self.berthssec.cars.count(car)>0)&(car.status=='halt'): 
            if  self.obj.netsim.del_car(car): # delete from netsim
                return 'WARNING:Netsim cannot delete carrier',car.id
            
            else: # netsim OK, delete car also from stop data structure
                #if self.berths_alloc.has_key(car.id):
                #    self.unallocate_berth(car)
                if self.cars[car.id]['berth_allocated']:
                    self.unalloc_berth(self.cars[car.id]['berth_allocated'],\
                                       car)
                    
                if self.cars[car.id]['berth_occupied']:
                    self.unoccupy_berth(self.cars[car.id]['berth_occupied'])
                
                self.delete_from_cars(car)
            
                return 'deleted'
            
        else: # carrier not in berth
            return 'posponed'
    
    def enter(self,user):
        ''' Add  user to this Track element'''

        # insert user into the stop's database
        # The field berth_ind corresponds to the queue at this berth  where the
        # user is queuing or will queue up.
        # berth_ind will be determined at get_walkto_position
        # when called with the 'queue_up' target place.
        
        self.users[user.id]={'berth_ind':-1}
        #if self.probe():
        #    print self.id,'enter',user.id,len(self.users)

        if self.obj.netmon: self.plot()
        return 0


    def get_walkto_position(self,user,args):
        """
        Return 2D coordinates of the position where the
        user should or wants to walk to.

        The name target place is defined in user.target_place

        It must be defined in terminate_walkto() what the user can do
        after he reached this position.
        """
        #print 'get_walkto_position for',user.id,user.target_place
        # goto shourest queue ?

        # user diameter
        d=0.8

        width_platform=2.0
        
        if user.target_place=='queue_up':
            # wants to walk to end of shortest queue of the stop 

            # search shortest queue...

            # in synchronous mode, only half of the berth are load berths
            # is stop in synchronous mode?
            if self.time_sync>0.0:
                ind_berth_min=self.ind_berth_firstalloc
            else:
                ind_berth_min=self.berthssec.localsec['berths_keep_clear']-1
                
            len_min=10**6
            
            for berth_ind in range(self.berthssec.berth_no-1,ind_berth_min,-1):
                len_queue= len(self.berths[berth_ind]['queue']) +\
                           self.berths[berth_ind]['users_approaching']
                if len_queue<len_min:
                    len_min=len_queue
                    berth_ind_shortest=berth_ind
                    
            # insert target queue in stop's user db
            self.users[user.id]['berth_ind']=berth_ind_shortest
            berth=self.berths[berth_ind_shortest]
            berth['users_approaching']+=1
            
            #print 'get_walkto_position for',user.id,\
            #      ' berth_ind',\
            #      berth_ind,self.berthssec.transform_polyline(\
            #    [berth['halt_point']-0.5*self.berthssec.berth_length,\
            #     -self.side*(1.5*self.obj.width+len_min*d)])
            
            return self.berthssec.transform_polyline(
                [berth['halt_point']-0.5*self.berthssec.berth_length,
                 -self.side*(1.1*self.obj.width+(len_min-1)*d)])

        # advance in queue ?
        if user.target_place=='queue':
            # get berth where user is queuing
            berth=self.berths[self.users[user.id]['berth_ind']]
            return self.berthssec.transform_polyline(
                [berth['halt_point']-0.5*self.berthssec.berth_length,
                 -self.side*(1.5*self.obj.width+args*d)])

        # walk to a waiting position on the frontend of the platform ?
        if user.target_place=='platform':
            return self.berthssec.transform_polyline(
                [self.berthssec.length*uniform(1.1,1.3),
                 -width_platform*self.side])
            # get current position
            #x,y=user.get_position()
            # keep x position and make him walking to a random y position
            # back on the platform
            #return self.berthssec.transform_polyline([x,uniform(
            #    -10*self.side,-15*self.side)])


        # goto random position, somehere around the stop
        elif user.target_place=='random':
            return self.berthssec.transform_polyline(\
                [uniform(-5,-1),
                 uniform(-1.2,-1.5)*self.side*width_platform])
        
            #return self.berthssec.transform_polyline([uniform(-1,\
            #                                           self.berthssec.length),
            #                                          uniform(-0.5*self.side,
            #                                                  -width_platform\
            #                                                  *self.side)])
        
        # walk to the exit position of the stop

        if user.target_place=='exit':
            # calculate  exit position at the very end of the stop,
            # a bit away from the berth
            return self.berthssec.transform_polyline(#
                [-0.5,-3*width_platform*self.side])


    def terminate_walkto(self,user):
        """
        This method is called when the user arrived at a target place
        on this track element.

        Dependent on the target place, the track element
        will trigger actions.
        """
        
        if  user.target_place=='queue_up':
            #print 'terminate_walkto for',user.id,user.target_place
            self.queue_user(user)

        # user arrived at a  waiting position on the backend of the platform ?

        elif user.target_place=='platform':
            # tell user that he is in a waiting position
            # He will the decide what to do--- for example making a new trip.
            user.waiting_at(self)

        elif user.target_place=='random':
            # tell user that he is in a waiting position
            # He will the decide what to do--- for example making a new trip.
            user.waiting_at(self)

        elif user.target_place=='exit':
            user.exit_at(self)
        
            
            
    def get_pos_entrance(self,x_offset=0,y_offset=0):
        """
        Get current position of stop entrance
        """
        # here, 3m behind end of platform and 5m away from platform
        return self.berthssec.transform_polyline(
            [self.berthssec.length+5+x_offset,\
             -(5+y_offset)*self.side])

    def get_pos_exit(self):
        """
        Get current position of stop exit
        """
        # here, 3m before platform and 5m away from platform
        return self.berthssec.transform_polyline([-5,-5*self.side])
    
    def get_dir_norm(self):
      """
      Get normalized vector of direction of stop on xy plane.
      """
      x0,y0=self.berthssec.transform_polyline([0.0,0.0])
      x1,y1=self.berthssec.transform_polyline([1.0,0.0])
      return (x1-x0,y1-y0)
      
    def exit(self,user):
        '''
        User exits this Track element.
        It is supposed that the user has already found a new location
        and the handover went OK.
        '''
        
        if 1:# not user.car: # user is not already assigned to a car
            # remove from queue if he is queuing
            self.unqueue_user(user)

            # remove from users db
            del self.users[user.id]

            #if self.probe():
            #    print self.id,'exit',user.id,len(self.users)
            
            # plot
            if self.obj.netmon: self.plot()
            return  0

        else:
            return 1

    def queue_user(self,user):
        '''Put user in the berth waiting queue'''
        #if self.probe():
        #    print self.id,'.queue_user',user.id
        if self.users.has_key(user.id):
            # get berth and queue assigned to user
            berth_ind=self.users[user.id]['berth_ind']
            if berth_ind>=0:
                berth=self.berths[berth_ind]
                #if self.probe():
                #    print '  at berth',berth['ind'],berth['halt_point'],berth['status']
                
                # decrease number of approaching users
                berth['users_approaching']-=1
                
            else: # be tolerant, if user did not approach properly
                berth=self.berths[-1]
                self.users[user.id]['berth_ind']=self.berthssec.berth_no-1
                #if self.probe():
                #    print '  at alternative berth',berth['ind'],berth['halt_point'],\
                #          berth['status']

            
            
            queue=berth['queue']
            
            queue.append(user)
            self.n_users_queuing+=1
            
            # found a berth with waiting carrier ?
            if berth['status']=='waiting':
                

                # get corresponding carrier
                car=berth['car']
                
                # tell waiting carrier to load
                # this will also terminate the current search-for-carrier task 
                #if self.probe():
                #    print '  load carrier',car.id
                self.cars[car.id]['state']='loading'
                #self.update_cars(car,{'state':'loading'})
                car.load()
                

                # set the search for item flag of the carrier task to zero
                if self.car_tasks.has_key(car.id):
                    # this task will be killed later by the carrier management
                    # when the new task for the item has been created
                    # take oldest task
                    self.car_tasks[car.id][0].mode='loaded'

                    #print 'queue_user:search_item=0 for task',\
                    #      self.car_tasks[car.id].id

            elif not berth['allocated']:
                # try to get a car allocated, but only in non-sync mode
                #if self.probe():
                #    print '  try to call alloc_car_for_berth',self.time_sync,self.time_sync==0.0
                if self.time_sync==0.0:
                    # there is no carrier waiting or allocated at this berth
                    self.alloc_car_for_berth(berth)

            return 0
        else:
            return 1
        
    def unqueue_user(self,user):
        '''Take user out off the berth waiting queue'''
        
        if self.users.has_key(user.id):
            berth_ind=self.users[user.id]['berth_ind']

            if berth_ind>=0: # user approaching or already in queue
                # get queue assigned to user
                queue=self.berths[berth_ind]['queue']
                if queue.count(user)>0: # user is in the queue
                    queue.remove(user)
                    self.n_users_queuing-=1
                else: # user is approaching but not yet in the queue
                    self.berths[berth_ind]['users_approaching']-=1

                self.users[user.id]['berth_ind']=-1
                return 0

            else: # user is not approaching neither in the queue
                return 0
                
                
        else: # unknown user
            return 1

    def queuing_user(self):
        '''
        Returns number of currently queuing users at this stop.
        '''
        n=0
        for berth in self.berths:
            n+=len(berth['queue'])
            
        return n

    #def get_waitingtime_users_queued(self):
    #    '''
    #    Returns waiting time of all currently queuing users at this stop.
    #    '''
    #    t=0.0
    #    for user in self.get_users_queued():
    #        t+=self.sim.time-user.time_start_waiting
    #        
    #    return t
            
        
    def get_users_queued(self):
        '''
        Returns a list with all queuing users at this stop.
        '''
        users_queuing=[]
        for berth in self.berths:
            for user in berth['queue']:
                users_queuing.append(user)
        return users_queuing

    def get_berths_with_queue(self,length_queue=1):
        '''
        Returns a list with berths that have a queue of users longer or
        equal to length_queue.
        '''
        
        berths_with_queue=[]
        for berth in self.berths:
            if len(berth['queue'])>=length_queue:
                berths_with_queue.append(berth)

        return berths_with_queue

    def terminate_unloading(self,car):
        """
        This is called by a carrier that has just unloaded an item.
        Here it will be decided wether to load the carrier with a user of the
        queue or to send it to a better position downstream.
        """
        #if self.probe():
        #print 'terminate_unloading of',car.id
        
            
        # do statistics
        self.n_unloads+=1
        

        # moved to control at mark...., but still needed for older versions
        # check if a task has been completed
        if self.car_tasks.has_key(car.id):
            self.car_tasks[car.id][0].haltin_sec(self.berthssec)
        #else:
        #    print 'WARNING: in terminate_unloading',self.id
        #    print '  car',car.id,'has terminated unloading,',\
        #          'but has no task??'

    
        # declare carrier unloaded, this has not been done yet
        # on the track element database
        self.update_cars(car,{'loaded': 0})
        
        # are we in synchronous mode?
        if self.time_sync>0.0:
            if self.cars[car.id]['berth_occupied']:
                berth=self.cars[car.id]['berth_occupied']
                
                # do statistics
                berth['n_unloads']+=1   

                # put berth from loading in waiting mode:
                berth['status']='waiting'
                self.update_cars(car,{'state': 'waiting'})

                # when berth is in unload area...the normal case when unloading
                if berth['ind']<=self.ind_berth_firstalloc:
                    #is there still queue in unload area
                    # this may happen during the transition from
                    # non-sync to sync mode
                    if len(berth['queue'])==0:
                        # move on to corresponding berth of load area
                        self.alloc_berth_with_car_sync(car)
                    else:
                        # try to load remaining user
                        car.load()

                else:  # berth is in load area 
                    # try to load
                    car.load()

            else:
                print 'WARNING in terminate_unloading',self.id
                print '  carrier',car.id,' has no more an occupied berth',\
                     self.cars[car.id]['berth_occupied'] 
                
                # try to load carrier in any case
                car.load()
                    
        else: # do non-synchronous mode
            # check if this carrier can be allocated
            # to a better downstream position
            if self.cars[car.id]['berth_occupied']:
                berth=self.cars[car.id]['berth_occupied']
                #print '  at',self.id,'berth',berth['ind'],berth['allocated']
                
                # do statistics
                berth['n_unloads']+=1 
                
                # put berth from loading in waiting mode:
                berth['status']='waiting'
                self.update_cars(car,{'state': 'waiting'})

                #if not self.cars[car.id].has_key('found_in'):
                #    print 'terminate_unloading: no key found_in in',car.id
                #    print self.cars[car.id]
                
                if self.cars[car.id]['found_in']=='berths_demand':
                    # before unload allocated demand berth, so...load now
                    car.load()

                else:
                    berth_found,found_in=self.find_berth_downstream(berth)
                
                    # is there a free demanding berth?
                    if found_in=='berths_demand':
                        # if the found berth is the current berth, we do load
                        if berth_found==berth:
                            #print '  found same berth',\
                            #      berth['ind'],berth['allocated']
                            car.load()
                        else:
                            # yes, send carrier away immediately
                            #if self.probe():
                            #    print '  found and alloc demand berth',\
                            #          berth_found['ind']
                            cd=self.unoccupy_berth(berth,{})
                            cd=self.alloc_berth(berth_found,car,cd,found_in)
                            #self.berths_demand.remove(berth_found)
                            car.control(cd)

                            # but there may still be a quque at this berth,
                            # so make sure it is in the demand list...
                            if len(berth['queue'])>0:
                                #print '  ..try to allocate a car for berth',\
                                #      berth['ind'],'at',self.id
                                self.alloc_car_for_berth(berth)

                    else:
                        # no berth found, try to load carrier here 
                        car.load()
                
            else:
                    print 'WARNING in terminate_unloading',self.id
                    print '  carrier',car.id,' has no more an occupied berth'

                    car.load()
        
        
    def load_item(self,car):
        '''
        In this case, make next user in the queue  board the carrier.
        It tells the respective user to board the carrier.
        It returns None if the queue is empty, otherwise the user. 
        '''
        user=None
        if self.cars[car.id]['berth_occupied']:
            # get queue of corresponding berth
            berth=self.cars[car.id]['berth_occupied']
            queue=berth['queue']
            
            #if self.probe():
            #    print 'load_item from queue at',berth['ind']
            #for u in queue: print '    ',u.id
            if len(queue)>0:
                # declare carrier as occupied
                berth['status']='occupied'
                self.update_cars(car,{'state':'loading'})
                
                user=queue.pop(0) # get first user in queue
                self.n_users_queuing-=1

                #if self.obj.netmon: self.plot()
                # print '  unqueued to load:',user.id
                self.users[user.id]['berth_ind']=-1
                user.start_boarding(car)

                # do statistics
                self.n_loads+=1
                berth['n_loads']+=1
                waitingtime=self.sim.time-user.time_start_waiting
                self.waitingtime_total+=waitingtime
                self.waitingtime2_total+=waitingtime**2
                self.waitingtime_min=min([self.waitingtime_min,waitingtime])
                self.waitingtime_max=max([self.waitingtime_max,waitingtime])
                
               
                
                
                # make users in queue walk to their new position
                #print '  queue to advanve:'
                #for u in queue: print '    in queue user:',u.id
                for i in range(len(queue)):
                    queue[i].walkto(self,'queue',i)

                # if there is still 
                if len(queue)>0:
                    # there are still users in the queue
                    # is stop in non-sync mode?
                    if self.time_sync<=0.0:
                        #...try to allocate a car
                        self.alloc_car_for_berth(berth)
                
                    
                        
                
            else:
                # this case should actually not happen in
                # async mode because load is only called with a non-empty queue
                #print 'load_item for',car.id,'at',self.id
                #print '  no user in the queue'
                berth['status']='waiting'
                self.update_cars(car,{'state':'waiting'})
                #cd=self.alloc_berth_with_car(car,{})
                #print '  cd=',cd
                #car.control(cd)
                
                
        else:
            #print 'no berth found'
            pass


        # check if conditions are satisfied to
        # switch between synch and non-synch mode
        if self.time_sync<=0.0:# we are in non-sync mode?
            # start sync mode if all berths have 
            # a queue of at least 2 users
            if self.is_start_sync():
                #print 'Switch',self.id,'to sync mode at time:',self.sim.time
                self.start_sync_mode()
        else: # we are in sync mode.
            # stop sync mode if half of the berths-1 have no more queues 
            if self.is_stop_sync():
                self.stop_sync_mode()
                
        return user

    def make_panel_specific(self,panel):
        """
        Add track element specific panel infos.
        To be overwritten.
        """
        # add a parameters page
        panel.make_page('Parameters',self.make_parampage,self.update_parampage)

    def make_parampage(self,parent,widgets):
        descriptors,dataset=self.format_params()
        widgets['param_data']=gui.DatasetEditor(descriptors,dataset,
                                               parent,
                                               applyentry_command=\
                                                   self.apply_param_gui)
        widgets['param_data'].pack(expand=1,fill='both')
        
    def update_parampage(self,panel,widgets):
        """
        Update of widgets on costs page. This method is called
        when cost-page needs to be updated.

        """
        descriptor,dataset=self.format_params()
        widgets['param_data'].update_contents(dataset)
        
    def apply_param_gui(self,paramdatawidget,param_data):
        """
        param_data comes from widchet-> contains all keys!!
        """
        if self.name!=param_data['name']:
            self.name=param_data['name']
            self.update_man()
        if self.zone!=param_data['zone']:
            self.zone=param_data['zone']
            self.update_man()
            
        if param_data['berths_keep_clear']<self.berthssec.berth_no:
            self.berthssec.localsec['berths_keep_clear']=\
                                 param_data['berths_keep_clear']

            self.berths_occupy=self.berthssec.berth_no-\
                                self.berthssec.localsec['berths_keep_clear']
        else:
            param_data['berths_keep_clear']=\
                                self.berthssec.localsec['berths_keep_clear']
        
        if self.status!='active':
            self.ind_berth_firstalloc=self.berthssec.berth_no\
                                       -param_data['n_berth_load']-1

        self.mode_exch=param_data['mode_exch']
      
        
        if self.obj.netmon:
            self.plot()
        descriptor,dataset=self.format_params()
        paramdatawidget.update_contents(dataset)

        
    def format_params(self):
        """
        Returns describtor and dataset of parameters.
        """
        descriptors=[
            {'key':'name',
             'name':'Name',
             'type':'text',
             'width':1,
             'mode':'write',
             },
            {'key':'zone',
             'name':'Zone',
             'type':'text',
             'width':1,
             'mode':'write',
             },
            {'key':'mode_exch',
             'name':'Mode exchange',
             'type':'text',
             'width':1,
             'mode':'write',
             },
            {'key':'berths_nr',
             'name':'Number of berths',
             'type':'integer',
             'printformat':'%4d',
             'width':1,
             'mode':'read',
             'valid':{'min':0,'max':1000},
             },
            {'key':'n_berth_load',
             'name':'Number of load-berths',
             'type':'integer',
             'printformat':'%3d',
             'width':1,
             'mode':'write',
             'valid':{'min':1,'max':self.berthssec.berth_no-1},
             },
            {'key':'berths_keep_clear',
             'name':'Berths to be kept clear',
             'type':'integer',
             'printformat':'%4d',
             'width':1,
             'mode':'write',
             'valid':{'min':0,'max':1000},
             },
             #{'key':'n_berths_load',
             #'name':'Berths in load area',
             #'type':'integer',
             #'printformat':'%4d',
             #'width':1,
             #'mode':'write',
             #'valid':{'min':0,'max':1000},
             #},
             #{'key':'time_sync',
             # 'name':'Fill-up time-out',
             # 'type':'real',
             # 'printformat':'%.2f',
             # 'units':{'s':1},
             # 'width':1,
             # 'mode':'write',
             # 'valid':{'min':0,'max':3600},
             # },
            ]
        
        dataset={'berths_nr':self.berthssec.berth_no,
                 'zone':self.zone,
                 'mode_exch':self.mode_exch,
                 'name':self.name,
                 'berths_keep_clear':\
                 self.berthssec.localsec['berths_keep_clear'],
                 'n_berth_load':self.berthssec.berth_no-\
                                   self.ind_berth_firstalloc-1,
                                   
                 #'time_sync':self.time_sync,
                 
                 }
        
        return descriptors,dataset

    def reset_statistics_mod_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular track element.

        To be overwritten.
        """
        self.n_unloads=0
        self.n_loads=0
        self.waitingtime_total=0.0
        self. waitingtime2_total =0.0
        self. waitingtime_min =10.0**8
        self. waitingtime_max =0.0
        
        if hasattr(self,'berth'):    
            for berth in self.berths: 
                berth['n_loads']=0
                berth['n_unloads']=0.0
        
    def format_statistics_specific(self,descriptors=[],dataset={}):
        """
        Return descriptor and dataset contents for statistical data
        about the specific module.

        To be overwritten.
        """
        # prepend name of stop!
        descriptors.insert(0,\
                           {'key':'name',
                            'name':'Name',
                            'type':'text',
                            'width':1,
                            'mode':'write',
                            })
        
        descriptors.append({'key':'n_unloads',
                            'name':'Number of unloads',
                            'type':'integer',
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'printformat':'%4d',
                            'mode':'read',
                            })
        descriptors.append({'key':'n_loads',
                            'name':'Number of loads',
                            'type':'integer',
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'printformat':'%4d',
                            'mode':'read',
                            })
        descriptors.append({'key':'n_unloads_per_h',
                            'name':'Number of unloads per h',
                            'type':'real',
                            'printformat':'%.2f',
                            'units':{'1/h':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
        descriptors.append({'key':'n_loads_per_h',
                            'name':'Nbr Loads/h',
                            #'name':'Number of loads per h',
                            'type':'real',
                            'printformat':'%.2f',
                            #'units':{'1/h':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
        descriptors.append({'key':'n_users_waiting',
                            'name':'Number waiting users',
                            'type':'integer',
                            'printformat':'%4d',
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
        descriptors.append({'key':'waitingtime_users_queued_av',
                            'name':'Avg Wait',
                            #'name':'Average waiting time',
                            'type':'real',
                            'printformat':'%.2f',
                            'units':{'s':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
                            
        descriptors.append({'key':'waitingtime_min',
                            'name':'Min Wait',
                            #'name':'Minimum waiting time',
                            'type':'real',
                            'printformat':'%.2f',
                            'units':{'s':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
        descriptors.append({'key':'waitingtime_max',
                            'name':'Max Wait',
                            #'name':'Maximum waiting time',
                            'type':'real',
                            'printformat':'%.2f',
                            'units':{'s':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
        descriptors.append({'key':'waitingtime_dev',
                            'name':'Dev Wait',
                            #'name':'Deviation of waiting time',
                            'type':'real',
                            'printformat':'%.2f',
                            'units':{'s':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            })
        
        
        if self.sim.time>1.0:
            n_unloads_per_h=3600.0*self.n_unloads/self.sim.time
            n_loads_per_h=3600.0*self.n_loads/self.sim.time
        else:
            n_unloads_per_h=0.0
            n_loads_per_h=0.0

        
        if self.n_loads>0:
            waitingtime_users_queued_av=self.waitingtime_total/self.n_loads
            waitingtime_dev=sqrt(self.waitingtime2_total/self.n_loads \
                                - waitingtime_users_queued_av**2)
        else:
            waitingtime_users_queued_av=0.0
            waitingtime_dev=0.0
            
        dataset.update({'name':self.name,
                        'n_unloads':self.n_unloads,
                        'n_loads':self.n_loads,
                        'n_unloads_per_h':n_unloads_per_h,
                        'n_loads_per_h':n_loads_per_h,
                        'n_users_waiting':self.queuing_user(),
                        'waitingtime_users_queued_av':\
                        waitingtime_users_queued_av,
                        'waitingtime_min':self.waitingtime_min,
                        'waitingtime_max':self.waitingtime_max,
                        'waitingtime_dev':waitingtime_dev})
        #print '======================',self.id
        #print descriptors
        #print
        
        
        
    def show_info_specific(self):
        """
        Shows track element specific information
        """
        print 'Entrance coordinates:',self.get_pos_entrance()
        print '  Users around stop:',len(self.users)
        
        print '\n'
        print '  Berths:'
        for ind in range(self.berthssec.berth_no):
            berth=self.berths[ind]
            if berth['allocated']: allocation='allocated'
            else: allocation='not allocated'

            
            
            print '    Berth',ind,'at',berth['halt_point'],berth['status'],\
                  allocation
            print '      Queue length:',len(berth['queue'])
            print '      Users approaching:',berth['users_approaching']
            if berth['car']:
                print '      Car id:',berth['car'].id
            else:
                print '      Car id: -'
        
        print 'time_sync',self.time_sync
        
        if self.time_sync>0:
            print '  extra info on sync mode:'
            print '     is_blocked_unloadzone=',self.is_blocked_unloadzone
            print '     is_blocked_loadzone=',self.is_blocked_loadzone
            print '     ind_berth_firstalloc=',self.ind_berth_firstalloc
            print '     ind_berth_alloc=',self.ind_berth_alloc
        else:
            print '  len(berths_demand)=',len(self.berths_demand)
            inds=[]
            for berth in self.berths_demand:
                inds.append(berth['ind'])
            print '    berths_demand indexes:',inds

        #print 'check find_berth_downstream: with load'
        #for berth in self.berths:
        #    berth_found,found_in=self.find_berth_downstream(berth,1)
        #    if berth_found:
        #        print 'start with berth',berth['ind'],\
        #              'foundberth',berth_found['ind'],'found_in',found_in
        #    else:
        #        print 'start with berth',berth['ind'],'-'
        
        if self.debug>5:

            print '  Users:'
            for id_user in self.users.keys():
                user=self.sim-get_mod(id_user)
                berth_ind=self.users[id_user]['berth_ind']
                if berth_ind>=0:
                    berth=self.berths[berth_ind]
                    if berth['queue'].count(user)>0:
                        status='queuing at berth at '+`berth['halt_point']`
                    else:
                        status='approaching berth at '+`berth['halt_point']`
                else:
                    status='walking to '+user.target_place
                    
                print '   ',user.id,status

        if self.debug>0:
            print '  All carriers on berths sec:'
            for car in self.berthssec.cars:
                id=car.id
                print '      car.id=',id
                print '          state=',self.cars[id]['state']
                if self.cars[id]['berth_occupied']:
                    ind_berth_occupied=self.cars[id]['berth_occupied']['ind']
                else:
                    ind_berth_occupied=-1
                if self.cars[id]['berth_allocated']:
                    ind_berth_allocated=self.cars[id]['berth_allocated']['ind']
                else:
                    ind_berth_allocated=-1
                print '          berth_allocated with ind',ind_berth_allocated
                print '          berth_occupied with ind',ind_berth_occupied
                print '          loaded=',self.cars[id]['loaded']

            
          
                    
     
            
    def plot_extra(self,canvas):

        # plot boxes around berth
        dy0=-self.side*1.0 
        dy1=-self.side*(1+self.obj.width_netmon) 

        
        color_platform='#001f91' 
        color_berth='#82a0ff'# #000080navy #4876ff RoyalBlue1
        
        # set section width
        width=self.obj.width_netmon
        if (canvas==self.obj.netmon):
            width_scaled=width*self.obj.netmon.canvas_scale
        else:
            width_scaled=width

        x0=self.berths[0]['halt_point']-self.berthssec.berth_length
        x1=self.berths[-1]['halt_point']+0.2
        #x0=self.berthssec.length-(self.berthssec.localsec['berth_no'])*\
        #    self.berthssec.berth_length-0.2
        #x1=self.berthssec.length
        id=canvas.create_line(
            self.berthssec.transform_polyline([x0,(dy0+dy1)/2,
                                               x1,(dy0+dy1)/2,
                                               ]),
            width=width_scaled,
            tag=self.id,fill=color_platform)
        
        if canvas==self.obj.netmon:
            # add section to canvas objects with width=x group
            # this means that width will be adjusted dynamically
            # when rescaling, using a fast tk aglorithm 
            self.obj.netmon.addtag_withtag('width=%.1f'%(width),id)
            
        
        for berth in self.berths:
            x0=berth['halt_point']-0.9*self.berthssec.berth_length
            id0=canvas.create_line(\
                self.berthssec.transform_polyline([x0,dy0, x0,dy1]),
                width=1,
                tag=self.id,fill=color_berth)
            x1=berth['halt_point']
            id1=canvas.create_line(\
                self.berthssec.transform_polyline([x1,dy0, x1,dy1]),
                width=1,
                tag=self.id,fill=color_berth)


        #if self.name:
        # put info on stop
        canvas.create_text(self.berthssec.transform_polyline(\
            [2,-self.side*10]),
                           text='Stop: '+self.name,
                           anchor='w',
                           tag=self.id,
                           font=('helvetica',8,'normal'),
                           fill='black')

        
        
        if self.debug>=5:
            # put name of berth below
            canvas.create_text(
                self.berthssec.transform_polyline([2,-self.side*15]),
                text=self.id,
                tag=self.id,
                font=('helvetica',8,'normal'),
                fill='black')
            
            

        

class ShortStopmod(Stopmod):
    """
    Stop with short acceleration line.
    Requires more conservative merge.
    """
    def control_car_on_enter_mergemod(self,sec,car,controldata):
        """
        Mergemod specific control_car_on_enter.

        Designed for 2 merge branches only.
        """
        #if (self.id=='test.track.merge_curve/1')&(sec==self.mergesecs[0]):
        #    print '========================================================'
        #    print 'control_car_on_enter_mergemod',sec.id,'@',self.obj.sim.time
        #    car.show_info()
        #    
        #    if car.car_infront:
        #        print '------------------>>'
        #        
        #        car.car_infront.show_info()
        
        
        if sec == self.mergesecs[1]: # on accel line
            #print '   msec 1 detected'
            if self.mergesecs[1].length>self.mergesecs[0].length:
                #'  do merge decision later under equal distance conditions'
                controldata['mark_at']=self.mergesecs[1].length-\
                                        self.mergesecs[0].length
            else:
                self.append_cars_noghost(car)
                #'   merge this carrier with carriers on section 0'
                if len(self.mergesecs[0].cars)>0:
                    # stop car here. Well be reaccelerated when carriers leve merge
                    controldata['velocity_line']=0.0
                #self.merge(car,self.mergesecs[0],controldata)
                
        elif sec == self.mergesecs[0]:
            #print '   msec 0 detected'
            if self.mergesecs[0].length>self.mergesecs[1].length:
                #'   do merge decision later under equal distance conditions'
                controldata['mark_at']=self.mergesecs[0].length-\
                                        self.mergesecs[1].length
            else:
                self.append_cars_noghost(car)
                #'    merge this carrier with carriers on section 1'
                self.merge(car,self.mergesecs[1],controldata)

        #else:
        #    # output section detected
        #    self.reassign_car_infront_after_merge()
            
            
        return controldata
    
    def control_car_on_leave_mergemod(self,sec,car,controldata):
        """
        Mergemod specific control_car_on_leave
        """
        #print 'control_car_on_leave_mergemod',sec.id,car.id
                
        # remove ghost carrier, if there is any
        controldata['ghost_infront']=None

        self.remove_cars_noghost(car)
        
        
        if sec == self.mergesecs[0]: # we are on main line
            # now accelerate cars on accel line, if possible
            if len(sec.cars)==0:
                if len(self.secs_dir['accel'].cars)>0:
                    #print '  accel',self.secs_dir['accel'].cars[-1].id
                    self.secs_dir['accel'].cars[-1].control(\
                        {'velocity_line':self.secs_dir['accel'].velocity_line})
 
 
    def control_car_on_enter(self,sec,car,controldata):
        '''
        Overwrites Stopmod function.
        Brakes down carrier alredy on main line
        
        Track element wide carrier control when carrier enters section sec.
        The suggested controldata from the respective section may be
        modified and MUST be returned.
        Here, command carrier to halt at halt_point.
        '''
        
        # is car entering the berth section ?
        #if sec.label=='entry':
        #    pass

        if sec.label=='berths':
            #if self.probe():
                #print 'control_car_on_enter',car.id,sec.label
                #print '   at',self.id,sec.id,sec.label
            #if self.probe():
            #    print 'control_car_on_enter',car.id,sec.label
            
            # insert car into track elemet wide cars dict
            # currently only used for berths section
            self.update_cars(car)
            
            # determine current state of car
            if self.car_tasks.has_key(car.id):
                # use always oldest  task
                task=self.car_tasks[car.id][0]
                #print 'control_car_on_enter:'
                # carrier has a load for this stop or is empty 
                if ((task.mode=='search_item') | task.target_flag):
                    #print '  make car searching for berth'
                    self.update_cars(car,{'state':'searching',
                                          'loaded': task.mode=='loaded'})
                    
                else:
                    #print '  carrier does not want to stop at a berth'
                    self.update_cars(car,{'state':'passthrough',
                                          'loaded':task.mode=='loaded' })

                
                
            else: # carrier has no task
                # make car searching
                self.update_cars(car,{'state':'searching'})


            #print 'enter',self.id,'car',car.id,\
            #      'state:',self.cars[car.id]['state'],\
            #      'loaded:',self.cars[car.id]['loaded']
                
            # is stop in synchronous mode?
            if self.time_sync>0.0:
                #print 'synchronous'
                controldata.update(self.alloc_berth_with_car_sync(\
                    car,controldata))
            else:
                #print 'aasynchronous'
                # allocate a berth for carrier, if possible
                controldata.update(self.alloc_berth_with_car(car,controldata))
            
        
        # check whether to diverge at carrier entrance at brake line
        # with a carrier on main_entry line
        elif (sec.label=='brake'):
            if (len(self.secs_dir['main_entry'].cars)>0):
                controldata.update(self.diverge(self.secs_dir['main_entry']))

            # hold on to main line speed untill out of clearing zone
            controldata['velocity_line']=\
                                  self.secs_dir['brake'].velocity_line
            controldata['mark_at']=sec.clearingzone

            # check out here already whether there is an empty carrier
            # at an empty berth, blocking the other berth.

            self.move_waiting_cars(0)

            #if self.queuing_user()==0:
            #    #if len(self.get_users_queued())==0:
            #    self.clear_berths(0)

                        
        # check whether to diverge at carrier entrance at the main_entry line
        # with a carrier on brake line
        elif (sec.label=='main_entry')&(len(self.secs_dir['brake'].cars)>0):
            controldata.update(self.diverge(self.secs_dir['brake']))

        elif  sec.label=='accel':
            # do here special merge for accelerating carrier
            self.append_cars_noghost(car)
            mergesec=self.secs_dir['main_exit']
            mergesec_cars_num=len(mergesec.cars)
            #mergesec_cars_noghost=self.cars_noghost[mergesec.id][:]
             
            if mergesec_cars_num==0:
                # go ahead and accelerate to line speed
                pass
            else:
                self.control_car_on_enter_mergemod(sec,car,controldata)

            #elif mergesec_cars_num==len(mergesec_cars_noghost):
            #    # mergecar is last entered in merge zone
            #    controldata.update(self.get_controldata_to_make_ghost(\
            #         car,mergesec_cars_noghost[-1]))
                
            #elif (self.time_sync>0):#&(self.ind_berth_alloc<0):
            #    #self.append_cars_noghost(car)
            #    #self.merge(car,mergesec,controldata,no_mergecar_behind=1)
            #    self.control_car_on_enter_mergemod(sec,car,controldata)
            #elif (self.time_sync>0)&\
            #     (((mergesec.length-mergesec.cars[0].get_path_sec()\
            #        -mergesec.cars[0].get_brakedist_comfort())-\
            #       (c)
            #      )
            
                
            #else:
            #    # stop car and wait untill it gets accelerated by a car on
            #    # the main line
            #    controldata['velocity_line']=0.0


                

        elif sec.label=='main_exit':
            d_mark=sec.length-self.secs_dir['accel'].length\
                    -car.time_brake*(sec.velocity_line-\
                                     self.secs_dir['berths'].velocity_line)\
                    -0.5/car.accel_comfort*(sec.velocity_line**2-\
                                     self.secs_dir['berths'].velocity_line**2)\
                    -sec.velocity_line**2*\
                    (0.5/car.accel_emergency-0.5/car.accel_failure)
            
                    
                                     
            # without TB
            #d_mark=sec.length-sec.velocity_line*(\
            #    (sec.velocity_line-self.secs_dir['berths'].velocity_line)
            #    /car.accel_comfort\
            #    +self.secs_dir['accel'].length/sec.velocity_line\
            #   -(sec.velocity_line-self.secs_dir['berths'].velocity_line)**2\
            #    /(2*car.accel_comfort*sec.velocity_line))\
            #    -0.5*sec.velocity_line**2/car.accel_emergency\
            #    +0.5*sec.velocity_line**2/car.accel_failure

            
            #print 'control_car_on_enter at',self.id,d_mark
            if d_mark>0:
                controldata['mark_at']=d_mark
            else:
                # check for mergecontrol instructions
                self.control_car_on_enter_mergemod(sec,car,controldata)

        return controldata
                           
class Track(Obj):
    """ Contains all functions for a generic track simulation object"""
    
    def init_obj(self):

        
        
        # basic parameters...
        self.width=2 # in meter

        # line speed for entire track element
        # if nothing else is specified for sections
        self.velocity_line=30/3.6 # in meters/second

        # appearance on the Netmon 2D visualizer
        #/usr/lib/X11/rgb.txt
        self.width_netmon=2.5
        self.color_netmon='blue'
        self.color_netmon_edit='green'
        
        # rotation angle 
        self.angle_rot=1
        
        # objects to be linked with in the link method
        self.netsim=None
        self.netmon=None
        self.netedit_gui=None
        
        # default management simulaton object
        self.man_id='test.trackman.experim'

        # common key map for edit mode
        self.keymap_edit={'n': self.swap_rootnode,
                          'a': self.rotate_clockwise,
                          'A': self.rotate_counterclockwise,}

        # section classes
        self.sec_classes={'line':Linesec,
                          'curve':Curvesec,
                          'scurve':Scurvesec,
                          'flexline':Flexlinesec
                          }
        

        # number_of_segments per section >3 (for curved sections only) 
        #self.number_of_segments=10

        # length of segment in meters
        self.length_seg=4.0
        
        # define initialy a None-module (here track element) to be edited 
        self.editmod=None

        # this defines the module datastructure of the track element
        # it may be overwritten in make_track
        self.Mod=Mod

        
        # initialize track specific parameters
        self.init_track()
        self.make_track()

    def init_track(self):
        """
        Initialise all track specific parameters except geometry of track
        itself, which is done in make_track().
        """
        pass
        
    def get_state(self):
        '''Prepare state image for save operation'''

        if self.netedit_gui:
            self.netedit_gui.default()
            

        state={'mods':self.mods,
               'mod_no_current':self.mod_no_current}
        state.update(self.get_state_specific())
        return state

    def get_state_specific(self):
        """
        Track object specific states.
        """
        return {}
        
    def set_state(self,state_image):
        '''Set state while opening simulation file'''
        self.mods=state_image['mods']
        self.mod_no_current=state_image['mod_no_current']

        for mod in self.mods.values():

            if state_image['version']=='0.0':
                mod.obj=self
                # the following lines are for downwardscompatibility with 0.0
                #if not hasattr(mod,'man_id'):
                # make sure that all sections of a element have a type 
                if hasattr(mod,'add_car'):
                    mod.secs[0].type='berth' # only sec type is berth
                    mod.secs[0].localsec['halt_point']=4.0
                    mod.placesec=mod.secs[0]
                    mod.halt_point=mod.placesec.localsec['halt_point']
                    mod.users={}
                    mod.queue=[]
                    mod.queue_max=5
                    
                else:
                    for sec in mod.secs:
                        # all others are simple input output
                        # elements, no berth no storage
                        sec.type='io'
                        
                mod.init_common()
                # restore all state of module that got lost during
                # save operation
                mod.set_state(self)

                # Do initializations concerning the management
                mod.car_tasks={} # controlled by Transport tasks
                mod.add_man()
                mod.update_man()
            

                for sec in mod.secs:
                    sec.clearingzone=10.0
                    sec.label=''
                    
            else:
                # restore all state of module that got lost during
                # save operation
                #print 'set_state: call mod.set_state for',mod.id
                mod.set_state(self)

        # Set track object specific state.
        self.set_state_specific(state_image)

        
                
    def set_state_specific(self,state_image):
        """
        Set track object specific state.
        """
        pass
        
    def link(self):
        self.netsim=self.try_link_with('base.net.netsim')
        self.netmon=self.try_link_with('base.net.netmon')
        self.netedit_gui=self.try_link_with('base.net.netedit_gui')

        # rebuild_pointers already here 
        if self.sim.init_flag:
            #print 'Track.link rebuild pointers of',self.id
            for mod in self.mods.values():
                mod.rebuild_pointers()
                if self.netmon:
                    mod.plot()

        self.link_specific()
        

    def link_specific(self):
        """
        Specific linking for stop claas for example...
        """
        pass

    def rebuild_pointers(self):
        """
        Rebuild pointers that have been removed before saving.
        This function is called AFTER all objects have been
        initialized and cross-linked.
        """
        #print 'rebuild_pointers',self.id,string.atof(self.get_version())
        if string.atof(self.get_version())<0.3:
            for mod in self.mods.values():
                #print '  call update_man'
                mod.update_man(0)
                    
    def report_statistics_specific(self,descriptor,dataset,
                                   dataset_sum,dataset_counter):
        """
        Return object specific descriptor and dataset contents for report on
        statistical data.
        """
        # compute length of all modules of this object
        length=0
        for mod in self.mods.values():
            length+=mod.get_length()

        descriptor.append({'key':'length',
                           'name':'Track length',
                           'type':'real',
                           'printformat':'%.2f',
                           'width':1,
                           'valid':{'min':0,'max':10**8},
                           'units':{'m':1},
                           'mode':'read',
                           })
        dataset['length']=length
        
    def report_costs_specific(self,descriptor,dataset,
                              dataset_sum,dataset_counter):
        """
        Return object specific descriptor and dataset contents for report on
        costs data.
        """
        # remove these because should not be in object report data
        #if dataset.has_key('length_total'):del dataset['length_total']
        #if dataset.has_key('length'):del dataset['length']
        pass
        
    def report_params(self):
        """
        Returns describtor and dataset of parameters.
        """
        descriptors=[{'key':'velocity_line_min',
                      'name':'Minimum line speed',
                      'type':'real',
                      'printformat':'%3.3f',
                      'width':1,
                      'mode':'read',
                      'valid':{'min':0,'max':10000},
                      'units':{'m/s':1},
                      },
                     {'key':'velocity_line_max',
                      'name':'Maximum line speed',
                      'type':'real',
                      'printformat':'%3.3f',
                      'width':1,
                      'mode':'read',
                      'valid':{'min':0,'max':10000},
                      'units':{'m/s':1},
                      },]
        
        dataset={'velocity_line_max':0.0,
                 'velocity_line_min':10.0**8}

        # just get maximum and minimum line speed of all modules and sections 
        for mod in self.mods.values():
            if mod.status=='active':
                for sec in mod.secs:
                    if sec.velocity_line<dataset['velocity_line_min']:
                        dataset['velocity_line_min']=sec.velocity_line
                    elif sec.velocity_line>dataset['velocity_line_max']:
                        dataset['velocity_line_max']=sec.velocity_line
        return descriptors,dataset
            
    def place_gui(self,x,y,z=0,rootnode=-1):
        mod=self.editmod
        if rootnode==-1:
            rootnode=self.swapnode_list[0]
            
        #print 'place_gui to:',x,y,z
        mod.set_root(rootnode,[x,y,z],self.editmod.get_angle(),swapnodes=1)

        # get the global open node that is closest to the global node,
        # corresponding to the root node of module mod
        node_closest=self.netsim.get_closest_node(
            mod.nodes_open[mod.rootnode] )
        
        if node_closest:
            self.editmod.set_root(self.swapnode_list[0],
                                  trans=node_closest.localnode['trans'],
                                  swapnodes=1)

        # put track element module into passive mode
        self.editmod.set_status('passive')


    def del_gui(self,event,tag):
        # get module with track element out of tag string
        self.editmod=self.sim.split_id(tag)['mod']

        # if there is no error setting module editmod to status edit...
        if not self.del_mod(self.editmod): # try to delete 
            # if successful, delete also from monitor
            self.editmod.del_plot()

        self.editmod=None
        self.netedit_gui.del_mod()

    def activate_gui(self,event,tag):
        # get module with track element out of tag string
        self.editmod=self.sim.split_id(tag)['mod']
        self.editmod.set_status('active')
        #print 'Activate',self.editmod.id
        #print 'Status= ',self.editmod.status
        self.editmod.plot()
        self.editmod=None
        self.netedit_gui.activate_mod()
        
    def inactivate_gui(self,event,tag):
        # get module with track element out of tag string
        self.editmod=self.sim.split_id(tag)['mod']
        self.editmod.set_status('passive')
        #print 'Deactivate',self.editmod.id
        #print 'Status= ',self.editmod.status
        self.editmod.plot()
        self.editmod=None
        self.netedit_gui.inactivate_mod()

    def show_info(self,event,tag):
        # get module with track element out of tag string
        self.editmod=self.sim.split_id(tag)['mod']
        self.editmod.show_info()
        self.netedit_gui.show_info()
        
    def move_gui(self,event,tag):
        # get module with track element out of tag string
        tag_data=self.sim.split_id(tag)
        self.editmod=tag_data['mod']

        node_id=tag_data['part_no']
        #print 'move_gui tag',tag,node_id,\
        #      self.editmod.set_status('edit')
        

        # if there is no error setting module editmod to status edit...
        if not self.editmod.set_status('edit'):

            canvas=event.widget

            # here x,y are the coordinates at 100% magnification
            x,y=self.netmon.get_unmagnified_canvas_coords(event)
            
            #print 'move_gui node',node_id
            #print '  rootnode  trans0=\n',\
            #      self.editmod.nodes[self.editmod.rootnode]['trans0']
            #print '  rootnode  trans=\n',\
            #      self.editmod.nodes[self.editmod.rootnode]['trans']
            
            # click on node -> try to stretch 
            if node_id:
                
                # search section with this node_id
                found=0
                for sec in self.editmod.secs:
                    if (sec.fromnode['id']==node_id)|\
                       (sec.tonode['id']==node_id):
                        sec.start_stretch_gui(event,node_id)
                        found=1
                        break
                    
                if not found:
                    self.start_move_gui(event,x,y)
            else:
                self.start_move_gui(event,x,y)
            
        else: # module editmod cannot be moved, maybe it is active
            self.editmod=None
            self.netedit_gui.move_mod()
            

    def start_move_gui(self,event,x,y):
        """
        Initialize callbacks to move  module around and to place it.
        x,y are coordinates at 100% magnification.
        """
        # reset list for changing the root of
        # the track element to be edited
        #event.x=x
        #event.y=y


        self.swapnode_list=self.nodes_connect.keys()
        self.swapnode_list.sort()
        
        
        if self.debug>=5:
            print "Move: ", self.editmod.id

        #x=event.x
        #y=event.y

        #self.netmon.canvas.scan_mark(x,y)

        # store coordinates in magnified system (used to compute diff only)
        event.x,event.y=self.netmon.magnify_xy(x,y)
        self.event_past = event

            
        self.editmod.set_root(self.swapnode_list[0],
                              [x,y,0],
                              self.editmod.get_angle(),swapnodes=1)
        self.editmod.plot()

        mod_x, mod_y, mod_z=self.editmod.get_pos()

        self.offset_x = mod_x-x
        self.offset_y = mod_y-y
        self.netmon.set_action('move track element',
                               callback_click = self.move_place,
                               callback_motion = self.add_motion,
                               callback_pressmotion = self.add_motion,
                               #callback_button2click=\
                               #                 self.rotate_counterclockwise,
                               callback_rightclick=self.move_end,
                               callback_keymap=self.keymap_edit,
                               cursor='right_ptr')
        
        
    def move_place(self,event):

        # Coordinates x,y are the coordinates at 100% magnification.
        # The magnified coordinates are event.x, event.y
        
        x,y=self.netmon.get_unmagnified_canvas_coords(event)

        # place with 100% magnification coordinates
        self.place_gui(x,y,z=0)
     
        self.netedit_gui.move_mod()
        

    def move_end(self,event):
        self.netedit_gui.default()
        
    def clone_gui(self,event,tag):
        """
        Clone a tagged module.
        """
        # get module with track element out of tag string
        tag_data=self.sim.split_id(tag)
        mod=tag_data['mod']

        # get clone
        self.editmod=self.clone(mod)
        
        if not self.editmod.set_status('edit'):
            canvas=event.widget
            # here x,y are the coordinates at 100% magnification
            x,y=self.netmon.get_unmagnified_canvas_coords(event)

            # start moving it around
            self.start_clone_gui(event,x,y)

        else: # module editmod cannot be moved, maybe it is active
            self.editmod=None
            self.netedit_gui.clone_mod()

    def start_clone_gui(self,event,x,y):
        """
        Initialize callbacks to move cloned module around and to place it.
        x,y are coordinates at 100% magnification.
        """
        # reset list for changing the root of
        # the track element to be edited
        #event.x=x
        #event.y=y
        
        self.swapnode_list=self.nodes_connect.keys()
        self.swapnode_list.sort()

        
        
        if self.debug>=5:
            print "Clone: ", self.editmod.id

        #x=event.x
        #y=event.y

        #self.netmon.canvas.scan_mark(x,y)
        event.x,event.y=self.netmon.magnify_xy(x,y)
        self.event_past = event

            
        self.editmod.set_root(self.swapnode_list[0],
                              [x,y,0],
                              self.editmod.get_angle(),swapnodes=1)
        self.editmod.plot()

        mod_x, mod_y, mod_z=self.editmod.get_pos()

        self.offset_x = mod_x-x
        self.offset_y = mod_y-y
        self.netmon.set_action('clone track element',
                               callback_click = self.clone_place,
                               callback_motion = self.add_motion,
                               callback_pressmotion = self.add_motion,
                               callback_rightclick=self.move_end,
                               callback_keymap=self.keymap_edit,
                               cursor='dotbox')
    def clone(self,mod):
        """
        Return a cloned version of module.
        """
        # clone...
        clonemod=self.Mod(self,mod.rootnode,
                          mod.get_pos(),
                          mod.get_angle())

        # do better cloning...not so easy. Problems with rootnode 
        
        # clone geometry
        #n=0
        #for sec in mod.secs:
        #    clonemod.secs[n].param=sec.param
        #    clonemod.secs[n].update_segs(sec.param)
            
        return clonemod
                
    def clone_place(self,event):
        """
        Place cloned module and call again netedit's clone module function.
        """
        
        canvas=event.widget
        if canvas==self.netmon.canvas:
            x,y=self.netmon.get_unmagnified_canvas_coords(event)
            self.place_gui(x,y,rootnode=self.editmod.rootnode)
            self.netedit_gui.clone_mod()
            
    def add_gui(self):
        
        # reset list for changing the root of the track element to be edited
        self.swapnode_list=self.nodes_connect.keys()
        self.swapnode_list.sort()

        
        self.netmon.set_action('put track element',
                               callback_click = self.add_click,
                               callback_motion = self.add_click,
                               callback_rightclick=self.add_end,
                               callback_termination = self.add_termination)
        
        
    def add_click(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:
            

            x,y=self.netmon.get_unmagnified_canvas_coords(event)

            self.editmod=self.Mod(self,self.swapnode_list[0],[x,y,0],0.0)
            self.editmod.plot()
            self.netmon.scrolledcanvas.resizescrollregion()
            
            if self.debug>=5:
                print 'Initialized module ' + self.editmod.id

            
            self.netmon.set_action('put track element',
                                   callback_doubleclick=self.show_pos,
                                   callback_click = self.add_place,
                                   callback_motion = self.add_motion,
                                   callback_pressmotion = self.add_motion,
                                   #callback_button2click=\
                                   #             self.rotate_counterclockwise,
                                   #callback_rightclick=self.rotate_clockwise,
                                   callback_rightclick=self.add_end,
                                   callback_keymap=self.keymap_edit,
                                   callback_termination = self.add_termination)

            # this is a bit dirty at the moment...improve also at
            # add_motion
            event.x=canvas.canvasx(event.x)
            event.y=canvas.canvasy(event.y)
            self.event_past = event
            
    def rotate(self,mod,angle,event):
        """
        Rotate editmod
        """

        canvas=event.widget
        # set new root with 100% magnification
        x,y=self.netmon.get_unmagnified_canvas_coords(event)
        mod.set_root(self.swapnode_list[0],[x,y,0],angle,swapnodes=1)
        
        
    def rotate_clockwise(self,event):
        """
        Rotate editmod rotate_clockwise
        """
        #alpha0=self.editmod.get_angle0()
        alpha=self.editmod.get_angle()+self.angle_rot#+alpha0
        #print 'rotate_clockwise',self.editmod.id,self.angle_rot,self.editmod.get_angle(),alpha
        
        self.rotate(self.editmod,alpha,event)
        self.editmod.plot()

    def rotate_counterclockwise(self,event):
        """
        Rotate editmod rotate_clockwise
        """
        #print 'rotate_counterclockwise',self.editmod.id,self.angle_rot,self.editmod.get_angle()
        self.rotate(self.editmod,self.editmod.get_angle()-self.angle_rot,event)
        self.editmod.plot()

    def swap_rootnode(self,event):
        # rotate list with local connecting node ids
        self.swapnode_list.append(self.swapnode_list[0])
        del self.swapnode_list[0]

        canvas=event.widget
        # set new root
        x,y=self.netmon.get_unmagnified_canvas_coords(event)
        self.editmod.set_root(self.swapnode_list[0],
                              [x,y,0],
                              self.editmod.get_angle(),swapnodes=1)
        self.editmod.plot()
        
   
    def show_pos(self,event):
        canvas=event.widget
        x=canvas.canvasx(event.x)
        y=canvas.canvasy(event.y)
        print '** Current pos: ',event.x,event.y,x,y
        
    def add_motion(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:
            if self.rescaled_flag:
                # reset root to current pointer position 
                x,y=self.netmon.get_unmagnified_canvas_coords(event)
                event.x,event.y=self.netmon.magnify_xy(x,y)
                self.event_past = event
                self.editmod.set_root(self.swapnode_list[0],
                              [x,y,0],
                              self.editmod.get_angle(),swapnodes=1)
                self.editmod.plot()
                self.rescaled_flag=0
            else:
                x=canvas.canvasx(event.x)
                y=canvas.canvasy(event.y)
                self.netmon.move([(self.editmod.id,
                                   self.event_past.x, x,
                                   self.event_past.y, y)],noscale=1
                                 )
                event.x=x
                event.y=y
                self.event_past=event
    
    def add_place(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:
            x,y=self.netmon.get_unmagnified_canvas_coords(event)
            self.place_gui(x,y)
            self.add_click(event)

    def add_end(self,event):
        
        self.add_termination(event)
        #self.netedit_gui.default()
        
        
    def add_termination(self,event=None):
        if self.editmod:
            
            if self.debug>=5:
                print 'Terminate Add module '+self.editmod.id

            if self.del_mod(self.editmod):
                return 1
            
            else:
                # delete from monitor
                self.editmod.del_plot()
                self.editmod=None
                self.netedit_gui.move_mod()
                return 0
            
    def add_mod(self,mod):
        """
        Insert new module in the module database mods of the simulation object
        Initial module values are already initialized by the  module mod.
        """
        self.mod_no_current=self.mod_no_current+1
        mod.no=self.mod_no_current
        mod.id=self.get_mod_id(mod)
            
        # add module to modules database
        self.mods[self.mod_no_current]=mod

        # add module with track element to netsim
        if self.netsim:
            self.netsim.add_elem(mod)

        self.add_mod_specific(mod)
        
        return 0

    def add_mod_specific(self,mod):
        pass
        
    def del_mod(self,mod):
        """
        Delete module
        """
        if self.editmod.set_status('edit'):
            print 'Could not set edit mode for Track element ',self.editmod.id
            return 1
        else:
            self.netsim.del_elem(mod)
            # delete elem from object internal array structures
            self.del_mod_specific(mod)
            del self.mods[mod.no]
            
            if len(self.mods)==0:
                # remove completely simulation object from current simulation
                self.sim.remove_obj(self)
            return 0
    def del_mod_specific(self,mod):
        """
        Delete module specific
        """
        pass

class Stop(Track):
    """
    Methods for stop objects.
    """
    def init_track(self):
        """
        Initialise all track specific parameters except geometry of track
        itself, which is done in make_track().
        """
        # track elements are usually event driven,
        # i.e. control functions are called when a carrier is entering,
        # leaving, e.t.c However stops do need to update internal data
        # For example to place waiting carriers on the track
        self.iterate_interval=1.0 # seconds
        
        

        # list of stops that are ready to put off-track carriers on track.
        self.stops_ready_to_add_offtrack_cars=[]

        # timer for switch between load/unload operation
        self.timer=Array_db()

        # make here already sure it is in the iterate cycle
        if self.netsim:
            self.sim.append_obj_iterate(self)
        
    def get_state_specific(self):
        """
        Track object specific states.
        """
        #print 'get_state_specific of',self.id
        #print '  timer=',self.timer.arrays
        
        stop_nos=[]
        for stop in  self.stops_ready_to_add_offtrack_cars:
            stop_nos.append(stop.no)
        del self.stops_ready_to_add_offtrack_cars

        

        return   {'stops_ready_to_add_offtrack_cars_nos':stop_nos,
                  'timer':self.timer}
    
        #return {'stops_ready_to_add_offtrack_cars':
        #        self.stops_ready_to_add_offtrack_cars}
        #return {}
    
    def set_state_specific(self,state_image):
        """
        Set track object specific state.
        """
        #print 'set_state_specific of',self.id
        self.stops_ready_to_add_offtrack_cars=[]
        for stop_no in state_image['stops_ready_to_add_offtrack_cars_nos']:
            self.stops_ready_to_add_offtrack_cars.append(self.mods[stop_no])
        
        if string.atof(self.get_version())<0.2:
            #print '  old version'
            self.iterate_interval=1.0 # seconds

        else:
            self.timer= state_image['timer']

        
        
    def ready_to_add_offtrack_cars(self,stop):
        """
        Indicate that stop is ready to put off-track carriers on track.
        During the iterate method these stops are called to add
        the off-track carriers.
        """
        # Cars that are beside the track, but ready to be placed on the track
        # Key is here the module no
        # Value is a directiry of cars, with car.id as key and
        # pointer to module as value.
        #print 'ready_to_add_offtrack_cars',stop.id
        self.stops_ready_to_add_offtrack_cars.append(stop)

        # cannot place carriers here during placing a carrier
        # Append simulation object to maincycle
        #self.sim.append_obj_iterate(self)
        # then carriers are added when calling iterate
        
        
    def iterate(self):
        """
        Update of internal states.
        """
        #print '-----------------------------------------------',self.sim.time
        # check if there are off track carriers ready to put back on track
        if self.stops_ready_to_add_offtrack_cars:
            # go throught all elem_ids that offer to place a car on the track
            for stop in self.stops_ready_to_add_offtrack_cars:
                stop.add_cars_offtrack()
                
            self.stops_ready_to_add_offtrack_cars=[]

            # Remove simulation object from maincycle
            #self.sim.remove_obj_iterate(self)

        # do stos in sync mode:
        # count down  timer 
        self.timer.arrays=self.timer.arrays-self.iterate_interval
        
        #print self.timer.arrays
        
        # get array index of those which terminated
        inds_term=nonzero(greater(0,self.timer.arrays))
        if len(inds_term)>0:
            for ind in inds_term:
                # tell stop to handle time-out event
                elem=self.mods[self.timer.get_key_with_index(ind)]
                #if elem.probe():
                #    print 'obj.iterate:'
                
                self.mods[self.timer.get_key_with_index(ind)].handle_timer()
                
        # do more stop specific stuff...used by dual stop
        self.iterate_specific()
        
    def iterate_specific(self):
        """
        Update of specific internal states.
        """
        pass

    def link_specific(self):
        """
        Specific linking for stop claas for example...
        """
        if self.sim.init_flag:
            
            if len(self.mods)>0:
                # print 'append_obj_iterate',self.id
                self.sim.append_obj_iterate(self)

                
    def add_timer_arrays(self,stop,time):
        '''
        Adding or setting time of stop in timer array.
        '''
        
        self.timer[stop.no]=time
        
    def add_mod_specific(self,mod):
        if self.netsim:
            self.sim.append_obj_iterate(self)
        
    def del_mod_specific(self,mod):
        """
        Delete module specific
        """
        self.del_timer_arrays(mod)
        if self.netsim:
            if len(self.mods)==0:
                self.sim.remove_obj_iterate(self)

        
    def del_timer_arrays(self,stop):
        '''
        Delete list of users from  array structure, used for boarding
        and unboarding.
        This is currently done after waiting.
        '''
        if self.timer.db.has_key(stop.no):
            del self.timer[stop.no]
            
class DualStopmod(Stopmod):
    """
    Methods for dual stop modules, using unloading and loading zone.
    """

    def handle_timer(self):
        """
        Called on timeout of timer during iterate.
        """
        pass
        
        
        
class Dualstop(Stop):
    """
    Methods for dual stop objects, using unloading and loading zone.
    """
    def init_track(self):
        """
        Initialise all track specific parameters except geometry of track
        itself, which is done in make_track().
        """
        # track elements are usually event driven,
        # i.e. control functions are called when a carrier is entering,
        # leaving, e.t.c However stops do need to update internal data
        # For example to place waiting carriers on the track
        self.iterate_interval=1.0 # seconds

        # list of stops that are ready to put off-track carriers on track.
        self.stops_ready_to_add_offtrack_cars=[]

        # timer for switch between load/unload operation
        self.timer=Array_db()


    def iterate_specific(self):
        """
        Update of specific internal states.
        """
        # count down  timer 
        self.timer.arrays-=self.iterate_interval
        
        # get array index of those which terminated
        inds_term=nonzero(greater(0,self.timer.arrays))
        if len(inds_term)>0:
            for ind in inds_term:
                # tell stop to handle time-out event
                self.mods[self.timer.get_key_with_index(ind)].handle_timer()
                
        
    def add_timer_arrays(self,stop,time):
        '''
        Adding or setting time of stop in timer array.
        '''
        self.timer[stop.no]=time
        
        
    def del_mod_specific(self,mod):
        """
        Delete module specific
        """
        self.del_timer_arrays(mod)

        
    def del_timer_arrays(self,stop):
        '''
        Delete list of users from  array structure, used for boarding
        and unboarding.
        This is currently done after waiting.
        '''
        if self.timer.db.has_key(stop.no):
            del self.state[stop.no]

    
